// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "application.h"

#include <QDebug>
#include <QLibraryInfo>
#include <QTimer>
#include <QTranslator>
#include <qcef_context.h>
#include <qcef_cookie_store.h>

#include "base/consts.h"
#include "service/album_cover_cache.h"
#include "service/dbus_manager.h"
#include "service/download_manager.h"
#include "service/feedback_manager.h"
#include "service/keybinder.h"
#include "service/local_music_manager.h"
#include "service/log_manager.h"
#include "service/lrc_manager.h"
#include "service/media_metadata.h"
#include "service/media_player.h"
#include "service/media_player_cache.h"
#include "service/popup_menu_manager.h"
#include "service/process_manager.h"
#include "service/signal_manager.h"
#include "service/skin_manager.h"
#include "service/storage_manager.h"
#include "service/system_tray_icon.h"
#include "service/thread_manager.h"
#include "service/url_manager.h"
#include "ui/window_manager.h"
#include "ui/views/desktop_lyrics.h"
#include "ui/views/mini_player.h"

namespace netease {

namespace {

// Currently only Simplified Chinese is supported.
const char kLocaleZhCNFile[] = ":/i18n/cloud-music_zh_CN.qm";

const char kAppIconName[] = ":/misc/netease-cloud-music.svg";

}  // namespace

Application::Application(int& argc, char** argv) : QApplication(argc, argv) {
  this->setI18n();

  // Sometimes no window is displayed, so switch this option off.
  this->setQuitOnLastWindowClosed(false);
  this->setAttribute(Qt::AA_UseHighDpiPixmaps);
}

Application::~Application() {
  if (this->dbus_manager) {
    delete this->dbus_manager;
    this->dbus_manager = nullptr;
  }

  if (this->feedback_manager) {
    delete this->feedback_manager;
    this->feedback_manager = nullptr;
  }

  if (this->keybinder) {
    delete this->keybinder;
    this->keybinder = nullptr;
  }

  if (this->local_music_manager) {
    delete this->local_music_manager;
    this->local_music_manager = nullptr;
  }

  if (this->lrc_manager) {
    delete this->lrc_manager;
    this->lrc_manager = nullptr;
  }

  if (this->desktop_lyrics) {
    delete this->desktop_lyrics;
    this->desktop_lyrics = nullptr;
  }

  if (this->mini_player) {
    delete this->mini_player;
    this->mini_player = nullptr;
  }

  if (this->window_manager) {
    delete this->window_manager;
    this->window_manager = nullptr;
  }

  if (this->media_player) {
    delete this->media_player;
    this->media_player = nullptr;
  }

  if (this->popup_menu_manager) {
    delete this->popup_menu_manager;
    this->popup_menu_manager = nullptr;
  }

  if (this->process_manager) {
    delete this->process_manager;
    this->process_manager = nullptr;
  }

  if (this->signal_manager) {
    delete this->signal_manager;
    this->signal_manager = nullptr;
  }

  if (this->skin_manager) {
    delete this->skin_manager;
    this->skin_manager = nullptr;
  }

  if (this->system_tray_icon) {
    delete this->system_tray_icon;
    this->system_tray_icon = nullptr;
  }

  if (this->url_manager) {
    delete this->url_manager;
    this->url_manager = nullptr;
  }

  if (this->media_metadata) {
    delete this->media_metadata;
    this->media_metadata = nullptr;
  }

  if (this->storage_manager) {
    delete this->storage_manager;
    this->storage_manager = nullptr;
  }

  if (this->album_cover_cache) {
    delete this->album_cover_cache;
    this->album_cover_cache = nullptr;
  }

  if (this->download_manager) {
    delete this->download_manager;
    this->download_manager = nullptr;
  }

  if (this->media_player_cache) {
    delete this->media_player_cache;
    this->media_player_cache = nullptr;
  }

  // Terminate background threads after all workers stopped.
  if (this->thread_manager) {
    delete this->thread_manager;
    this->thread_manager = nullptr;
  }

  if (this->log_manager) {
    delete this->log_manager;
    this->log_manager = nullptr;
  }
}

void Application::initChildren() {
  // Initialize base objects. These objects do not depend on any other
  // manager objects.
  this->signal_manager = new SignalManager();
  this->log_manager = new LogManager();
  this->thread_manager = new ThreadManager();
  this->process_manager = new ProcessManager();
  this->url_manager = new UrlManager();

  this->album_cover_cache = new AlbumCoverCache();
  this->album_cover_cache->moveToThread(
      this->thread_manager->album_cover_cache_thread);

  // DownloadManager depends on StorageManager.
  this->download_manager = new DownloadManager();
  this->download_manager->moveToThread(
      this->thread_manager->download_manager_thread);

  // Feedback depend on UrlManager.
  this->feedback_manager = new FeedbackManager();
  this->feedback_manager->moveToThread(
      this->thread_manager->resources_manager_thread);

  this->local_music_manager = new LocalMusicManager();
  this->local_music_manager->moveToThread(
      this->thread_manager->local_music_thread);

  this->media_metadata = new MediaMetadata();
  this->media_metadata->moveToThread(
      this->thread_manager->resources_manager_thread);

  this->storage_manager = new StorageManager();
  this->storage_manager->moveToThread(this->thread_manager->storage_thread);

  // MediaPlayerCache depends on DownloadManager.
  // MediaPlayerCache depends on StorageManager.
  this->media_player_cache = new MediaPlayerCache();
  this->media_player_cache->moveToThread(
      this->thread_manager->media_player_cache_thread);

  // MediaPlayer depends on MediaPlayerCache.
  this->media_player = new MediaPlayer();

  // DBusManager depends on MediaPlayer.
  this->dbus_manager = new DBusManager();

  // LrcManager depends on MediaPlayer.
  this->lrc_manager = new LrcManager();

  // Keybinder depends on MiniPlayer.
  this->keybinder = new Keybinder();

  // SkinManager is required by PopupManager.
  this->skin_manager = new SkinManager();
  // SystemTrayIcon depends on SignalManager.
  this->system_tray_icon = new SystemTrayIcon();
  this->popup_menu_manager = new PopupMenuManager();
  this->desktop_lyrics = new DesktopLyrics();
  this->mini_player = new MiniPlayer();
  this->window_manager = new WindowManager();
}

void Application::setI18n() {
  // Load system translations resources.
  QTranslator* sys_translator = new QTranslator(this);
  const QString qt_trans = QLibraryInfo::location(
      QLibraryInfo::TranslationsPath);
  if (sys_translator->load(QDir(qt_trans).filePath("qt_zh_CN.qm"))) {
    this->installTranslator(sys_translator);
  }

  QTranslator* translator = new QTranslator(this);
  if (translator->load(kLocaleZhCNFile)) {
    this->installTranslator(translator);
  } else {
    qCritical() << "[main] Failed to load zh cn qm file: " << kLocaleZhCNFile;
  }
  this->setApplicationName(kAppName);
  this->setApplicationVersion(kAppVersion);

  this->setApplicationDisplayName(QObject::tr(kAppDisplayName));

  QApplication::setWindowIcon(QIcon(kAppIconName));
}

void Application::quitApplication() {
  QCefFlushCookies();
  // Stop player first.
  media_player->stop();

  window_manager->closeAllWebWindows();

  QTimer::singleShot(500, this, &QApplication::quit);
}

}  // namespace netease
