// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "base/device_util.h"

#include <arpa/inet.h>
#include <net/if.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/statvfs.h>

#include <QCryptographicHash>
#include <QDateTime>
#include <QDebug>
#include <QFile>
#include <QSettings>
#include <QTextStream>

#include "base/file_util.h"

namespace netease {

namespace {

const char kCPUInfoFile[] = "/proc/cpuinfo";
const char kMemFile[] = "/proc/meminfo";
const char kNetworkDevFile[] = "/proc/net/dev";
const char kLinuxVersionFile[] = "/proc/version";

const char kLsbReleaseFile[] = "/etc/lsb-release";

// Name of unknown linux distribution.
const char kDistroUnknown[] = "unknown";

const int kDeviceIdLength = 52;

}  // namespace

QString GetCPUModelName() {
  QFile file(kCPUInfoFile);
  if (! file.open(QIODevice::ReadOnly)) {
    qCritical() << "[base]::GetCPUModelName() failed to open /proc/cpuinfo";
    return "";
  }
  QTextStream stream(&file);
  QString line;
  QString model_name_prefix("model name\t");
  line = stream.readLine();
  while (line.size() > 0) {
    if (line.startsWith(model_name_prefix)) {
      file.close();
      return line.remove(model_name_prefix);
    }
    line = stream.readLine();
  }
  file.close();
  return "";
}

QString GetDeviceId() {
  QString id(GetCPUModelName());
  id += GetMemoryTotal();
  id += GetMacAddress();
  if (id.size() == 0) {
    id = QString::number(QDateTime::currentDateTime().toMSecsSinceEpoch());
  }
  QString hash(QCryptographicHash::hash(id.toLatin1(),
                                        QCryptographicHash::Sha256).toHex());
  hash.resize(kDeviceIdLength);
  return hash;
}

bool GetDiskUsage(const QString& path, qint64* total_size, qint64* freespace) {
  struct statvfs buf;
  // Note that QString shall be converted to std::string first.
  // Or else error will occur if there are non ASCII chars in |path|.
  if (statvfs(path.toStdString().c_str(), &buf) == -1) {
    return false;
  } else {
    *total_size = static_cast<qint64>(buf.f_bsize * buf.f_blocks);
    *freespace = static_cast<qint64>(buf.f_bsize * buf.f_bavail);
    return true;
  }
}

QString GetDistroName() {
  QSettings settings(kLsbReleaseFile, QSettings::IniFormat);
  return settings.value("DISTRIB_ID", kDistroUnknown).toString();
}

QString GetDistroNameAndVersion() {
  QSettings settings(kLsbReleaseFile, QSettings::IniFormat);
  return settings.value("DISTRIB_DESCRIPTION", kDistroUnknown).toString();
}

QString GetMacAddress() {
  QStringList device_list(GetNetworkDeviceList());

  for (const QString& device_name : device_list) {
    QString device_mac = GetMacAddress(device_name);
    // Mac address of hysical network device does not ends with "00:00".
    if (! device_mac.endsWith(QString("00:00"))) {
      return device_mac;
    }
  }

  return "";
}

QString GetMacAddress(const QString& device_name) {
  struct ifreq ifr;
  int sock_fd;
  char mac_address[30];

  sock_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (sock_fd == -1) {
    qWarning() << "[base]::GetMacAddress() failed to open socket!";
    return "";
  }
  memset(&ifr, 0, sizeof(ifr));

  strncpy(ifr.ifr_name, device_name.toStdString().c_str(), device_name.size());
  if ((ioctl(sock_fd, SIOCGIFHWADDR, &ifr)) < 0) {
    qWarning() << "[base]::GetMaAddress() calls ioctl() error!";
    return "";
  }

  sprintf(mac_address, "%02x:%02x:%02x:%02x:%02x:%02x",
          (unsigned char)ifr.ifr_hwaddr.sa_data[0],
          (unsigned char)ifr.ifr_hwaddr.sa_data[1],
          (unsigned char)ifr.ifr_hwaddr.sa_data[2],
          (unsigned char)ifr.ifr_hwaddr.sa_data[3],
          (unsigned char)ifr.ifr_hwaddr.sa_data[4],
          (unsigned char)ifr.ifr_hwaddr.sa_data[5]
          );
  close(sock_fd);

  return mac_address;
}

QString GetMemoryTotal() {
  QFile file(kMemFile);
  if (! file.open(QIODevice::ReadOnly)) {
    qCritical() << "[base]::GetMemoryTotal() Failed to open /proc/meminfo.";
    return "";
  }
  QTextStream stream(&file);
  QString line;
  QString mem_total_prefix("MemTotal:\t");
  line = stream.readLine();
  while (line.size() > 0) {
    if (line.startsWith(mem_total_prefix)) {
      file.close();
      // Remove prefix and spaces.
      return line.remove(mem_total_prefix).remove(QChar(' '));
    }
    line = stream.readLine();
  }
  file.close();
  return "";
}

QStringList GetNetworkDeviceList() {
  QFile file(kNetworkDevFile);
  QStringList device_list;

  if (! file.open(QIODevice::ReadOnly)) {
    qCritical() << "[base]::GetNetworkDeviceList() failed to open"
                << kNetworkDevFile;
    return device_list;
  }

  QTextStream stream(&file);

  QString line;
  line = stream.readLine();
  while (line.size() > 0) {
    // device id does not contain '|' mark.
    if (! line.contains("|")) {
      int index = line.indexOf(QChar(':'));
      if (index > 0) {
        device_list.append(line.left(index).remove(QChar(' ')));
      }
    }
    line = stream.readLine();
  }
  file.close();

  return device_list;
}

QString GetLinuxVersion() {
  return ReadTextFileContent(kLinuxVersionFile);
}

bool IsUbuntuUnity() {
  return qgetenv("XDG_CURRENT_DESKTOP") == "Unity" ||
    qgetenv("XDG_CURRENT_DESKTOP") == "ubuntu:GNOME";
}

}  // namespace netease
