// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "base/file_util.h"

#include <QCryptographicHash>
#include <QDebug>
#include <QFile>
#include <QTextCodec>
#include <QTextStream>

#include "base/consts.h"

namespace netease {

namespace {

// Folder name of user cache.
const char kCacheDirName[] = ".cache";

// Folder name of user configuration.
const char kConfigDirName[] = ".config";

}  // namespace

QDir ConcateDir(const QDir& parent_dir, const QString& folder_name) {
  if (! parent_dir.exists(folder_name)) {
    // TODO(xushaohua): Handles permission error.
    parent_dir.mkpath(folder_name);
  }
  return QDir(parent_dir.filePath(folder_name));
}

bool CreateParentDirs(const QString& filepath) {
  return QFileInfo(filepath).absoluteDir().mkpath(".");
}

QString GetFileBasename(const QString& filepath) {
  const int slash_index = filepath.lastIndexOf(QDir::separator());
  QString tmp_filepath;
  if (slash_index > -1) {
    tmp_filepath = filepath.mid(slash_index + 1);
  } else {
    tmp_filepath = filepath;
  }

  const int dot_index = tmp_filepath.lastIndexOf(QChar('.'));
  if (dot_index > 0) {
    return tmp_filepath.left(dot_index);
  } else {
    return tmp_filepath;
  }
}

QString GetFileExtension(const QString& filepath) {
  const int dot_index = filepath.lastIndexOf(QChar('.'));
  if (dot_index > 0) {
    // Ignores hidden file.
    return filepath.mid(dot_index + 1).toLower();
  }

  return "";
}

qint64 GetFileSize(const QString& filepath) {
  QFileInfo info(filepath);
  if (info.exists()) {
    return info.size();
  } else {
    return 0;
  }
}

QString GetUrlHash(const QString& url) {
  const QByteArray content(url.toUtf8());
  return QString(
      QCryptographicHash::hash(content, QCryptographicHash::Md5).toHex());
}

// Alternatively, we can use QStandardPaths::standardLocations().
QDir GetUserCacheDir() {
  QDir cache_dir(QDir::home().filePath(kCacheDirName));
  return ConcateDir(cache_dir, kAppName);
}

QDir GetConfigDir() {
  return QDir::home().filePath(kConfigDirName);
}

QDir GetUserConfigDir() {
  QDir config_dir(QDir::home().filePath(kConfigDirName));
  return ConcateDir(config_dir, kAppName);
}

bool IsMediaFile(const QString& path) {
  const QString ext = GetFileExtension(path);

  return ext == "aac" || ext == "ape" || ext == "cda" ||
      ext == "flac" || ext == "m4a" || ext == "mp3" ||
      ext == "ogg" || ext == "wma" || ext =="wav";
}

QString ReadTextFileContent(const QString& path) {
  QFile file(path);
  if (file.exists()) {
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
      qWarning() << "[file_util].ReadTextFileContent() failed to open"
                 << path;
      return "";
    }
    QTextStream text_stream(&file);
    QString str = text_stream.readAll();
    file.close();
    return str;
  } else {
    qWarning() << "[file_util].ReadFileContent() file not found: " << path;
    return "";
  }
}

QString ReadGBKTextFileContent(const QString& path) {
  QFile file(path);
  if (file.exists()) {
    if (!file.open(QIODevice::ReadOnly)) {
      qWarning() << "[file_util].ReadGBKTextFileContent() failed to open"
                 << path;
      return "";
    }
    const QByteArray file_data = file.readAll();
    QTextCodec* codec = QTextCodec::codecForName("GB18030");
    file.close();
    return codec->toUnicode(file_data);
  } else {
    qWarning() << "[file_util].ReadGBKTextFileContent() file not found:"
               << path;
    return "";
  }
}

bool ReadTextFile(const QString& path, QString& content) {
  QFile file(path);
  if (file.exists()) {
    file.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream text_stream(&file);
    content = text_stream.readAll();
    file.close();
    return true;
  } else {
    qWarning() << "[base]::ReadTextFile() failed!"
                << ", path: " << path;
    return false;
  }
}

// Rename file if it exists.
QString RenameDuplicatedFile(const QString& filepath) {
  if (!QFile::exists(filepath)) {
    // No need to rename file.
    return filepath;
  }

  const int dot_index = filepath.lastIndexOf(QChar('.'));
  QString extname;
  QString basename;
  if (dot_index > -1) {
    extname = filepath.mid(dot_index);
    basename = filepath.left(dot_index);
  } else {
    basename = filepath;
  }

  // basename + (#) + extname.
  QString renamed_filepath;
  const QString placeholder("(%1)");
  for (int i = 1; true; ++i) {
    renamed_filepath = basename + placeholder.arg(i) + extname;
    if (!QFile::exists(renamed_filepath)) {
      break;
    }
  }
  return renamed_filepath;
}

bool WriteTextFile(const QString& path, const QString& content) {
  QFile file(path);
  if (file.open(QIODevice::WriteOnly | QIODevice::Text)) {
    QTextStream text_stream(&file);
    text_stream << content;
    text_stream.flush();
    file.close();
    return true;
  }else {
    qCritical() << "[base]::WriteTestFile() failed!"
                << ", path:" << path;
    return false;
  }
}

}  // namespace netease
