// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "channel/orca_reply.h"

#include <QFile>
#include <QMap>
#include <QNetworkRequest>

#include "application.h"
#include "base/file_util.h"
#include "service/storage_manager.h"

namespace netease {

namespace {

// Host name of cache/?xxx.
// e.g. cache/?http://music.163.com/api/song/lyric?os=osx&id=37955128&lv=-1
// Lyrics/mp3/mv are cached in local storage.
const char kHostCache[] = "cache";

// Host name of orpheus/webapp resource.
// e.g. orpheus://orpheus/src/html/app.html.
const char kHostOrpheus[] = "orpheus";

// Host name of skin resource.
// e.g. orpheus://skin/default/mac/web/css/skin_light.css.
const char kHostSkin[] = "skin";

// Root name of storage request.
// e.g. orpheus://orpheus/storage/local?file=preview/font.png
const char kStoragePath[] = "/storage/";
const char kLocalStoragePath[] = "/storage/local";

// Prefix to local skin.
// Local skins are located at webapp/style/skin/ directory.
const char kSkinLocal[] = "/default";

// Prefix name of skin file in qrc format.
const char kSkinPrefixName[] = ":/webapp/style/skin%1";

// Prefix to remote skin.
//const char kSkinRemote[] = "/remote";

// Host name of local request.
//const char kHostLocal[] = "local";

const char kDefaultMimeType[] = "text/plain";

const QMap<QString, QString> mime_types = {
  {"html", "text/html"},
  {"svg", "image/svg+xml"},
  {"jpg", "image/jpeg"},
  {"jpeg", "image/jpeg"},
  {"gif", "image/gif"},
  {"png", "image/png"},
  {"css", "text/css"},
  {"less", "text/less"},
  {"js", "application/javascript"},
};

QString GetMimeType(const QString& filepath) {
  const QString extname = GetFileExtension(filepath);
  if (extname.length() > 0) {
    return mime_types.value(extname, kDefaultMimeType);
  } else {
    qWarning() << "GetMimeType() unknown type:" << filepath;
    return kDefaultMimeType;
  }
}

}  // namespace

OrcaReply::OrcaReply(const QNetworkRequest& request,
                     const QNetworkAccessManager::Operation operation,
                     QObject* parent)
    : QNetworkReply(parent),
      content_(),
      offset_(0) {

  this->setRequest(request);
  this->setUrl(request.url());
  this->setOperation(operation);

  QMetaObject::invokeMethod(this, "route", Qt::QueuedConnection);
}

void OrcaReply::abort() {
  // NOTE(xushaohua): Do nothing currently.
  QNetworkReply::close();
}

qint64 OrcaReply::bytesAvailable() const {
  return content_.length() + QIODevice::bytesAvailable() - offset_;
}

void OrcaReply::close() {
  QNetworkReply::close();
}

void OrcaReply::handle404() {
  qWarning() << "[OrcaReply]::handle404()" << this->url();

  emit error(NetworkError::ContentNotFoundError);
  // Abort connection forcefully, or else webview will block other requests.
//  emit finished();
  QMetaObject::invokeMethod(this, "finished", Qt::QueuedConnection);
}

void OrcaReply::handleAlbumCoverDownloaded(const QString& url,
                                           const QString& filepath) {
  if (url == this->url().query()) {
    sendFile(filepath);
  }
}

void OrcaReply::handleAlbumCoverError(const QString& url,
                                      const QString& filepath) {
  Q_UNUSED(filepath);
  if (url == this->url().query()) {
    handle404();
  }
}

void OrcaReply::handleCache() {
  QObject::connect(app->storage_manager,
                   &StorageManager::onAlbumCoverDownloaded,
                   this, &OrcaReply::handleAlbumCoverDownloaded);
  QObject::connect(app->storage_manager, &StorageManager::onAlbumCoverError,
                   this, &OrcaReply::handleAlbumCoverError);
  const QString filepath(app->storage_manager->getAlbumCover(url().query()));
  if (filepath.length() > 0) {
    this->sendFile(filepath);
  }
}

void OrcaReply::handleOrpheus() {
  this->sendFile(OrcaToQrc(url()));
}

void OrcaReply::handleSkin() {
  QString path = url().path();
  if (path.startsWith(kSkinLocal)) {
    // Remove skin path prefix.
    const QString relative_path(path.remove(0, QString(kSkinLocal).size()));
    const QString filepath(QString(kSkinPrefixName).arg(relative_path));
    this->sendFile(filepath);
  } else {
    qWarning() << "[OrcaReply]::handleSkin() remote skin not support!";
    handle404();
  }
}

void OrcaReply::handleStorage() {
  if (url().path().startsWith(kLocalStoragePath)) {
    // Removes "file=" query name.
    const QString filename(url().query().mid(5));
    const QString filepath(
        app->storage_manager->getLocalStorageFilePath(filename));
    this->sendFile(filepath);
  } else {
    qWarning() << "[OrcaReply]::handleStorage() current type not supported!"
               << this->url();
    this->handle404();
  }
}

bool OrcaReply::isSequential() const {
  return true;
}

// static
QString OrcaReply::OrcaToQrc(const QUrl& url) {
  return (QString(":/webapp%1").arg(url.path()));
}

qint64 OrcaReply::readData(char* data, qint64 max_size) {
  if (offset_ < content_.size()) {
    qint64 number = qMin(max_size, content_.size() - offset_);
    memcpy(data, content_.constData() + offset_, number);
    offset_ += number;
    return number;
  } else {
    // Returns -1 on failed.
    return -1;
  }
}

void OrcaReply::route() {
  const QString host = url().host();

  if (host == kHostOrpheus) {
    if (url().path().startsWith(kStoragePath)) {
      this->handleStorage();
    } else {
      this->handleOrpheus();
    }
  } else if (host == kHostSkin) {
    this->handleSkin();
  } else if (host == kHostCache) {
    this->handleCache();
  } else {
    qCritical() << "[OrcaReply]::router() not supported scheme: " << url();
    handle404();
  }
}

void OrcaReply::sendFile(const QString& filepath) {
  QFile file(filepath);
  if (file.exists()) {
    file.open(QIODevice::ReadOnly);
    content_ = file.readAll();

    this->open(QIODevice::ReadOnly);

    this->setHeader(QNetworkRequest::ContentTypeHeader,
                    GetMimeType(filepath));

    this->setHeader(QNetworkRequest::ContentLengthHeader,
                    QVariant(content_.size()));

    this->setRawHeader("Cache-Control", "max-age=2592000");

//    if (filepath.contains("svg")) {
//      qDebug() << "[OrcaReply]::sendFile()" << filepath
//               << content_.size();
//    }

    QMetaObject::invokeMethod(this, "readyRead", Qt::QueuedConnection);
    QMetaObject::invokeMethod(this, "finished", Qt::QueuedConnection);
    file.close();

  } else {
    qWarning() << "[OrcaReply]::snedFile() file not found: " << filepath;
    handle404();
  }
}

qint64 OrcaReply::size() const {
  return content_.size();
}

}  // namespace netease
