// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "dbus/mpris_adapter.h"

#include <QDebug>
#include <QtDBus/QDBusConnection>

#include "application.h"
#include "base/consts.h"
#include "service/media_player.h"
#include "service/signal_manager.h"
#include "ui/window_manager.h"

namespace netease {

// static
const char MprisAdapter::kServiceName[] =
    "org.mpris.MediaPlayer2.netease-cloud-music";
// static
const char MprisAdapter::kServicePath[] = "/org/mpris/MediaPlayer2";
// static
const char MprisAdapter::kRootInterface[] = "org.mpris.MediaPlayer2";
// static
const char MprisAdapter::kPlayerInterface[] =
    "org.mpris.MediaPlayer2.Player";

MprisAdapter::MprisAdapter(QObject* parent)
    : QObject(parent),
      mpris_player_(new ::MprisPlayer(this)) {
  this->setObjectName("mpris_adapter");

  this->initPlayerDBus();
}

MprisAdapter::~MprisAdapter() {

}

void MprisAdapter::initPlayerDBus() {
  // Service name is the same as kServiceName[].
  mpris_player_->setServiceName(kAppName);
  mpris_player_->setCanRaise(true);
  mpris_player_->setCanQuit(true);
  mpris_player_->setCanSetFullscreen(false);
  mpris_player_->setHasTrackList(false);
  mpris_player_->setDesktopEntry(kAppName);
  mpris_player_->setIdentity(kAppDisplayName);
  // Supported audio types:
  //  * flac
  //  * mp3
  //  * ogg
  //  * wav
  //  * wma
  mpris_player_->setSupportedMimeTypes({"audio/aac",
                                        "audio/m4a",
                                        "audio/mp3",
                                        "audio/wav",
                                        "audio/wma",
                                        "audio/x-ape",
                                        "audio/x-flac",
                                        "audio/x-ogg",
                                        "audio/x-oggflac",
                                        "audio/x-vorbis"});
  mpris_player_->setSupportedUriSchemes({"http", "file", "smb"});

  mpris_player_->setCanControl(true);
  mpris_player_->setCanGoPrevious(true);
  mpris_player_->setCanGoNext(true);
  mpris_player_->setCanPause(true);
  mpris_player_->setCanPlay(true);
  mpris_player_->setCanSeek(false);
  mpris_player_->setShuffle(false);
  mpris_player_->setLoopStatus(Mpris::LoopStatus::Playlist);

  connect(app->media_player, &MediaPlayer::mediaChanged,
          this, &MprisAdapter::handlePlayerMediaChanged);
  connect(app->media_player, &MediaPlayer::stateChanged,
          this, &MprisAdapter::handlePlayerStateChanged);
  connect(app->media_player, &MediaPlayer::positionChanged,
          this, &MprisAdapter::handlePositionChanged);
  connect(app->media_player, &MediaPlayer::volumeChangedWithAnimation,
          this, &MprisAdapter::handleVolumeChanged);
  connect(app->media_player, &MediaPlayer::trackInfoChanged,
          this, &MprisAdapter::handlePlayerMediaChanged);
  connect(app->signal_manager, &SignalManager::onAlbumCoverUpdated,
          this, &MprisAdapter::handleAlbumCoverUpdated);

  connect(mpris_player_, &MprisPlayer::quitRequested,
          app->signal_manager, &SignalManager::onRequestMainWindowClose);
  connect(mpris_player_, &MprisPlayer::raiseRequested,
          app->signal_manager, &SignalManager::onRequestChangeToMain);
  connect(mpris_player_, &MprisPlayer::previousRequested,
          app->signal_manager, &SignalManager::onPlayPrev);
  connect(mpris_player_, &MprisPlayer::nextRequested,
          app->signal_manager, &SignalManager::onPlayNext);
  connect(mpris_player_, &MprisPlayer::playRequested,
          app->signal_manager, &SignalManager::onRequestResume);
  connect(mpris_player_, &MprisPlayer::pauseRequested,
          app->signal_manager, &SignalManager::onRequestPause);
  connect(mpris_player_, &MprisPlayer::stopRequested,
          app->signal_manager, &SignalManager::onRequestPause);
  connect(mpris_player_, &MprisPlayer::seekRequested,
          this, &MprisAdapter::onSeekRequested);
  connect(mpris_player_, &MprisPlayer::volumeRequested,
          this, &MprisAdapter::onVolumeChangedRequested);

  connect(mpris_player_, &MprisPlayer::openUriRequested,
          this, &MprisAdapter::onOpenUriRequested);
}

void MprisAdapter::handleAlbumCoverUpdated(const QString& filepath) {
  album_url_ = QUrl::fromLocalFile(filepath);
  this->updateMetadata();
}

void MprisAdapter::handlePlayerReset() {
//  album_url_.clear();
}

void MprisAdapter::handlePlayerMediaChanged() {
  // Reset metadata.
  this->updateMetadata();
}

void MprisAdapter::handlePlayerStateChanged() {
  switch (app->media_player->state()) {
    case MediaPlayer::PlayingState: {
      mpris_player_->setPlaybackStatus(Mpris::Playing);
      break;
    }
    case MediaPlayer::PausedState: {
      mpris_player_->setPlaybackStatus(Mpris::Paused);
      break;
    }
    case MediaPlayer::StoppedState: {
      mpris_player_->setPlaybackStatus(Mpris::Stopped);
      break;
    }
    default: {
      break;
    }
  }
}

void MprisAdapter::handlePositionChanged(qlonglong position) {
  // TODO(LiuLang): Emit this signal one and all.
  mpris_player_->setPosition(position * 1000);
}

void MprisAdapter::handleVolumeChanged(int volume) {
  mpris_player_->setVolume(volume * 1.0 / 100.0);
}

void MprisAdapter::updateMetadata() {
  auto track_info = app->media_player->track_info();
  Q_ASSERT(track_info != nullptr);
  QVariantMap metadata;
  if (!track_info->album_name.isEmpty()) {
    metadata.insert(Mpris::metadataToString(Mpris::Album),
                    track_info->album_name);
  }
  if (!track_info->artist_name.isEmpty()) {
    metadata.insert(Mpris::metadataToString(Mpris::Artist),
                    track_info->artist_name);
  }
  if (!track_info->song_name.isEmpty()) {
    metadata.insert(Mpris::metadataToString(Mpris::Title),
                    track_info->song_name);
  }
  if (track_info->duration > 0) {
    metadata.insert(Mpris::metadataToString(Mpris::Length),
                    track_info->duration * 1000);
  }
  if (!album_url_.isEmpty()) {
    metadata.insert(Mpris::metadataToString(Mpris::ArtUrl), album_url_);
  }

  metadata.insert(Mpris::metadataToString(Mpris::TrackNumber), 1);

  mpris_player_->setMetadata(metadata);
}

void MprisAdapter::onSeekRequested(qint64 offset) {
  const qint64 position = app->media_player->position();
  app->media_player->setPosition(position + offset);
}

void MprisAdapter::onVolumeChangedRequested(double volume) {
  const int int_volume = static_cast<int>(volume * 100);
  app->media_player->setVolume(int_volume);
}

void MprisAdapter::onOpenUriRequested(const QUrl& url) {
  emit app->signal_manager->onPlayLocalMusic(url.toLocalFile());
}

}  // namespace netease