// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by General Public License that can be found
// in the LICENSE file.

#include <signal.h>
#include <QDebug>
#include <QFile>
#include <QTimer>
#include <qcef_context.h>

#include "application.h"
#include "base/consts.h"
#include "base/file_util.h"
#include "service/command_line.h"
#include "service/custom_scheme_handler.h"
#include "service/dbus_manager.h"
#include "service/process_manager.h"
#include "service/settings_manager.h"
#include "third_party/ne_enigma/ne_enigma.h"
#include "ui/window_manager.h"

namespace {

const char kChannelJsFile[] = ":/channel/channel_preload.js";
const char kChannelJsUrl[] = "qrc://channel/channel_preload.js";

const char kDisableGpu[] = "--disable-gpu";
const char kEnableDomStorageFlush[] = "--enable-aggressive-domstorage-flushing";

bool IsGPUEnabled() {
  return netease::SettingsGetValue("setting", "hardware-acceleration").toBool();
}

// TODO(LiuLang): Fix segfault in Application::quitApplication().
void __attribute__((unused)) HandleSegFault(int sig) {
  Q_UNUSED(sig);

  QCefShutdown();
  netease::DoRestart();
  exit(0);
}

QString __attribute__((unused)) GetQcefOverridePath() {
  const QString exe("/proc/self/exe");
  QDir dir(QFileInfo(exe).canonicalPath());
  dir.cdUp();
  return dir.filePath("lib/netease-cloud-music/lib/qcef");
}

QString __attribute__((unused)) GetQtPluginPath() {
  const QString exe("/proc/self/exe");
  QDir dir(QFileInfo(exe).canonicalPath());
  dir.cdUp();
  return dir.filePath("lib/netease-cloud-music/plugins");
}

void __attribute__((unused)) LoadXcbPlugin() {
#ifdef EMBED_QCEF
  const QString plugin = GetQtPluginPath();
  qputenv("QT_QPA_PLATFORM_PLUGIN_PATH", plugin.toLocal8Bit());
#endif
}

}  // namespace

netease::Application* app;

#ifndef N_DEBUG
std::ostream& operator<<(std::ostream& os, wchar_t const* c) {
  os << c;
  return os;
}
#endif

int main(int argc, char* argv[]) {
  // Fix xcb plugin missing issue.
  LoadXcbPlugin();

  QCefGlobalSettings settings;

#ifdef EMBED_QCEF
  const QString override_path = GetQcefOverridePath();
  settings.setOverridePath(override_path);
#endif

  settings.setCustomSchemeHandler(netease::OrpheusSchemeHandler);
  QVector<QUrl> src_urls = {
      QUrl("orpheus://cache/"),
      QUrl("orpheus://image/"),
      QUrl("orpheus://localmusic/"),
      QUrl("orpheus://native/"),
      QUrl("orpheus://orpheus/"),
      QUrl("orpheus://skin/"),
      QUrl("orpheus://storage"),
      QUrl("orpheus://upload/"),
  };
  for (const QUrl& src : src_urls) {
    settings.addCustomScheme(src);
  }
  settings.registerSyncMethod("encodeAnonymousId",
                              ne_enigma::EncodeAnonymousId);
  settings.registerSyncMethod("encryptId",
                              ne_enigma::EncryptId);
  settings.registerSyncMethod("serialData",
                              ne_enigma::SerialData);

  settings.setPersistSessionCcookies(true);
  QDir cef_dir = netease::ConcateDir(netease::GetUserCacheDir(),
                                     netease::kCefDirName);
  settings.setCachePath(cef_dir.filePath("Cache"));
  settings.setUserDataPath(cef_dir.filePath("UserData"));
  settings.setNoSandbox(true);
#ifndef NDEBUG
  settings.setRemoteDebug(true);
  settings.setRemoteDebugPort(9223);
  settings.setLogSeverity(QCefGlobalSettings::LogSeverity::Info);
#else
  settings.setRemoteDebug(false);
  settings.setLogSeverity(QCefGlobalSettings::LogSeverity::Error);
#endif

  const QString console_log = cef_dir.filePath("console.log");
  QFile::remove(console_log);
  settings.setLogFile(console_log);

  // Register channel-preload.js
  settings.registerUserScript(kChannelJsFile, QUrl(kChannelJsUrl));

  // Set hardware accelerator state.
  if (!IsGPUEnabled()) {
    settings.addCommandLineSwitch(kDisableGpu, "");
  }

  // Enable aggressive storage commit to minimize data loss.
  // See public/common/content_switches.cc.
  settings.addCommandLineSwitch(kEnableDomStorageFlush, "");

  int status = QCefInit(argc, argv, settings);
  if (status >= 0) {
    return 0;
  }

  // FIXME(zccrs): Disable the X session manager
  //               Fix the application blocked at IceProcessMessages on QSmSocketReceiver::socketActivated
  const QByteArray session_manager = qgetenv("SESSION_MANAGER");
  qunsetenv("SESSION_MANAGER");

  app = new netease::Application(argc, argv);

  if (!session_manager.isEmpty()) {
    qputenv("SESSION_MANAGER", session_manager);
  }

  if (netease::IsApplicationRunning()) {
    const QString local_file = netease::GetMediaFileFromArgs();
    qDebug() << "Local file:" << local_file << app->arguments();
    if (!local_file.isEmpty()) {
      netease::NotifyPlayingLocalFile(local_file);
    }
    // Exit current process after 500ms.
    QTimer::singleShot(500, app, &QApplication::quit);
    return app->exec();
  } else {
    QCefBindApp(app);
    app->initChildren();
    app->window_manager->initMainWindow();
  }

#ifndef N_DEBUG
  // Catch exceptions in release mode.
  signal(SIGSEGV, HandleSegFault);

  // Ignores SIGTERM.
  signal(SIGTERM, SIG_IGN);
#endif

  status = app->exec();
  netease::DoRestart();
  return status;
}
