// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "native/app.h"

#include <QColor>
#include <QDebug>

#include "application.h"
#include "base/consts.h"
#include "base/device_util.h"
#include "base/process_util.h"
#include "service/auto_start.h"
#include "service/default_client.h"
#include "service/dbus_manager.h"
#include "service/feedback_manager.h"
#include "service/log_manager.h"
#include "service/url_manager.h"
#include "service/process_manager.h"
#include "service/settings_manager.h"
#include "service/signal_manager.h"
#include "service/skin_manager.h"
#include "service/storage_manager.h"
#include "ui/window_manager.h"

namespace netease {

namespace {

const char kExitModeMoveRun[] = "moverun";
const char kExitModeInstall[] = "install";
const char kExitModeQuitRun[] = "quitrun";
const char kExitModeRestart[] = "restart";

// Start type, automatically startup with desktop session.
const char kStartTypeAutorun[] = "autorun";

// If key in "key=value" pair in ini file is empty, replace it with "default".
// The second parameter in getLocalConfig() and setLocalConfig() might be
// replaced by "default" if it is empty because key cannot be empty in
// ini file.
const char kDefaultSettingsKey[] = "default-key";

// Name of cooroperation.
const char kMainCooperationName[] = "linux";

}  // namespace

App::App(QObject* parent) : QObject(parent) {
  // If this object name is setup, this object is visible in webview as the
  this->setObjectName("app");

  this->initConnections();
}

void App::cancelAutoRun(const QString& name) const {
  Q_UNUSED(name);
  SetAutoStart(false);
}

void App::chooseColor(const QString& element_id, const QString& color) {
  qDebug() << Q_FUNC_INFO << element_id << color;
  emit app->window_manager->chooseColor(element_id, color);
}

void App::daEvent(const QVariantList& events) const {
  qDebug() << "TODO [App]::daEvent()" << events;
}

void App::discernMusic() const {
  qDebug() << "TODO [App]::discernMusic()";
}

void App::exit() {
  qDebug() << Q_FUNC_INFO;
  app->process_manager->exit();
}

void App::exitWithMode(const QString& mode) {
  qDebug() << Q_FUNC_INFO << mode;
  QStringList arguments;
  if (mode == kExitModeInstall) {
    arguments.append(kArgsInstall);
  } else if (mode == kExitModeRestart) {
    arguments.append(kArgsRestart);
  } else if (mode == kExitModeQuitRun) {
    arguments.append(kArgsQuitRun);
  }
  app->process_manager->restart(arguments);
}

void App::exitWithMove(const QString& mode, const QString& move_from,
                       const QString& move_to, const QString& require_value) {
  qDebug() << Q_FUNC_INFO << mode << move_from << move_to << require_value;
  if (mode != kExitModeMoveRun) {
    qWarning() << "[App]::exit() unsupported exit mode: " << mode;
    return;
  }

  QStringList arguments;
  arguments.append(kArgsMoveRun);

  QString require(kArgsRequire);
  require.append(require_value);
  arguments.append(require);

  QString move_run_src(kArgsMoveRunSrc);
  move_run_src.append(move_from);
  arguments.append(move_run_src);

  QString move_run_dst(kArgsMoveRunDst);
  move_run_dst.append(move_to);
  arguments.append(move_run_dst);

  app->process_manager->restart(arguments);
}

void App::forceExit() {
  KillAllProcesses();
}

void App::getAppStartCommand() const {
  qDebug() << "TODO [App]::getAppStartCommand()";
}

QString App::getAppStartType() const {
  qDebug() << "TODO [App].getAppStartType()";

  // TODO(xushaohua): Check process startup type.
  return kStartTypeAutorun;
}

bool App::getAutoRunState(const QString& name) const {
  Q_UNUSED(name);

  return GetAutoStart();
}

QVariantMap App::getCooperation() const {
  QVariantMap result;
  result.insert("main", QString(kMainCooperationName));
  result.insert("sub", GetDistroName());
  return result;
}

QString App::getDefaultMusicPlayPath() const {
  qDebug() << "TODO [App]::getDefaultMusicPlayPath()";

  return "";
}

QVariant App::getLocalConfig(const QString& group_name, const QString& key) {
  QString new_key(key);
  if (new_key.length() == 0) {
    new_key = kDefaultSettingsKey;
  }

  QVariant result = SettingsGetValue(group_name, new_key);
  return result;
}

void App::initCDNs(const QString& config) {
  // TODO(LiuLang):
  qDebug() << Q_FUNC_INFO << config;
}

void App::initConfig(const QVariantMap& options) const {
  qDebug() << "TODO [App]::initConfig()" << options;
}

void App::initUrls(const QVariantMap& url_info) const {
  emit app->url_manager->initUrls(url_info);
}

void App::initConnections() {
  connect(app->signal_manager, &SignalManager::onPlayLocalMusic,
          this, &App::onplaylocalmusic);
  connect(app->feedback_manager, &FeedbackManager::onSendStatFile,
          this, &App::onsendstatis);

  connect(app->window_manager, &WindowManager::onChooseColor,
          this, &App::onchooseColorResult);
  connect(app->window_manager, &WindowManager::onSaveFile,
          this, &App::onSaveFileDialog);
  connect(app->window_manager, &WindowManager::onSelectFiles,
          this, &App::onSelectFileLimitCount);
  connect(app->window_manager, &WindowManager::onSelectFolder,
          this, &App::onselectsystemfile);
}

bool App::isRegisterDefaultClient() const {
  return IsDefaultClient();
}

bool App::loadSkinPackets(const QString& skin_type,
                          const QString& skin_name,
                          const QVariantMap& options) const {
  // TODO(xushaohua): Move to SkinManager.
  const QVariantMap btn_color = options.value("btn_color").toMap();

  if (btn_color.isEmpty()) {
    QColor invalid_color;
    return app->skin_manager->changeSkin(skin_type, skin_name, invalid_color);
  } else {
    const double h = btn_color.value("h").toInt() / 360.0;
    const double s = btn_color.value("s").toInt() / 100.0;
    const double l = btn_color.value("l").toInt() / 100.0;
    QColor button_color(QColor::fromHslF(h, s, l));
    return app->skin_manager->changeSkin(skin_type, skin_name, button_color);
  }
}

void App::log(const QString& msg) const {
#ifndef N_DEBUG
  qDebug() << Q_FUNC_INFO << msg;
#endif
  emit app->log_manager->log(msg);
}

void App::login(const QVariantMap& options) const {
  qDebug() << "TODO [App]::login()" << options;
}

void App::openSaveFileDialog(const QString& keyword, const QString& title,
                             const QString& filter, const QString& filename) {
  emit app->window_manager->saveFile(keyword, title, filter, filename);
}

void App::querySkinPackets() {
  QVariantList local_skins;

  for (const SkinManager::Skin& skin : app->skin_manager->getLocalSkins()) {
    QVariantMap local_skin;
    local_skin.insert("name", skin.name);
    local_skin.insert("info", skin.info);
    local_skins.append(local_skin);
  }

  QVariantMap result;
  result.insert("default", local_skins);
  emit onqueryskinpackets(result);
}

bool App::registerDefaultClient() const {
  return SetDefaultClient(true);
}

void App::selectSystemDir(const QString& keyword, const QString& title,
                          const QString& message, const QString& path) {
  emit app->window_manager->selectFolder(keyword, title, message, path);
}

void App::selectSystemFileLimitCount(const QString& arg1, const QString& arg2,
                                     const QString& arg3, const QString& filter,
                                     int limit) {
  emit app->window_manager->selectFiles(arg1, arg2, arg3, filter, limit);
}

void App::sendFeedback(const QVariantMap& info) {
  emit app->feedback_manager->sendFeedback(info);
}

void App::sendStatis(const QString& msg) {
  emit app->feedback_manager->sendStatFile(msg);
}

void App::setAutoRun(const QString& name, const QString& param) const {
  Q_UNUSED(name);

  if (SetAutoStart(true)) {
    SetAutoStartOption(param);
  } else {
    qCritical() << "[App]::setAutoRun() failed to setup autostart with param:"
                << param;
  }
}

void App::setLocalConfig(const QString& group_name, const QString& key,
                         const QVariant& value) const {
  QString new_key(key);
  if (new_key.length() == 0) {
    new_key = kDefaultSettingsKey;
  }

  SettingsSetValue(group_name, new_key, value);
}

void App::setThumbnail(const QVariantMap& conf) {
  qDebug() << "TODO:" << Q_FUNC_INFO << conf;
  Q_UNUSED(conf);
}

void App::setWatchNetStatus(bool watch) const {
  if (watch) {
    connect(app->dbus_manager, &DBusManager::connectivityChanged,
            this, &App::onnetstatus);
  } else {
    disconnect(app->dbus_manager, &DBusManager::connectivityChanged,
               this, &App::onnetstatus);
  }
}

void App::showProxySetting() const {
  qDebug() << "TODO: [App]::showProxySetting()";
}

void App::statis(const QString& msg) const {
  emit app->log_manager->statMsg(msg);
}

bool App::unRegisterDefaultClient() const {
  return SetDefaultClient(false);
}

}  // namespace netease
