// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_APP_H_
#define NATIVE_APP_H_

#include <QObject>
#include <QVariant>

namespace netease {

// App class implements channel.app interface.
class App : public QObject {
  Q_OBJECT

 public:
  explicit App(QObject* parent);

 signals:
  // Emitted after calling chooseColor() slot.
  // |element_id| is the same as in chooseColor().
  // |red|, |green| and |blue| are color values.
  void onchooseColorResult(const QString& element_id, int red,
                           int green, int blue);

  // Notify web window to quit program forcefully.
  void onexitmessage();

  // Raise web window.
  void onforegroundpreinstance();

  // TODO(xushaohua): Add comments.
  void onmusicdiscern();

  // TODO(xushaohua): Add comments.
  void onmusicdiscern3();

  // Emitted when network connection state changed.
  void onnetstatus(bool is_online);

  // Notify web window to play local music at |media_path|.
  void onplaylocalmusic(const QString& media_path);

  // Emitted after calling querySkinPackets.
  // |skins| is an object containing:
  //   * default: a list of local skin packets.
  void onqueryskinpackets(const QVariantMap& skins);

  // Emitted after calling openSaveFileDialog()
  void onSaveFileDialog(const QString& keyword, int err, const QString& path);

  // Emitted after calling selectSystemFileAndDir().
  // Used in cloud upload page.
  void onSelectFileAndDir();

  // Emitted after calling selectSystemDir() slot.
  // |ok| is used to mark folder chooser operation is discarded or not.
  // |keyword| i s the same as in selectSystemDir().
  // |new_path| is the new absolute path if succeeded, or empty if failed.
  void onselectsystemfile(bool ok, const QString& keyword,
                          const QString& new_path);

  // Emitted after calling selectSystemFileLimitCount().
  void onSelectFileLimitCount(const QString& keyword, int ok,
                              const QVariantList& files);

  // This signal is emitted after calling sendStatis() method.
  void onsendstatis();

  // Taskbar is not implemented on Linux platform.
  void ontaskbarrestart();

  // TODO(xushaohua): Add comments.
  void ontestproxy();

  // Display web message.
  // Not implemented yet.
  void onwebmessage(const QString& msg);

 public slots:
  // Disable startup with desktop session.
  // |name| is ignored on Linux platform.
  void cancelAutoRun(const QString& name) const;

  // Popup a color chooser dialog and emit onchooseColorResult() signal.
  // This slot is used in settings page when changing font color of desktop
  // lyrics.
  // |element_id| is used to mark DOM element.
  // |color| is current color, like "020040".
  void chooseColor(const QString& element_id, const QString& color);

  // TODO(xushaohua): Call SDK api.
  void daEvent(const QVariantList& events) const;

  // TODO(xushaohua): Add comments.
  void discernMusic() const;

  // Exit whole process.
  // |mode| might be:
  //   * empty, exit process.
  //   * "quitrun", restart without installing new version.
  //   * "install", restart with new version.
  //   * "restart", restart the whole process.
  //   * "moverun", when changing download folder to new one. e.g.
  //     ["moverun", "/home/xushaohua-vnc/Music/ClousMusic",
  //       "/home/xushaohua-vnc/Desktop/hello", "1"]
  // Note that when |mode| is "quitrun", "moverun" or "install",
  // pass that parameter to // --orpheus-startup parameter list,
  // because webview needs to know how to handle restart operations.
  void exit();
  void exitWithMode(const QString& mode);
//  void exit(const QString& mode, const QString& option);
  void exitWithMove(const QString& mode, const QString& move_from,
                    const QString& move_to, const QString& require_value);

  // Kill current process forcefully.
  void forceExit();

  // TODO(xushaohua): Add comments.
  void getAppStartCommand() const;

  // Process startup type might be "autorun" or empty.
  // TODO(xushaohua): List available start types.
  QString getAppStartType() const;

  // Check autostart is enabled or not.
  // |name| is ignored on Linux platform.
  bool getAutoRunState(const QString& name) const;

  // Get current distribution vendor name.
  // Returns an object containing:
  //   * main: "linux".
  //   * sub: linux-distro-name.
  QVariantMap getCooperation() const;

  // If local music was played last time when program exited, returns its
  // absolute path.
  QString getDefaultMusicPlayPath() const;

  // Read local configuration.
  // |group_name| is group name of QSettings.
  // |key| is key of settings.
  QVariant getLocalConfig(const QString& group_name, const QString& key);

  // TODO(LiuLang): Support CND.
  void initCDNs(const QString& config);

  // DEPRECATED.
  // NOTE: use setLocalConfig() instead.
  void initConfig(const QVariantMap& options) const;

  // Store alternative resource host names.
  // Those hostnames and path are used to construct resource url.
  // |url_info| is an object like:
  //   * discern: "http://mno.igame.163.com/api/music/matcher"
  //   * discern_uri: "/api/music/matcher"
  //   * e_batch_url: "http://mno.igame.163.com/eapi/batch"
  //   * e_url: "http://mno.igame.163.com/eapi"
  //   * fixdiscern: "http://mno.igame.163.com/api/music/fixmatcher"
  //   * fixdiscern_uri: "/api/music/multimatcher"
  //   * lyric: "http://mno.igame.163.com/api/song/media?"
  //   * music: ["http://m1.music.126.net/", "http://m2.music.126.net/"]
  //   * refer: "http://mno.igame.163.com/di"
  //   * statis: "http://mno.igame.163.com/api/feedback/client/log"
  void initUrls(const QVariantMap& url_info) const;

  // Check whether this application is the default audio player on current
  // desktop environment.
  bool isRegisterDefaultClient() const;

  // Change skin. Note that this all changes skin of mini player.
  // |skin_type| is "default", meaning local skin.
  // |options| is used to set button color in mini player, might be empty
  // or contains "btn_color" item.
  // Returns true if that skin packets exists.
  bool loadSkinPackets(const QString& skin_type,
                       const QString& skin_name,
                       const QVariantMap& options) const;

  // Log |message| to system.
  void log(const QString& msg) const;

  // Does nothing currently.
  void login(const QVariantMap& options) const;

  void openSaveFileDialog(const QString& keyword, const QString& title,
                          const QString& filter, const QString& filename);

  // Query available skin packets, emit onqueryskinpackets() signal.
  void querySkinPackets();

  // Set this application as default audio player on current desktop
  // environment.
  // Returns true on success.
  bool registerDefaultClient() const;

  // Popup a native folder chooser dialog.
  // |keyword| is useless currently.
  // |title| is the title of dialog.
  // |message| is display at the top of dialog.
  // |path| is the default folder of dialog, might be empty.
  void selectSystemDir(const QString& keyword, const QString& title,
                       const QString& message, const QString& path);

  void selectSystemFileLimitCount(const QString& arg1, const QString& arg2,
                                  const QString& arg3, const QString& filter,
                                  int limit);

  // Send feedback to server.
  // |info| is an object containing:
  //   * contact
  //   * content, message content.
  //   * cookie, browser cookies.
  //   * resolution, screen resolution.
  //   * sendPrivateMsg, default is true.
  //   * title
  //   * url, send this message to.
  void sendFeedback(const QVariantMap& info);

  // Send application usage statistics to server.
  void sendStatis(const QString& cookie);

  // Setup autostart options.
  // |name| is ignored on Linux platform.
  // |param| is passed to process startup params.
  // e.g. if param = "foo", then "--orca-startup=foo" shall be appended to
  // startup params.
  // TODO(xushaohua): Encode |param|.
  void setAutoRun(const QString& name, const QString& param) const;

  // Update local settings.
  // |group_name| and |key| is the same as in getLocalConfig().
  // If |key| is empty, set it to default key (kDefaultSettingsKey).
  // |value|, by default, is string only, but now it is declared as
  // QVariant.
  void setLocalConfig(const QString& group_name, const QString& key,
                      const QVariant& value) const;

  // Setup thumbnail information displayed at taskbar.
  // When buttons in thumbnail is clicked, emit Player::onthumbnailaction().
  // Currently this feature only works on Windows system.
  // |conf| is an object containing:
  //   * tooltip: tooltip text.
  //   * defaultCover: default album cover.
  //   * btnMiddle: an object, like { id: 301, tooltip: "Play", url: xxx }.
  //   * albumCoverUrl: current album cover.
  void setThumbnail(const QVariantMap& conf);

  // Watch network connection state change.
  // Will emit onnetstatus() signal.
  // TODO(xushaohua): Connect to network-manager dbus signal.
  void setWatchNetStatus(bool watch) const;

  // Popup system proxy setup dialog.
  // TODO(xushaohua): Call system dbus interface.
  void showProxySetting() const;

  // Save web view statistics.
  void statis(const QString& msg) const;

  // Unset this application as default audio player on current desktop
  // environment.
  bool unRegisterDefaultClient() const;

 private:
  void initConnections();
};

}  // namespace netease

#endif  // NATIVE_APP_H_
