// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "native/audio_player.h"

#include <QDebug>

#include "application.h"
#include "service/signal_manager.h"

namespace netease {

namespace {

// Media player is currently playing content.
const int kPlayStatePlay = 1;

// Media player has paused playback.
const int kPlayStatePause = 2;

// State used in onVolume() signal.
// Change volume of player successfully.
const int kVolumeStateSuccess = 0;

// Has enough buffer to play.
const int kBufferingStatePlayable = 0;

// Has not enough media buffer to play and player shall be paused.
const int kBufferingStateUnPlayable = 1;

// State used in onSeek() signal.
// Seek to new position successfully.
const int kSeekStateSuccess = 0;

// Metadata of media resource is loaded successfully.
const int kLoadStateSuccess = 0;

// Failed to load metadata info of media resource.
// This may occur when network connection is available or that media
// resource is an accessible.
const int kLoadStateFail = 1;

const int kLoadStateNetworkError = 2;

// State used in onEnd() signal.
// Reaches end of media source.
const int kPlayingEndStateNormal = 0;

}  // namespace

AudioPlayer::AudioPlayer(QObject* parent)
    : QObject(parent),
      is_seeking_(false),
      on_load_emitted_(false),
      play_id_(),
      resume_or_pause_id_(),
      seek_id_(),
      volume_id_() {
  this->setObjectName("audioplayer");

  initConnections();
}

double AudioPlayer::duration() const {
  return app->media_player->duration() / 1000.0;
}

void AudioPlayer::enmeratorDevices(const QString& command) {
  qDebug() << "TODO [AudioPlayer]::enumeratorDevices()" << command;
  // TODO(xushaohua): Emit onEnuratorDevices() signal.
}

double AudioPlayer::getCurrentTime() const {
  return app->media_player->position() / 1000.0;
}

double AudioPlayer::getDownloadSchedule() const {
  // TODO(xushaohua): Check buffer status.
  return app->media_player->bufferStatus() / 100.0;
}

double AudioPlayer::getVolume() const {
  return app->media_player->volume() / 100;
}

void AudioPlayer::handleLoadError(int error_code) {
  QVariantMap result;
  result.insert("code", error_code);
  emit onLoad(play_id_, result);
}

void AudioPlayer::handleMediaStatusChanged(MediaPlayer::MediaStatus status) {
  qDebug() << Q_FUNC_INFO << status;
  QVariantMap result;
  switch (status) {
    case MediaPlayer::MediaStatus::BufferingMedia:  // fall through
    case MediaPlayer::MediaStatus::BufferedMedia:  // fall through
    case MediaPlayer::MediaStatus::LoadedMedia: {
      if (on_load_emitted_) {
        return;
      }

      const qint64 duration = app->media_player->duration();
      if (duration < 1) {
        return;
      }
      on_load_emitted_ = true;

      // Metadata, including duration loaded.
      result.insert("code", kLoadStateSuccess);
      result.insert("duration", duration / 1000.0);
      result.insert("errorCode", 1);
      result.insert("errorString", "");
      result.insert("activeCode", 1);
      qDebug() << "result:" << result;
      emit onLoad(play_id_, result);
      break;
    }

    case MediaPlayer::MediaStatus::EndOfMedia: {
      // Reaches end of current media.
      result.insert("code", QVariant(kPlayingEndStateNormal));
      emit onEnd(play_id_, result);
      break;
    }

    case MediaPlayer::MediaStatus::InvalidMedia: {
      qCritical() << Q_FUNC_INFO << "Invalid Media:" << play_id_;
      // Notify network error.
      result.insert("code", kLoadStateNetworkError);
      emit onLoad(play_id_, result);
      break;
    }

    default: {
      // Ignores any other status.
      break;
    }
  }
}

void AudioPlayer::handlePositionChanged(qint64 position) {
  double position_sec = position / 1000.0;
  double buffering_percent = app->media_player->bufferStatus() / 100.0;

  // If seek() slot is called, emit onSeek() signal.
  if (is_seeking_) {
    is_seeking_ = false;
    emit onSeek(play_id_, seek_id_, kSeekStateSuccess, position_sec);
  }
  emit onPlayProgress(play_id_, position_sec, buffering_percent);
}

void AudioPlayer::handleSeekableChanged(bool seekable) {
  if (seekable) {
    emit onBuffering(play_id_, kBufferingStatePlayable);
  } else {
    emit onBuffering(play_id_, kBufferingStateUnPlayable);
  }
}

void AudioPlayer::handleStateChanged(MediaPlayer::State state) {
  switch (state) {
  case MediaPlayer::State::PausedState:
    emit onPlayState(play_id_, resume_or_pause_id_, kPlayStatePause);
    break;

  case MediaPlayer::State::PlayingState:
    emit onPlayState(play_id_, resume_or_pause_id_, kPlayStatePlay);
    break;

  case MediaPlayer::State::StoppedState:
    emit app->signal_manager->onUpdateSystemTrayContextMenu();
    break;

  default:
    // Ignore other states.
    break;
  }
}

void AudioPlayer::handleVolumeChanged() {
  // Set play_id to empty.
  double volume = app->media_player->volume() / 100.0;
  emit onVolume("", volume_id_, kVolumeStateSuccess , volume);
}

void AudioPlayer::init(const QString& command,
                       const QVariantMap& device_info) const {
  qDebug() << "TODO [AudioPlayer]::init()"
           << ", command: " << command
           << ", device_info: " << device_info;
}

void AudioPlayer::initConnections() const {
  connect(app->media_player, &MediaPlayer::onLoadError,
          this, &AudioPlayer::handleLoadError);
  connect(app->media_player, &MediaPlayer::mediaStatusChanged,
          this, &AudioPlayer::handleMediaStatusChanged);
  connect(app->media_player, &MediaPlayer::mutedChanged,
          this, &AudioPlayer::handleVolumeChanged);
  connect(app->media_player, &MediaPlayer::positionChanged,
          this, &AudioPlayer::handlePositionChanged);
  connect(app->media_player, &MediaPlayer::volumeChangedWithAnimation,
          this, &AudioPlayer::handleVolumeChanged);
  connect(app->media_player, &MediaPlayer::seekableChanged,
          this, &AudioPlayer::handleSeekableChanged);
  connect(app->media_player, &MediaPlayer::stateChanged,
          this, &AudioPlayer::handleStateChanged);
}

bool AudioPlayer::isMuted() const {
  return app->media_player->isMuted();
}

void AudioPlayer::load(const QString& play_id, const QVariantMap& track_info) {
  qDebug() << "[AudioPlayer]::load()" << track_info;
  on_load_emitted_ = false;
  play_id_ = play_id;
  app->media_player->load(track_info);
}

void AudioPlayer::mute(bool mute) const {
  app->media_player->setMuted(mute);
}

void AudioPlayer::pause(const QString& play_id,
                        const QString& resume_or_pause_id) {
  play_id_ = play_id;
  resume_or_pause_id_ = resume_or_pause_id;
  app->media_player->pause();
  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void AudioPlayer::play(const QString& play_id,
                       const QString& resume_or_pause_id) {
  play_id_ = play_id;
  resume_or_pause_id_ = resume_or_pause_id;
  app->media_player->play();
  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void AudioPlayer::seek(const QString& play_id, const QString& seek_id,
                     double new_position) {
  qDebug() << Q_FUNC_INFO << play_id << seek_id << new_position;
  is_seeking_ = true;
  play_id_ = play_id;
  seek_id_ = seek_id;
  const qint64 position_ms = static_cast<qint64>(new_position * 1000);
  app->media_player->setPosition(position_ms);
}

void AudioPlayer::setPlayMode(const QString& mode) const {
  qDebug() << "TODO [AudioPlayer]::setPlayMode(): " << mode;
}

void AudioPlayer::setVolume(const QString& play_id, const QString& volume_id,
                            double volume) {
  Q_UNUSED(play_id);
  volume_id_ = volume_id;
  app->media_player->setVolume(int(volume * 100));
}

void AudioPlayer::stop(const QString& play_id) {
  play_id_ = play_id;
  app->media_player->stop();
}

void AudioPlayer::switchEffect(const QString& id, const QVariantList& options) {
  if (id == "id") {
    if (options.length() > 0) {
      // Read fade in effect switch.
      const QMap<QString, QVariant> option_map = options.at(0).toMap();
      const QString fade = option_map.value("name").toString();
      if (fade == "fade") {
        const bool state = option_map.value("on").toBool();
        app->media_player->setCrossFade(state);
      }
    }
  }
}

}  // namespace netease
