// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_AUDIO_PLAYER_H_
#define NATIVE_AUDIO_PLAYER_H_

#include <QObject>
#include <QVariant>

#include "service/media_player.h"

namespace netease {

// AudioPlayer controls media player playback state.
class AudioPlayer : public QObject {
  Q_OBJECT

 public:
  explicit AudioPlayer(QObject* parent);

 signals:
  // Buffering state of media resource changed.
  // Metadata is already loaded.
  void onBuffering(const QString& play_id, int state);

  // Emit when current media resource reaches end.
  // |info| is an object containing this field:
  //    * code: PlayingEndState.
  void onEnd(const QString& play_id, const QVariantMap& info);

  // Emitted after calling enmeratorDevices() slot.
  // |command| is the same as in enmeratorDevices().
  // |devices| is a map with pair of { "type" => [AudioOutputDevice, ..] }.
  // |current| is an AudioOutputDevice object refering to currently used
  // audio output device.
  void onEnmeratorDevices(const QString& command, const QVariantMap& devices,
                          const QVariantMap& current_device);

  // Emitted after calling load() slot.
  // |info| is an object containing this field:
  //   * code: LoadState.
  //   * duration: media resource duration, in second.
  //   * errorCode: 1, unused currently.
  //   * activeCode: 1, unused currently.
  //   * errorString: "", unused currently.
  // Note that only emit this signal only once for each song. Or else
  // seek() would be called in web ui infinitely.
  void onLoad(const QString& play_id, const QVariantMap& info);

  // Emitted when media player is playing and play progress moves forward.
  // |current_progress| is current position of playback, in second.
  // |buffering_percent| is percent of temporary buffer, between 0.0 and 1.0.
  void onPlayProgress(const QString& play_id, double current_progress,
                      double buffering_percent);

  // Emitted when playing state changed.
  // |resume_or_pause_id| is the same as in pause() or play() slot.
  void onPlayState(const QString& play_id, const QString& resume_or_pause_id,
                   int state);

  // Emitted when playback is seeked, by calling seek() slot.
  // |play_id| and |seek_id| are used in seek() slot.
  // |position| is the new playback position.
  void onSeek(const QString& play_id, const QString& seek_id,
              int code, double new_position);

  // Emitted when volume of player changed.
  void onVolume(const QString& play_id, const QString& volume_id,
                int code, double new_volume);

 public slots:
  // Get current media duration.
  double duration() const;

  // List audio output devices and emit onEnumratorDevices() signal.
  // TODO(xushaohua): Call pauseaudio interface.
  // |command| is "getOutDevices" currently.
  void enmeratorDevices(const QString& command);

  // Get current playback position, in seconds.
  double getCurrentTime() const;

  // TODO(xushaohua): Add comments.
  double getDownloadSchedule() const;

  // Get current volume.
  // Volume ranges from 0.0 to 1.0.
  double getVolume() const;

  // Initialize audio output device.
  // |command| might be "device".
  // |device_info| is an object with following fields:
  // {
  //   type: xx,
  //   device: {
  //     id: xx,
  //     deviceId: xx,
  //     name: xx
  //   }
  // }
  // Does nothing on linux platform currently.
  void init(const QString& command, const QVariantMap& device_info) const;

  // Check whether player is in mute state.
  bool isMuted() const;

  // Stop playing last media resource and load a new music resource.
  // All necessary information about that resource is stored in |track_info|,
  // including:
  //   * "songId", stored in |song_id_|.
  //   * "dfsId", stored in |dfs_id_|.
  //   * "albumId", stored in |album_id_|.
  //   * "url", stored in |cover_url_|.
  //   * "file", stored in |local_path_|.
  //   * "type", stored in |music_type_|.
  // The abbreviated process is:
  // webview calls Player.load(track_info), and |player_| starts loading
  // media resource, when metadata of that resource is downloaded, |player_|
  // emits onLoad() signal and Player emits that signal too.
  // When webview receives onLoad signal, it calls Player::play() slot.
  // |play_id| is stored in |track_info|.playId, plus bitrate.
  // e.g. 368740_1_940034-bitrate-192.
  void load(const QString& play_id, const QVariantMap& track_info);

  // Set/unset mute state of player.
  void mute(bool mute) const;

  // Pause playing the current source.
  void pause(const QString& play_id, const QString& resume_or_pause_id);

  // Start or resume playing the current source.
  void play(const QString& play_id, const QString& resume_or_pause_id);

  // Seek to new playback position and emit onSeek() signal.
  void seek(const QString& play_id, const QString& seek_id,
            double new_position);

  // Play mode might be:
  //   * "playorder"
  //   * "playcycle"
  //   * "playrandom"
  // TODO(xushaohua): Why change play mode in native side?
  void setPlayMode(const QString& mode) const;

  // Change volume of media player.
  // Note that |play_id| is empty and should not be stored be |play_id_|.
  // Because even if no media resource is being played, it should still be
  // able to change volume of player.
  void setVolume(const QString& play_id, const QString& volume_id,
                 double volume);

  // Stop playing.
  void stop(const QString& play_id);

  // Switch audio effect.
  void switchEffect(const QString& id, const QVariantList& options);

 private:
  // Connect signals to slots.
  void initConnections() const;

  // Set this flag to true when seek() slot is called, and reset it to false
  // after emitting onSeek() signal.
  bool is_seeking_;

  // This flag is used to make sure that onLoad() signal is only emitted only
  // once for each song.
  bool on_load_emitted_;

  QString play_id_;
  QString resume_or_pause_id_;
  QString seek_id_;
  QString volume_id_;

  // This struct is used on onEnumerateDevices() signal.
  struct AudioOutputDevice {
    QString name;
    QString id;
    QString deviceId;
    QString type;
  };

 private slots:
  void handleLoadError(int error_code);

  // Handles media buffering status changes.
  void handleMediaStatusChanged(MediaPlayer::MediaStatus status);

  // Convert unit in milliseconds to seconds when media position changed.
  // And emit ontimeupdate() signal.
  void handlePositionChanged(qint64 position);

  // When loading a new media resource, player starts to load metadata of that
  // resource and caches some buffer.
  // seekableChanged() signal is emitted when enough buffer has been filled
  // and audio is seekable.
  // At this time, emit AudioPlayer::onBuffering() to info frontend that it
  // is able to play current media resource.
  void handleSeekableChanged(bool seekable);

  // Handles media playing state changes.
  void handleStateChanged(MediaPlayer::State state);

  // Handles player volume changed signal.
  void handleVolumeChanged();
};

}  // namespace netease

#endif  // NATIVE_AUDIO_PLAYER_H_
