// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_MUSIC_LIBRARY_H_
#define NATIVE_MUSIC_LIBRARY_H_

#include <QObject>

namespace netease {

// MusicLibrary is used to handle offline music library.
class MusicLibrary : public QObject {
  Q_OBJECT

 public:
  explicit MusicLibrary(QObject* parent);

 signals:
  // Emitted after calling addLibrary() ended.
  // |error| is error number, might be 0 or 1.
  // If |error| != 0, set error reason to |reason|.
  void onaddend(const QVariantMap& path, int error, const QString& reason);

  // Emitted after calling addLibrary().
  // |song_num| is the number of items added to library.
  void onaddprogress(const QString& path, int song_num, const QString& reason);

  void onexecsql(const QVariantMap& result);

  // Emitted when files changed in music library.
  void onobserveLibrary();

  // Emitted after calling parseCueInfo().
  void onparsecueinfo(const QString& id, int err, const QVariantMap& info,
                      const QString& reason);

  void onquerylibraryversion(const QString& current_version,
                             const QString& last_version);

  // Emitted after calling readMusicInfo().
  void onreadmusicinfo(const QString& id, const QString& filepath, int err,
                       const QVariantMap& info);

  // Emitted after calling removeLibrary().
  // If |error| != 1, set |reason|.
  void onremovelibrary(const QString& library_path, int error,
                       const QString& reason);

  // Emitted after calling removeLibraryItems().
  // |reason| may contain a field "fail", which is a list of files which were
  // failed to be deleted. { "fail": [...] }
  void onremovelibraryitems(const QVariantMap& reason);

 public slots:
  // Add |library_path| to libraries.
  // |song_limit| is the number of items used in onAddProgress().
  void addLibrary(const QString& library_path, int song_limit);

  // Execute a SQl |statement| with |id|.
  // |id| is a string to identifies current SQL |statement|,
  // like "checkTableExist" or "createTable".
  // When this statement is executed, emit onexecsqldone() signal.
  void execSql(const QString& id, const QString& statement);

  // Exec sql transactions.
  void execSqls(const QString& id, const QStringList& statements);

  // Get absolute library path based on |library_path|.
  QString getLibraryPath(const QString& library_path);

  // Monitor file changes in music library.
  // Emit onobservelibrary() signal when file changed.
  void observeLibrary(const QString& library_path);

  // Get offline database version and emit onquerylibraryversion().
  void queryLibraryVersion();

  // Parse cue file and emit onparsecueinfo() signal.
  void parseCueInfo(const QString& id, const QString& filepath, bool unused);

  // Parse Music metadata and emit onreadmusicinfo() signal.
  void readMusicInfo(const QString& id, const QString& filepath, bool unused);

  // Remove Music library at |library_path|.
  void removeLibrary(const QString& library_path);

  // Remove items from local library.
  void removeLibraryItems(const QVariantList& items);

  // Remove |library_path| from file watcher manager.
  void removeObserveLibrary(const QString& library_path);

 private:
  void initConnections();
};

}  // namespace netease

#endif  // NATIVE_MUSIC_LIBRARY_H_
