// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_OS_H_
#define NATIVE_OS_H_

#include <QObject>
#include <QVariantMap>

namespace netease {

// OS is used to retrieve system information.
class OS : public QObject {
  Q_OBJECT

 public:
  explicit OS(QObject* parent);

 signals:
  // Emitted after calling isOnline().
  void onisonline(bool online);

  // Emitted after calling checkNativeSupportFonts().
  void onCheckNativeSupportFonts(const QString& is_success,
                                 const QStringList& fonts);

  // Emitted after calling querySystemFonts().
  void onQuerySystemFonts(const QString& is_success, const QStringList& fonts);

 public slots:
  // Check whether font in |fons| are supported in system.
  // Returns ["success", [font1, font2,,,]].
  void checkNativeSupportFonts(const QStringList& fonts);

  // Like getDeviceId(), get global unique identifier of current device.
  QString getADDeviceID();

  // Get global unique identifier of current device.
  QString getDeviceId();

  // Check disk space |path| refers to.
  // |path| might be a filepath or a folder path.
  // Returns a json string including:
  //   * "total", total space of device, in byte.
  //   * "free", freespace of device, in byte.
  QString getDiskSpace(const QString& path);

  // TODO(xushaohua): Add comments.
  void getKeyboardState() const;

  // Get version of operating system.
  // Current it returns 4 (windows vista).
  int getOsVer() const;

  // If |type| is "monitor", returns screen geometry.
  // If |type| is "lastInputInfo", returns time delta (in microseconds) to
  // last input event, including mouse event and keyboard event.
  // Else returns Object{}.
  // NOTE(xushaohua): Currently only "monitor" type is supported.
  QVariantMap getSystemInfo(const QString& type) const;

  // Check if main window is in fullscreen mode.
  bool hasFullscreenWindow() const;

  // Check whether Aero style is enabled or not.
  // NOTE: this method always returns false on linux platform.
  bool isAeroStyleEnabled() const;

  // Check whether |filepath| exists.
  bool isFileExist(const QString& filepath) const;

  // Check network connection status. Emit onisonline() signal.
  // Note: another api is app::setWatchNetStatus().
  void isOnLine();

  // Open |url| with desktop default browser.
  void navigateExternal(const QString& url) const;

  // Get version number of Operating system.
  QString queryOsVer();

  // Get Simplified Chinese fonts installed in system.
  // Returns [string, [string, string,,,]]
  // Emit onQuerySystemFonts().
  void querySystemFonts();

  // Open folder in which |filepath| is placed with default file manager.
  void shellExplor(const QString& filepath) const;

  // Open |path| with default application.
  void shellOpen(const QString& path) const;
};

}  // namespace netease

#endif  // NATIVE_OS_H_
