// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "native/player.h"

#include <QDebug>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>

#include "application.h"
#include "service/lrc_manager.h"
#include "service/media_player.h"
#include "service/signal_manager.h"
#include "service/storage_manager.h"
#include "ui/models/song_info.h"
#include "ui/views/desktop_lyrics.h"
#include "ui/views/mini_player.h"

namespace netease {

namespace {

// Action name mini player sends to webview.
const char kActionPrev[] = "prev";
const char kActionNext[] = "next";
const char kActionPlay[] = "play";
const char kActionPause[] = "pause";
//const char kActionStop[] = "stop";
//const char kActionVolume[] = "volume";
//const char kActionMode[] = "mode";
// Add current played song to favor list.
//const char kActionLove[] = "love";

// Action name desktop lyric sends to webview.
// Close desktop lyrci window.
const char kLrcActionClose[] = "close";
// Switch to detail page in main window.
const char kLrcActionDetail[] = "detail";
// Move lrc content backward slightly.
const char kLrcActionBackward[] = "offset_back";
// Move lrc content forward slightly.
const char kLrcActionForward[] = "offset_forward";
// Lock desktop lyric window.
const char kLrcActionLock[] ="lock";
// Unlock desktop lyric window.
const char kLrcActionUnlock[] = "unlock";
// Request to popup a menu when settings-button is clicked.
const char kLrcActionSetting[] = "setting";

}  // namespace

Player::Player(QObject* parent) : QObject(parent) {
  this->setObjectName("player");
  initConnections();
}

void Player::addListElement(const QVariant& songs) {
  QJsonDocument song_doc(QJsonDocument::fromJson(songs.toByteArray()));
  if (! song_doc.isArray()) {
    qCritical() << "[Player]::addListElement() failed to parse song list: "
                << songs;
  }

  QList<SongInfo> song_list;
  QJsonArray song_array(song_doc.array());
  for (const QJsonValue& song_value : song_array) {
    QJsonObject song_obj(song_value.toObject());
    SongInfo* song_info = new SongInfo;
    song_info->album = song_obj.value("album").toString();
    song_info->alias = song_obj.value("alias").toString();
    song_info->artist = song_obj.value("artist").toString();
    song_info->cloud = song_obj.value("cloud").toInt();
    song_info->from = song_obj.value("from").toString();
    song_info->mv = song_obj.value("mv").toString();
    song_info->play_id = song_obj.value("id").toString();
    song_info->program = song_obj.value("program").toInt();
    song_info->title = song_obj.value("title").toString();
    song_info->track_id = song_obj.value("track_id").toInt();
    song_list.append(*song_info);
  }
  app->mini_player->addSongs(song_list);
}

void Player::deleteListElement(const QVariantList& songs) {
  QStringList play_id_list;
  for (const QVariant& item : songs) {
    play_id_list.append(item.toString());
  }
  app->mini_player->deleteSongs(play_id_list);
}

void Player::handleLrcBackward() {
  emit ondesktoplyricaction(kLrcActionBackward);
}

void Player::handleLrcClose() {
  emit ondesktoplyricaction(kLrcActionClose);
}

void Player::handleLrcForward() {
  emit ondesktoplyricaction(kLrcActionForward);
}

void Player::handleLrcLock() {
  emit ondesktoplyricaction(kLrcActionLock);
}

void Player::handleLrcSetting() {
  emit ondesktoplyricaction(kLrcActionSetting);
}

void Player::handleLrcUnlock() {
  emit ondesktoplyricaction(kLrcActionUnlock);
}

void Player::handlePause() {
  emit onaction(kActionPause);
}

void Player::handlePlayNext() {
  emit onaction(kActionNext);
}

void Player::handlePlayPrev() {
  emit onaction(kActionPrev);
}

void Player::handleResume() {
  emit onaction(kActionPlay);
}

void Player::handleShowDetail() {
  emit ondesktoplyricaction(kLrcActionDetail);
}

void Player::handleVolumeChange(int volume) {
  emit onminivolumechange(volume / 100.0);
}

void Player::initConnections() const {
  SignalManager* signal_manager = app->signal_manager;
//  connect(signal_manager, &SignalManager::onPlayMode,
//          this, &Player::onplaymode);

  connect(signal_manager, &SignalManager::onPlayNext,
          this, &Player::handlePlayNext);
  connect(signal_manager, &SignalManager::onPlayPrev,
          this, &Player::handlePlayPrev);

  connect(signal_manager, &SignalManager::onRequestPause,
          this, &Player::handlePause);
  connect(signal_manager, &SignalManager::onRequestResume,
          this, &Player::handleResume);

  connect(signal_manager, &SignalManager::onRequestChangeToMain,
          this, &Player::onrequestchangetomain);

  connect(signal_manager, &SignalManager::onRequestPlay,
          this, &Player::onrequestplay);

  connect(signal_manager, &SignalManager::onRequestChangeVolume,
          this, &Player::handleVolumeChange);

  connect(signal_manager, &SignalManager::onMiniPlayerContextMenu,
          this, &Player::oncontextmenu);
  connect(signal_manager, &SignalManager::onMiniPlayerMenu,
          this, &Player::onmenu);

  connect(signal_manager, &SignalManager::onRequestFavorSong,
          this, &Player::onfavour);
  connect(signal_manager, &SignalManager::onRequestLikeSong,
          this, &Player::onlikeclick);

  connect(signal_manager, &SignalManager::onLrcRequestBackward,
          this, &Player::handleLrcBackward);
  connect(signal_manager, &SignalManager::onLrcRequestClose,
          this, &Player::handleLrcClose);
  connect(signal_manager, &SignalManager::onLrcRequestForward,
          this, &Player::handleLrcForward);
  connect(signal_manager, &SignalManager::onLrcRequestLock,
          this, &Player::handleLrcLock);
  connect(signal_manager, &SignalManager::onLrcRequestSettingMenu,
          this, &Player::handleLrcSetting);
  connect(signal_manager, &SignalManager::onLrcRequestUnlock,
          this, &Player::handleLrcUnlock);
  connect(signal_manager, &SignalManager::onLrcRequestFontSize,
          this, &Player::onFontSize);
  connect(signal_manager, &SignalManager::onLrcRequestMouseScroll,
          this, &Player::ondesktopmousewheel);

  connect(signal_manager, &SignalManager::onRequestShowDetail,
          this, &Player::handleShowDetail);

  connect(signal_manager, &SignalManager::onRenderLrcImageResult,
          this, &Player::onRenderLRCImageResult);
  connect(signal_manager, &SignalManager::onRequestDeleteFm,
          this, &Player::onrecyle);
}

void Player::mute(bool mute) const {
  app->media_player->setMuted(mute);
}

void Player::preload(const QVariantMap& options) const {
  qDebug() << "TODO [Player]::preload()" << options;

  // TODO(xushaohua): Call MediaPlayerCache.
}

void Player::removeAll() const {
  app->mini_player->clearPlaylist();
}

void Player::renderLRCImage(const QString& text,
                            const QString& filename) const {
  const QString filepath(
      app->storage_manager->getLocalStorageFilePath(filename));
  emit app->signal_manager->onRequestRenderLrcImage(text, filepath);
}

void Player::setAlwaysOnTop(bool topmost) const {
  app->mini_player->setAlwaysOnTop(topmost);
}

void Player::setCover(const QString& url, const QString& type) const {
  Q_UNUSED(type);
  app->mini_player->setAlbumCover(url);
}

void Player::setCoverDefault() const {
  app->mini_player->setUseDefaultCover();
}

void Player::setCurrentPlay(const QString& play_id) const {
  app->mini_player->setCurrentPlay(play_id);
}

void Player::setDesktopLyricTopMost(bool topmost) const {
  app->desktop_lyrics->setTopMost(topmost);
}

void Player::setFavour(int favour) const {
  emit app->signal_manager->onSetFavorMark(favour == 1);
}

void Player::setFont(const QString& option, const QString& font_size) const {
  qDebug() << "TODO [Player]::setFont()"
           << ", option:" << option
           << ", font_size:" << font_size;
}

void Player::setInfo(const QVariantMap& track_info) const {
  app->media_player->setInfo(track_info);
}

void Player::setLikeMark(int state) {
  emit app->signal_manager->onSetLikeMark(state);
}

void Player::setLineMode(bool is_single_line) const {
  emit app->signal_manager->onSetLrcLineMode(is_single_line);
}

void Player::setLock(bool locked) const {
  app->desktop_lyrics->setLocked(locked);
  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void Player::setLRCColor(const QString& base_color1,
                         const QString& base_color2,
                         const QString& played_color1,
                         const QString& played_color2) const {
  emit app->signal_manager->onSetLrcColor(base_color1, base_color2,
                                          played_color1, played_color2);
}

void Player::setLRCDJ() const {
  app->lrc_manager->setIsDj();
}

void Player::setLRCEmpty() const {
  app->lrc_manager->setLrcEmpty();
}

void Player::setLRCFont(const QString& font_size, const QString& unused,
                        const QString& font_family) const {
  emit app->signal_manager->onSetLrcFont(font_size, unused, font_family);
}

void Player::setLRCSlogan(const QString& slogan) const {
  app->lrc_manager->slogan = slogan;
}

void Player::setLyrics(const QVariantMap& lyrics) {
  const QString lrc = lyrics.value("lrc").toString();
  const QString krc = lyrics.value("krc").toString();
  app->lrc_manager->setLyrics(lrc, krc);
}

void Player::setMiniPlayerState(const QVariantMap& options) {
  Q_UNUSED(options);
}

void Player::setMusicOnly() const {
  app->lrc_manager->setIsLightMusic();
}

void Player::setOffset(qint64 offset) const {
  app->lrc_manager->setOffset(offset);
}

void Player::setOutlineColor(const QString& color1,
                             const QString& color2) const {
  qDebug() << "TODO [Player]::setOutlineColor() " << color1 << color2;
  Q_UNUSED(color1);
  Q_UNUSED(color2);
}

void Player::setPlayMode(const QString& mode) const {
  qDebug() << "TODO [Player]::setPlayMode()" << mode;
  Q_UNUSED(mode);
  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void Player::setTextAlign(const QString& primary_alignment,
                          const QString& secondary_alignment) const {
  emit app->signal_manager->onSetLrcTextAlignment(primary_alignment,
                                                  secondary_alignment);
}

void Player::setTotalTime(unsigned int totalTime) {
  qDebug() << "TODO [Player]::setTotalTime()" << totalTime;
}

void Player::showVolume(double volume) const {
  qDebug() << "TODO [Player]::showVolume()" << volume;
}

void Player::simulateKRC(const bool& on) const {
  qDebug() << "TODO [Player]::simulateKRC()" << on;

  // Do nothing.
}

void Player::updateDesktopLrcTooltips(const QString& prev_tooltip,
                                      const QString& play_pause_tooltip,
                                      const QString& next_tooltip) {
  app->desktop_lyrics->updateTooltips(prev_tooltip,
                                      play_pause_tooltip,
                                      next_tooltip);
}

void Player::updateTooltips(const QString& tooltip) {
  app->mini_player->setShowMainWindowButtonToolTip(tooltip);
}

}  // namespace netease
