// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_PLAYER_H_
#define NATIVE_PLAYER_H_

#include <QObject>
#include <QVariant>

namespace netease {

// Player controls desktop lyric and mini player state.
class Player : public QObject {
  Q_OBJECT

 public:
  explicit Player(QObject* parent);

 signals:
  // Handles mini player action.
  void onaction(const QString& action);

  // TODO(xushaohua): Add comments.
  void onaddlog();

  // Mini Player requests to popup context menu on top panel.
  void oncontextmenu();

  // Adjust lrc time offset by scrolling mouse wheel.
  // |button| is left button, middle button or right button.
  // Note that only |offset| is used and all other arguments are ignored by
  // web player. And |offset| can be a positive or negative integer.
  void ondesktopmousewheel(int button, int offset, int offset_x, int offset_y);

  // Desktop lrics window requests actions.
  void ondesktoplyricaction(const QString& action);

  // TODO(xushaohua): Add comments.
//  void ondurationloaded(int duration);

  // Toggle current program favorite state.
  void onfavour(int like);

  // Emitted when font size desktop lyrics changed by resize lyrics window.
  void onFontSize(int font_size);

  // Add current played song to red-heart playlist or remove from it.
  void onlikeclick();

  // TODO(xushaohua): Add comments.
  void onlyricsupdate();

  // Mini player requests to popup context menu on playlist.
  // |play_id| is the row id on which mouse click is right clicked.
  void onmenu(const QString& play_id);

  // Emitted when volume changed in mini player.
  void onminivolumechange(double volume);

  // Emitted when network connection status changed.
  // |status| might be:
  //   * 0, means login succeeded.
  //   * 1, means login failed.
  //   * 2, means it is ready to retry login process.
  //   * 3, means network disconnected.
  void onnotify(int status);

  // TODO(xushaohua): Add comments.
  void onNotifyWindow();

  // Emitted when mini player changes |mode| of playing process.
  // |mode| is the same as defined in setMode().
  void onplaymode(const QString& mode);

  // TODO(xushaohua): Add comments.
  // TODO(xushaohua): Remove onplaymode() signal.
  void onplaymodechange();

  // TODO(xushaohua): Add comments.
  void onplaystate(int play_state);

  // TODO(xushaohua): Add comments.
//  void onpositionchange(double position);

  // Delete current song from fm list and switch to next one.
  void onrecyle();

  // Emitted after calling renderLRCImage().
  // |ok| if true if lrc image is generated successfully.
  void onRenderLRCImageResult(bool ok, int image_width, int image_height);

  // Change to standard mode from mini-player mode.
  void onrequestchangetomain();

  // Request to close main process.
  void onrequestclose();

  // Mini player requests to play song with id |play_id|.
  void onrequestplay(const QString& play_id);

  // TODO(xushaohua): Add comments.
  void onshortcut();

  // Emitted when buttons in thumbnail is clicked.
  // See App::setThumbnail for more information.
  // |id| is button id.
  void onthumbnailaction(int id);

  // Emitted when the playback position of the current media changed.
  // TODO(xushaohua): onpositionchange() ?
//  void ontimeupdate(double position, double download_percent);

 public slots:
  // Notify mini player to add songs to playlist.
  void addListElement(const QVariant& songs);

  // Notify mini player to delete songs from playlist.
  void deleteListElement(const QVariantList& songs);

  // Set/unset mute state of player.
  // This slot is the same as setMute().
  void mute(bool mute) const;

  // Preload next media source in playlist.
  void preload(const QVariantMap& options) const;

  // Notify mini player to clear playlist.
  void removeAll() const;

  // Generate lrc preview image, used in settings page. And emit
  // onRenderLRCImageResult() signal.
  // |text| is the text content to be renderer.
  // |filename| is "preview/font.png", and that path is requested as
  // "orpheus://orpheus/storage/local?file=preview/font.png".
  void renderLRCImage(const QString& text, const QString& filename) const;

  // Set/unset mini player window topmost.
  void setAlwaysOnTop(bool topmost) const;

  // Set current album cover.
  void setCover(const QString& url, const QString& type) const;

  // If not album cover available, use the default one.
  void setCoverDefault() const;

  // Notify mini player to update current song.
  // |play_id| is used to identify songs in playlist.
  void setCurrentPlay(const QString& play_id) const;

  // Keep desktop lyric window on top or not.
  void setDesktopLyricTopMost(bool topmost) const;

  // Set favorite mark (+1).
  void setFavour(int favour) const;

  // TODO(xushaohua): Add comments.
  void setFont(const QString& option, const QString& font_size) const;

  // Set current media track info.
  // TODO(xushaohua): Remove this slot.
  void setInfo(const QVariantMap& track_info) const;

  // Show red heart.
  void setLikeMark(int state);

  // Set line mode of lyric, single line or dual lines.
  void setLineMode(bool is_single_line) const;

  // Lock/unlock desktop lyrics.
  void setLock(bool locked) const;

  // Change font color of desktop lyrics.
  void setLRCColor(const QString& base_color1,
                   const QString& base_color2,
                   const QString& played_color1,
                   const QString& played_color2) const;

  // Set lyric for DJ, no lyric.
  void setLRCDJ() const;

  // If not lyric is available, cleanup previous lyric content and display
  // the default lyric slogan.
  void setLRCEmpty() const;

  // Set font name for desktop lyric.
  void setLRCFont(const QString& font_size, const QString& unused,
                  const QString& font_family) const;

  // Setup the default slogan displayed in desktop lyrics window when no
  // lyric is loaded.
  void setLRCSlogan(const QString& slogan) const;

  // Set lrc and krc.
  // Added in webapp2.0.
  void setLyrics(const QVariantMap& lyrics);

  // TODO(xushaohua): add comments.
  void setMiniPlayerState(const QVariantMap& options);

  // No lyric available in light music (easy music). And change slogan to
  // "Enjoy Light Music".
  void setMusicOnly() const;

  // Set lyric offset. Also update time line in desktop lyrics.
  void setOffset(qint64 offset) const;

  // TODO(xushaohua): Add comments.
  void setOutlineColor(const QString& color1, const QString& color2) const;

  // UNUSED.
  // Play mode might be:
  //   * "playorder"
  //   * "playcycle"
  //   * "playrandom"
  void setPlayMode(const QString& mode) const;

  // Set alignment of text in lyric window.
  // |primary_alignment| and |secondary_alignment| might be:
  //   * "left"
  //   * "center"
  //   * "right"
  void setTextAlign(const QString& primary_alignment,
                    const QString& secondary_alignment) const;

  // UNUSED.
  // Update total media resource duration.
  void setTotalTime(unsigned int totalTime);

  // UNUSED.
  // Show volume slide bar in mini player.
  void showVolume(double volume) const;

  // DEPRECATED
  // Simulate lrc file to krc file.
  void simulateKRC(const bool& on) const;

  // Update tooltips in desktop lyric.
  void updateDesktopLrcTooltips(const QString& prev_tooltip,
                                const QString& play_pause_tooltip,
                                const QString& next_tooltip);

  // Update tooltips in mini player.
  void updateTooltips(const QString& tooltip);

 private:
  // Connect signals to slots.
  void initConnections() const;

 private slots:
  void handleLrcBackward();
  void handleLrcClose();
  void handleLrcForward();
  void handleLrcLock();
  void handleLrcUnlock();
  void handleLrcSetting();

  void handlePause();
  void handlePlayPrev();
  void handlePlayNext();
  void handleResume();
  void handleShowDetail();
  void handleVolumeChange(int volume);
};

}  // namespace netease

#endif  // NATIVE_PLAYER_H_
