// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "native/storage.h"

#include <QDebug>
#include <QFileInfo>

#include "application.h"
#include "service/media_metadata.h"
#include "service/storage_manager.h"

namespace netease {

Storage::Storage(QObject* parent) : QObject(parent) {
  this->setObjectName("storage");

  initConnections();
}

void Storage::addid3(const QString& task_id, const QString& media_path,
              const QString& album_cover_path, const QString& media_info,
              const QVariantMap& tag) {
  qDebug() << Q_FUNC_INFO << media_info;
  emit app->media_metadata->addId3(task_id, media_path, album_cover_path,
                                   media_info, tag);
}

void Storage::cancelMoveStorage() const {
  qDebug() << "TODO [Storage]::cancelMoveStorage()";
}

void Storage::checkFilesExist(const QString& task_id, const QVariantList& files,
                              const QString& download_path) {
  emit app->storage_manager->checkFilesExist(task_id, files, download_path);
}

void Storage::clearCache(const QStringList& folders) const {
  Q_UNUSED(folders);
  emit app->storage_manager->clearCache();
}

void Storage::copyfiles() {
  qDebug() << "TODO [Storage]::copyfiles()";
}

void Storage::deletefile(const QString& task_id, const QString& path_type,
                         const QString& tag, const QString& filepath,
                         bool delete_empty_dir) {
  emit app->storage_manager->deleteFile(task_id, path_type, tag, filepath,
                                        delete_empty_dir);
}

void Storage::downloadscanner(const QString& dirpath, bool delete_file,
                              const QString& tag, int limit,
                              const QStringList& exist_files) {
  qDebug() << "[Storage]::downloadscanner()" << dirpath << exist_files << tag << limit << delete_file;
  Q_UNUSED(delete_file);
  Q_UNUSED(tag);
  emit app->storage_manager->downloadScanner(dirpath, limit, exist_files);
}

void Storage::execsql(const QString& id, const QString& statement) const {
  emit app->storage_manager->execSql(id, statement);
}

void Storage::exectransaction(const QString& id,
                              const QString& statement) const {
 emit app->storage_manager->execTransaction(id, statement);
}

void Storage::getTempFile(const QString& url) {
  emit app->storage_manager->getTempFile(url);
}

void Storage::imagesInfo(const QString &id, const QStringList &files) {
  emit app->storage_manager->imagesInfo(id, files);
}

void Storage::init(const QStringList& download_paths, double cache_capacity,
                   const QString& cache_path) {
  qDebug() << Q_FUNC_INFO << download_paths << cache_capacity << cache_path;
  QString escaped_download_path;
  // Remove backslash, which is appended in web page.
  if (!download_paths.isEmpty()) {
    const QString& download_path = download_paths.first();
    if (download_path.endsWith('\\')) {
      escaped_download_path = download_path.left(download_path.length() - 1);
    } else {
      escaped_download_path = download_path;
    }
  }

  if (escaped_download_path.isEmpty()) {
    escaped_download_path = GetDefaultMusicDownloadPath();
  }

  // Append path separator
  if (!escaped_download_path.endsWith(QChar('/'))) {
    escaped_download_path.append(QChar('/'));
  }

  emit app->storage_manager->init(escaped_download_path, cache_capacity);
  emit this->onInit(escaped_download_path, cache_path);
}

void Storage::initConnections() const {
  connect(app->storage_manager, &StorageManager::onCheckFilesExist,
          this, &Storage::oncheckfilesexist);
  connect(app->storage_manager, &StorageManager::onClearCache,
          this, &Storage::onclearcache);
  connect(app->storage_manager, &StorageManager::onDownloadScanner,
          this, &Storage::ondownloadscanner);
  connect(app->storage_manager, &StorageManager::onFileDeleted,
          this, &Storage::onfiledeleted);
  connect(app->storage_manager, &StorageManager::onExecSqlDone,
          this, &Storage::onexecsqldone);
  connect(app->storage_manager, &StorageManager::onGetTempFile,
          this, &Storage::ongettempfile);
  connect(app->storage_manager, &StorageManager::onImagesInfo,
          this, &Storage::onImagesInfo);
  connect(app->storage_manager, &StorageManager::onListFile,
          this, &Storage::onlistfile);
  connect(app->storage_manager, &StorageManager::onMoveProcess,
          this, &Storage::onmoveprocess);
  connect(app->storage_manager, &StorageManager::onQueryCacheTracks,
          this, &Storage::onquerycachetracks);
  connect(app->storage_manager, &StorageManager::onQueryNewCacheTracks,
          this, &Storage::onquerynewcachetracks);
  connect(app->storage_manager, &StorageManager::onReadFromFileDone,
          this, &Storage::onreadfromfiledone);
  connect(app->storage_manager, &StorageManager::onSaveToFileDone,
          this, &Storage::onsavetofiledone);
  connect(app->storage_manager, &StorageManager::onScaleImage,
          this, &Storage::onScaleImage);
  connect(app->storage_manager, &StorageManager::onScanDirectory,
          this, &Storage::onscanDirectory);
  connect(app->storage_manager, &StorageManager::onUpdateTempFile,
          this, &Storage::onupdatetemp);
  connect(app->media_metadata, &MediaMetadata::onAddId3,
          this, &Storage::onaddid3done);
}

void Storage::listFile(const QString& task_id, const QString& path_type,
                       const QString& tag, const QString& filepath) {
  emit app->storage_manager->listFile(task_id, path_type, tag, filepath);
}

void Storage::movefiles(const QString& operation, const QString& src_path_type,
                        const QString& arg1, const QStringList& src_files,
                        const QString&  dest_path_type, const QString& arg2,
                        const QStringList& dest_files) {
  qDebug() << Q_FUNC_INFO << src_files << dest_files;
  emit app->storage_manager->moveFiles(operation, src_path_type,
      arg1, src_files, dest_path_type, arg2, dest_files);
}

void Storage::moveStorage(const QString& src, const QString& dest,
                          int capacity) const {
  qDebug() << "TODO [Storage]::moveStorage()";
  Q_UNUSED(src);
  Q_UNUSED(dest);
  Q_UNUSED(capacity);
}

void Storage::observeOffline() {
  qDebug() << "TODO [Storage]::observeOffline()";
}

void Storage::queryCacheTracks() {
  qDebug() << Q_FUNC_INFO;
  emit app->storage_manager->queryCacheTracks();
}

void Storage::queryNewCacheTracks(const QString& task_id,
                                  const QVariantMap& song_info) {
  qDebug() << Q_FUNC_INFO << task_id << song_info;
  emit app->storage_manager->queryNewCacheTracks(task_id, song_info);
}

//void Storage::queryNewCacheTracks(const QString& task_id,
//                                  const QVariantMap& song_info) {
//  QString bitrate = song_info.value("bitrate").toString();
//  QString md5 = song_info.value("md5").toString();
//  QString song_id = song_info.value("songId").toString();

//  QVariantList result;
//  if (app->media_player_cache->isCached(song_id, bitrate, md5)) {
//    QVariantMap info;
//    info.insert("cache", 100);
//    result.append(info);
//  }
//  emit onquerynewcachetracks(task_id, result);
//}

void Storage::readfromfile(const QString& task_id, const QString& filepath) {
  emit app->storage_manager->readFromFile(task_id, filepath);
}

void Storage::removetemp(const QString& url) {
  emit app->storage_manager->removeTempFile(url);
}

void Storage::savetofile(const QString& task_id, const QString& content,
                         const QString& tag, const QString& filepath) {
  emit app->storage_manager->saveToFile(task_id, content, tag, filepath);
}

void Storage::scaleImage(const QString& id, const QVariantMap& info) {
  qDebug() << "[Storage]::scaleImage()";
  emit app->storage_manager->scaleImage(id, info);
}

void Storage::scanDirectory(const QString& task_id,const QString& path_type,
                            int scan_type, const QString& path) {
  qDebug() << Q_FUNC_INFO << task_id << path_type << scan_type << path;
  emit app->storage_manager->scanDirectory(task_id, path_type, scan_type, path);
}

bool Storage::testwriteable(const QString& path) {
  const QFileInfo info(path);
  return info.exists() && info.isWritable();
}

void Storage::updatetemp(const QString& url, const QString& content,
                         const QString& mime) {
  emit app->storage_manager->updateTempFile(url, content, mime);
}

}  // namespace netease
