// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_STORAGE_H_
#define NATIVE_STORAGE_H_

#include <QObject>
#include <QVariantList>

namespace netease {

// Storage is used to manager disk cache.
class Storage : public QObject {
  Q_OBJECT

 public:
  explicit Storage(QObject* parent);

 signals:
  // Emitted after calling addid3().
  // |success| is 1 or 0.
  void onaddid3done(const QString& task_id, int success);

  // Emitted when local cache reaches its capacity.
  void oncachelack();

  // Emitted after calling checkFilesExist().
  // |task_id| is the same as used in checkFilesExist().
  // |success| is 0 or 1.
  // |result| is a list of queried files, each file contains field
  // "id", "path" and "exist".
  void oncheckfilesexist(const QString& task_id, int success,
                         const QVariantList& result);

  // Emitted after calling clearCache().
  void onclearcache(bool ok);

  // Emitted when new media files are added to |download_dir_|.
  void ondownloadscanner(const QVariantList& items);

  // Emit this signal after a SQL statement executation is done.
  // |id| is SQL statement identifier used in execsql().
  // |err| is the result of SQL statement execution, 0 means ok.
  // |ret|, which is an empty string, is unused.
  void onexecsqldone(const QString& id, int err, const QVariantList& ret);

  // Emitted when |files| are deleted by user.
  // Relative to |download_dir_| defined in StorageManager.
  void onfiledeleted(const QString& files);

  // Emitted after getTempFile() is called.
  // |url| is the same as used in getTempFile().
  // |content| is file content, might be empty.
  void ongettempfile(const QString& url, const QString& content);

  // Emitted after calling imagesInfo().
  void onImagesInfo(const QString& id, const QVariantList& result);

  void onInit(const QString& download_path, const QString& cache_path);

  // Emitted after calling listFile().
  void onlistfile(const QString& task_id, int err, const QVariantList& result);

  // Emitted after calling movefiles().
  void onmoveprocess(const QVariantMap& state);

  // Get moveStorage() status.
  // |percent| 0.0~1.0, 1.0 means moving storage finished.
  // |est_time| estimated time in seconds. It should be 0 when |percent|
  // is 1.0.
  void onmovingstorage(double percent, int est_time);

  // Emitted after calling queryCacheTracks().
  // |tracks| is a list of track object. See onquerynewcachetracks().
  void onquerycachetracks(const QVariantList& tracks);

  // Emitted after calling queryNewCacheTracks().
  // |task_id| is the same as in queryNewCacheTracks().
  // |tracks| is a list of track object. A track object includes these fields:
  //   * cached, 0-100, percent of cached data.
  void onquerynewcachetracks(const QString& task_id,
                             const QVariantList& tracks);

  // TODO(xushaohua): Add comments.
  void onqueryofflinetrackpath();

  // Emitted after calling scaleImage.
  // |result|, 0 means success.
  // |path| is absolute path to temporary image file.
  void onScaleImage(const QString& task_id, int result, const QString& path);

  // Emitted after slot readfromfile() is done.
  // |content| is file content.
  void onreadfromfiledone(const QString& task_id, int err,
                          const QString& content);

  // Emitted after slot savetofile() is done.
  void onsavetofiledone(const QString& task_id, int err);

  // Emitted after slot scanDirectory() is done.
  void onscanDirectory(const QString& task_id);

  // Emitted after calling updatetemp().
  void onupdatetemp(const QString& url, bool ok);

 public slots:
  // Write ID3 tag to media resource file.
  // |tag| contains:
  //   * talb: "21"
  //   * tit2: "He Won't Go"
  //   * tpe1: "Adele"
  //   * trck: "6
  void addid3(const QString& task_id, const QString& media_path,
              const QString& album_cover_path, const QString& media_info,
              const QVariantMap& tag);

  // Cancel move storage operation.
  // This function is useful when moving resources storage to a different
  // filesystem.
  // This slot does nothing on linux platform.
  void cancelMoveStorage() const;

  // Check files existence at local storage.
  // It is possible that downloaded music files were deleted by user.
  // |task_id| might be "syncoffline" or "synccloud".
  // |files| is a list of file info with fields "id" and "path".
  // |download_path| is empty if |task_id| is "synccloud".
  // Note that "synccloud" is not supported currently.
  // Emits oncheckfilesexist() signal.
  void checkFilesExist(const QString& task_id, const QVariantList& files,
                       const QString& download_path);

  // Clear cache folder.
  void clearCache(const QStringList& folders) const;

  // Copy files to new place.
  // This slot is used when moving storage cache to another filesystem.
  // This slot does nothing on linux platform.
  void copyfiles();

  void deletefile(const QString& task_id, const QString& path_type,
                  const QString& tag, const QString& filepath,
                  bool delete_empty_dir);

  // Scan |download_dir_|
  // |exist_files| is a list of files which are already stored in db.
  // They are relative to |download_dir_| and starts with a slash.
  void downloadscanner(const QString& dirpath, bool delete_file,
                       const QString& tag, int limit,
                       const QStringList& exist_files);

  // Execute a SQl |statement| with |id|.
  // |id| is a string to identifies current SQL |statement|,
  // like "checkTableExist" or "createTable".
  // When this statement is executed, emit onexecsqldone() signal.
  void execsql(const QString& id, const QString& statement) const;

  // Execute sql transaction. For better performance.
  void exectransaction(const QString& id, const QString& statement) const;

  // Read local cached file and emit ongettempfile() signal.
  void getTempFile(const QString& url);

  // Query image info.
  void imagesInfo(const QString& id, const QStringList& files);

  // Initialize |download_path| and |cache_path|.
  // If |download_paths| is empty, returns $HOME/Music.
  // |cache_capacity| is space used as local cache, in GB. When user moves
  // capacity slide in webview page, this value is updated and stored to
  // local settings file. If more space is used when application is running,
  // clear some old files.
  // |cache_path| is always set to $HOME/.cache/cloud-music.
  // Returns appropriate |download_path| and |cache_path|.
  void init(const QStringList& download_paths, double cache_capacity,
            const QString& cache_path);

  // List files in a folder.
  void listFile(const QString& task_id, const QString& path_type,
                const QString& tag, const QString& filepath);

  // Move files to new folder in the same filesystem.
  // This slot is called when moving storage cache to new place.
  // |operation| might be "move".
  // |src_path_type| might be "abs" or "rel".
  // Emit onmoveprocess() signal.
  void movefiles(const QString& operation, const QString& src_path_type,
                 const QString& arg1, const QStringList& src_files,
                 const QString&  dest_path_type, const QString& arg2,
                 const QStringList& dest_files);

  // Move resources storage from |src| to |dest|.
  // |src| might be empty for the first time.
  // |capacity| is in unit of GB.
  // When move operation finished, emit/call onmovedstorage().
  // This slot does nothing on linux platform.
  void moveStorage(const QString& src, const QString& dest, int capacity) const;

  // Observe file changing in download-folder.
  // Does nothing on Linux platform.
  void observeOffline();

  // Query all local cached songs and emit onquerycachetracks() signal.
  // Used in offline mode.
  void queryCacheTracks();

  // Query whether song with |song_info| is cached at local disk.
  // Emit onquerynewcachetracks() signal.
  // |song_info| is an object containing:
  //   * songId.
  //   * bitrate.
  //   * md5.
  void queryNewCacheTracks(const QString& task_id,
                           const QVariantMap& song_info);

  // Read local cached file, and emit onreadfromfiledone() signal.
  void readfromfile(const QString& task_id, const QString& filepath);

  // Remove local cached file.
  void removetemp(const QString& url);

  // Save temporary data to file.
  // |task_id| is used to mark current task, like this:
  // "flush history track|bwfGl|1449560245046".
  // |content| is the content to be written to file.
  // |tag| is not used currently and is empty.
  // |filepath| is the relative file path in cache directory, like:
  // "webdata\file\history".
  void savetofile(const QString& task_id, const QString& content,
                  const QString& tag, const QString& filepath);

  // Scale image and emit onScaleImage() signal.
  // |info| is an object containing:
  //   * path: absolute path to original image.
  //   * width: new width
  //   * height: new height
  //   * scale_type: 1 -> based on width; 2 -> based on height.
  void scaleImage(const QString& id, const QVariantMap& info);

  // Scan directory and emit onscanDirectory().
  void scanDirectory(const QString& task_id,const QString& path_type,
                     int scan_type, const QString& path);

  // Check whether |path| is writable.
  bool testwriteable(const QString& path);

  // Update local cached file.
  // |url| will be hashed and converted to filename.
  // |content| is data to be stored.
  // |mime| content type, like "text/plain".
  // Emit onupdatetemp() signal after file is cached.
  void updatetemp(const QString& url, const QString& content,
                  const QString& mime);

 private:
  // Initialize signal connections.
  void initConnections() const;
};

}  // namespace netease

#endif  // NATIVE_STORAGE_H_
