// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "native/win_helper.h"

#include <QClipboard>
#include <QDebug>
#include <QMimeData>
#include <QTimer>
#include <QUrl>

#include "application.h"
#include "base/consts.h"
#include "service/command_line.h"
#include "service/keybinder.h"
#include "service/popup_menu_manager.h"
#include "service/signal_manager.h"
#include "ui/views/web_view.h"
#include "ui/window_manager.h"

namespace netease {

namespace {

//const int kDefaultMainWindowWidth = 1022;
//const int kDefaultMainWindowHeight = 670;

// Show window.
const char kWindowStateShow[] = "show";

// Hide window.
const char kWindowStateHide[] = "hide";

// Show maximized window.
const char kWindowStateMaximized[] = "maximize";

// Show minimized window.
const char kWindowStateMinimized[] = "minimize";

// Show minimized window, on MacOS X.
const char kWindowStateZoom[] = "zoom";

// Show normal window.
const char kWindowStateRestore[] = "restore";

QString StringifyState(const Qt::WindowStates states) {
    if (states & Qt::WindowMaximized) {
        return kWindowStateMaximized;
    }
    if (states & Qt::WindowMinimized) {
        return kWindowStateMinimized;
    }
    if (states & Qt::WindowNoState) {
        return kWindowStateRestore;
    }
    return kWindowStateRestore;
}

enum MenuType {
  // Popup menu in main window.
  kMenuTypeMain = 0,

  // Popup menu in mini player.
  kMenuTypeMini = 1,

  // Popup menu at taskbar.
  kMenuTypeTaskbar = 2,

  // Popup menu in desktop lyric window.
  kMenuTypeDesktopLyric = 3,
};

}  // namespace

WinHelper::WinHelper(QObject* parent, WebView* web_view)
    : QObject(parent),
      web_view_(web_view),
      web_loading_finished_(false) {
  this->setObjectName("winhelper");

  this->initConnections();
}

void WinHelper::bringWindowToTop() const {
  web_view_->raiseWindow();
}

void WinHelper::destroyWindow() const {
//  web_view_->closeWindow();
  app->window_manager->destroyWindow(web_view_);
}

void WinHelper::dragWindow() const {
  web_view_->moveWindow();
}

void WinHelper::enterFullscreen() const {
  web_view_->enterFullscreen();
}

void WinHelper::exitFullscreen() const {
  web_view_->exitFullscreen();
}

QString WinHelper::getClipBoardData() const {
  QClipboard* clipboard = QApplication::clipboard();
  return clipboard->text();
}

QVariantMap WinHelper::getNativeWindowRect(const QString& window_name) const {
  QVariantMap result;
  QRect rect = app->window_manager->getNativeWindowRect(window_name);

  result.insert("x", QVariant(rect.x()));
  result.insert("y", QVariant(rect.y()));
  result.insert("width", QVariant(rect.width()));
  result.insert("height", QVariant(rect.height()));
  return result;
}

double WinHelper::getWindowOpacity() const {
  return web_view_->windowOpacity();
}

QVariantMap WinHelper::getWindowInfo(const QString& info) const {
  QVariantMap result;
  if (info == "status") {
    result.insert("visible", QVariant(web_view_->isVisible()));
    if (web_view_->isMinimized()) {
      result.insert("status", QVariant("minimize"));
    } else if (web_view_->isMaximized()) {
      result.insert("status", QVariant("maximize"));
    } else if (web_view_->isHidden()) {
      result.insert("status", QVariant("hidden"));
    } else {
      result.insert("status", QVariant("restore"));
    }
  }

  return result;
}

QVariantMap WinHelper::getWindowPosition() const {
  QVariantMap result;
  result.insert("x", QVariant(web_view_->x()));
  result.insert("y", QVariant(web_view_->y()));
  result.insert("width", QVariant(web_view_->width()));
  result.insert("height", QVariant(web_view_->height()));
  bool always_on_top = web_view_->windowFlags() | Qt::WindowStaysOnTopHint;
  result.insert("topmost", QVariant(always_on_top));
  return result;
}

void WinHelper::handleOnSizeWindowDone() {
  QVariantMap result;

  result.insert("left", web_view_->x());
  result.insert("top", web_view_->y());
  result.insert("right", web_view_->x() + web_view_->width());
  result.insert("bottom", web_view_->y() + web_view_->height());

  emit onsizeWindowDone(result);
}

void WinHelper::handleWindowStateSizeChanged() {
  // Window state of web-view might require extra time to update
  QTimer::singleShot(50, this, SLOT(sendWindowState()));
}

void WinHelper::sendWindowState() {
  emit this->onSizeStatus(StringifyState(this->web_view_->windowState()),
                          this->web_view_->width(),
                          this->web_view_->height());
}

void WinHelper::hidePopupMenu() {
  emit app->popup_menu_manager->hidePopupMenu();
}

void WinHelper::initConnections() const {
  if (web_view_ != nullptr) {
    connect(web_view_, &WebView::onClose, this, &WinHelper::onclose);
    connect(web_view_, &WebView::onFocusIn, this, &WinHelper::onfocus);
    connect(web_view_, &WebView::onFocusOut, this, &WinHelper::onlosefocus);
    connect(web_view_, &WebView::onSizeWindowDone,
            this, &WinHelper::handleOnSizeWindowDone);
    connect(web_view_, &WebView::windowStateSizeChanged,
            this, &WinHelper::handleWindowStateSizeChanged);
  } else {
    qWarning() << "WinHelper::initConnections(), web_view_ is nullptr!";
  }

  connect(app->keybinder, &Keybinder::onHotkey,
          this, &WinHelper::onHotkey);
  connect(app->keybinder, &Keybinder::onRegisterHotkeyResult,
          this, &WinHelper::onRegisterHotkeyResult);
  connect(app->keybinder, &Keybinder::onUnregisterHotkeyResult,
          this, &WinHelper::onUnregisterHotkeyResult);

  connect(app->popup_menu_manager, &PopupMenuManager::onMenuItemClicked,
          this, &WinHelper::onmenuclick);
  connect(app->popup_menu_manager, &PopupMenuManager::onMenuHide,
          this, &WinHelper::onmenuhide);

  connect(app->signal_manager, &SignalManager::onRequestMainWindowClose,
          this, &WinHelper::onclose);
}

void WinHelper::initMainWindow() const {
  web_view_->is_main_window = true;
}

void WinHelper::launchWindow(const QString& url,
                             const QVariantMap& position,
                             const QVariantMap& options) {
  Q_UNUSED(position);
  // TODO(xushaohua): move window to |position|.
  bool visible = options.value("visible", QVariant(false)).toBool();
  bool resizable = options.value("resizable", QVariant(false)).toBool();
  bool spec_window = options.value("spec_window", QVariant(false)).toBool();

  if (spec_window) {
    app->window_manager->launchWindow(url, visible, resizable, web_view_);
  } else {
    app->window_manager->launchWindow(url, visible, resizable, 0);
  }
}

void WinHelper::loadingFinished() {
  web_loading_finished_ = true;
  web_view_->show();

  // Play local music if needed.
  if (web_view_->is_main_window) {
    const QString local_file = GetMediaFileFromArgs();
    if (!local_file.isEmpty()) {
      emit app->signal_manager->onPlayLocalMusic(local_file);
    }
  }
}

void WinHelper::popupMenu(const QVariantMap& conf, int menu_type) {
  emit app->popup_menu_manager->popupMenu(conf, menu_type);
}

void WinHelper::registerHotkey(const QString& action_name,
                               const QVariantList& keys, bool is_global) {
  emit app->keybinder->registerHotkey(action_name, keys, is_global);
}

void WinHelper::setClipBoardData(const QString& content) const {
  QClipboard* clipboard = QApplication::clipboard();
  clipboard->setText(content);
}

void WinHelper::setFilesToClipboard(const QString& action,
                                    const QStringList& files) {
  QClipboard* clipboard = QApplication::clipboard();
  if (action == "copy") {
    QList<QUrl> uris;
    for (const QString& file : files) {
      uris.append(QUrl::fromLocalFile(file));
    }
    QMimeData* mime_data = new QMimeData();

    // Plain text
    mime_data->setText(files.join(QChar('\n')));

    // Normal file manager.
    mime_data->setUrls(uris);

    // For Gnome.
    QByteArray gnome_format("copy");
    for (const QUrl& uri : uris) {
      gnome_format.append(QChar('\n'));
      gnome_format.append(uri.toEncoded());
    }
    mime_data->setData("x-special/gnome-copied-files", gnome_format);
    clipboard->setMimeData(mime_data);
  }
}

void WinHelper::setNativeWindowShow() {
  qDebug() << Q_FUNC_INFO;
  web_view_->show();
  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void WinHelper::setNativeSubWindowShow(const QString& window_name,
                                       bool show,
                                       const QVariantMap& options) const {
  qDebug() << Q_FUNC_INFO << window_name << show << options;
  Q_UNUSED(options);
  app->window_manager->setNativeWindowShow(window_name, show);
  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void WinHelper::setWindowIconFromLocalFile(const QString& url) const {
  qDebug() << "TODO [WinHelper]::setWindowIconFromLocalFile():" << url;
}

void WinHelper::setWindowOpacity(double opacity) const {
  web_view_->setWindowOpacity(opacity);
}

void WinHelper::setWindowPosition(const QVariantMap& position) const {
  QVariant default_value(-1);
  const int x = position.value("x", default_value).toInt();
  const int y = position.value("y", default_value).toInt();
  const int width = position.value("width", default_value).toInt();
  const int height = position.value("height", default_value).toInt();
  const bool topmost = position.value("topmost").toBool();

  web_view_->move(x, y);

  // Ignore larger window resolution.
//  if (width > kDefaultMainWindowWidth) {
//    web_view_->moveToCenter();
//    width = qMin(kDefaultMainWindowWidth, width);
//    height = qMin(kDefaultMainWindowHeight, height);
//  }

  if (width > 0 && height > 0) {
    web_view_->resize(width, height);
  }

  web_view_->setTopMost(topmost);
}

void WinHelper::setWindowSizeLimit(const QVariantMap& min_size,
                                   const QVariantMap& max_size) const {
  QVariant default_value(-1);
  int min_width = min_size.value("x", default_value).toInt();
  int min_height = min_size.value("y", default_value).toInt();
  int max_width = max_size.value("x", default_value).toInt();
  int max_height = max_size.value("y", default_value).toInt();

  if (min_width > -1 && min_height > -1) {
    web_view_->setMinimumSize(min_width, min_height);
  }
  if (max_width > -1 && max_height > -1) {
    web_view_->setMaximumSize(max_width, max_height);
  }
}

void WinHelper::setWindowTitle(const QString& title) const {
  web_view_->setWindowTitle(title);
}

void WinHelper::setWindowZoomLevel(double level) const {
  web_view_->setZoomLevel(level);
}

void WinHelper::showWindow(const QString& state) {
  if (!web_loading_finished_) {
    qDebug() << "web page not fully loaded, ignore window state";
    return;
  }

  if (state == kWindowStateShow) {
    web_view_->show();
  } else if (state == kWindowStateHide) {
    web_view_->hide();
  } else if (state == kWindowStateMaximized) {
    web_view_->showMaximized();
  } else if (state == kWindowStateMinimized || state == kWindowStateZoom) {
    web_view_->showMinimized();
  } else if (state == kWindowStateRestore) {
    if (!first_restore_activated_) {
      first_restore_activated_ = true;
      web_view_->showMaximized();
    } else {
      web_view_->showNormal();
    }
  } else {
    qWarning() << "[WinHelper]::showWindow() unknown state: " << state;
    return;
  }

  emit app->signal_manager->onUpdateSystemTrayContextMenu();
}

void WinHelper::sizeWindow(const QString& position) const {
  web_view_->resizeWindow(position);
}

void WinHelper::unregisterHotkey(const QString& action_name, bool is_global) {
  emit app->keybinder->unregisterHotkey(action_name, is_global);
}

}  // namespace netease
