// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef NATIVE_WIN_HELPER_H_
#define NATIVE_WIN_HELPER_H_

#include <QObject>
#include <QVariant>

namespace netease {

class WebView;

// WinHelper is used to control native window and retrieve its properties,
// and creates new popup webview dialog.
// window.open() is disabled in webview by default, so call launchWindow() to
// open a new webview window and call destroyWindow() to close that window.
class WinHelper : public QObject {
  Q_OBJECT

 public:
  WinHelper(QObject* parent, WebView* web_view);

 signals:
  // Emitted when aero window style changed.
  // This signal is ignoed on linux plat`rm.
  void onAeroStyleChanged();

  // Emitted when main webview window is requested to close.
  void onclose();

  // Emitted when current webview window grabs focus.
  void onfocus();

  // Show preferences page.
  void ongotopreference();

  // Emitted when a registered hotkey is triggered.
  void onHotkey(const QString& action_name, bool is_global);

  // Emitted when current webview window loses focus.
  void onlosefocus();

  // Emitted when a menu item is clicked.
  void onmenuclick(const QString& menu_id);

  // Emitted when menu is hidden.
  void onmenuhide();

  // Emitted after calling registerHotkey().
  // |error_code| is an int error. 0 means success.
  void onRegisterHotkeyResult(const QString& action_name, bool is_global,
                              int error_code);

  // Emitted after calling showWindow().
  // |state| is the same as in showWinodw().
  // |width| is window width.
  // |height| is window height.
  void onSizeStatus(const QString& state, int width, int height);

  // Emitted after sizeWindow() is done.
  // |gemoetry| is an object containing:
  //   * top:
  //   * left:
  //   * bottom:
  //   * right:
  void onsizeWindowDone(const QVariantMap& gemoetry);

  // Emitted after calling unregisterHotkey().
  void onUnregisterHotkeyResult(const QString& action_name, bool is_global,
                                int error_code);

 public slots:
  // Raise main window.
  void bringWindowToTop() const;

  // Destroys this webview window.
  // NOTE(xushaohua): Only popup webview window calls this slot.
  // Main webview window calls "app.exit".
  void destroyWindow() const;

  // Make web_view window draggable.
  // It will emit ondragbegin() and ondragend() signals.
  void dragWindow() const;

  // Add fullscreen flag to current webview window.
  void enterFullscreen() const;

  // Remove fullscreen flag from current webview window.
  void exitFullscreen() const;

  // Get text content in clipboard.
  QString getClipBoardData() const;

  // Get properties of native window. |window_name| might be "mini_player" or
  // "desktop_lyrics".
  QVariantMap getNativeWindowRect(const QString& window_name) const;

  // Get window attributes.
  // If |info| is "status", returns an object including these properties:
  //   * "visible", window visibility.
  //   * "status", status might be "minimize", "maximize" or "restore".
  QVariantMap getWindowInfo(const QString& info) const;

  // Get current level of opacity of main window.
  double getWindowOpacity() const;

  // Get position of main window. Including:
  // e.g. {"x": 0, "y": 0, "width": 100, "height": 100, "topmost": false}.
  // "topmost" means stacking on top or not.
  QVariantMap getWindowPosition() const;

  void hidePopupMenu();

  // Setup this window as main window.
  void initMainWindow() const;

  // Open a new web page in a new window. Might use this method to open a
  // new popup dialog.
  // |url| is the url to open.
  // |position| is the new window position property, including x, y,
  // width and height.
  // |options| defines other properties of window, including:
  //   * "visible": window is visible or not, default is false.
  //   * "resizable": window is resiable, default is false.
  //   * "taskButton": stack in taskbar or not, default is false.
  //   * "spec_window": set transient to current window.
  void launchWindow(const QString& url, const QVariantMap& position,
                    const QVariantMap& options);

  // Show main window.
  void loadingFinished();

  // Popup context menu.
  void popupMenu(const QVariantMap& conf, int menu_type);

  // Register keyboard shortcut and emit onRegisterHotkeyResult() signal.
  // |action_name| is the action to be trigger when that key is pressed, like
  //   "play_pause".
  // |keys| is keycodes of keybinder group, like [17, 80] => "Ctrl+P".
  // If |is_global| is true, register it as global keybinding.
  void registerHotkey(const QString& action_name, const QVariantList& keys,
                      bool is_global);

  // Update data in clipboard with |content|.
  // Note that javascript in web context can also access system clipboard.
  void setClipBoardData(const QString& content) const;

  // Set |files| as contents of clipboard.
  // |files| are converted to URIs.
  void setFilesToClipboard(const QString& action, const QStringList& files);

  // Show current web view window.
  void setNativeWindowShow();

  // Change visibility of native window.
  // |window| might be "mini_player" or "desktop_lyrics".
  void setNativeSubWindowShow(const QString& window_name, bool show,
                              const QVariantMap& options) const;

  // Set window icon.
  // That icon is displayed in task bar and window navigation view.
  // |url| is a link with orpheus:// as scheme.
  void setWindowIconFromLocalFile(const QString& url) const;

  // Setup the level of opacity for main window.
  void setWindowOpacity(double opacity) const;

  // Update position properties of window.
  // |position| is described in getWindowPosition().
  void setWindowPosition(const QVariantMap& position) const;

  // Set size limitation of window.
  // |min_size| and |max_size| might be an object including
  // properties "x" and "y", or undefined if not set.
  void setWindowSizeLimit(const QVariantMap& min_size,
                          const QVariantMap& max_size) const;

  // Setup title of main window.
  void setWindowTitle(const QString& title) const;

  void setWindowZoomLevel(double level) const;

  // Update window display status and emit onSizeStatus() signal.
  // |state| might be:
  //   * "show", show window.
  //   * "hide", hide window.
  //   * "maximize", set window maximized.
  //   * "zoom", set window maximized on MacOS X.
  //   * "minimize", set window minimized.
  //   * "restore", restore window to normal size.
  void showWindow(const QString& state);

  // Start resizing window and emit onsizeWindowDone() signal.
  void sizeWindow(const QString& position) const;

  // Unregister keyboard shortcut and emit onUnergisterHotkeyResult() signal.
  void unregisterHotkey(const QString& action_name, bool is_global);

 private:
  // Initialize signal and slots.
  void initConnections() const;

  // WebView this object belongs to.
  WebView* web_view_ = nullptr;

  // Set this flag to true when web page is fully loaded.
  bool web_loading_finished_;

  bool first_restore_activated_ = false;

 private slots:
  // Handle WebView::onSizeWindowDone() signal, and emit onSizeWindowDone()
  // signal.
  void handleOnSizeWindowDone();

  void handleWindowStateSizeChanged();

  void sendWindowState();
};

}  // namespace netease

#endif  // NATIVE_WIN_HELPER_H_
