// Copyright (c) 2015 LiuLang. All rights reserved.
// Use of this source is governed by General Public License that can be found
// in the LICENSE file.

#include "service/backend/database_worker.h"

#include <QDebug>

#include "service/backend/chinese2pinyin.h"

namespace netease {

namespace {

int PinYinAsc(void* p_arg, int len1, const void* word1,
              int len2, const void* word2) {
  Q_UNUSED(p_arg);
  const QString w1(QByteArray(static_cast<const char*>(word1), len1));
  const QString w2(QByteArray(static_cast<const char*>(word2), len2));
  const QString c1 = Chinese2Pinyin(w1);
  const QString c2 = Chinese2Pinyin(w2);

  return c1.compare(c2);
}

int PinYinDesc(void* p_arg, int len1, const void* word1,
              int len2, const void* word2) {
  return (0 - PinYinAsc(p_arg, len1, word1, len2, word2));
}

}  // namespace

DatabaseWorker::DatabaseWorker(QObject* parent) : QObject(parent) {
}

DatabaseWorker::~DatabaseWorker() {
  commit();
  if (db_) {
    sqlite3_close(db_);
    db_ = nullptr;
  }
}

void DatabaseWorker::bindCollate() {
  if (db_) {
    sqlite3_create_collation(db_, "pinyin_asc", SQLITE_UTF8, nullptr, PinYinAsc);
    sqlite3_create_collation(db_, "pinyin_desc", SQLITE_UTF8, nullptr, PinYinDesc);
  }
}

void DatabaseWorker::commit() {
  if (db_) {
    sqlite3_exec(db_, "COMMIT", nullptr, nullptr, nullptr);
  }
}

bool DatabaseWorker::open(const QString &db_file) {
  if (db_) {
    sqlite3_close(db_);
  }
  const int state = sqlite3_open(db_file.toStdString().c_str(), &db_);
  if (state == SQLITE_OK) {
    bindCollate();
  }

  return state == SQLITE_OK;
}

QVariantList DatabaseWorker::rawExec(const QString& statements, bool* ok) {
  const QStringList statements_list = statements.split(QChar(';'));
  return exec(statements_list, ok);
}

QVariantList DatabaseWorker::exec(const QStringList& statements, bool* ok) {
  Q_ASSERT(ok);
  QVariantList result;

  if (!db_) {
    qWarning() << "[DatabaseWorker]::exec() db not open";
    *ok = false;
    return result;
  }

  sqlite3_stmt* stmt = nullptr;

  for (const QString& statement : statements) {
    const std::string statement_str = statement.toStdString();
    const char* query = statement_str.c_str();
    const int state = sqlite3_prepare_v2(db_, query, -1, &stmt, 0);

    if (state != SQLITE_OK) {
      // error.
      qWarning() << "[DatabaseWorker]::exec() error:" << statement
                 << QString(sqlite3_errmsg(db_));
      *ok = false;
      sqlite3_reset(stmt);
      commit();
      return result;
    }

    // Read columns.
    const int columns = sqlite3_column_count(stmt);
    while (sqlite3_step(stmt) == SQLITE_ROW) {
      QVariantMap row;
      for (int column = 0; column < columns; ++column) {
        const char* column_name = sqlite3_column_name(stmt, column);
        switch (sqlite3_column_type(stmt, column)) {
          case SQLITE_INTEGER: {
            const double value = sqlite3_column_double(stmt, column);
            row.insert(column_name, value);
            break;
          }

          case SQLITE_FLOAT: {
            const int value = sqlite3_column_int(stmt, column);
            row.insert(column_name, value);
            break;
          }

          case SQLITE_TEXT: {
            const unsigned char* value = sqlite3_column_text(stmt, column);
            row.insert(column_name, QString((const char*)value));
            break;
          }

          case SQLITE_BLOB: {
            const void* value = sqlite3_column_blob(stmt, column);
            row.insert(column_name, QString(static_cast<const char*>(value)));
            break;
          }

          case SQLITE_NULL: {
            row.insert(column_name, QVariant());
            break;
          }

          default: {
            // Ignored.
            break;
          }
        }
      }
      result.append(row);
    }
  }

  sqlite3_finalize(stmt);
  commit();
  *ok = true;

  return result;
}

void DatabaseWorker::execTransaction(const QString& statement, bool* ok) {
  if (!db_) {
    qWarning() << "[DatabaseWorker]::execTransaction() db not open";
    *ok = false;
    return;
  }
  QString transaction("begin transaction;");
  transaction.append(statement);
  transaction.append("commit transaction;");
  sqlite3_exec(db_, transaction.toStdString().c_str(),
               nullptr, nullptr, nullptr);
  *ok = true;
}

}  // namespace netease