// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/backend/download_worker.h"

#include <QDebug>
#include <QFile>
#include <QFileInfo>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QTimer>

#include "application.h"
#include "base/file_util.h"
#include "service/storage_manager.h"
#include "service/backend/download_song_info.h"

namespace netease {

namespace {

// Refresh data record each 1s.
const int kUpdateSpeedInterval = 1000;

const int kDownloadTypeOk = 0;
//const int kDownloadTypeDiskError = 2;  // No free space to write file.

}  // namespace

DownloadWorker::DownloadWorker(QObject* parent)
    : QObject(parent),
      download_status_(DownloadStatus::kStatusReady),
      task_(),
      task_info_() {
  this->setObjectName("download_worker");

  download_timer_ = new QTimer();
  download_timer_->setInterval(kUpdateSpeedInterval);
  download_timer_->start();

  this->initConnections();
}

DownloadWorker::~DownloadWorker() {
  // TODO(xushaohua): Pause tasks.
  // TODO(xushaohua): dump download_song_info.
  delete request_;
  request_ = nullptr;
  delete reply_;
  reply_ = nullptr;

  // Delete network access manager at end.
  delete network_manager_;
  network_manager_ = nullptr;

  delete file_handler_;
  file_handler_ = nullptr;

  download_timer_->stop();
  delete download_timer_;
  download_timer_ = nullptr;
}

void DownloadWorker::compareFileSize(qint64 server_file_size) {
  // File size shall be a positive integer.
  if (server_file_size <= 0) {
    qWarning() << "server file size is invalid:" << server_file_size;
    errorOccurred();
    return;
  }

  if (task_info_.exists() && QFile::exists(task_.temp_abs_rel_path)) {
    // Read old task info from disk.
    task_info_.read();
    // Both media resource file and info file exists.

    // File size matches.
    QFileInfo file_info(task_.temp_abs_rel_path);

    if (file_info.size() != task_info_.downloaded_size) {
      // Size mismatch, download from start.
      task_info_.downloaded_size = 0;

      // Remove old media resource file.
      QFile::remove(task_.temp_abs_rel_path);
      task_info_.remove();

      this->downloadFileContent();

    } else {
      // Size of downloaded data matches, start from current point.
      this->downloadFileContent();
    }
  } else {
    // Rename real filename if needed.
    if (QFile::exists(task_.abs_rel_path)) {
      task_.abs_rel_path = RenameDuplicatedFile(task_.abs_rel_path);
      const QString dirname =
          app->storage_manager->getDownloadDir().absolutePath();
      task_.rel_path = task_.abs_rel_path.mid(dirname.length());
    }

//    task_info_.setSongFilepath(task_.abs_rel_path);

    // Download from start.
    task_info_.total_size = server_file_size;
    task_info_.downloaded_size = 0;

    this->downloadFileContent();
  }
}

void DownloadWorker::downloadFileContent() {
  // Start timer.
  //download_timer_->start();

  delete file_handler_;
  file_handler_ = new QFile(task_.temp_abs_rel_path);
  bool ok = file_handler_->open(QIODevice::Append);
  if (! ok) {
    qCritical() << "[DownloadWorker]::startDownload()"
                << "Failed to open file: " << task_.id;
    errorOccurred();
    return;
  }

  request_->setUrl(QUrl(task_.url));
  QString range = "bytes=";
  range.append(QString::number(task_info_.downloaded_size));
  range.append("-");
  request_->setRawHeader(QByteArray("Range"), range.toUtf8());

  delete reply_;
  reply_ = nullptr;
  reply_ = network_manager_->get(*request_);
  connect(reply_, &QNetworkReply::finished,
          this, &DownloadWorker::handleFinished);
  connect(reply_, &QNetworkReply::readyRead,
          this, &DownloadWorker::handleReadyRead);
}

void DownloadWorker::errorOccurred() {
  errorOccurred(task_.id);
}

void DownloadWorker::errorOccurred(const QString &task_id) {
  download_status_ = DownloadStatus::kStatusError;
  emit this->onError(task_id);
  download_status_ = DownloadStatus::kStatusReady;
  // TODO(xushaohua): Clear resources.
}

void DownloadWorker::handleCancel(const QString& task_id) {
  if (task_id == task_.id) {
    download_status_ = DownloadStatus::kStatusCancel;
  }
}

void DownloadWorker::handleDownload(const DownloadTask& task) {
  if (!task.isValid()) {
    qCritical() << "[DownloadWorker]::handleDownload() invalid task"
                << task.id;
    errorOccurred(task.id);
    return;
  }

  if (!this->isReady()) {
    qCritical() << "[DownloadWorker]::doStartDownload():"
                << "current worker is still on";
    errorOccurred(task.id);
    return;
  }

  // Reset status.
  download_status_ = DownloadStatus::kStatusInProcess;
  task_ = task;
  download_speed_ = 0;

  // TODO(xushaohua): Check media resource file exists.

  task_info_.setSongFilepath(task_.temp_abs_rel_path);
  task_info_.md5 = task_.md5;

  this->queryRemoteFileSize();
}

void DownloadWorker::handleFinished() {
//  if (download_status_ == DownloadStatus::kStatusInProcess) {
//    file_handler_->close();
//    if (task_info_.downloaded_size == task_info_.total_size) {
//      task_info_.remove();
//      emit onDownloaded(task_.id);
//      download_status_ = DownloadStatus::kStatusFinished;
//    } else {
//      errorOccurred();
//    }
//  }
}

void DownloadWorker::handleInit() {
  Q_ASSERT(network_manager_ == nullptr);
  Q_ASSERT(request_ == nullptr);
  network_manager_ = new QNetworkAccessManager();
  request_ = new QNetworkRequest();
}

void DownloadWorker::handlePause(const QString& task_id) {
  if (task_id == task_.id) {
    download_status_ = DownloadStatus::kStatusPaused;
  }
}

void DownloadWorker::handleReadyRead() {
  const QByteArray data(reply_->readAll());

  switch (download_status_) {
  case DownloadStatus::kStatusInProcess: {
    const qint64 data_received = data.size();

    task_info_.downloaded_size += data_received;

    task_info_.write();

    file_handler_->write(data);

    // Update download speed.
    download_speed_ += data_received;

    // Send the last batch of data has been received.
    if (task_info_.downloaded_size >= task_info_.total_size) {
      this->sendDownloadProcessInfo(true);

      // Stop timer.
      //download_timer_->stop();
    }
    break;
  }

  case DownloadStatus::kStatusCancel: {
    // Remove temporary files.
    file_handler_->close();
    file_handler_->remove();
    task_info_.remove();

    // Close http request.
    reply_->abort();
    emit onCanceled(task_.id);
    download_status_ = DownloadStatus::kStatusReady;
    break;
  }

  case DownloadStatus::kStatusPaused: {
    file_handler_->close();
    task_info_.write();

    // Close http request.
    reply_->abort();
    download_status_ = DownloadStatus::kStatusReady;
    break;
  }

  default:
    // Ignored.
    download_status_ = DownloadStatus::kStatusReady;
    break;
  }
}

void DownloadWorker::handleRequestRemoteFileSize() {
  if (reply_->error() == QNetworkReply::NoError) {
    const int status =
        reply_->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

    switch (status) {
    case 200:
    case 206: {
      // Reset last signal handler.
      QObject::disconnect(reply_, &QNetworkReply::metaDataChanged,
                          this, &DownloadWorker::handleRequestRemoteFileSize);
      const qint64 server_file_size =
          reply_->header(QNetworkRequest::ContentLengthHeader).toLongLong();
      this->compareFileSize(server_file_size);
      break;
    }

    case 301:
    case 302:
    case 307: {
      // Handle http redirection.
      const QUrl url = reply_->attribute(
          QNetworkRequest::RedirectionTargetAttribute).toUrl();
      // Update new url location.
      task_.url = url.toString();

      QObject::disconnect(reply_, &QNetworkReply::metaDataChanged,
                          this, &DownloadWorker::handleRequestRemoteFileSize);

      // Send HEAD request again.
      request_->setUrl(QUrl(task_.url));
      delete reply_;
      reply_ = nullptr;
      reply_ = network_manager_->get(*request_);
      connect(reply_, &QNetworkReply::metaDataChanged,
              this, &DownloadWorker::handleRequestRemoteFileSize);
      break;
    }

    default: {
      // Unknown error.
      errorOccurred();
      break;
    }
    }

  } else {
    errorOccurred();
  }
}

void DownloadWorker::handleTimeout() {
  sendDownloadProcessInfo(false);
}

void DownloadWorker::initConnections() {
  connect(this, &DownloadWorker::cancel, this, &DownloadWorker::handleCancel);
  connect(this, &DownloadWorker::init, this, &DownloadWorker::handleInit);
  connect(this, &DownloadWorker::pause, this, &DownloadWorker::handlePause);
  connect(this, &DownloadWorker::start, this, &DownloadWorker::handleDownload);
  connect(download_timer_, &QTimer::timeout,
          this, &DownloadWorker::handleTimeout);
}

bool DownloadWorker::isReady() const {
  return download_status_ == DownloadStatus::kStatusReady;
}

void DownloadWorker::queryRemoteFileSize() {
  request_->setUrl(QUrl(task_.url));
  delete reply_;
  reply_ = nullptr;
  reply_ = network_manager_->get(*request_);
  connect(reply_, &QNetworkReply::metaDataChanged,
          this, &DownloadWorker::handleRequestRemoteFileSize);
}

void DownloadWorker::sendDownloadProcessInfo(bool force) {
  // FIXME(xushaohua): might cause segfault here.
  if (!force && download_status_ != DownloadStatus::kStatusInProcess) {
    return;
  }

  QVariantMap result;
  result.insert("type", kDownloadTypeOk);
  result.insert("down", task_info_.downloaded_size);
  result.insert("total", task_info_.total_size);
  result.insert("path", task_.rel_path);

  result.insert("speed", download_speed_);

  // Reset download speed.
  download_speed_ = 0;

  if ((task_info_.downloaded_size >= task_info_.total_size) &&
      (task_info_.total_size > 0)) {
    result.insert("islast", 1);
    download_status_ = DownloadStatus::kStatusReady;

    file_handler_->close();
    // Remove task info file.
    task_info_.remove();

    // Move media file from temp folder to download folder.
    QFile::rename(task_.temp_abs_rel_path, task_.abs_rel_path);
  } else {
    result.insert("islast", 0);
  }

  emit this->onProcess(task_.id, result);
}

}  // namespace netease
