// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_BACKEND_DOWNLOAD_WORKER_H_
#define SERVICE_BACKEND_DOWNLOAD_WORKER_H_

#include <QObject>
#include <QDataStream>
#include <QVariantMap>
class QFile;
class QNetworkAccessManager;
class QNetworkReply;
class QNetworkRequest;
class QTimer;

#include "service/backend/download_song_info.h"

namespace netease {

struct DownloadTask {
  QString id;  // task id.
  QString url;
  QString rel_path;
  QString abs_rel_path;  // Absolute filepath.
  QString temp_abs_rel_path;  // Absolute filepath to temp file.
  QString pre_path;
  QString abs_pre_path;
  QString md5;

  bool isValid() const {
    return (id.length() > 0 && url.length() > 0 && rel_path.length() > 0);
  }
};

// Download large files, like mp3 or mp4 files.
// Support breakpoint resumption.
class DownloadWorker : public QObject {
  Q_OBJECT

 public:
  explicit DownloadWorker(QObject* parent = nullptr);
  ~DownloadWorker() override;

  bool isReady() const;

 signals:
  // Emitted after task is canceled and temporary files are removed.
  void onCanceled(const QString& task_id);

  // Emitted when error occurs.
  void onError(const QString& id);

  // Emitted when new piece of data received.
  // |percent| is in range 0.0~1.0.
  void onProcess(const QString& id, const QVariantMap& result);

  // Emit this signal in UI thread to cancel current task.
  void cancel(const QString& task_id);

  void init();

  // Emit this signal in UI thread to pause current downloading task.
  void pause(const QString& task_id);

  // Emit this signal in UI thread to start downloading process.
  void start(const DownloadTask& task);

 private:
  void compareFileSize(qint64 server_file_size);

  // Download content of media resource file.
  void downloadFileContent();

  // TODO(xushaohua): handles error state.
  void errorOccurred();
  void errorOccurred(const QString &task_id);

  void initConnections();

  // Query file size at server.
  // If that size does not match file-size in info file, remove local old
  // resource file and download from start again.
  // TODO(xushaohua): file size is already existed in |task_info_|.
  void queryRemoteFileSize();

  // Send onProcess signal.
  // If |force| is true, ignores download status.
  void sendDownloadProcessInfo(bool force);

  enum class DownloadStatus {
    kStatusCancel = 0,
    kStatusInProcess = 1,
    kStatusPaused = 2,
    kStatusReady = 3,
    kStatusError = 4,
    kStatusFinished = 5,
  };

  DownloadStatus download_status_;

  DownloadTask task_;
  DownloadSongInfo task_info_;

  // File handler refering to rel_path_.
  QFile* file_handler_ = nullptr;

  QNetworkAccessManager* network_manager_ = nullptr;
  QNetworkReply* reply_ = nullptr;
  QNetworkRequest* request_ = nullptr;

  QTimer* download_timer_ = nullptr;
  qint64 download_speed_;

 private slots:
  // Handles onCancel() signal.
  void handleCancel(const QString& task_id);

  // Handles QNetworkReply::finished() signal.
  void handleFinished();

  void handleInit();

  // Handles onPause() siganl.
  void handlePause(const QString& task_id);

  // Handles QNetworkReply::readyRead() signal.
  // Also handles download status changing.
  void handleReadyRead();

  // Handles QNetworkAccessManager::head() in requestRemoteFileSize().
  void handleRequestRemoteFileSize();

  // Handles onStart() signal.
  // 1. Get server file size.
  // 2. Read local song info file.
  //    a. If info file does not exist, download from start.
  //    b. Else, compare server file size and file size in info file.
  //       * If these two values match, download from that point.
  //       * Else use the smaller value as downloading point.
  // 3. Start download file content.
  void handleDownload(const DownloadTask& task);

  // Send download process info.
  void handleTimeout();
};

}  // namespace netease

#endif  // SERVICE_BACKEND_DOWNLOAD_WORKER_H_
