// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/backend/filesystem_watcher.h"

#include <fcntl.h>
#include <sys/inotify.h>
#include <unistd.h>

#include <QDebug>
#include <QSocketNotifier>

namespace netease {

namespace {

// Read 10 inotify events each time.
const size_t kBufLength = 10 * (sizeof(struct inotify_event) + NAME_MAX + 1);

// Watch flags used for sub folders.
const uint32_t kWatchFlags = IN_DELETE  | IN_CREATE | IN_MOVE;

// Watch flags used for root folders.
const uint32_t kRootDirWatchFlags = kWatchFlags | IN_DELETE_SELF |
                                    IN_MOVE_SELF | IN_UNMOUNT;

}  // namespace

FilesystemWatcher::FilesystemWatcher(QObject* parent)
    : QObject(parent),
      inotify_fd_(-1),
      watch_items_() {
  this->setObjectName("filesystem_watcher");

  connect(this, &FilesystemWatcher::watchFolder,
          this, &FilesystemWatcher::handleWatchFolder);
  connect(this, &FilesystemWatcher::rmWatchFolder,
          this, &FilesystemWatcher::handleRemoveWatchFolder);
}

FilesystemWatcher::~FilesystemWatcher() {
  if (socket_notifier_ != nullptr) {
    delete socket_notifier_;
    socket_notifier_ = nullptr;
  }

  // Close inotify watchers.
  this->removeAllWatchItems();

  if (inotify_fd_ > 0) {
    close(inotify_fd_);
  }
}

void FilesystemWatcher::addWatchItem(int watch_id, const QString& root_dir,
                                     const QString& dir, bool recursive) {
  WatchItem item;
  item.watch_id = watch_id;
  item.root_dir = root_dir;
  item.dir = dir;
  item.recursive = recursive;
  watch_items_.append(item);
}

int FilesystemWatcher::getWatchItemById(int watch_id) {
  for (int i = 0; i < watch_items_.size(); i++) {
    if (watch_items_.at(i).watch_id == watch_id) {
      return i;
    }
  }
  return -1;
}

void FilesystemWatcher::removeAllWatchItems() {
  if (inotify_fd_ < 0) {
    return;
  }

  for (const auto& watch_item : watch_items_) {
    inotify_rm_watch(inotify_fd_, watch_item.watch_id);
  }
}

void FilesystemWatcher::removeWatchItemById(int watch_id) {
  Q_ASSERT(inotify_fd_ > 0);
  for (int i = 0; i < watch_items_.size(); i++) {
    if (watch_items_.at(i).watch_id == watch_id) {
      const int status =
          inotify_rm_watch(inotify_fd_, watch_id);
      if (status != 0) {
        qWarning() << Q_FUNC_INFO << watch_id << strerror(errno);
      }
      watch_items_.removeAt(i);
      return;
    }
  }
}

void FilesystemWatcher::removeWatchItemByPath(const QString& path) {
  Q_ASSERT(inotify_fd_ > 0);
  for (int i = 0; i < watch_items_.size(); ) {
    if (watch_items_.at(i).dir == path) {
      const int status =
          inotify_rm_watch(inotify_fd_, watch_items_.at(i).watch_id);
      if (status != 0) {
        qWarning() << Q_FUNC_INFO << path << strerror(errno);
      }
      watch_items_.removeAt(i);
      return;
    } else {
      i++;
    }
  }
}

void FilesystemWatcher::removeWatchItemsByParentDir(const QString& parent_dir) {
  Q_ASSERT(inotify_fd_ > 0);
  int state;
  const QString real_parent_dir(parent_dir + "/");
  for (int i = 0; i < watch_items_.size(); ) {
    if (watch_items_.at(i).dir.startsWith(real_parent_dir)) {
      state = inotify_rm_watch(inotify_fd_, watch_items_.at(i).watch_id);
      if (state != 0) {
        qWarning() << Q_FUNC_INFO << parent_dir << watch_items_.at(i).dir;
      }
      watch_items_.removeAt(i);
    } else {
      i++;
    }
  }
}

void FilesystemWatcher::removeWatchItemsByRootDir(const QString& root_dir) {
  Q_ASSERT(inotify_fd_ > 0);
  for (int i = 0; i < watch_items_.size(); ) {
    if (watch_items_.at(i).root_dir == root_dir) {
      int status = inotify_rm_watch(inotify_fd_, watch_items_.at(i).watch_id);
      if (status != 0) {
        qWarning() << Q_FUNC_INFO << "failed to rm inotify:" << root_dir;
      }
      watch_items_.removeAt(i);
    } else {
      i++;
    }
  }
}
void FilesystemWatcher::handleWatchFolder(const QString& root_dir,
                                          bool recursive) {
  this->doWatchFolder(root_dir, root_dir, recursive, true);
}

void FilesystemWatcher::doWatchFolder(const QString& root_dir,
                                      const QString& current_dir,
                                      bool recursive, bool initial) {
  this->initInotify();

  // Do nothing if root_dir is already monitored.
  for (const WatchItem& item : watch_items_) {
    if (item.dir == current_dir) {
      return;
    }
  }

  QDir dir(current_dir);
  if (!dir.exists()) {
    qWarning() << Q_FUNC_INFO << "Invalid folder:" << current_dir;
    if (initial) {
      emit this->onWatchFolder(root_dir, false);
    }
    return;
  }

  // Add to watch list.
  const int watch_id = inotify_add_watch(inotify_fd_,
      QFile::encodeName(current_dir),
      initial ? kRootDirWatchFlags : kWatchFlags);
  if (watch_id == -1) {
    qCritical() << Q_FUNC_INFO << "Failed to watch:" << root_dir
                << ", reason:" << strerror(errno);
    if (initial) {
      emit this->onWatchFolder(root_dir, false);
    }
    return;
  } else {
    this->addWatchItem(watch_id, root_dir, current_dir, recursive);

    if (recursive) {
      const bool ok = this->watchSubFolders(root_dir, current_dir);
      if (initial) {
        emit this->onWatchFolder(root_dir, ok);
      }
    } else if (initial) {
      emit this->onWatchFolder(root_dir, true);
    }
  }
}

void FilesystemWatcher::handleRemoveWatchFolder(const QString& root_dir) {
  Q_ASSERT(inotify_fd_ > 0);
  this->removeWatchItemsByRootDir(root_dir);
}

bool FilesystemWatcher::watchSubFolders(const QString& root_dir,
                                        const QString& parent_dir) {
  const QFileInfoList folder_list =
      QDir(parent_dir).entryInfoList(QDir::Dirs | QDir::NoDotAndDotDot);

  for (const QFileInfo& item : folder_list) {
    const QString item_dir = item.absoluteFilePath();
    const int watch_id = inotify_add_watch(inotify_fd_,
                                           item_dir.toStdString().c_str(),
                                           kWatchFlags);
    if (watch_id < 0) {
      qCritical() << Q_FUNC_INFO << "Failed to watch:" << item_dir
                  << ", reason:" << strerror(errno);
      return false;
    }

    // This method is called when in recursive mode.
    this->addWatchItem(watch_id, root_dir, item_dir, true);

    // Recursively.
    if (!this->watchSubFolders(root_dir, item_dir)) {
      return false;
    }
  }
  return true;
}

void FilesystemWatcher::handleInotifyActivated() {
  Q_ASSERT(inotify_fd_ > 0);
  char buf[kBufLength];
  char* barrier;
  ssize_t num_read = 0;
  struct inotify_event* event;

  num_read = read(inotify_fd_, buf, kBufLength);

  // If read() returns -1, no inotify_event is available currently.
  if (num_read <= 0) {
    return;
  }

  for (barrier = buf; barrier < buf + num_read; ) {
    event = (struct inotify_event*) barrier;
    const int index = getWatchItemById(event->wd);
    if (index == -1) {
//      if (event->mask & IN_IGNORED) {
//        qDebug() << "ignored:" << event->wd;
//      }
      barrier += sizeof(struct inotify_event) + event->len;
      continue;
    }

    const WatchItem watch_item = watch_items_.at(index);
    QDir dir(watch_item.dir);
    const QString filename = event->name;
    const QString abs_path = dir.absoluteFilePath(filename);

    if ((event->mask & IN_DELETE) || (event->mask & IN_MOVED_FROM)) {
      if (event->mask & IN_ISDIR) {
        // Remove current folder and sub-folders.
        this->removeWatchItemsByParentDir(abs_path);
        this->removeWatchItemByPath(abs_path);

        emit this->folderDeleted(watch_item.root_dir, abs_path);
      } else {
        // File is deleted.
        emit this->fileDeleted(watch_item.root_dir, abs_path);
      }
    } else if ((event->mask & IN_CREATE) || (event->mask & IN_MOVED_TO)) {
      if (event->mask & IN_ISDIR) {
        // Add current folder to watch list.
        if (watch_item.recursive) {
          // Watch current folder and sub folders.
          this->doWatchFolder(watch_item.root_dir, abs_path, true);
        }
        emit this->folderAdded(watch_item.root_dir, abs_path);
      } else {
        emit this->fileAdded(watch_item.root_dir, abs_path);
      }
    } else if (event->mask & (IN_MOVE_SELF | IN_DELETE_SELF | IN_UNMOUNT)) {
      this->removeWatchItemsByRootDir(watch_item.root_dir);
      emit this->folderDeleted(watch_item.root_dir, watch_item.root_dir);
    }

    barrier += sizeof(struct inotify_event) + event->len;
  }
}

bool FilesystemWatcher::initInotify() {
  if (inotify_fd_ < 0) {
    inotify_fd_ = inotify_init1(O_NONBLOCK);
    if ((inotify_fd_ > 0) && (socket_notifier_ == nullptr)) {
      socket_notifier_ = new QSocketNotifier(inotify_fd_,
                                             QSocketNotifier::Read);
      socket_notifier_->setEnabled(true);
      connect(socket_notifier_, &QSocketNotifier::activated,
              this, &FilesystemWatcher::handleInotifyActivated);
    }
  }
  Q_ASSERT(inotify_fd_ > 0);
  return inotify_fd_ > 0;
}

QDebug& operator<<(QDebug& debug, const FilesystemWatcher::WatchItem& item) {
  debug << "{" << item.watch_id << item.root_dir << item.dir << "}";
  return debug;
}

}  // namespace netease
