/*
 * Copyright (C) 2017 ~ 2017 Deepin Technology Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "service/backend/local_music_scanner.h"

#include <QDateTime>
#include <QDebug>
#include <QFileInfoList>

#include "base/file_util.h"
#include "service/media_metadata.h"

namespace netease {

namespace {

// Duration of media file shall be greater than 30s.
const long kMinimumMediaDuration = 30000;

// Separate artist names.
QString SplitArtistNames(const QString& artist) {
  QString result(',');
  // Artists are separated by '/' or '&' in most media files.
  for (const QString& item : artist.split(QRegExp("/|&"))) {
    result.append(item.trimmed());
    result.append(',');
  }
  return result;
}

// Supported media files:
// * ".aac",
// * ".ape",
// * ".cda",
// * ".flac",
// * ".m4a",
// * ".mp3",
// * ".ogg",
// * ".wav",
// * ".wma",
// * ".cue",
int ConvertMetadata(const QString& root_dir,
                    const QString& filepath,
                    LocalMediaRecordList& list) {
  if (IsMediaFile(filepath)) {
    AudioMetadata metadata = ReadAudioInfo(filepath);

    // Check media duration.
    if (metadata.duration < kMinimumMediaDuration) {
      qDebug() << "media duration too short, ignore";
      return 0;
    }

    QFileInfo fileinfo(filepath);
    const QString parent_dir(fileinfo.dir().absolutePath());
    LocalMediaRecord media_record;
    media_record.file = filepath;
    media_record.dir = root_dir;
    media_record.title = metadata.title;
    if (media_record.title.isEmpty()) {
      const QString filename = fileinfo.fileName();
      const int dot_index = filename.lastIndexOf(QChar('.'));
      if (dot_index > 0) {
        // Ignores hidden file.
        media_record.title = filename.left(dot_index);
      } else {
        media_record.title = filename;
      }
    }

    media_record.artist = SplitArtistNames(metadata.artist);
    media_record.album = metadata.album;
    media_record.duration = metadata.duration;
    media_record.timestamp = fileinfo.lastModified().toTime_t();
    media_record.bitrate = metadata.bitrate;
    media_record.file_size = fileinfo.size();
    media_record.id = GetUrlHash(filepath);
    media_record.parent_dir = parent_dir;
    media_record.library_path = root_dir;
    media_record.track_id = metadata.track_id;
    list.append(media_record);
    return 1;
  } else if (GetFileExtension(filepath) == "cue") {
    CueMetadata metadata = ReadCueInfo(filepath);

    int n_items_inserted = 0;
    const QFileInfo fileinfo(filepath);
    const QString parent_dir(fileinfo.dir().absolutePath());
    const uint timestamp = fileinfo.lastModified().toTime_t();
    for (const CueTrack& track : metadata.tracks) {
      // Check media duration.
      if (track.duration < kMinimumMediaDuration) {
        continue;
      }
      LocalMediaRecord media_record;
      media_record.album = metadata.title;
      media_record.artist = SplitArtistNames(track.performer);
      media_record.duration = track.duration;
      media_record.file = metadata.data_path;
      media_record.file_size = track.size;
      media_record.dir = root_dir;
      media_record.id = GetUrlHash(metadata.data_path + track.title);
      media_record.title = track.title;
      media_record.timestamp = timestamp;
      media_record.parent_dir = parent_dir;
      media_record.bitrate = 0;
      media_record.library_path = root_dir;
      media_record.source = metadata.cue_path;
      media_record.track_id = track.track_id;
      list.append(media_record);
    }
    return n_items_inserted;
  }
  return 0;
}

}  // namespace

LocalMusicScanner::LocalMusicScanner(QObject* parent) : QObject(parent) {
  connect(this, &LocalMusicScanner::scan,
          this, &LocalMusicScanner::doScan);
  connect(this, &LocalMusicScanner::scanAll,
          this, &LocalMusicScanner::doScanAll);
  qRegisterMetaType<LocalMediaRecord>("LocalMediaRecord");
  qRegisterMetaType<LocalMediaRecordList>("LocalMediaRecordList");
}

LocalMusicScanner::~LocalMusicScanner() {

}

void LocalMusicScanner::doScan(const QString& library_path,
                               const QString& root_dir,
                               const QStringList& files) {
  LocalMediaRecordList record_list;
  for (const QString& file : files) {
    ConvertMetadata(root_dir, file, record_list);
  }
  emit this->onScan(library_path, root_dir, files.size(), record_list);
}

void LocalMusicScanner::doScanAll(const QString& root_dir,
                                  const QStringList& files) {
  LocalMediaRecordList record_list;
  for (const QString& file : files) {
    ConvertMetadata(root_dir, file, record_list);
  }
  emit this->onScanAll(root_dir, record_list);
}

}  // namespace netease
