/*
 * Copyright (C) 2017 ~ 2017 Deepin Technology Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "service/backend/media_metadata_reader.h"

#include <attachedpictureframe.h>
#include <commentsframe.h>
#include <fileref.h>
#include <flacfile.h>
#include <id3v1tag.h>
#include <id3v2tag.h>
#include <mp4coverart.h>
#include <mp4file.h>
#include <mpegfile.h>
#include <textidentificationframe.h>
#include <xiphcomment.h>

#include <memory>
#include <QDebug>
#include <QTextCodec>
#include <QtCore/QRegularExpression>

#include "base/file_util.h"
#include "service/backend/aes.h"

extern "C" {
#include "third_party/libcue/libcue.h"
#include "third_party/libcue/cue_time.h"
}

namespace netease {

namespace {

const char kCommentPrefix[] = "163 key(Don't modify):";

// Convert frame number to time delta, used in cue file.
long CueFrameToTime(long frame) {
  int m, s, f;
  time_frame_to_msf(frame, &m, &s, &f);
  return (m * 60 + s) * 1000 + f * 40 / 3;
}

}  // namespace

QString EncryptMediaComment(const QString& comment) {
  QString result(kCommentPrefix);
  result.append(AES_Encrypt(comment));
  return result;
}

QString DecryptMediaComment(QString cipher_comment) {
  if (cipher_comment.startsWith(kCommentPrefix)) {
    const int pref_len = static_cast<int>(strlen(kCommentPrefix));
    cipher_comment = cipher_comment.mid(pref_len);
  }

  return AES_Decrypt(cipher_comment);
}

bool IsValidMediaComment(const QString& comment) {
  return comment.startsWith(kCommentPrefix);
}

AudioMetadata ReadAudioInfo(const QString& filepath) {
  AudioMetadata metadata;
  metadata.size = GetFileSize(filepath);

  const std::string std_filepath(filepath.toStdString());
  TagLib::FileRef tag_file(std_filepath.c_str());
  std::unique_ptr<TagLib::MP4::File> tag_mp4_file;

  TagLib::Tag* tag = nullptr;
  TagLib::AudioProperties* audioProperties = nullptr;

  if (!tag_file.isNull()) {
    if (tag_file.tag()) {
      tag = tag_file.tag();
    }
    if (tag_file.audioProperties()) {
      audioProperties = tag_file.audioProperties();
    }
  } else {
    tag_mp4_file.reset(new TagLib::MP4::File(std_filepath.c_str()));
    if (tag_mp4_file->isValid()) {
      if (tag_mp4_file->tag()) {
        tag = tag_mp4_file->tag();
      }
      if (tag_mp4_file->audioProperties()) {
        audioProperties = tag_mp4_file->audioProperties();
      }
    }
  }

  if (tag != nullptr) {
    // UTF8 encoded.
    metadata.album = TStringToQString(tag->album());
    metadata.artist = TStringToQString(tag->artist());
    metadata.title = TStringToQString(tag->title());
    metadata.comment = TStringToQString(tag->comment());

    // Normal CJK Unicode range:
    // See https://en.wikipedia.org/wiki/Unicode#4.0
//    QRegExp cjk("[\\x4e00-\\x9fff]");
//    // If filename contains Chinese, but characters in tag does not contain
//    // any Chinese character, use GB18030 encoding.
//    if (QFileInfo(filepath).fileName().contains(cjk) &&
//        ((!metadata.album.isEmpty() && !metadata.album.contains(cjk)) ||
//        (!metadata.artist.isEmpty() && !metadata.artist.contains(cjk)) ||
//        (!metadata.title.isEmpty() && !metadata.title.contains(cjk)))) {
//      // Localized encode, current only GB18030 is used.
//      QTextCodec* codec = QTextCodec::codecForName("GB18030");
//      metadata.album = codec->toUnicode(tag->album().toCString());
//      metadata.artist = codec->toUnicode(tag->artist().toCString());
//      metadata.title = codec->toUnicode(tag->title().toCString());
//      metadata.comment = codec->toUnicode(tag->comment().toCString());
//    }
  }

  if (metadata.title.isEmpty()) {
    // Use filename as title.
    metadata.title = GetFileBasename(filepath);
  }

  if (audioProperties != nullptr) {
    // seconds to milli seconds.
    metadata.duration = audioProperties->length() * 1000;
    metadata.bitrate = audioProperties->bitrate();
    metadata.sample_rate = audioProperties->sampleRate();
  }

  return metadata;
}

CueMetadata ReadCueInfo(const QString& filepath) {
  QDir parent = QFileInfo(filepath).absoluteDir();
  CueMetadata metadata;
  metadata.cue_path = filepath;

  Cd* cd = cue_parse_string(
      ReadGBKTextFileContent(filepath).toStdString().c_str());
  if (cd) {
    Cdtext* cd_text = cd_get_cdtext(cd);
    if (cd_text) {
      metadata.title = cdtext_get(PTI_TITLE, cd_text);
      metadata.performer = cdtext_get(PTI_PERFORMER, cd_text);
      metadata.genre = cdtext_get(PTI_GENRE, cd_text);
    }

    // TODO(xushaohua): Read bitrate info.

    Rem* cd_rem = cd_get_rem(cd);
    if (cd_rem) {
      metadata.date = rem_get(REM_DATE, cd_rem);
    }

    const int n_tracks = cd_get_ntrack(cd);
    for (int track_id = 0; track_id < n_tracks; ++track_id) {
      Track* track = cd_get_track(cd, track_id);

      if (track) {
        if (metadata.data_filename.isEmpty()) {
          metadata.data_filename = track_get_filename(track);
          metadata.data_path = parent.absoluteFilePath(metadata.data_filename);
          metadata.size = GetFileSize(metadata.data_path);
        }
        CueTrack cue_track;
        cue_track.track_id = track_id;
        cue_track.start = CueFrameToTime(track_get_start(track));
        cue_track.length = track_get_length(track);
        cue_track.duration = CueFrameToTime(cue_track.length);

        Cdtext* track_text = track_get_cdtext(track);
        if (track_text) {
          cue_track.performer = cdtext_get(PTI_PERFORMER, track_text);
          cue_track.title = cdtext_get(PTI_TITLE, track_text);
        }

        metadata.tracks.append(cue_track);
      }
      metadata.duration = 0;
      if (metadata.tracks.length() > 0) {
        CueTrack last_track = metadata.tracks.last();
        metadata.duration = last_track.start + last_track.duration;
      }

      // Update durations of each track.
      for (auto iter = metadata.tracks.begin(); iter != metadata.tracks.end();
           ++iter) {
        if (metadata.duration > 0) {
          (*iter).size = metadata.size * (*iter).duration / metadata.duration;
        }
      }
    }

    cd_delete(cd);
  } else {
    qCritical() << "ReadCueInfo() failed:" << filepath;
  }

  return metadata;
}

}  // namespace netease