// Copyright (c) 2015 Deepin. All rights reserved.
// Use of this source is governed by General Public License that can be found
// in the LICENSE file.

#include "service/backend/simple_download_manager.h"

#include <QDebug>
#include <QFile>

#include "service/backend/simple_download_worker.h"

namespace netease {

namespace {

const int kMaximumTasks = 5;
const int kDefaultTasks = 3;

}  // namespace

SimpleDownloadManager::SimpleDownloadManager(QObject* parent)
    : QObject(parent),
      tasks_(),
      workers_(),
      workers_inited_(false) {
  connect(this, &SimpleDownloadManager::download,
          this, &SimpleDownloadManager::handleDownload);
}

SimpleDownloadManager::~SimpleDownloadManager() {
}

void SimpleDownloadManager::initWorkers(int num_workers) {
  if (workers_inited_) {
    qWarning() << "[SimpleDownloadManager]::initWorkers() already inited!";
    return;
  }

  if (num_workers < 1 || num_workers > kMaximumTasks) {
    qWarning() << "[SimpleDownloadManager]::initWorkers() invalid num_workers:"
               << num_workers;
    num_workers = kDefaultTasks;
  }

  workers_inited_ = true;
  for (int i = 0; i < num_workers; ++i) {
    // |download_worker| will be freed in destructor of current class.
    SimpleDownloadWorker* download_worker = new SimpleDownloadWorker();
    connect(download_worker, &SimpleDownloadWorker::onDownloaded,
            this, &SimpleDownloadManager::handleWorkerDownloaded);
    connect(download_worker, &SimpleDownloadWorker::onError,
            this, &SimpleDownloadManager::handleWorkerError);

    Worker worker;
    worker.worker = download_worker;
    worker.ready = true;
    workers_.append(worker);
  }
}

void SimpleDownloadManager::handleWorkerDownloaded(const QString& url,
                                                   const QString& filepath) {
  this->resetWorkerByUrl(url, filepath);
  emit this->onDownloaded(url, filepath);
  this->loadNextTask();
}

void SimpleDownloadManager::handleWorkerError(const QString& url,
                                              const QString& filepath) {
  this->resetWorkerByUrl(url, filepath);
  emit this->onError(url, filepath);
  this->loadNextTask();
}

void SimpleDownloadManager::handleDownload(const QString& url,
                                           const QString& filepath) {
  if (!workers_inited_) {
    qWarning() << "[SimpleDownloadManager]::handleDownload() workers are not"
                << "initialized!";
    return;
  }

  // Ignores duplicated tasks.
  for (int i = 0; i < workers_.length(); ++i) {
    if ((url == workers_.at(i).url) && (filepath == workers_.at(i).filepath)) {
      return;
    }
  }

  for (int i = 0; i < tasks_.length(); ++i) {
    if ((url == tasks_.at(i).url) && filepath == tasks_.at(i).filepath) {
      return;
    }
  }

  // If file exists, do nothing.
  if (QFile::exists(filepath)) {
    emit this->onDownloaded(url, filepath);
    return;
  }

  // Add a new task.
  Task task;
  task.url = url;
  task.filepath = filepath;
  tasks_.append(task);

  this->loadNextTask();
}

void SimpleDownloadManager::loadNextTask() {
  for (int i = 0; i < workers_.length(); ++i) {
    if (tasks_.isEmpty()) {
      return;
    }

    if (workers_.at(i).ready) {
      // Consumes the oldest task.
      workers_[i].ready = false;
      workers_[i].filepath = tasks_.first().filepath;
      workers_[i].url = tasks_.first().url;
      emit workers_[i].worker->download(workers_[i].url,
                                        workers_[i].filepath);
      tasks_.removeFirst();
    }
  }
}

bool SimpleDownloadManager::resetWorkerByUrl(const QString& url,
                                             const QString& filepath) {
  for (int i = 0; i < workers_.length(); ++i) {
    if ((workers_.at(i).url == url) && workers_.at(i).filepath == filepath) {
      // Reset worker state.
      workers_[i].ready = true;
      workers_[i].url = "";
      return true;
    }
  }
  return false;
}

}  // namespace netease