// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/backend/simple_download_worker.h"

#include <QDebug>
#include <QFile>
#include <QNetworkAccessManager>
#include <QNetworkRequest>
#include <QThread>

namespace netease {

SimpleDownloadWorker::SimpleDownloadWorker(QObject* parent)
    : QObject(parent),
      filepath_(),
      url_() {

  network_manager_ = new QNetworkAccessManager();
  request_ = new QNetworkRequest();

  this->initConnections();
}

SimpleDownloadWorker::~SimpleDownloadWorker() {
  delete file_handler_;
  file_handler_ = nullptr;

  delete request_;
  request_ = nullptr;
  delete reply_;
  reply_ = nullptr;

  delete network_manager_;
  network_manager_ = nullptr;

  this->reset();
}

void SimpleDownloadWorker::handleError(QNetworkReply::NetworkError code) {
  qWarning() << "[SimpleDownloadWorker]::handleError()"
             << ", url: " << url_
             << ", code: " << code;

  emit onError(url_, filepath_);
  this->reset();
}

void SimpleDownloadWorker::handleFinished() {
  if (reply_->error() == QNetworkReply::NoError) {
    const int status =
        reply_->attribute(QNetworkRequest::HttpStatusCodeAttribute).toInt();

    switch (status) {
    case 200:
    case 206: {
      file_handler_->close();
      if (QFile::rename(tmp_filepath_, filepath_)) {
        emit onDownloaded(url_, filepath_);
        tmp_filepath_ = "";
      } else {
        qWarning() << "[SimpleDownloadWorker]::handleFinished()"
                   << "Failed to rename tmp file:" << tmp_filepath_;
        emit onError(url_, filepath_);
      }
      this->reset();
      break;
    }

    case 301:
    case 302:
    case 307: {
      // Handle http redirection.
      const QUrl url = reply_->attribute(
          QNetworkRequest::RedirectionTargetAttribute).toUrl();
      request_->setUrl(url);

      delete reply_;
      reply_ = network_manager_->get(*request_);
      connect(reply_, &QNetworkReply::finished,
              this, &SimpleDownloadWorker::handleFinished);
      connect(reply_, &QNetworkReply::readyRead,
              this, &SimpleDownloadWorker::handleReadyRead);
      break;
    }

    default: {
      qWarning() << "[SimpleDownloadWorker]::handleFinished() invalid http"
                 << " status: " << status;
      emit onError(url_, filepath_);
      this->reset();
    }
    }
  } else {
    qWarning() << "[SimpleDownloadWorker]::handleFinished() reply error:"
               << reply_->error();
    emit onError(url_, filepath_);
    this->reset();
  }
}

void SimpleDownloadWorker::handleReadyRead() {
  Q_ASSERT(reply_);
  Q_ASSERT(file_handler_);
  const QByteArray data(reply_->readAll());
  if (file_handler_) {
    file_handler_->write(data);
  }
}

void SimpleDownloadWorker::handleStartDownload(const QString& url,
                                               const QString& filepath) {
  if (url.isEmpty() || filepath.isEmpty()) {
    // Invalid task.
    emit this->onError(url, filepath);
    this->reset();
    return;
  }

  url_ = url;
  filepath_ = filepath;
  tmp_filepath_ = filepath + ".tmp";
  this->startDownload();
}

void SimpleDownloadWorker::initConnections() {
  connect(this, &SimpleDownloadWorker::download,
          this, &SimpleDownloadWorker::handleStartDownload);
}

void SimpleDownloadWorker::startDownload() {
  Q_ASSERT(network_manager_);
  Q_ASSERT(request_);

  delete file_handler_;
  file_handler_ = new QFile(tmp_filepath_);
  const bool ok = file_handler_->open(QIODevice::Append | QIODevice::Truncate);
  if (! ok) {
    qWarning() << "[SimpleDownloadWorker]::startDownload()"
               << "Failed to open file:" << tmp_filepath_;
    // No permission to write.
    emit onError(url_, filepath_);
    this->reset();
    return;
  }

  request_->setUrl(QUrl(url_));

  delete reply_;
  reply_ = network_manager_->get(*request_);
  connect(reply_, &QNetworkReply::finished,
          this, &SimpleDownloadWorker::handleFinished);
  connect(reply_, &QNetworkReply::readyRead,
          this, &SimpleDownloadWorker::handleReadyRead);
}

void SimpleDownloadWorker::reset() {
//  if (file_handler_ && file_handler_->isOpen()) {
//    file_handler_->close();
//    delete file_handler_;
//    file_handler_ = nullptr;
//  }

  if (!tmp_filepath_.isEmpty()) {
    QFile::remove(tmp_filepath_);
  }

  // Reset task info.
  tmp_filepath_ = "";
  filepath_ = "";
  url_ = "";
}

}  // namespace netease
