// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/backend/vlc_media_player.h"

#include <QDebug>

#include "base/consts.h"

namespace netease {

namespace {

const int kBufferingMaximum = 100;

}  // namespace

VlcMediaPlayer::VlcMediaPlayer(QObject* parent)
  : QObject(parent),
    vlc_instance_(libvlc_new(0, nullptr)),
    vlc_media_player_(nullptr),
    vlc_current_media_(nullptr),
    buffering_(0),
    media_(),
    event_types_({
      libvlc_MediaPlayerMediaChanged,
      libvlc_MediaPlayerNothingSpecial,
      libvlc_MediaPlayerOpening,
      libvlc_MediaPlayerBuffering,
      libvlc_MediaPlayerPlaying,
      libvlc_MediaPlayerPaused,
      libvlc_MediaPlayerStopped,
      libvlc_MediaPlayerForward,
      libvlc_MediaPlayerBackward,
      libvlc_MediaPlayerEndReached,
      libvlc_MediaPlayerEncounteredError,
      libvlc_MediaPlayerTimeChanged,
      libvlc_MediaPlayerPositionChanged,
      libvlc_MediaPlayerSeekableChanged,
      libvlc_MediaPlayerPausableChanged,
      libvlc_MediaPlayerTitleChanged,
      libvlc_MediaPlayerSnapshotTaken,
      libvlc_MediaPlayerLengthChanged,
      libvlc_MediaPlayerVout,
      libvlc_MediaPlayerScrambledChanged,
      libvlc_MediaPlayerUnmuted,
      libvlc_MediaPlayerMuted,
      libvlc_MediaPlayerAudioVolume,
    }) {

  Q_ASSERT(vlc_instance_ != nullptr);
  libvlc_set_user_agent(vlc_instance_, kAppDisplayName, kAppName);
  libvlc_set_app_id(vlc_instance_, kAppName, kAppVersion, kAppName);

  // Initialize media player.
  // TODO(xushaohua): returns NULL on error.
  vlc_media_player_ = libvlc_media_player_new(vlc_instance_);

  qRegisterMetaType<MediaStatus>("MediaStatus");
  qRegisterMetaType<State>("State");

  this->registerVlcEventHandlers();
}

VlcMediaPlayer::~VlcMediaPlayer() {
  this->unregisterVlcEventHandlers();

  libvlc_media_release(vlc_current_media_);
  libvlc_media_player_release(vlc_media_player_);
  libvlc_release(vlc_instance_);
}

bool VlcMediaPlayer::audioAvailable() const {
  // TODO(xushaohua):
  libvlc_state_t state = libvlc_media_player_get_state(vlc_media_player_);
  return state == libvlc_Playing;
}

int VlcMediaPlayer::bufferStatus() const {
  return buffering_;
}

qint64 VlcMediaPlayer::duration() const {
  return libvlc_media_player_get_length(vlc_media_player_);
}

bool VlcMediaPlayer::isMuted() const {
  return libvlc_audio_get_mute(vlc_media_player_);
}

bool VlcMediaPlayer::isSeekable() const {
  return libvlc_media_player_is_seekable(vlc_media_player_);
}

const QUrl& VlcMediaPlayer::media() const {
  return media_;
}

VlcMediaPlayer::MediaStatus VlcMediaPlayer::mediaStatus() const {
  libvlc_state_t state = libvlc_media_player_get_state(vlc_media_player_);
  // TODO(xushaohua): handles other state.
  switch (state) {
    case libvlc_NothingSpecial: return MediaStatus::NoMedia;
    case libvlc_Ended:          return MediaStatus::EndOfMedia;
    case libvlc_Error:          return MediaStatus::InvalidMedia;
    case libvlc_Buffering:      return MediaStatus::BufferingMedia;
    case libvlc_Opening:        return MediaStatus::BufferingMedia;
    default:                    return MediaStatus::InvalidMedia;
  }
}

void VlcMediaPlayer::pause() {
  libvlc_media_player_pause(vlc_media_player_);
}

void VlcMediaPlayer::play() {
  qDebug() << Q_FUNC_INFO;
  libvlc_media_player_play(vlc_media_player_);
}

qint64 VlcMediaPlayer::position() const {
  const qint64 pos = libvlc_media_player_get_time(vlc_media_player_) * 1000;
  return pos;
}

void VlcMediaPlayer::registerVlcEventHandlers() {
  libvlc_event_manager_t* event_manager =
      libvlc_media_player_event_manager(vlc_media_player_);
  if (event_manager != nullptr) {
    for (const libvlc_event_e& event_type : event_types_) {
      libvlc_event_attach(event_manager, event_type, vlcEventCallback, this);
    }
  } else {
    qCritical() << "[VlcMediaPlayer]::registerVlcEventHandler()"
                << "invalid event manager!";
  }
}

void VlcMediaPlayer::setMedia(const QUrl& url) {
  qDebug() << Q_FUNC_INFO << url;
  media_ = url;
  buffering_ = 0;

  libvlc_media_release(vlc_current_media_);
  const std::string url_str(url.url().toStdString());
  vlc_current_media_ = libvlc_media_new_location(vlc_instance_,
                                                 url_str.c_str());
  libvlc_media_player_set_media(vlc_media_player_, vlc_current_media_);
}

void VlcMediaPlayer::setMuted(bool muted) {
  qDebug() << "[VlcMediaPlayer]::setMuted()" << muted;
  libvlc_audio_set_mute(vlc_media_player_, muted);
}

void VlcMediaPlayer::setPosition(qint64 position) {
  qDebug() << Q_FUNC_INFO << position << media_;
  libvlc_media_player_set_time(vlc_media_player_,
                               static_cast<libvlc_time_t>(position));
}

void VlcMediaPlayer::setVolume(int volume) {
  libvlc_audio_set_volume(vlc_media_player_, volume);
}

VlcMediaPlayer::State VlcMediaPlayer::state() const {
  libvlc_state_t state = libvlc_media_player_get_state(vlc_media_player_);
  switch (state) {
  case libvlc_Playing:   return VlcMediaPlayer::PlayingState;
  case libvlc_Opening:   return VlcMediaPlayer::PlayingState;
  case libvlc_Buffering: return VlcMediaPlayer::PlayingState;
  case libvlc_Paused:    return VlcMediaPlayer::PausedState;
  default:               return VlcMediaPlayer::StoppedState;
  }
}

void VlcMediaPlayer::stop() {
  libvlc_media_player_stop(vlc_media_player_);
}

void VlcMediaPlayer::unregisterVlcEventHandlers() {
  libvlc_event_manager_t* event_manager =
      libvlc_media_player_event_manager(vlc_media_player_);
  if (event_manager != nullptr) {
    for (const libvlc_event_e& event_type : event_types_) {
      libvlc_event_detach(event_manager, event_type, vlcEventCallback, this);
    }
  } else {
    qCritical() << Q_FUNC_INFO << "invalid event manager!";
  }
}

// static
void VlcMediaPlayer::vlcEventCallback(const libvlc_event_t* event, void* data) {
  VlcMediaPlayer* player = static_cast<VlcMediaPlayer*>(data);
  Q_ASSERT(player);

  switch (event->type) {
    case libvlc_MediaPlayerMediaChanged: {
      emit player->mediaStatusChanged(MediaStatus::BufferingMedia);
      break;
    }

    case libvlc_MediaPlayerOpening: {
      emit player->mediaStatusChanged(MediaStatus::BufferingMedia);
      break;
    }

    case libvlc_MediaPlayerBuffering: {
      player->buffering_ = static_cast<int>(
          event->u.media_player_buffering.new_cache);
      emit player->bufferStatusChanged(player->buffering_);

      if (player->buffering_ >= kBufferingMaximum) {
        emit player->mediaStatusChanged(MediaStatus::BufferedMedia);
      } else {
        emit player->mediaStatusChanged(MediaStatus::BufferingMedia);
      }
      break;
    }

    case libvlc_MediaPlayerPlaying: {
      emit player->stateChanged(State::PlayingState);
      break;
    }

    case libvlc_MediaPlayerPaused: {
      emit player->stateChanged(State::PausedState);
      break;
    }

    case libvlc_MediaPlayerStopped: {
      emit player->stateChanged(State::StoppedState);
      break;
    }

    case libvlc_MediaPlayerEndReached: {
      emit player->mediaStatusChanged(MediaStatus::EndOfMedia);
      break;
    }

    case libvlc_MediaPlayerEncounteredError: {
      qDebug() << Q_FUNC_INFO << "Invalid Media";
      emit player->mediaStatusChanged(MediaStatus::InvalidMedia);
      break;
    }

    case libvlc_MediaPlayerMuted: {
      emit player->mutedChanged(true);
      break;
    }

    case libvlc_MediaPlayerUnmuted: {
      emit player->mutedChanged(false);
      break;
    }

    case libvlc_MediaPlayerAudioVolume: {
      emit player->volumeChanged(player->volume());
      break;
    }

    case libvlc_MediaPlayerLengthChanged: {
      emit player->durationChanged(
          event->u.media_player_length_changed.new_length);
      break;
    }

//    case libvlc_MediaPlayerPositionChanged: {
//      const float percent = event->u.media_player_position_changed.new_position;
//      const qint64 position = static_cast<qint64>(
//          percent * player->duration() / 100);
//      emit player->positionChanged(position);
//      break;
//    }

    case libvlc_MediaPlayerSeekableChanged: {
      emit player->seekableChanged(
          event->u.media_player_seekable_changed.new_seekable != 0);
      break;
    }

    case libvlc_MediaPlayerTimeChanged: {
      emit player->positionChanged(event->u.media_player_time_changed.new_time);
      break;
    }

    default: {
      break;
    }
  }
}

int VlcMediaPlayer::volume() const {
  return libvlc_audio_get_volume(vlc_media_player_);
}

}  // namespace netease
