// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_BACKEND_VLC_MEDIA_PLAYER_H_
#define SERVICE_BACKEND_VLC_MEDIA_PLAYER_H_

#include <QObject>
#include <QUrl>
#include <QVector>
#include <vlc/vlc.h>

namespace netease {

// This class wraps up libvlc.
class VlcMediaPlayer : public QObject {
  Q_OBJECT
  Q_PROPERTY(bool audioAvailable READ audioAvailable
                 NOTIFY audioAvailableChanged)
  Q_PROPERTY(int bufferStatus READ bufferStatus NOTIFY bufferStatusChanged)
  Q_PROPERTY(qint64 duration READ duration NOTIFY durationChanged)
  Q_PROPERTY(MediaStatus mediaStatus READ mediaStatus
                 NOTIFY mediaStatusChanged)
  Q_PROPERTY(bool muted READ isMuted WRITE setMuted NOTIFY mutedChanged)
  Q_PROPERTY(qint64 position READ position WRITE setPosition
                 NOTIFY positionChanged)
  Q_PROPERTY(bool seekable READ isSeekable NOTIFY seekableChanged)
  Q_PROPERTY(State state READ state NOTIFY stateChanged)
  Q_PROPERTY(int volume READ volume WRITE setVolume NOTIFY volumeChanged)
  Q_PROPERTY(QUrl media READ media WRITE setMedia NOTIFY mediaChanged)

 public:
  explicit VlcMediaPlayer(QObject* parent = nullptr);
  ~VlcMediaPlayer();

  enum MediaStatus {
    // No media specified.
    NoMedia,

    // Current media resource is being buffered.
    BufferingMedia,

    // Current media resource is fully buffered.
    BufferedMedia,

    // Reaches end of media resource.
    EndOfMedia,

    // TODO(LiuLang):
    LoadedMedia,

    // Failed to parse media resource.
    InvalidMedia
  };
  Q_ENUM(MediaStatus)

  // Keep compatible with QMediaPlayer::State
  enum State {
    StoppedState,
    PlayingState,
    PausedState
  };
  Q_ENUM(State)

  bool audioAvailable() const;
  int bufferStatus() const;
  qint64 duration() const;
  bool isMuted() const;
  bool isSeekable() const;
  const QUrl& media() const;
  MediaStatus mediaStatus() const;
  qint64 position() const;
  State state() const;
  int volume() const;

 signals:
  void audioAvailableChanged(bool available);
  void bufferStatusChanged(int percent_filled);
  void durationChanged(qint64 duration);
  //void error(Error error);
  void mediaChanged(const QUrl& url);
  void mediaStatusChanged(MediaStatus status);
  void mutedChanged(bool muted);
  void positionChanged(qint64 position);
  void seekableChanged(bool seekable);
  void stateChanged(State state);
  void volumeChanged(int volume);

 public slots:
  void pause();
  void play();
  void setMedia(const QUrl& url);
  void setMuted(bool muted);

  // Seek to new position.
  void setPosition(qint64 position);

  void setVolume(int volume);

  void stop();

 private:
  // Attach vlc event handlers.
  void registerVlcEventHandlers();
  // Detach vlc event handlers.
  void unregisterVlcEventHandlers();
  // vlc event handler.

  // |data| is a VlcMediaPlayer instance.
  static void vlcEventCallback(const libvlc_event_t* event, void* data);

  libvlc_instance_t* vlc_instance_;
  libvlc_media_player_t* vlc_media_player_;
  libvlc_media_t* vlc_current_media_;

  // Percent of media data buffered, 0~100.
  int buffering_;

  QUrl media_;

  const QVector<libvlc_event_e> event_types_;
};

}  // namespace netease

#endif  // SERVICE_BACKEND_VLC_MEDIA_PLAYER_H_
