// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/dbus_manager.h"

#include "dbus/mpris_adapter.h"
#include "dbus/network_manager_interface.h"

namespace netease {

namespace {

//  NM_STATE_UNKNOWN = 0
//    Networking state is unknown.
//  NM_STATE_ASLEEP = 10
//    Networking is inactive and all devices are disabled.
//  NM_STATE_DISCONNECTED = 20
//    There is no active network connection.
//  NM_STATE_DISCONNECTING = 30
//    Network connections are being cleaned up.
//  NM_STATE_CONNECTING = 40
//    A network device is connecting to a network and there is no other available network connection.
//  NM_STATE_CONNECTED_LOCAL = 50
//    A network device is connected, but there is only link-local connectivity.
//  NM_STATE_CONNECTED_SITE = 60
//    A network device is connected, but there is only site-local connectivity.
//  NM_STATE_CONNECTED_GLOBAL = 70
//    A network device is connected, with global network connectivity.
const int kNMStateConnected = 70;
const int kNMStateDisconnected = 20;

}  // namespace

bool IsApplicationRunning() {
  QDBusInterface interface(MprisAdapter::kServiceName,
                           MprisAdapter::kServicePath,
                           MprisAdapter::kRootInterface,
                           QDBusConnection::sessionBus());
  if (interface.isValid()) {
    interface.asyncCall("Raise");
    return true;
  }
  return false;
}

bool NotifyPlayingLocalFile(const QString& filepath) {
  QDBusInterface interface(MprisAdapter::kServiceName,
                           MprisAdapter::kServicePath,
                           MprisAdapter::kPlayerInterface,
                           QDBusConnection::sessionBus());
  if (interface.isValid()) {
    QFileInfo info(filepath);
    const QUrl url(QUrl::fromLocalFile(info.canonicalFilePath()));
    interface.asyncCall("OpenUri", url.toString());
    return true;
  }
  return false;
}

DBusManager::DBusManager(QObject* parent)
    : QObject(parent),
      mpris_adapter_obj_(new MprisAdapter(this)),
      nm_iface_(new NetworkManagerInterface(this)) {
  setObjectName("dbus_manager");

  connect(nm_iface_, &NetworkManagerInterface::StateChanged,
          this, &DBusManager::handleNMStateChanged);
}

void DBusManager::handleNMStateChanged(uint state) {
  if (state == kNMStateConnected) {
    emit connectivityChanged(true);
  } else if (state == kNMStateDisconnected) {
    emit connectivityChanged(false);
  }
  // Ignores other states.
}

bool DBusManager::isOnline() {
  QDBusPendingReply<uint> reply = nm_iface_->state();
  reply.waitForFinished();
  if (reply.isValid()) {
    return reply.value() == kNMStateConnected;
  }
  // Ignores other states.
  return true;
}

void DBusManager::unRegister() {
  if (mpris_adapter_obj_ != nullptr) {
    delete mpris_adapter_obj_;
    mpris_adapter_obj_ = nullptr;
  }
}

}  // namespace netease