// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/default_client.h"

#include <glib.h>
#include <QCoreApplication>
#include <QDir>
#include <QSettings>

#include "base/consts.h"
#include "base/file_util.h"
#include "third_party/simpleini/SimpleIni.h"

namespace netease {

namespace {

// Config file used to setup default applications for current desktop.
// That is $HOME/.config/mimeapps.list.
const char kMimeAppFileName[] = "mimeapps.list";

// Section names in mimeapps.list.
const char kDefaultApplicationsSection[] = "Default Applications";
const char kAddedAssociationsSection[] = "Added Associations";

// Mime list supported by player.
const int kAudioMimeNameLength = 13;
const char kAudioMimeNames[kAudioMimeNameLength][20] = {
    "audio/aac",
    "audio/flac",
    "audio/mp3",
    "audio/mp4",  // m4a
    "audio/mpeg",
    "audio/ogg",
    "audio/x-ape",
    "audio/x-flac",
    "audio/x-mp3",
    "audio/x-mpeg",
    "audio/x-ms-wma",  // wma
    "audio/x-vorbis",
    "audio/x-wav",
};

QString GetMimeAppPath() {
#ifndef GLIB_VERSION_2_42
  // Placed at $HOME/.local/share/applications/mimeapps.list in glib (<=2.40)
  QDir config_dir(QDir::home().filePath(".local/share/applications"));
#else
  // Placed at $HOME/.config/mimeapps.list in glib (>= 2.42).
  QDir config_dir(QDir::home().filePath(".config"));
#endif
  return config_dir.absoluteFilePath(kMimeAppFileName);
}

}  // namespace

bool IsDefaultClient() {
  const QString mime_app_filepath(GetMimeAppPath());

  // If mimeapps.list file not found, returns false.
  if (!QFile::exists(mime_app_filepath)) {
    return false;
  }

  bool state = true;

  QSettings settings(mime_app_filepath, QSettings::IniFormat);

  settings.beginGroup(kDefaultApplicationsSection);
  for (int i = 0; i < kAudioMimeNameLength; ++i) {
    const QString app_name = settings.value(kAudioMimeNames[i]).toString();
    state &= (app_name == kAppDesktopFile);
  }
  settings.endGroup();
  return state;
}

// SimpleIni is used instead of QSettings, because QSettings escapes
// section names and keys, which is not what we need.
bool SetDefaultClient(bool is_default) {
  const QString mime_app_filepath(GetMimeAppPath());

  // If mimeapp.list file not found, returns false.
  if (!is_default && !QFile::exists(mime_app_filepath)) {
    return false;
  }

  CSimpleIniA settings;
  settings.SetUnicode(true);
  QString content(ReadTextFileContent(mime_app_filepath));
  settings.LoadData(content.toStdString().c_str(), content.length());

  for (int i = 0; i < kAudioMimeNameLength; ++i) {
    const char* mime_name = kAudioMimeNames[i];
    if (is_default) {
      // Set as default music client.
      settings.SetValue(kDefaultApplicationsSection,
                        mime_name,
                        kAppDesktopFile);
      settings.SetValue(kAddedAssociationsSection,
                        mime_name,
                        kAppDesktopFile);
    } else {
      const QString app_name =
          settings.GetValue(kDefaultApplicationsSection, mime_name);
      // Reset to system default music program.
      if (app_name == kAppDesktopFile) {
        settings.Delete(kDefaultApplicationsSection, mime_name);
      }

      const QString app_name2 =
          settings.GetValue(kAddedAssociationsSection, mime_name);
      if (app_name2 == kAppDesktopFile) {
        settings.Delete(kAddedAssociationsSection, mime_name);
      }
    }
  }
  std::string strData;
  settings.Save(strData);
  return WriteTextFile(mime_app_filepath, QString::fromStdString(strData));
}

}  // namespace netease
