// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/download_manager.h"

#include <QDebug>
#include <QFileInfo>
#include <QThread>

#include "application.h"
#include "base/file_util.h"
#include "service/storage_manager.h"
#include "service/backend/download_worker.h"

namespace netease {

DownloadManager::DownloadManager(QObject* parent)
    : QObject(parent),
      workers_() {
  this->setObjectName("download_manager");
  qRegisterMetaType<DownloadTask>("DownloadTask");

  this->initConnections();
}

DownloadManager::~DownloadManager() {
  for (DownloadWorker* worker : workers_) {
    delete worker;
  }
  workers_.clear();
}

void DownloadManager::handleDownload(const QVariantMap& task_info) {
  qDebug() << Q_FUNC_INFO << task_info;
  DownloadTask task;
  task.id = task_info.value("id", "").toString();
  task.url = task_info.value("url", "").toString();

  // Concat download path.
  task.rel_path = task_info.value("rel_path", "").toString();
  task.abs_rel_path =
      app->storage_manager->getDownloadDir().absoluteFilePath(task.rel_path);
  task.pre_path = task_info.value("pre_path", "").toString();
  if (task.pre_path.length() > 0) {
    task.abs_pre_path =
        app->storage_manager->getDownloadDir().absoluteFilePath(task.pre_path);
  } else {
    task.abs_pre_path = "";
  }

  // Create parent folders if needed.
  CreateParentDirs(task.abs_rel_path);

  task.md5 = task_info.value("md5", "").toString();

  task.temp_abs_rel_path =
      app->storage_manager->getTempFilepath(task.abs_rel_path);

  if (!task.isValid()) {
    qCritical() << "[DownloadManager]::download() invalid task:" << task.id;
    emit this->onError(task.id);
    return;
  }

  for (DownloadWorker* worker : workers_) {
    if (worker->isReady()) {
      emit worker->start(task);
      return;
    }
  }

  // Create a new worker if all of current workers are busy.
  DownloadWorker* worker = new DownloadWorker();
  workers_.append(worker);

  // Use high priority thread.
  // FIXME(LiuLang): Move to background thread.
//  worker->moveToThread(app->thread_manager->high_priority_thread);

  // Initialize internal objects.
  emit worker->init();

  connect(worker, &DownloadWorker::onCanceled,
          this, &DownloadManager::onCanceled);
  connect(worker, &DownloadWorker::onError,
          this, &DownloadManager::onError);
  connect(worker, &DownloadWorker::onProcess,
          this, &DownloadManager::onProcess);
  connect(this, &DownloadManager::cancel,
          worker, &DownloadWorker::cancel);
  connect(this, &DownloadManager::pause,
          worker, &DownloadWorker::pause);

  emit worker->start(task);
}

void DownloadManager::initConnections() {
  connect(this, &DownloadManager::download,
          this, &DownloadManager::handleDownload);
  connect(this, &DownloadManager::queryDownloadSchedule,
          this, &DownloadManager::handleQueryDownloadSchedule);
}

void
DownloadManager::handleQueryDownloadSchedule(const QVariantList& task_list) {
  QDir dir = app->storage_manager->getDownloadDir();
  DownloadSongInfo info_manager;
  QVariantList result;
  for (const QVariant& task : task_list) {
    const auto map = task.toMap();
    const QString id = map.value("id").toString();
    const QString path = map.value("path").toString();
    if (id.isEmpty() || path.isEmpty()) {
      continue;
    }

    const QString abspath = dir.absoluteFilePath(path);
    const QString temp_path = app->storage_manager->getTempFilepath(abspath);
    info_manager.setSongFilepath(temp_path);
    info_manager.read();
    const double percent = info_manager.downloaded_size * 1.0 /
        info_manager.total_size;

    QVariantMap info_map;
    info_map.insert("id", id);
    info_map.insert("shecdule", percent);
    result.append(info_map);
  }
  emit this->onQueryDownloadSchedule(result);
}

}  // namespace netease
