// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/feedback_manager.h"

#include <QDebug>
#include <QFile>
#include <QNetworkAccessManager>
#include <QHttpPart>
#include <QNetworkReply>
#include <QNetworkRequest>

#include "application.h"
#include "base/consts.h"
#include "base/device_util.h"
#include "service/backend/zip_wrapper.h"
#include "service/log_manager.h"
#include "service/url_manager.h"

namespace netease {

namespace {

const char kFormDataClientVersion[] = "form-data; name=\"clientVersion\"";
const char kFormDataContact[] = "form-data; name=\"contact\"";
const char kFormDataContent[] = "form-data; name=\"content\"";
const char kFormDataCpu[] = "form-data; name=\"cpu\"";
const char kFormDataOsVersion[] = "form-data; name=\"osVersion\"";
const char kFormDataResolution[] = "form-data; name=\"resolution\"";
const char kFormDataTitle[] = "form-data; name=\"title\"";

const char kFormDataLogFile[] =
    "form-data;name=\"file\";filename=\"log.zip\"";
const char kFormDataStatisFile[] =
    "form-data; name=\"file\"; filename=\"statis.zip\"";

const char kMimeTypeZip[] = "application/zip";

const char kZipFileExtensionName[] = ".zip";

QString GetZipFilepath(const QString& filepath) {
  return filepath + kZipFileExtensionName;
}

}  // namespace

FeedbackManager::FeedbackManager(QObject* parent)
    : QObject(parent),
      network_manager_(new QNetworkAccessManager(this)) {
  this->setObjectName("feedback_manager");

  this->initConnections();
}

FeedbackManager::~FeedbackManager() {
  delete multi_part_;
  multi_part_ = nullptr;
}

// TODO(xushaohua):
// * truncate log file
// * rename zip file
// * remove zip file on uploaded.
void FeedbackManager::handleSendFeedback(const QVariantMap& info) {
  const QString contact(info.value("contact", QString()).toString());
  const QString content(info.value("content", QString()).toString());
  const QString cookie(info.value("cookie", QString()).toString());
  const QString resolution(info.value("resolution", QString()).toString());
//  const bool private_msg(info.value("sendPrivateMsg", QVariant(true)).toBool());
  const QString title(info.value("title", QString()).toString());
  const QString url(info.value("url", QString()).toString());

  delete multi_part_;
  multi_part_ = new QHttpMultiPart(QHttpMultiPart::FormDataType);

  QHttpPart title_part;
  title_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                       QVariant(kFormDataTitle));
  title_part.setBody(title.toUtf8());
  multi_part_->append(title_part);

  QHttpPart resolution_part;
  resolution_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                            QVariant(kFormDataResolution));
  resolution_part.setBody(resolution.toUtf8());
  multi_part_->append(resolution_part);

  QHttpPart contact_part;
  contact_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                         QVariant(kFormDataContact));
  contact_part.setBody(contact.toUtf8());
  multi_part_->append(contact_part);

  QHttpPart content_part;
  content_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                         QVariant(kFormDataContent));
  content_part.setBody(content.toUtf8());
  multi_part_->append(content_part);

  QHttpPart os_version_part;
  os_version_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                            QVariant(kFormDataOsVersion));
  os_version_part.setBody(GetDistroNameAndVersion().toUtf8());
  multi_part_->append(os_version_part);

  QHttpPart client_version_part;
  client_version_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                                QVariant(kFormDataClientVersion));
  client_version_part.setBody(QByteArray(kAppVersion));
  multi_part_->append(client_version_part);

  QHttpPart cpu_part;
  cpu_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                     QVariant(kFormDataCpu));
  cpu_part.setBody(GetCPUModelName().toUtf8());
  multi_part_->append(cpu_part);

  QHttpPart file_part;
  file_part.setHeader(QNetworkRequest::ContentTypeHeader,
                      QVariant(kMimeTypeZip));
  file_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                      QVariant(kFormDataLogFile));
  const QString log_filepath(app->log_manager->getTmpLogFilepath());

  const QString zip_filepath(GetZipFilepath(log_filepath));
  if (! backend::Zip(zip_filepath, log_filepath)) {
    qCritical() << "[FeedbackManager]::sendFeedback() failed to create zip file"
                << zip_filepath;
    return;
  }

  // Clear log file
  app->log_manager->resetLogFile();

  // TODO(LiuLang: Release |file| resources
  QFile* file = new QFile(zip_filepath);
  file->open(QIODevice::ReadOnly);
  file_part.setBodyDevice(file);
  multi_part_->append(file_part);
  // Delete file with multi_part_, because it cannot be deleted now.
  file->setParent(multi_part_);

  const QUrl request_url(url);
  QNetworkRequest request(request_url);
  request.setRawHeader(QByteArray("Cookie"), cookie.toUtf8());

  QNetworkReply* reply = network_manager_->post(request, multi_part_);
  multi_part_->setParent(reply);
}

void FeedbackManager::handleSendStatFile(const QString& cookie) {
  delete multi_part_;
  multi_part_ = new QHttpMultiPart(QHttpMultiPart::FormDataType);

  QHttpPart file_part;
  file_part.setHeader(QNetworkRequest::ContentTypeHeader,
                      QVariant(kMimeTypeZip));
  file_part.setHeader(QNetworkRequest::ContentDispositionHeader,
                      QVariant(kFormDataStatisFile));
  const QString stat_filepath(app->log_manager->getTmpStatFilepath());
  if (QFile(stat_filepath).size() == 0) {
    qWarning() << "handleSendStatFile() stat file is empty";
    return;
  }

  const QString zip_filepath(GetZipFilepath(stat_filepath));
  if (! backend::Zip(zip_filepath, stat_filepath)) {
    qCritical() << "handleSendStatFile() failed to create zip file"
                << zip_filepath;
    return;
  }

  app->log_manager->resetStatFile();

  QFile* file = new QFile(zip_filepath);
  file->open(QIODevice::ReadOnly);
  file_part.setBodyDevice(file);
  multi_part_->append(file_part);
  file->setParent(multi_part_);

  const QString url(app->url_manager->url("statis"));
  const QUrl request_url(url);
  QNetworkRequest request(request_url);
  request.setRawHeader(QByteArray("Cookie"), cookie.toUtf8());

  QNetworkReply* reply = network_manager_->post(request, multi_part_);
  multi_part_->setParent(reply);

  emit this->onSendStatFile();
}

void FeedbackManager::initConnections() {
  connect(this, &FeedbackManager::sendFeedback,
          this, &FeedbackManager::handleSendFeedback);
  connect(this, &FeedbackManager::sendStatFile,
          this, &FeedbackManager::handleSendStatFile);
}

}  // namespace netease
