// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/keybinder.h"

#include <QDebug>
#include <QKeySequence>
#include <QList>
#include <QShortcut>

#include "application.h"
#include "service/backend/global_shortcut.h"
#include "ui/views/mini_player.h"

namespace netease {

namespace {

// State code used in onRegisterHotkeyResult() and onUnRegisterHotkeyResult().
const int kErrorCodeSucceeded = 0;
const int kErrorCodeFailed = 1;

const char kActionName[] = "actionName";

struct KeySeq {
  int keycode;
  int qt_key;
};

// Convert keycode in browser to Qt key.
const static QList<KeySeq> keysyms = {
  {  3, Qt::Key_Cancel},
  {  6, Qt::Key_Help},
  {  8, Qt::Key_Backspace},
  {  9, Qt::Key_Tab},
  { 12, Qt::Key_Clear},
  { 13, Qt::Key_Return},
  { 14, Qt::Key_Enter},
  { 16, Qt::Key_Shift},
  { 17, Qt::Key_Control},
  { 18, Qt::Key_Alt},
  { 19, Qt::Key_Pause},
  { 20, Qt::Key_CapsLock},
  { 27, Qt::Key_Escape},
  { 32, Qt::Key_Space},
  { 33, Qt::Key_PageUp},
  { 34, Qt::Key_PageDown},
  { 35, Qt::Key_End},
  { 36, Qt::Key_Home},
  { 37, Qt::Key_Left},
  { 38, Qt::Key_Up},
  { 39, Qt::Key_Right},
  { 40, Qt::Key_Down},
  { 42, Qt::Key_Print},
  { 45, Qt::Key_Insert},
  { 46, Qt::Key_Delete},
  { 48, Qt::Key_0},
  { 49, Qt::Key_1},
  { 50, Qt::Key_2},
  { 51, Qt::Key_3},
  { 52, Qt::Key_4},
  { 53, Qt::Key_5},
  { 54, Qt::Key_6},
  { 55, Qt::Key_7},
  { 56, Qt::Key_8},
  { 57, Qt::Key_9},
  { 59, Qt::Key_Semicolon},
  { 61, Qt::Key_Equal},
  { 65, Qt::Key_A},
  { 66, Qt::Key_B},
  { 67, Qt::Key_C},
  { 68, Qt::Key_D},
  { 69, Qt::Key_E},
  { 70, Qt::Key_F},
  { 71, Qt::Key_G},
  { 72, Qt::Key_H},
  { 73, Qt::Key_I},
  { 74, Qt::Key_J},
  { 75, Qt::Key_K},
  { 76, Qt::Key_L},
  { 77, Qt::Key_M},
  { 78, Qt::Key_N},
  { 79, Qt::Key_O},
  { 80, Qt::Key_P},
  { 81, Qt::Key_Q},
  { 82, Qt::Key_R},
  { 83, Qt::Key_S},
  { 84, Qt::Key_T},
  { 85, Qt::Key_U},
  { 86, Qt::Key_V},
  { 87, Qt::Key_W},
  { 88, Qt::Key_X},
  { 89, Qt::Key_Y},
  { 90, Qt::Key_Z},
  { 93, Qt::Key_Menu},
  { 95, Qt::Key_Sleep},
  {106, Qt::Key_multiply},
  {107, Qt::Key_Plus},
  {109, Qt::Key_Minus},
  {112, Qt::Key_F1},
  {113, Qt::Key_F2},
  {114, Qt::Key_F3},
  {115, Qt::Key_F4},
  {116, Qt::Key_F5},
  {117, Qt::Key_F6},
  {118, Qt::Key_F7},
  {119, Qt::Key_F8},
  {120, Qt::Key_F9},
  {121, Qt::Key_F10},
  {122, Qt::Key_F11},
  {123, Qt::Key_F12},
  {124, Qt::Key_F13},
  {125, Qt::Key_F14},
  {126, Qt::Key_F15},
  {127, Qt::Key_F16},
  {128, Qt::Key_F17},
  {129, Qt::Key_F18},
  {130, Qt::Key_F19},
  {131, Qt::Key_F20},
  {132, Qt::Key_F21},
  {133, Qt::Key_F22},
  {134, Qt::Key_F23},
  {135, Qt::Key_F24},
  {144, Qt::Key_NumLock},
  {145, Qt::Key_ScrollLock},
  {176, Qt::Key_MediaNext},
  {177, Qt::Key_MediaPrevious},
  {178, Qt::Key_MediaStop},
  {179, Qt::Key_MediaPlay},
  {188, Qt::Key_Comma},
  {190, Qt::Key_Period},
  {191, Qt::Key_Slash},
  {192, Qt::Key_QuoteLeft},
  {219, Qt::Key_BracketLeft},
  {220, Qt::Key_Backslash},
  {221, Qt::Key_BracketRight},
  {222, Qt::Key_Apostrophe},
  {224, Qt::Key_Meta},
};

bool operator==(const KeySeq& key_seq, int key) {
  return key_seq.keycode == key;
}

int GetQtKey(int keycode, bool* ok) {
  Q_ASSERT(ok);
  auto iter = std::find(keysyms.cbegin(), keysyms.cend(), keycode);
  if (iter == keysyms.cend()) {
    *ok = false;
    return 0;
  } else {
    *ok = true;
    int qt_key = (*iter).qt_key;

    // Handles modifiers.
    switch (qt_key) {
      case Qt::Key_Control: return Qt::CTRL;
      case Qt::Key_Alt:     return Qt::ALT;
      case Qt::Key_Shift:   return Qt::SHIFT;
      case Qt::Key_Meta:    return Qt::META;
      default:              return qt_key;
    }
  }
}

// Convert key code list to QKeySequence.
QKeySequence GetKeySequence(const QVariantList& keys, bool* ok) {
  int qt_key = 0;
  Q_ASSERT(ok);

  for (const QVariant& key : keys) {
    int keycode = key.toInt();
    qt_key += GetQtKey(keycode, ok);
    if (!(*ok)) {
      break;
    }
  }

  if (*ok) {
    return QKeySequence(qt_key);
  }
  return QKeySequence();
}

}  // namespace

Keybinder::Keybinder(QObject* parent)
    : QObject(parent),
      global_shortcut_(),
      local_shortcut_() {
  this->setObjectName("keybinder");

  initConnections();
}

Keybinder::~Keybinder() {
  for (auto it = global_shortcut_.begin(); it != global_shortcut_.end(); ++it) {
    if (it.value()) {
      delete it.value();
      it.value() = nullptr;
    }
  }
}

void Keybinder::handleShortcutActivated() {
  const QString action_name =
      QObject::sender()->property(kActionName).toString();
  // QShortcut has "autoRepeat" property.
  const bool is_local = QObject::sender()->property("autoRepeat").isValid();
  qDebug() << "[Keybinder]::handleShortcutActivated()"
           << " action_name:" << action_name << ",is_global:" << is_local;
  emit onHotkey(action_name, !is_local);
}

void Keybinder::handleRegisterHotkey(const QString& action_name,
                                     const QVariantList& keys,
                                     bool is_global) {
  qDebug() << "register hotkey:" << action_name << keys << is_global;
  if (is_global) {
    registerGlobalHotkey(action_name, keys);
  } else {
    registerLocalHotkey(action_name, keys);
  }
}

void Keybinder::handleUnregisterHotkey(const QString& action_name,
                                       bool is_global) {
  qDebug() << "unregister hotkey:" << action_name << is_global;
  if (is_global) {
    unregisterGlobalHotkey(action_name);
  } else {
    unregisterLocalHotkey(action_name);
  }
}

void Keybinder::initConnections() {
  connect(this, &Keybinder::registerHotkey,
          this, &Keybinder::handleRegisterHotkey);
  connect(this, &Keybinder::unregisterHotkey,
          this, &Keybinder::handleUnregisterHotkey);
}

void Keybinder::registerGlobalHotkey(const QString &action_name,
                                     const QVariantList &keys) {
  GlobalShortcut* shortcut;
  if (global_shortcut_.contains(action_name)) {
    shortcut = global_shortcut_.value(action_name);
    global_shortcut_.remove(action_name);
    delete shortcut;
  }

  bool ok = true;
  QKeySequence key_sequence = GetKeySequence(keys, &ok);
  if (ok) {
    shortcut = new GlobalShortcut(key_sequence);
    ok = shortcut->registerNow();
    if (ok) {
      global_shortcut_.insert(action_name, shortcut);
      connect(shortcut, &GlobalShortcut::activated,
              this, &Keybinder::handleShortcutActivated);
      shortcut->setProperty(kActionName, action_name);
    } else {
      qWarning() << "[Keybinder]::registerGlobalHotkey() failed to call"
                 << "registerNow()" << key_sequence;
    }
  }
  const int error_code = ok ? kErrorCodeSucceeded : kErrorCodeFailed;
  emit onRegisterHotkeyResult(action_name, true, error_code);
}

void Keybinder::registerLocalHotkey(const QString &action_name,
                                    const QVariantList &keys) {
  QShortcut* shortcut;
  if (local_shortcut_.contains(action_name)) {
    shortcut = local_shortcut_.value(action_name);
  } else {
    // Local shortcut needs to bind to a target widget.
    shortcut = new QShortcut(app->mini_player);
    // Set context of shortcut to ApplicationShortcut so that this shortcut
    // can be triggered when one of application windows is visible.
    shortcut->setContext(Qt::ApplicationShortcut);
    local_shortcut_.insert(action_name, shortcut);
    connect(shortcut, &QShortcut::activated,
            this, &Keybinder::handleShortcutActivated);
  }
  bool ok = true;
  shortcut->setKey(GetKeySequence(keys, &ok));
  shortcut->setProperty(kActionName, action_name);
  shortcut->setEnabled(true);
  emit onRegisterHotkeyResult(action_name, false, kErrorCodeSucceeded);
}

void Keybinder::unregisterGlobalHotkey(const QString &action_name) {
  if (global_shortcut_.contains(action_name)) {
    GlobalShortcut* shortcut = global_shortcut_.value(action_name);
    emit onUnregisterHotkeyResult(action_name, true,
                                  kErrorCodeSucceeded);
    global_shortcut_.remove(action_name);
    delete shortcut;
  } else {
    // Always emit success state if not found.
    emit onUnregisterHotkeyResult(action_name, true, kErrorCodeSucceeded);
  }
}

void Keybinder::unregisterLocalHotkey(const QString &action_name) {
  if (local_shortcut_.contains(action_name)) {
    QShortcut* shortcut = local_shortcut_.value(action_name);
    shortcut->setEnabled(false);
    emit onUnregisterHotkeyResult(action_name, false,
                                  kErrorCodeSucceeded);
    local_shortcut_.remove(action_name);
  } else {
    emit onUnregisterHotkeyResult(action_name, false, kErrorCodeSucceeded);
  }
}

}  // namespace netease
