// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/local_music_manager.h"

#include <QDebug>
#include <QDir>
#include <QDirIterator>
#include <QStandardPaths>
#include <QThread>

#include "base/consts.h"
#include "base/file_util.h"
#include "service/backend/database_worker.h"
#include "service/backend/filesystem_watcher.h"

namespace netease {

namespace {

const char kOfflineDBSchemeFile[] = ":/misc/offline-db.sql";

const char kAliasMyMusic[] = "<mymusic>";
const char kAliasDownload[] = "<download>";
//const char kAliasWindowsMedia[] = "<windowsmedia>";
//const char kAliasItunes[] = "<itunes>";

//const char kAddLibraryTaskId[] = "local-music-manager-add-library";

const char kAddRecordStatement[] = "INSERT INTO `track` ("
    "file, dir, title, album, artist, duration, timestamp, bitrate, filesize, "
    "id, parentdir, librarypath, tracknumber) values ("
    "'%1', '%2', '%3', '%4', '%5', %6, %7, %8, "
    "%9, '%10', '%11', '%12', %13)";

const char kDeleteRecordByIdStatement[] = "DELETE FROM `track` WHERE id='%1'";
const char kDeleteRecordByDirStatement[] = "DELETE FROM `track` WHERE dir='%1'";
const char kDeleteRecordByFileStatement[] =
    "DELETE FROM `track` WHERE INSTR(file, '%1') > 0";
const char kUpdateIgnoredStatement[] =
    "UPDATE `track` SET ignored=1 WHERE id='%1'";
const char kDeleteRecordByFilepath[] = "DELETE FROM `track` WHERE file='%1'";

void ScanLibraryFiles(const QString& dir, QStringList& list) {
  QDirIterator dir_iter(dir, QDirIterator::Subdirectories);
  while (dir_iter.hasNext()) {
    const QString item(dir_iter.next());
    if (QFileInfo(item).isFile()) {
      list.append(item);
    }
  }
}

}  // namespace

QString GetRealLibraryPath(const QString& library_path) {
  if (library_path.startsWith('<')) {
    // Library alias starts with '<'
    if (library_path == kAliasMyMusic) {
      return QStandardPaths::writableLocation(QStandardPaths::MusicLocation);
    }

    if (library_path == kAliasDownload) {
      return QStandardPaths::writableLocation(QStandardPaths::DownloadLocation);
    }
    // Ignores other alias names.
    return "";
  }

  return library_path;
}

LocalMusicManager::LocalMusicManager(QObject* parent)
    : QObject(parent),
      db_worker_(new DatabaseWorker(this)),
      filesystem_watcher_(new FilesystemWatcher()),
      scanner_(new LocalMusicScanner()),
      inotify_thread_(new QThread()),
      scanner_thread_(new QThread()),
      tasks_(),
      scanner_busy_(false),
      song_limit_(50) {
  this->setObjectName("local_music_library");

  inotify_thread_->start();
  scanner_thread_->start();
  filesystem_watcher_->moveToThread(inotify_thread_);
  scanner_->moveToThread(scanner_thread_);

  this->initConnections();

  this->initDB();
}

LocalMusicManager::~LocalMusicManager() {
  if (filesystem_watcher_ != nullptr) {
    delete filesystem_watcher_;
    filesystem_watcher_ = nullptr;
  }
  if (scanner_ != nullptr) {
    delete scanner_;
    scanner_ = nullptr;
  }

  if (inotify_thread_->isRunning()) {
    inotify_thread_->quit();
    inotify_thread_->wait(3);
    delete inotify_thread_;
    inotify_thread_ = nullptr;
  }

  if (scanner_thread_->isRunning()) {
    scanner_thread_->quit();
    scanner_thread_->wait(3);
    delete scanner_thread_;
    scanner_thread_ = nullptr;
  }
}

void LocalMusicManager::handleAddLibrary(const QString& library_path,
                                         int song_limit) {
  qDebug() << Q_FUNC_INFO << library_path << song_limit;

  // Scan libraries.
  QString root_dir(GetRealLibraryPath(library_path));

  // Current library already in scanning task list.
  if (tasks_.contains(root_dir)) {
    qDebug() << root_dir << "is in tasks";
    return;
  }

  song_limit_ = song_limit;

  if (!QFileInfo(root_dir).isDir()) {
    qDebug() << "[LocalMusicLibrary]::handleAddLibrary() invalid path:"
               << library_path;
    // |reason| is empty.
    QVariantMap result;
    result.insert("library", library_path);
    result.insert("result", 0);
    result.insert("reason", "");
    QStringList dirs;
    result.insert("dirs", dirs);
    emit this->onAddEnd(result, 1, "Folder not found");
    return;
  }

  // |result| is empty.
  emit this->onAddProgress(library_path, 0, "");

  QStringList files;
  ScanLibraryFiles(root_dir, files);
  if (files.isEmpty()) {
    QVariantMap result;
    result.insert("library", library_path);
    result.insert("result", 0);
    result.insert("reason", "");
    QStringList dirs;
    result.insert("dirs", dirs);
    emit this->onAddEnd(result, 0, "");
    return;
  } else {
    tasks_.insert(library_path, files);
    this->notifyScanner();
  }
}

bool LocalMusicManager::insertRecords(const LocalMediaRecordList& media_list) {
  // Insert as a new record.
  sqlite3* db = db_worker_->db();
  if (!db) {
    qWarning() << "LocalMusicManager::insertMediaRecord() db is null while"
               << "inserting a new record";
    return false;
  }

  sqlite3_exec(db, "BEGIN TRANSACTION", nullptr, nullptr, nullptr);
  for (const LocalMediaRecord& record : media_list) {
    QString sql(kAddRecordStatement);
    sql = sql.arg(record.file)
        .arg(record.dir)
        .arg(record.title)
        .arg(record.album)
        .arg(record.artist)
        .arg(record.duration)
        .arg(record.timestamp)
        .arg(record.bitrate)
        .arg(record.file_size)
        .arg(record.id)
        .arg(record.parent_dir)
        .arg(record.library_path)
        .arg(record.track_id);

    sqlite3_exec(db, sql.toStdString().c_str(), nullptr, nullptr, nullptr);
  }
  sqlite3_exec(db, "END TRANSACTION", nullptr, nullptr, nullptr);
  db_worker_->commit();

  return true;
}

bool LocalMusicManager::removeRecords(const LocalMediaRecordList& media_list) {
  // Insert as a new record.
  sqlite3* db = db_worker_->db();
  if (!db) {
    qWarning() << "LocalMusicManager::removeRecords() db is null while"
               << "inserting a new record";
    return false;
  }

  sqlite3_exec(db, "BEGIN TRANSACTION", nullptr, nullptr, nullptr);
  for (const LocalMediaRecord& record : media_list) {
    QString sql(kDeleteRecordByFilepath);
    sql = sql.arg(record.file);
    sqlite3_exec(db, sql.toStdString().c_str(), nullptr, nullptr, nullptr);
  }
  sqlite3_exec(db, "END TRANSACTION", nullptr, nullptr, nullptr);
  db_worker_->commit();

  return true;
}

void LocalMusicManager::notifyScanner() {
  qDebug() << Q_FUNC_INFO << tasks_.size();
  if (!scanner_busy_) {
    // Scanner is ready, send another beach of tasks.
    if (!tasks_.empty()) {
      const QString& library_path = tasks_.firstKey();
      const QStringList& library_tasks = tasks_[library_path];
      const int size = qMin(song_limit_, library_tasks.size());
      QStringList list;
      for (int i = 0; i < size; ++i) {
        list.append(library_tasks.at(i));
      }
      const QString root_dir = GetRealLibraryPath(library_path);
      emit this->scanner_->scan(library_path, root_dir, list);
    }
  }
}

void LocalMusicManager::handleObserveLibrary(const QString& library_path) {
  const QString real_path(GetRealLibraryPath(library_path));
  if (real_path.isEmpty()) {
    qDebug() << "Ignores library path alias:" << library_path;
  } else {
    emit filesystem_watcher_->watchFolder(real_path, true);
  }
}

void LocalMusicManager::handleRemoveLibrary(const QString& library_path) {
  const QString root_dir(GetRealLibraryPath(library_path));

  // Remove library from inotify watch list.
  emit filesystem_watcher_->rmWatchFolder(root_dir);

  // Also remove scan_tasks_ within library.
  if (tasks_.contains(root_dir)) {
    tasks_.remove(root_dir);
  }

  // Remove records from database.
  const QString query(kDeleteRecordByDirStatement);
  bool ok = true;
  db_worker_->rawExec(query.arg(root_dir), &ok);
  emit this->onRemoveLibrary(library_path, 0, "");
}

void LocalMusicManager::handleRemoveLibraryItems(const QVariantList& items) {
  QVariantMap result;

  // Items failed to delete.
  QVariantList fail;
  bool ok = true;

  for (const QVariant& item : items) {
    const QMap<QString, QVariant> map = item.toMap();
    const QString id = map.value("id").toString();
    const bool remove_file = map.value("removefile").toBool();
    const QString path = map.value("path").toString();
    // |remove_by_id| is always true.
    //const bool remove_by_id = map.value("removebyid").toBool();

    if (remove_file) {
      // Remove local media file.
      ok = QFile::remove(path);
      if (!ok) {
        qWarning() << "[LocalMusicManager]::handleRemoveLibraryItems()"
                   << "Failed to remove file:" << path;
      }
      // Also remove record from db.
      const QString query(kDeleteRecordByIdStatement);
      db_worker_->rawExec(query.arg(id), &ok);
      if (!ok) {
        fail.append(id);
      }
    } else {
      // Only set `ignored` field in `track` table to true.
      const QString query(kUpdateIgnoredStatement);
      db_worker_->rawExec(query.arg(id), &ok);
      if (!ok) {
        fail.append(id);
      }
    }

  }
  result.insert("fail", fail);

  emit this->onRemoveLibraryItems(result);
}

void LocalMusicManager::handleRemoveObserveLibrary(
    const QString& library_path) {
  qDebug() << Q_FUNC_INFO << library_path;
  emit filesystem_watcher_->rmWatchFolder(library_path);
}

void LocalMusicManager::handleFileAdded(const QString& root_dir,
                                        const QString& filepath) {
  qDebug() << Q_FUNC_INFO << root_dir << filepath;
  emit scanner_->scanAll(root_dir, {filepath});
}

void LocalMusicManager::handleFileDeleted(const QString& root_dir,
                                          const QString& filepath) {
  Q_UNUSED(root_dir);
  const QString query(kDeleteRecordByFilepath);
  bool ok = true;
  db_worker_->rawExec(query.arg(filepath), &ok);
  emit this->onObserveLibrary();
}

void LocalMusicManager::handleFolderAdded(const QString& root_dir,
                                          const QString& dir) {
  QStringList list;
  ScanLibraryFiles(dir, list);
  if (!list.isEmpty()) {
    emit scanner_->scanAll(root_dir, list);
  }
}

void LocalMusicManager::handleFolderDeleted(const QString& root_dir,
                                            const QString& dir) {
  Q_UNUSED(root_dir);
  const QString query(kDeleteRecordByFileStatement);
  bool ok = true;
  db_worker_->rawExec(query.arg(dir), &ok);
  emit this->onObserveLibrary();
}

void LocalMusicManager::initConnections() {
  connect(this, &LocalMusicManager::addLibrary,
          this, &LocalMusicManager::handleAddLibrary);
  connect(this, &LocalMusicManager::observeLibrary,
          this, &LocalMusicManager::handleObserveLibrary);
  connect(this, &LocalMusicManager::removeLibrary,
          this, &LocalMusicManager::handleRemoveLibrary);
  connect(this, &LocalMusicManager::removeLibraryItems,
          this, &LocalMusicManager::handleRemoveLibraryItems);
  connect(this, &LocalMusicManager::removeObserveLibrary,
          this, &LocalMusicManager::handleRemoveObserveLibrary);

  connect(filesystem_watcher_, &FilesystemWatcher::fileAdded,
          this, &LocalMusicManager::handleFileAdded);
  connect(filesystem_watcher_, &FilesystemWatcher::fileDeleted,
          this, &LocalMusicManager::handleFileDeleted);
  connect(filesystem_watcher_, &FilesystemWatcher::folderAdded,
          this, &LocalMusicManager::handleFolderAdded);
  connect(filesystem_watcher_, &FilesystemWatcher::folderDeleted,
          this, &LocalMusicManager::handleFolderDeleted);

  connect(this, &LocalMusicManager::execSql,
          this, &LocalMusicManager::handleExecSql);
  connect(this, &LocalMusicManager::execTransaction,
          this, &LocalMusicManager::handleExecTransaction);

  connect(scanner_, &LocalMusicScanner::onScan,
          this, &LocalMusicManager::handleTaskScanned);
  connect(scanner_, &LocalMusicScanner::onScanAll,
          this, &LocalMusicManager::handleAllTasksScanned);
}

void LocalMusicManager::initDB() {
  db_worker_->open(GetUserConfigDir().filePath(kOfflineDB));

  bool ok = true;
  db_worker_->rawExec(ReadTextFileContent(kOfflineDBSchemeFile), &ok);
  if (!ok) {
    qWarning() << "[LocalMusicManager]::initDB() failed to init local db";
  }
}

void LocalMusicManager::handleExecSql(const QString& id,
                                      const QString& statement) {
  bool ok = true;
  QVariantList value;
  value = db_worker_->rawExec(statement, &ok);
  const int err = ok ? 0 : 1;
  QVariantMap result;
  result.insert("id", id);
  result.insert("error", err);
  result.insert("value", value);

  emit this->onExecSqlDone(result);
}

void LocalMusicManager::handleExecTransaction(const QString& id,
                                              const QStringList& statements) {
  QVariantList value;
  bool ok = true;
  value = db_worker_->exec(statements, &ok);
  const int err = ok ? 0 : 1;
  QVariantMap result;
  result.insert("id", id);
  result.insert("error", err);
  result.insert("value", value);

  emit this->onExecSqlDone(result);
}

void LocalMusicManager::handleTaskScanned(const QString& library_path,
                                          const QString& root_dir,
                                          int task_count,
                                          const LocalMediaRecordList& list) {
  Q_UNUSED(root_dir);
  scanner_busy_ = false;
  // Remove tasks from task list.
  if (tasks_.contains(library_path)) {
    QStringList& library_task = tasks_[library_path];
    for (int i = 0; (i < task_count) && (!library_task.isEmpty()); ++i) {
      library_task.removeFirst();
    }

    // Remove empty library path from task list.
    if (library_task.isEmpty()) {
      tasks_.remove(library_path);
    }

    // Insert into database.
    this->removeRecords(list);
    this->insertRecords(list);
  }

  // Emit onProgress() signal.
  emit this->onAddProgress(library_path, list.size(), "");

  if (tasks_.isEmpty()) {
    QVariantMap result;
    result.insert("library", library_path);
    result.insert("result", 0);
    result.insert("reason", "");
    QStringList dirs;
    result.insert("dirs", dirs);
    emit this->onAddEnd(result, 0, "");
  } else {
    this->notifyScanner();
  }
}

void LocalMusicManager::handleAllTasksScanned(const QString& root_dir,
    const LocalMediaRecordList& media_list) {
  qDebug() << Q_FUNC_INFO << root_dir << media_list.size();
  Q_UNUSED(root_dir);
  // Insert into database.
  this->removeRecords(media_list);
  this->insertRecords(media_list);
  emit this->onObserveLibrary();
}

}  // namespace netease
