// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_LOCAL_MUSIC_MANAGER_H_
#define SERVICE_LOCAL_MUSIC_MANAGER_H_

#include <QObject>
#include <QFileInfo>
#include <QMap>

#include "service/backend/local_music_scanner.h"

namespace netease {

class DatabaseWorker;
class FilesystemWatcher;

// Get real path of library alias.
QString GetRealLibraryPath(const QString& library_path);

// Manages local music resources.
// Scanning operation works in a separator thread.
// Database operation works in the same thread as LocalMusicManager.
class LocalMusicManager : public QObject {
  Q_OBJECT

 public:
  explicit LocalMusicManager(QObject* parent = nullptr);
  ~LocalMusicManager() override;

 signals:
  // |paths| is an object, containing:
  // * library: "<mymusic>",
  // * result: 0 for success,
  // * reason: "", error reason,
  void onAddEnd(const QVariantMap& path, int error, const QString& reason);
  void onAddProgress(const QString& path, int song_num, const QString& reason);

  void onObserveLibrary();
  void onRemoveLibrary(const QString& library_path, int error,
                       const QString& reason);
  void onRemoveLibraryItems(const QVariantMap& reason);

  // |library_path| might be path alias, like <mymusic> or <download>.
  void addLibrary(const QString& library_type, int song_limit);

  void observeLibrary(const QString& library_path);
  void removeLibrary(const QString& library_path);
  void removeLibraryItems(const QVariantList& items);
  void removeObserveLibrary(const QString& library_path);

  void onExecSqlDone(const QVariantMap& result);
  void execSql(const QString& id, const QString& statement);
  void execTransaction(const QString& id, const QStringList& statements);

 private:
  void initConnections();
  void initDB();

  // Returns true when |media_list| is inserted or updated in db.
  bool insertRecords(const LocalMediaRecordList& media_list);

  bool removeRecords(const LocalMediaRecordList& media_list);

  void notifyScanner();

  DatabaseWorker* db_worker_ = nullptr;
  FilesystemWatcher* filesystem_watcher_ = nullptr;
  LocalMusicScanner* scanner_ = nullptr;
  QThread* inotify_thread_ = nullptr;
  QThread* scanner_thread_ = nullptr;
  // library-path => file-list
  QMap<QString, QStringList> tasks_;
  bool scanner_busy_;
  int song_limit_;

 private slots:
  void handleAddLibrary(const QString& library_path, int song_limit);
  void handleObserveLibrary(const QString& library_path);
  void handleRemoveLibrary(const QString& library_path);

  // Remove records from database. Might remove local files too.
  void handleRemoveLibraryItems(const QVariantList& items);
  void handleRemoveObserveLibrary(const QString& library_path);

  // Read metadata of |filepath|, insert into database and emit
  // onObserveLibrary() signal.
  void handleFileAdded(const QString& root_dir, const QString& filepath);

  // Delete record matching |filepath| and emit onObserveLibrary() signal.
  void handleFileDeleted(const QString& root_dir, const QString& filepath);

  // Scan files in |dir| recursively, insert into database and emit
  // onObserveLibrary() signal.
  void handleFolderAdded(const QString& root_dir, const QString& dir);

  // Delete records in database matching |dir| and emit onObserveLibrary()
  // signal.
  void handleFolderDeleted(const QString& root_dir, const QString& dir);

  void handleExecSql(const QString& id, const QString& statement);
  void handleExecTransaction(const QString& id, const QStringList& statements);

  void handleTaskScanned(const QString& library_path,
                         const QString& root_dir,
                         int task_count,
                         const LocalMediaRecordList& list);

  void handleAllTasksScanned(const QString& root_dir,
                             const LocalMediaRecordList& media_list);
};

}  // namespace netease

#endif  // SERVICE_LOCAL_MUSIC_MANAGER_H_
