// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/log_manager.h"

#include <unistd.h>
#include <sys/types.h>
#include <QDebug>
#include <QDir>
#include <QFile>
#include "base/consts.h"
#include "base/file_util.h"
#include "third_party/CuteLogger/CuteLogger/ConsoleAppender.h"
#include "third_party/CuteLogger/CuteLogger/Logger.h"
#include "third_party/CuteLogger/CuteLogger/FileAppender.h"

namespace netease {

namespace {

// Defines log format.
const char kLogFormat[] =
    "%{time}{MM-dd, HH:mm:ss} [%{type:-7}] [%{file:-25} %{line}] %{message}\n";

// Folder name to keep log file.
// That is $HOME/.cache/cloud-music/Logs/.
const char kLogDirName[] = "Logs";

// File name of log file.
// That is $HOME/.cache/cloud-music/Logs/webview.log.
const char kLogFileName[] = "webview.log";
const char kTmpLogFileName[] = "webview-tmp.log";

// File name of native log file.
// That is $HOME/.cache/cloud-music/Logs/native.log
const char kNativeLogFileName[] = "native.log";

// File name of statistics file.
// That is $HOME/.cache/cloud-music/Logs/web-statis.log.
const char kStatisFileName[] = "web-statis.log";
const char kTmpStatisFileName[] = "web-statis-tmp.log";

// Truncate log file to 1MB on application startup.
const qint64 kFileSize1MB = 1048576;
const qint64 kFileSize256KB = 262144;
// 64KB
const qint64 kFileSize64KB = 65536;

bool CopyTruncatedFile(const QString& src, const QString& dst, qint64 size) {
  QFile src_file(src);
  if (!src_file.exists()) {
    return false;
  }
  if (!src_file.open(QIODevice::ReadOnly)) {
    return false;
  }

  QFile dst_file(dst);
  if (!dst_file.open(QIODevice::WriteOnly)) {
    return false;
  }

  qint64 offset = 0;
  if (src_file.size() > size) {
    // Copy from middle of file.
    offset = src_file.size() - size;
  }
  src_file.seek(offset);
  qint64 bytes_read = 0;
  const int kBufferSize = kFileSize64KB;
  char buf[kBufferSize];

  while (true) {
    bytes_read = src_file.read(buf, kBufferSize);
    if (bytes_read == -1) {
      qWarning() << "CopyTruncatedFile() failed to read file";
      break;
    } else if (bytes_read == 0) {
      // Reaches end of file.
      break;
    }
    dst_file.write(buf, bytes_read);
  }
  src_file.close();
  dst_file.close();
  return (bytes_read > -1);
}

// Truncate log file from middle.
void TruncateLogFile(const QString& filepath, qint64 size) {
  const QString tmp_file(filepath + "tmp-file");
  CopyTruncatedFile(filepath, tmp_file, size);
  QFile::remove(tmp_file);
}

}  // namespace

LogManager::LogManager(QObject* parent)
    : QObject(parent),
      log_filepath_(),
      tmp_log_filepath_(),
      native_log_filepath_(),
      statis_filepath_(),
      tmp_stat_filepath_(),
      log_file_(nullptr),
      statis_file_(nullptr) {
  this->setObjectName("log_manager");
  this->init();
  this->initConnections();
}

LogManager::~LogManager() {
  if (log_file_ && log_file_->isOpen()) {
    log_file_->close();
    delete log_file_;
    log_file_ = nullptr;
  }
  if (statis_file_ && statis_file_->isOpen()) {
    statis_file_->close();
    delete statis_file_;
    statis_file_ = nullptr;
  }
}

QString LogManager::getTmpLogFilepath() {
  if (QFile::exists(tmp_log_filepath_)) {
    QFile::remove(tmp_log_filepath_);
  }
  CopyTruncatedFile(log_filepath_, tmp_log_filepath_, kFileSize256KB);
  return tmp_log_filepath_;
}

QString LogManager::getTmpStatFilepath() {
  if (QFile::exists(tmp_stat_filepath_)) {
    QFile::remove(tmp_stat_filepath_);
  }
  CopyTruncatedFile(statis_filepath_, tmp_stat_filepath_, kFileSize256KB);
  return tmp_stat_filepath_;
}

void LogManager::resetLogFile() {
  log_file_->close();
  log_file_->open(QIODevice::Truncate | QIODevice::WriteOnly);
}

void LogManager::resetStatFile() {
  statis_file_->close();
  statis_file_->open(QIODevice::Truncate | QIODevice::WriteOnly);
}

void LogManager::init() {
  QDir log_dir(ConcateDir(GetUserCacheDir(), kLogDirName));
  log_filepath_ = log_dir.absoluteFilePath(kLogFileName);
  tmp_log_filepath_ = log_dir.absoluteFilePath(kTmpLogFileName);
  native_log_filepath_ = log_dir.absoluteFilePath(kNativeLogFileName);
  statis_filepath_ = log_dir.absoluteFilePath(kStatisFileName);
  tmp_stat_filepath_ = log_dir.absoluteFilePath(kTmpStatisFileName);

  // Truncate log file.
  TruncateLogFile(native_log_filepath_, kFileSize1MB);
  TruncateLogFile(log_filepath_, kFileSize64KB);
  TruncateLogFile(statis_filepath_, kFileSize64KB);

  // Initialize console logger.
  ConsoleAppender* console_appender = new ConsoleAppender();
#ifndef NDEBUG
  console_appender->setDetailsLevel(Logger::Debug);
#else
  // In release mode, Logger::Info is used because app::log() use that level.
  console_appender->setDetailsLevel(Logger::Error);
#endif
  console_appender->setFormat(kLogFormat);
  logger->registerAppender(console_appender);

  // Initialize file logger.
  FileAppender* file_appender = new FileAppender(native_log_filepath_);
#ifndef NDEBUG
  file_appender->setDetailsLevel(Logger::Debug);
#else
  file_appender->setDetailsLevel(Logger::Error);
#endif
  file_appender->setFormat(kLogFormat);
  logger->registerAppender(file_appender);

  // Init web log service.
  log_file_ = new QFile(log_filepath_);
  if (!log_file_->open(QIODevice::Append | QIODevice::WriteOnly)) {
    qWarning() << "[LogManager]::init() failed to init log file!"
                << log_file_->errorString();
  }

  // Init statistics service.
  statis_file_ = new QFile(statis_filepath_);
  if (!statis_file_->open(QIODevice::Append | QIODevice::WriteOnly)) {
    qWarning() << "[LogManager]::init() failed to init stat file!"
                << statis_file_->errorString();
  }
}

void LogManager::handleLog(const QString& msg) {
  if (log_file_->isOpen()) {
    log_file_->write(qPrintable(msg));
    const char separator = '\n';
    log_file_->write(&separator, 1);
    log_file_->flush();
  } else {
    qWarning() << Q_FUNC_INFO << msg;
  }
}

void LogManager::handleStatMsg(const QString& msg) {
  if (statis_file_->isOpen()) {
    statis_file_->write(qPrintable(msg));
    statis_file_->flush();
  } else {
    qWarning() << Q_FUNC_INFO << msg;
  }
}

void LogManager::initConnections() {
  connect(this, &LogManager::log, this, &LogManager::handleLog);
  connect(this, &LogManager::statMsg, this, &LogManager::handleStatMsg);
}

}  // namespace netease
