// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/lrc_manager.h"

#include <QDebug>

#include "application.h"
#include "service/media_player.h"

namespace netease {

namespace {

const int kLrcElementNil = -1;

// Add 500ms ahead of lrc content to make it more natural.
const qint64 kLrcHumanOffset = -500;

// Check square element exists in |lrc| or not.
bool HasNoTimeTag(const QString& lrc) {
  return lrc.indexOf("[00:") == -1;
}

}  // namespace

bool operator<(const qint64 pos, const LrcManager::WordElement& word_element) {
  return pos < word_element.time;
}

bool operator<(const LrcManager::WordElement& word_element, const qint64 pos) {
  return word_element.time < pos;
}

bool operator<(const qint64 start, const LrcManager::LrcElement& lrc_element) {
  return start < lrc_element.start;
}

bool operator<(const LrcManager::LrcElement& lrc_element, const qint64 start) {
  return lrc_element.start < start;
}

LrcManager::LrcManager(QObject* parent)
    : QObject(parent),
      slogan(),
      lrc_type_(LrcType::kLrcTypeNotSet),
      current_element_(kLrcElementNil),
      metadata_(),
      lrc_elements_(),
      lrc_raw_() {
  this->setObjectName("lrc_manager");
  this->initConnections();
}

LrcManager::~LrcManager() {

}

LrcManager::LrcElement LrcManager::currentElement() const {
  if (lrc_elements_.isEmpty() || current_element_ == kLrcElementNil) {
    return LrcElement();
  } else {
    return lrc_elements_.at(current_element_);
  }
}

LrcManager::LrcElement LrcManager::firstElement() const {
  if (lrc_elements_.isEmpty()) {
    return LrcElement();
  } else {
    return lrc_elements_.first();
  }
}

void LrcManager::initConnections() {
  connect(this, &LrcManager::setIsDj, this, &LrcManager::handleSetIsDj);
  connect(this, &LrcManager::setIsLightMusic,
          this, &LrcManager::handleSetIsLightMusic);
  connect(this, &LrcManager::setLrcEmpty, this, &LrcManager::handleSetLrcEmpty);
  connect(this, &LrcManager::setLyrics, this, &LrcManager::handleSetLyrics);
  connect(this, &LrcManager::setOffset, this, &LrcManager::handleSetOffset);
  connect(app->media_player, &MediaPlayer::trackInfoChanged,
          this, &LrcManager::handleMediaInfoChanged);
}

bool LrcManager::isDj() const {
  return lrc_type_ & LrcType::kLrcTypeDj;
}

bool LrcManager::isFirstElement() const {
  return (lrc_elements_.length() > 0) && (current_element_ == 0);
}

bool LrcManager::isKrc() const {
  return lrc_type_ & LrcType::kLrcTypeKrc;
}

bool LrcManager::isLastElement() const {
  return (lrc_elements_.length() > 0) &&
         (current_element_ == lrc_elements_.length() - 1);
}

bool LrcManager::isLightMusic() const {
  return lrc_type_ & LrcType::kLrcTypeLightMusic;
}

bool LrcManager::isLrc() const {
  return lrc_type_ & LrcType::kLrcTypeLrc;
}

bool LrcManager::isLrcEmpty() const {
  return lrc_type_ & LrcType::kLrcTypeEmpty;
}

bool LrcManager::isNoTimeLrc() const {
  return lrc_type_ & LrcType::kLrcTypeNoTimeLrc;
}

LrcManager::LrcElement LrcManager::lastElement() const {
  if (lrc_elements_.isEmpty()) {
    return LrcElement();
  } else {
    return lrc_elements_.last();
  }
}

LrcManager::LrcElement LrcManager::nextElement() const {
  if (lrc_elements_.isEmpty()) {
    return LrcElement();
  } else {
    if (current_element_ < lrc_elements_.size() - 2) {
      return lrc_elements_.at(current_element_ + 1);
    } else {
      return LrcElement();
    }
  }
}

void LrcManager::parseKrc() {
  QString lrc(lrc_raw_.replace(QString("\\n"), QChar('\n')));

  // Parse each line of content.
  for (const QString& line : lrc.split(QChar('\n'))) {

    // Ignore empty line.
    if (line.length() == 0) {
      continue;
    }

    // Check if current line is translated into Simplified Chinese.
    if (line.startsWith("[-dd]")) {
      if (lrc_elements_.length() > 0) {
        auto iter = lrc_elements_.end();
        (*iter).zh_cn = line.mid(5);
      }
      continue;
    }

    // Match [*] tag.
    QRegExp square_reg("\\[([^\\]]*)\\]");

    // Caps time delta in krc.
    QRegExp delta_reg("\\((\\d+)\\s*,\\s*(\\d+)\\)");

    int krc_start_time = 0;
    int krc_duration = 0;
    int krc_line_number = 0;

    bool ok = true;

    // Match and capture tags.
    int pos = 0;
    while (square_reg.indexIn(line, pos) > -1) {
      // Removes left square bracket.
      int tag_start = pos + 1;
      // Removes right square bracket.
      int tag_len = square_reg.matchedLength() - 2;

      QString tag(line.mid(tag_start, tag_len));

      if (tag.startsWith("al:", Qt::CaseInsensitive)) {
        metadata_.album = tag.mid(3);
      } else if (tag.startsWith("ar:", Qt::CaseInsensitive)) {
        metadata_.artist = tag.mid(3);
      } else if (tag.startsWith("by:", Qt::CaseInsensitive)) {
        metadata_.editor = tag.mid(3);
      } else if (tag.startsWith("co:", Qt::CaseInsensitive)) {
        metadata_.composer = tag.mid(3);
      } else if (tag.startsWith("cr:", Qt::CaseInsensitive)) {
        metadata_.creator = tag.mid(3);
      } else if (tag.startsWith("dt:", Qt::CaseInsensitive)) {
        metadata_.datetime = tag.mid(3);
      } else if (tag.startsWith("offset:", Qt::CaseInsensitive)) {
        metadata_.offset = tag.mid(7).toLongLong();
      } else if (tag.startsWith("re:", Qt::CaseInsensitive)) {
        metadata_.repeat = tag.mid(3).toInt();
      } else if (tag.startsWith("ti:", Qt::CaseInsensitive)) {
        metadata_.title = tag.mid(3);
      } else if (tag.startsWith("#:")) {
        metadata_.url = tag.mid(2);
      } else {
        // Parse krc time tag. Like [28058,8710] or [28058,8710, 3]
        QStringList time_tags(tag.split(QChar(',')));
        if (time_tags.length() > 1) {
          krc_start_time = time_tags.at(0).toLongLong(&ok);
          if (ok) {
            krc_duration = time_tags.at(1).toLongLong(&ok);
            if (time_tags.length() == 3 && ok) {
              krc_line_number = time_tags.at(2).toInt(&ok);
            }
          }
        }
      }

      // Go forward.
      pos += square_reg.matchedLength();
    }

    // If end-square-bracket is not the end of line, assumes that
    // remaining content is words of song.
    if (pos < line.length() && ok) {

      // Parse words and duration of krc content.
      QString words(line.mid(pos));
      QString content;

      LrcElement lrc_element;
      lrc_element.start = krc_start_time;
      lrc_element.duration = krc_duration;
      lrc_element.line_number = krc_line_number;

      int pos = 0;
      while (delta_reg.indexIn(words, pos) > -1) {
        pos = delta_reg.matchedLength() + delta_reg.pos();

        bool ok = true;
        QStringList captured_texts(delta_reg.capturedTexts());
        if (captured_texts.length() == 3) {
          WordElement word_element;
          word_element.time = captured_texts.last().toLongLong(&ok);

          int current_index = delta_reg.indexIn(words, pos + 1);

          if (current_index == -1) {
            // Current word reaches end of current line.
            word_element.word = words.mid(pos);
          } else {
            word_element.word = words.mid(pos, current_index - pos);
          }

          // Rebuild content in lrc_element.
          // If this word element contains multimple chars, prepend a
          // space character.
          if (word_element.word.length() > 1) {
            content.append(QChar(' '));
          }
          content.append(word_element.word);

          if (ok) {
            lrc_element.words.append(word_element);
          } else {
            qWarning() << "invalid word_element: " << word_element.word;
          }
        } else if (captured_texts.length() == 2) {
          // Read the sustainable time.
          lrc_element.sustain_time = captured_texts.at(1).toLongLong();
        }
      }
      lrc_element.content = content;

      lrc_elements_.append(lrc_element);
    }
  }

  // Sort lrc_elements.
  std::sort(lrc_elements_.begin(), lrc_elements_.end());

  // Add offset to each lrc_element.
  if (metadata_.offset != 0) {
    for (auto iter = lrc_elements_.begin(); iter != lrc_elements_.end();
         ++iter) {
      (*iter).start += metadata_.offset;

      // Set sustain_time if it is krc.
      qint64 sustain_time = 0;
      for (const WordElement& word_element : (*iter).words) {
        sustain_time += word_element.time;
      }
      (*iter).sustain_time = sustain_time;
    }
  }

  this->updateMetadata();

  // FIXME(xushaohua): time tag is incorrect.
  LrcElement title_lrc_element;
  title_lrc_element.start = 0;

  QString title_and_artist(metadata_.title);
  if (metadata_.artist.length() > 0) {
    title_and_artist.append(" - ");
    title_and_artist.append(metadata_.artist);
  }
  title_lrc_element.content = title_and_artist;

  // If it is krc, add word_element.
  const qint64 start_time = lrc_elements_.at(0).start;
  const qint64 time_delta = start_time / title_and_artist.length();
  for (const QString& word : title_and_artist) {
    WordElement word_element;
    word_element.word = word;
    word_element.time = time_delta;
    title_lrc_element.words.append(word_element);
  }

  // Prepend title element.
  lrc_elements_.prepend(title_lrc_element);
}

void LrcManager::parseLrc() {
  QString lrc(lrc_raw_.replace(QString("\\n"), QChar('\n')));

  if (HasNoTimeTag(lrc)) {
    lrc_type_ = LrcType::kLrcTypeNoTimeLrc;
    return;
  }

  // Parse each line of content.
  for (const QString& line : lrc.split(QChar('\n'))) {
    if (line.length() == 0) {
      continue;
    }

    // Match [*] tag.
    QRegExp square_reg("\\[([^\\]]*)\\]");

    // To store time tags of current line.
    QList<int> lrc_time_tags;

    bool ok = true;

    // Match and capture tags.
    int pos = 0;
    while (square_reg.indexIn(line, pos) > -1) {

      // Removes left square bracket.
      int tag_start = pos + 1;
      // Removes right square bracket.
      int tag_len = square_reg.matchedLength() - 2;

      QString tag(line.mid(tag_start, tag_len));

      if (tag.startsWith("al:", Qt::CaseInsensitive)) {
        metadata_.album = tag.mid(3);
      } else if (tag.startsWith("ar:", Qt::CaseInsensitive)) {
        metadata_.artist = tag.mid(3);
      } else if (tag.startsWith("by:", Qt::CaseInsensitive)) {
        metadata_.editor = tag.mid(3);
      } else if (tag.startsWith("co:", Qt::CaseInsensitive)) {
        metadata_.composer = tag.mid(3);
      } else if (tag.startsWith("cr:", Qt::CaseInsensitive)) {
        metadata_.creator = tag.mid(3);
      } else if (tag.startsWith("offset:", Qt::CaseInsensitive)) {
        metadata_.offset = tag.mid(7).toLongLong();
      } else if (tag.startsWith("ti:", Qt::CaseInsensitive)) {
        metadata_.title = tag.mid(3);
      } else {
        // Parse time tag. Like [03:21.428]
        int colon_index = tag.indexOf(QChar(':'));

        if (colon_index > 0) {
           int minutes = tag.left(colon_index).toLongLong(&ok);
          if (ok) {
            double seconds = tag.mid(colon_index + 1).toFloat(&ok);
            if (ok) {
              int time_tag = minutes * 60 * 1000 + seconds * 1000;
              lrc_time_tags.append(time_tag);
            }
          }
        }
      }

      // Go forward.
      pos += square_reg.matchedLength();
    }

    // If end-square-bracket is not the end of line, assumes that
    // remaining content is words of song.
    if (pos < line.length() && ok && lrc_time_tags.length() > 0) {
      // Add lrc_element.
      QString words(line.mid(pos));

      // Store all time tags.
      for (int time_tag : lrc_time_tags) {
        LrcElement lrc_element;
        lrc_element.start = time_tag;
        lrc_element.content = words;
        lrc_elements_.append(lrc_element);
      }
    }
  }

  // Sort lrc_elements.
  std::sort(lrc_elements_.begin(), lrc_elements_.end());

  // TODO(xushaohua):Add durations to lrc_element.
  for (auto iter = lrc_elements_.begin(); iter != lrc_elements_.end();
       ++iter) {
    if ((iter + 1) != lrc_elements_.end()) {
      (*iter).duration = (* (iter + 1)).start - (*iter).start;
    } else {
      (*iter).duration = 0;
    }
  }

  // Add offset to each lrc_element.
  if (metadata_.offset != 0) {
    for (auto iter = lrc_elements_.begin(); iter != lrc_elements_.end();
         ++iter) {
      (*iter).start += metadata_.offset;
    }
  }

  this->updateMetadata();

  // FIXME(xushaohua): time tag is incorrect.
  LrcElement title_lrc_element;
  title_lrc_element.start = 0;

  QString title_and_artist(metadata_.title);
  if (metadata_.artist.length() > 0) {
    title_and_artist.append(" - ");
    title_and_artist.append(metadata_.artist);
  }
  title_lrc_element.content = title_and_artist;

  // Add sustain_time to title_lrc_element.
//  if (lrc_elements_.size() > 0) {
//    title_lrc_element.sustain_time = lrc_elements_.at(0).start - 100;
//  }

  lrc_elements_.prepend(title_lrc_element);
}

void LrcManager::reset() {
  qDebug() << "[LrcManager]::reset()";
  lrc_type_ = LrcType::kLrcTypeNotSet;

  lrc_raw_ = "";
  metadata_ = {};
  offset_ = 0;
  lrc_elements_.clear();
  current_element_ = kLrcElementNil;
}

LrcManager::LrcElement LrcManager::searchElement(qint64 position) {
  if (!isLrcEmpty()) {
    // Handles lrc offset time.
    const qint64 position_delta = position - offset_ - kLrcHumanOffset;
    // NOTE(xushaohua): Currently offset in metadata is ignored.
//    const qint64 position_delta = position - offset_ - metadata_.offset;

    auto iter = std::upper_bound(lrc_elements_.cbegin(),
                                 lrc_elements_.cend(),
                                 position_delta);
    if (iter != lrc_elements_.cend()) {
      current_element_ = std::distance(lrc_elements_.cbegin(), iter) - 1;
    } else {
      current_element_ = lrc_elements_.size() - 1;
    }
    if (0 <= current_element_ && current_element_ < lrc_elements_.size()) {
//      auto element = lrc_elements_.at(current_element_);
//      if (element.sustain_time > 0 &&
//          element.start + element.sustain_time < position) {
//        // Read next element.
//        current_element_ += 1;
//      }
//      if (current_element_ < lrc_elements_.size()) {
//        return lrc_elements_.at(current_element_);
//      }
        return lrc_elements_.at(current_element_);
    }
  }

  return LrcElement();
}

void LrcManager::handleSetIsDj() {
  reset();
  lrc_type_ = LrcType::kLrcTypeDj;
  this->updateMetadata();

  emit this->onLrcUpdated();
}

void LrcManager::handleSetIsLightMusic() {
  reset();
  lrc_type_ = LrcType::kLrcTypeLightMusic;
  this->updateMetadata();

  emit this->onLrcUpdated();
}

void LrcManager::setKrc(const QString& krc) {
  reset();

  lrc_raw_ = krc;
  lrc_type_ = LrcType::kLrcTypeKrc;
  parseKrc();
}

void LrcManager::setLrc(const QString& lrc) {
  reset();

  lrc_raw_ = lrc;
  lrc_type_ = LrcType::kLrcTypeLrc;
  parseLrc();
}

void LrcManager::handleSetLrcEmpty() {
  reset();
  lrc_type_ = LrcType::kLrcTypeNotSet;

  this->updateMetadata();
  emit this->onLrcUpdated();
}

void LrcManager::handleSetLyrics(const QString& lrc, const QString& krc) {
  if (!krc.isEmpty()) {
    this->setKrc(krc);
  } else if (!lrc.isEmpty()) {
    this->setLrc(lrc);
  } else {
    qWarning() << "[LrcManager]::setLyrics() lyrics is empty";
  }

  emit this->onLrcUpdated();
}

void LrcManager::handleSetOffset(qint64 offset) {
  qDebug() << Q_FUNC_INFO << offset;
  offset_ = offset;
  // There is no need to notify desktop lyrics to update current lrc element.
}

void LrcManager::handleMediaInfoChanged() {
  updateMetadata();
}

LrcManager::LrcElement LrcManager::titleElement() {
  LrcElement lrc_element;
  lrc_element.start = 0;

  if (metadata_.title.isEmpty()) {
    updateMetadata();
  }

  QString title_and_artist(metadata_.title);
  if (!metadata_.artist.isEmpty()) {
    title_and_artist.append(" - ");
    title_and_artist.append(metadata_.artist);
  }
  lrc_element.content = title_and_artist;
  return lrc_element;
}

void LrcManager::updateMetadata() {
  MediaPlayer::TrackInfo* track_info = app->media_player->track_info();
  if (metadata_.album.isEmpty()) {
    metadata_.album = track_info->album_name;
  }
  if (metadata_.title.isEmpty()) {
    metadata_.title = track_info->song_name;
  }
  if (metadata_.artist.isEmpty()) {
    metadata_.artist = track_info->artist_name;
  }
}

}  // namespace netease

