// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_LRC_MANAGER_H_
#define SERVICE_LRC_MANAGER_H_

#include <QObject>

namespace netease {

// Parse lrc file and notify content update.
class LrcManager : public QObject {
  Q_OBJECT

 public:
  explicit LrcManager(QObject* parent = nullptr);
  ~LrcManager() override;

  struct WordElement {
    // Sustain time.
    qint64 time;

    QString word;

   public:
    bool operator<(const WordElement& other) {
      return this->time < other.time;
    }
  };

  struct LrcElement {
    // Start time of current sentence, in milliseconds.
    qint64 start = 0;

    // Set current time. Used in translations.
    // Used in krc.
    int line_number = 0;

    // Duration of current sentence, in milliseconds.
    // This value is 0 if current element is the last one.
    // Used in lrc.
    qint64 duration = 0;

    // current lrc sentence, without any time tags.
    QString content;

    // Time to keep displaying current line before switch to next one, in
    // milliseconds.
    // Used in krc.
    qint64 sustain_time = 0;

    // Each word and its time delta.
    // Used in krc.
    QList<WordElement> words;

    // Translated to Simplified Chinese. Default is empty.
    // Used in krc.
    QString zh_cn;

   public:
    bool operator<(const LrcElement& other) const {
      return this->start < other.start;
    }
  };

  LrcElement currentElement() const;
  LrcElement firstElement() const;
  LrcElement lastElement() const;
  LrcElement nextElement() const;
  // Get lrc_element which is appropriated to |position|.
  // Update |current_element_| is necessary.
  LrcElement searchElement(qint64 position);

  // Get title - artist LrcElement.
  // If no lrc or krc is available, display this element in desktop lyrics
  // window and mini player window.
  LrcElement titleElement();

  bool isFirstElement() const;
  bool isLastElement() const;

  bool isDj() const;
  bool isKrc() const;
  bool isLightMusic() const;
  bool isLrc() const;
  // Some lyric might have no time tag.
  bool isNoTimeLrc() const;

  // Check whether lrc is available.
  bool isLrcEmpty() const;

  // Displays slogan when no lrc is availabel in desktop lyrics window.
  QString slogan;

  // Update metadata based on TrackInfo in mediaplayer.
  // And prepend title-artist to |lrc_elements_|.
  void updateMetadata();

 signals:
  // Emitted when lrc content updated.
  void onLrcUpdated();

  void setIsDj();

  void setIsLightMusic();

  void setLrcEmpty();

  // Set both lrc and krc.
  void setLyrics(const QString& lrc, const QString& krc);

  void setOffset(qint64 offset);

 private:
  struct Metadata {
    QString album;
    QString artist;

    QString composer;

    // Creator of song text.
    QString creator;

    // Time when this lrc file is created.
    QString datetime;

    // Editor of lrc file.
    QString editor;

    // Song name.
    QString title;

    // Song link address to which this krc file refers.
    QString url;

    // Repeat time.
    int repeat = 0;

    // Overall timestamp adjustment in milliseconds,
    // + shifts time up, - shifts down.
    // TODO(xushaohua): put offset into lrc_elems_.
    qint64 offset = 0;
  };

  enum LrcType {
    kLrcTypeNotSet = 1,
    kLrcTypeDj = 2,
    kLrcTypeLightMusic = 4,
    kLrcTypeNoTimeLrc = 8,
    kLrcTypeEmpty = kLrcTypeNotSet | kLrcTypeDj | kLrcTypeLightMusic |
                    kLrcTypeNoTimeLrc,

    // Lrc content is not empty only if its type is lrc or krc.
    kLrcTypeLrc = 32,
    kLrcTypeKrc = 64,
    kLrcTypeNotEmpty = kLrcTypeLrc | kLrcTypeKrc,
  };

  void initConnections();

  void parseKrc();
  void parseLrc();

  // Reset lrc states.
  void reset();

  void setKrc(const QString& krc);
  void setLrc(const QString& lrc);

  LrcType lrc_type_;

  int current_element_;

  Metadata metadata_;
  QList<LrcElement> lrc_elements_;

  QString lrc_raw_;

  // Lrc time offset, controlled by web player.
  qint64 offset_;

 private slots:
  void handleSetIsDj();
  void handleSetIsLightMusic();
  void handleSetLrcEmpty();
  void handleSetLyrics(const QString& lrc, const QString& krc);
  void handleSetOffset(qint64 offset);

  void handleMediaInfoChanged();
};

bool operator<(const qint64 pos, const LrcManager::WordElement& word_element);
bool operator<(const LrcManager::WordElement& word_element, const qint64 pos);
bool operator<(const qint64 start, const LrcManager::LrcElement& lrc_element);
bool operator<(const LrcManager::LrcElement& lrc_element, const qint64 start);

}  // namespace netease


#endif  // SERVICE_LRC_MANAGER_H_
