// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by General Public License that can be found
// in the LICENSE file.

#include "service/media_metadata.h"

#include <fstream>
#include <attachedpictureframe.h>
#include <commentsframe.h>
#include <fileref.h>
#include <flacfile.h>
#include <id3v1tag.h>
#include <id3v2tag.h>
#include <mp4coverart.h>
#include <mp4file.h>
#include <mpegfile.h>
#include <textidentificationframe.h>
#include <xiphcomment.h>

#include <QDebug>

#include "application.h"
#include "base/file_util.h"

#ifdef QStringToTString
#undef QStringToTString
#endif
#define QStringToTString(s) TagLib::String(s.toUtf8().data(), TagLib::String::UTF8)

namespace netease {

namespace {

const int kAddId3Success = 1;
const int kAddId3Failed = 0;
const char kExtNameMp3[] = "mp3";
const char kExtNameMp4[] = "mp4";
const char kExtNameAac[] = "aac";
const char kExtNameFlac[] = "flac";
const char kMimeTypeJpeg[] = "image/jpeg";

const char kMP3TagCoverName[] = "APIC";
const char kMp4TagCoverName[] = "covr";

TagLib::ByteVector getPictureData(const char* filename) {
  std::ifstream is;
  is.open(filename, std::ios::binary);

  is.seekg(0, std::ios::end);
  const int length = is.tellg();
  is.seekg(0, std::ios::beg);

  char *buffer = new char[length];

  is.read(buffer, length);
  is.close();

  TagLib::ByteVector result(buffer, length);
  delete[] buffer;

  return result;
}


}  // namespace

MediaMetadata::MediaMetadata(QObject* parent) : QObject(parent) {
  setObjectName("media_metadata");
  initConnections();
}

MediaMetadata::~MediaMetadata() {

}

void MediaMetadata::handleAddId3(const QString& task_id,
                                 const QString& media_path,
                                 const QString& album_cover_path,
                                 const QString& media_info,
                                 const QVariantMap& tag) {
  qDebug() << Q_FUNC_INFO << media_info << media_path;
  AudioMetadata metadata;
  metadata.title = tag.value("tit2").toString();
  metadata.album = tag.value("talb").toString();
  metadata.artist = tag.value("tpe1").toString();
  metadata.track_id = tag.value("trck").toInt();
  metadata.comment = media_info;
  metadata.cover_filepath = album_cover_path;

  if (!QFile::exists(media_path)) {
    qWarning() << "[MediaMetadata]::handleAddId3() file not found:" << media_path;
    emit this->onAddId3(task_id, kAddId3Failed);
    return;
  }
  const QString ext_name = GetFileExtension(media_path);
  bool ok;
  if (ext_name.compare(kExtNameMp3, Qt::CaseInsensitive) == 0) {
    ok = writeMp3Tag(media_path, metadata);
  } else if (ext_name.compare(kExtNameMp4, Qt::CaseInsensitive) == 0 ||
             ext_name.compare(kExtNameAac, Qt::CaseInsensitive) == 0) {
    ok = writeMp4Tag(media_path, metadata);
  } else if (ext_name.compare(kExtNameFlac, Qt::CaseInsensitive) == 0) {
    ok = writeFlacTag(media_path, metadata);
  } else {
    qWarning() << "[MediaMetadata]::addId3() unsupported media type:"
               << media_path;
    ok = false;
  }

  emit this->onAddId3(task_id, ok ? kAddId3Success : kAddId3Failed);
}

void MediaMetadata::handleReadAudioInfo(const QString& id,
                                        const QString& filepath) {
  const AudioMetadata metadata = ReadAudioInfo(filepath);
  QVariantMap info;
  info.insert("title", metadata.title);
  info.insert("album", metadata.album);
  info.insert("artist", metadata.artist);
  info.insert("bitrate", metadata.bitrate);
  info.insert("filesize", metadata.size);
  info.insert("duration", static_cast<qlonglong>(metadata.duration));
  info.insert("index", 0);
  emit this->onReadAudioInfo(id, filepath, 0, info);
}

void MediaMetadata::handleReadCueInfo(const QString& id,
                                      const QString& filepath) {
  const CueMetadata metadata = ReadCueInfo(filepath);
  QVariantMap info;
  info.insert("audiopath", metadata.data_path);
  info.insert("cuepath", metadata.cue_path);
  info.insert("title", metadata.title);
  info.insert("performer", metadata.performer);
  info.insert("bitrate", metadata.bitrate);
  QVariantList tracks;

  for (const CueTrack& cue_track : metadata.tracks) {
    QVariantMap track;
    track.insert("start", static_cast<qlonglong>(cue_track.start));
    track.insert("title", cue_track.title);
    track.insert("filesize", cue_track.size);
    track.insert("duration", static_cast<qlonglong>(cue_track.duration));
    track.insert("index", cue_track.track_id);
    tracks.append(track);
  }

  info.insert("tracks", tracks);
  emit this->onReadCueInfo(id, 0, info, "");
}

void MediaMetadata::initConnections() {
  connect(this, &MediaMetadata::addId3,
          this, &MediaMetadata::handleAddId3);
  connect(this, &MediaMetadata::readAudioInfo,
          this, &MediaMetadata::handleReadAudioInfo);
  connect(this, &MediaMetadata::readCueInfo,
          this, &MediaMetadata::handleReadCueInfo);
}

bool MediaMetadata::writeFlacTag(const QString& media_path,
                                 const AudioMetadata& metadata) {
  TagLib::FLAC::File file(media_path.toStdString().c_str());
  if (!file.isValid() || !file.audioProperties()) {
    return false;
  }

  auto tag = file.xiphComment();
  if (!tag) {
    return false;
  }

  if (!metadata.cover_filepath.isEmpty() &&
      QFile::exists(metadata.cover_filepath)) {
    TagLib::ByteVector picture =
        getPictureData(metadata.cover_filepath.toStdString().c_str());
    if (picture.size() > 0) {
      auto frame = new TagLib::FLAC::Picture();
      if (frame) {
        file.removePictures();
        frame->setType(TagLib::FLAC::Picture::FrontCover);
        frame->setMimeType(kMimeTypeJpeg);
        frame->setData(picture);
        file.addPicture(frame);
      }
    }
  }

  tag->setAlbum(QStringToTString(metadata.album));
  tag->setArtist(QStringToTString(metadata.artist));
  tag->setTitle(QStringToTString(metadata.title));
  tag->setComment(QStringToTString(EncryptMediaComment(metadata.comment)));
  tag->setTrack(static_cast<uint>(metadata.track_id));

  return file.save();
}

bool MediaMetadata::writeMp3Tag(const QString& media_path,
                                const AudioMetadata& metadata) {
  TagLib::MPEG::File file(media_path.toStdString().c_str());
  if (!file.isValid() || !file.audioProperties()) {
    return false;
  }

  auto tag = file.ID3v2Tag();
  if (!tag) {
    return false;
  }

  TagLib::ByteVector field = kMP3TagCoverName;
  TagLib::ID3v2::FrameList apic_list = tag->frameListMap()[field];
  TagLib::ID3v2::AttachedPictureFrame *front_cover = NULL;

  // remove covers
  for (TagLib::ID3v2::FrameList::ConstIterator it = apic_list.begin();
       it != apic_list.end(); ++it) {
    TagLib::ID3v2::AttachedPictureFrame* currFrame =
        static_cast<TagLib::ID3v2::AttachedPictureFrame*>(*it);
    tag->removeFrame(currFrame, false);
  }

  if (!metadata.cover_filepath.isEmpty() &&
      QFile::exists(metadata.cover_filepath)) {
    // add new cover
    front_cover = new TagLib::ID3v2::AttachedPictureFrame(field);
    front_cover->setMimeType(kMimeTypeJpeg);
    TagLib::ByteVector picture_data =
        getPictureData(metadata.cover_filepath.toStdString().c_str());
    front_cover->setPicture(picture_data);
    front_cover->setType(TagLib::ID3v2::AttachedPictureFrame::FrontCover);
    tag->addFrame(front_cover);
  }

  tag->setAlbum(QStringToTString(metadata.album));
  tag->setArtist(QStringToTString(metadata.artist));
  tag->setTitle(QStringToTString(metadata.title));
  tag->setComment(QStringToTString(EncryptMediaComment(metadata.comment)));
  tag->setTrack(static_cast<uint>(metadata.track_id));

  return file.save(TagLib::MPEG::File::ID3v2, true, 3);
}

bool MediaMetadata::writeMp4Tag(const QString& media_path,
                                const AudioMetadata& metadata) {
  TagLib::MP4::File file(media_path.toStdString().c_str());
  if (!file.isValid() || !file.audioProperties()) {
    return false;
  }

  TagLib::MP4::Tag* tag = file.tag();
  if (!tag) {
    return false;
  }

  tag->setAlbum(QStringToTString(metadata.album));
  tag->setArtist(QStringToTString(metadata.artist));
  tag->setTitle(QStringToTString(metadata.title));
  tag->setComment(QStringToTString(EncryptMediaComment(metadata.comment)));
  tag->setTrack(static_cast<uint>(metadata.track_id));

  if (!metadata.cover_filepath.isEmpty() &&
      QFile::exists(metadata.cover_filepath)) {
    TagLib::ByteVector image_data =
        getPictureData(metadata.cover_filepath.toStdString().c_str());
    if (image_data.size() > 0) {
      TagLib::MP4::CoverArt art(TagLib::MP4::CoverArt::JPEG, image_data);
      TagLib::MP4::CoverArtList cover_art_list = TagLib::MP4::CoverArtList();
      cover_art_list.append(art);
      tag->itemListMap().insert(kMp4TagCoverName,
                                TagLib::MP4::Item(cover_art_list));
    }
  }

  return file.save();
}

}  // namespace netease
