// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/media_player.h"

#include <QDebug>
#include <QTimer>
#include <QTimerEvent>

#include "application.h"
#include "service/media_player_cache.h"

namespace netease {

namespace {

//const int kLoadDriverError = 1;
const int kLoadNetworkError = 2;
//const int kLoadCodecError = 3;
const int kLoadLocalFileError = 4;
const int kLoadParamError = 10003;
//const int kLoadUnknownError = 10004;

const int kCrossFadeDuration = 500;

// Refers to local resource file.
const int kMediaTypeLocal = 0;
// Refers to web resource file.
const int kMediaTypeStream = 4;

// Wait for 15s before emitting network error.
const int kPlaybackStuckTimeout = 15000;

}  // namespace

MediaPlayer::MediaPlayer(QObject* parent)
    : VlcMediaPlayer(parent),
      track_info_(new TrackInfo()),
      original_volume_(),
      cross_fade_enabled_(false),
      is_fade_out_mode_(false),
      cross_fade_factor_(0.0),
      cross_fade_animation_(new QPropertyAnimation(this, "cross_fade_factor")),
      loading_timeout_timer_(new QTimer(this)) {
  this->setObjectName("media_player");

  connect(this, &VlcMediaPlayer::durationChanged,
          this, &MediaPlayer::handleDurationChanged);
  connect(cross_fade_animation_, &QPropertyAnimation::finished,
          this, &MediaPlayer::onCrossFadeAnimationFinished);
  connect(loading_timeout_timer_, &QTimer::timeout,
          this, &MediaPlayer::onLoadingMediaTimeout);
  connect(this, &MediaPlayer::positionChanged,
          this, &MediaPlayer::onPositionChanged);
  connect(this, &MediaPlayer::volumeChanged,
          this, &MediaPlayer::onVolumeChanged);
  loading_timeout_timer_->setInterval(kPlaybackStuckTimeout);

  cross_fade_animation_->setDuration(kCrossFadeDuration);
}

MediaPlayer::~MediaPlayer() {
  delete track_info_;
  track_info_ = nullptr;
}

bool MediaPlayer::isCrossFadeEnabled() const {
  return cross_fade_enabled_;
}

void MediaPlayer::setCrossFade(bool on) {
  cross_fade_enabled_ = on;
}

void MediaPlayer::load(const QVariantMap& track_info) {
  qDebug() << Q_FUNC_INFO << track_info;

  emit this->trackInfoReset();

  const QVariant empty("");
  // Parse track info.
  track_info_->bitrate = track_info.value("bitrate", empty).toString();
  track_info_->duration =
      qint64(track_info.value("duration", empty).toDouble() * 1000000);
  track_info_->file_size = track_info.value("fileSize", empty).toString();
  track_info_->format = track_info.value("format", empty).toString();
  track_info_->md5 = track_info.value("md5", empty).toString();
  track_info_->musicurl = track_info.value("musicurl", empty).toString();
  track_info_->path = track_info.value("path", empty).toString();
  track_info_->playbrt = track_info.value("playbrt", empty).toString();
  track_info_->play_id = track_info.value("playId", empty).toString();
  track_info_->song_id = track_info.value("songId", empty).toString();
  track_info_->start_time = track_info.value("starttime", empty).toString();
  track_info_->type = track_info.value("type", empty).toInt();
  track_info_->volume_delta = track_info.value("volumeDelta", empty).toString();

  emit this->trackInfoChanged();

  if (track_info_->type == kMediaTypeLocal) {
    // Play local music resource.
    if (track_info_->path.length() > 0 && QFile::exists(track_info_->path)) {
      this->stop();
      this->setMedia(QUrl::fromLocalFile(track_info_->path));
      this->play();
    } else {
      qCritical() << "[MediaPlayer]::load() invalid resource path:"
                  << track_info_->path;
      emit mediaStatusChanged(MediaStatus::InvalidMedia);
      emit onLoadError(kLoadLocalFileError);
    }

  } else if (track_info_->type == kMediaTypeStream) {
    if ((!track_info_->musicurl.isEmpty()) &&
        (!track_info_->song_id.isEmpty())) {
      // Now, media is buffered automatically.
      const QString filepath(app->media_player_cache->getCachedPath(
          track_info_->song_id, track_info_->bitrate, track_info_->md5));
      const QFileInfo cache_info(filepath);
      if (cache_info.exists() && cache_info.isFile() && cache_info.size() > 0) {
        // Use cache for fade in and fade out
        this->stop();
        this->setMedia(QUrl::fromLocalFile(filepath));
        this->play();
      } else {
        // Use cache for fade in and fade out
        this->stop();
        this->setMedia(track_info_->musicurl);
        this->play();

        // Cache remote playback resource.
        app->media_player_cache->load(track_info_->musicurl,
                                      track_info_->song_id,
                                      track_info_->bitrate,
                                      track_info_->md5,
                                      track_info_->format);
      }
    } else {
      qWarning() << "[MediaPlayer]::load() invalid track_info!";
      // Emit invalid media error.
      emit mediaStatusChanged(MediaStatus::InvalidMedia);
      emit onLoadError(kLoadParamError);
    }
  }
}

void MediaPlayer::setInfo(const QVariantMap& track_info) {
  qDebug() << "[MediaPlayer]::setInfo()" << track_info;
  const QVariant empty("");

  track_info_->album_id = track_info.value("albumId", empty).toString();
  track_info_->album_name = track_info.value("albumName", empty).toString();
  track_info_->artist_name = track_info.value("artistName", empty).toString();
  track_info_->song_name = track_info.value("songName", empty).toString();
  track_info_->song_type = track_info.value("songType", empty).toString();
  track_info_->url = track_info.value("url", empty).toString();
  if (track_info_->artist_name.isEmpty()) {
    track_info_->artist_name = QObject::tr("Unknown Artist");
  }
  if (track_info_->song_name.isEmpty()) {
    track_info_->song_name = QObject::tr("Unknown Song");
  }

  emit this->trackInfoChanged();
}

void MediaPlayer::pause() {
  loading_timeout_enabled_ = false;
  loading_timeout_timer_->stop();

  if (!cross_fade_enabled_) {
    VlcMediaPlayer::pause();
    return;
  }
  this->fadeOut();
}

void MediaPlayer::play() {
  loading_timeout_enabled_ = true;
  loading_timeout_timer_->start();

  if (!cross_fade_enabled_) {
    VlcMediaPlayer::play();
    return;
  }

  this->fadeIn();
}

void MediaPlayer::stop() {
  loading_timeout_enabled_ = false;
  loading_timeout_timer_->stop();

  if (!cross_fade_enabled_) {
    VlcMediaPlayer::stop();
    return;
  }

  this->fadeOut();
}

void MediaPlayer::togglePlayPause() {
  if (this->state() == State::PlayingState) {
    this->pause();
  } else if (this->state() == State::PausedState) {
    this->play();
  }
}

MediaPlayer::TrackInfo* MediaPlayer::track_info() const {
  return track_info_;
}

void MediaPlayer::handleDurationChanged(qint64 duration) {
  track_info_->duration = duration;
  emit this->trackInfoChanged();
}

void MediaPlayer::setCrossFadeFactor(double factor) {
  cross_fade_factor_ = factor;
  this->blockSignals(true);
  const int vol = static_cast<int>(original_volume_ * cross_fade_factor_);
  VlcMediaPlayer::setVolume(vol);
  this->blockSignals(false);
}

double MediaPlayer::crossFadeFactor() const {
  return cross_fade_factor_;
}

void MediaPlayer::fadeOut() {
  if (this->media().isEmpty()) {
    VlcMediaPlayer::stop();
    return;
  }
  if (this->state() == State::PlayingState) {
    if (cross_fade_animation_->state() != QAbstractAnimation::Running) {
      // Update original volume only if cross fade animation is not running.
      original_volume_ = this->volume();
    }
    is_fade_out_mode_ = true;
    cross_fade_animation_->setStartValue(1.0);
    cross_fade_animation_->setEndValue(0.1);
    cross_fade_animation_->start();
  }
}

void MediaPlayer::fadeIn() {
  if (this->media().isEmpty()) {
    VlcMediaPlayer::stop();
    return;
  }
  if (this->state() != State::PlayingState) {
    is_fade_out_mode_ = false;
    if (cross_fade_animation_->state() != QAbstractAnimation::Running) {
      // Update original volume only if cross fade animation is not running.
      const int new_volume = this->volume();
      // Default volume of vlc player is -1.
      if (new_volume > 0) {
        original_volume_ = new_volume;
      }
    }
    cross_fade_animation_->setStartValue(0.1);
    cross_fade_animation_->setEndValue(1.0);
    cross_fade_animation_->start();
    VlcMediaPlayer::play();
  }
}

void MediaPlayer::setVolume(int volume) {
  original_volume_ = volume;
  VlcMediaPlayer::setVolume(volume);
}

void MediaPlayer::onCrossFadeAnimationFinished() {
  if (is_fade_out_mode_) {
    VlcMediaPlayer::pause();
  }
  this->blockSignals(true);
  VlcMediaPlayer::setVolume(original_volume_);
  this->blockSignals(false);
}

void MediaPlayer::onLoadingMediaTimeout() {
  if (loading_timeout_enabled_) {
    emit this->mediaStatusChanged(MediaStatus::InvalidMedia);
    emit this->onLoadError(kLoadParamError);
  }
}

void MediaPlayer::onPositionChanged() {
  // Reset loading timer when media position changed.
  if (loading_timeout_enabled_) {
    loading_timeout_timer_->start();
  }
}

void MediaPlayer::onVolumeChanged(int volume) {
  if (cross_fade_animation_->state() != QAbstractAnimation::Running) {
    // |volume| might be -1 in libvlc.
    if (volume >= 0) {
      emit this->volumeChangedWithAnimation(volume);
    }
  }
}

}  // namespace netease
