// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_MEDIA_PLAYER_H_
#define SERVICE_MEDIA_PLAYER_H_

#include "service/backend/vlc_media_player.h"

#include <QPropertyAnimation>
#include <QUrl>
#include <QVariantMap>

class QTimer;

namespace netease {

// AudioPlayer is used to play an audio music source.
// Currently it wraps up libvlc.
class MediaPlayer : public VlcMediaPlayer {
  Q_OBJECT

  Q_PROPERTY(double cross_fade_factor
                 READ crossFadeFactor
                 WRITE setCrossFadeFactor)

 public:
  explicit MediaPlayer(QObject* parent = nullptr);
  ~MediaPlayer() override;

  // Set volume with animation
  void pause();
  void play();
  void stop();

  bool isCrossFadeEnabled() const;
  void setCrossFade(bool on);
  double crossFadeFactor() const;

  // TrackInfo is used to record current song item, including song id,
  // album cover url, media format, etc.
  struct TrackInfo {
    QString artist_name;

    // Album id is like "album2263099".
    QString album_id;

    QString album_name;

    // Bitrate is like "192".
    QString bitrate;

    // Duration is like "295.131".
    // Converted to ns.
    qint64 duration;

    // File size is like "7078808".
    QString file_size;

    // Media format is like "mp3".
    QString format;

    // MD5 is like "94c0c1975ee8c41afbc855ee79cb9571".
    QString md5;

    // Music resource url is like "http://m7.music.126.net/20151207122503/" +
    // "fa865c7c0a73cb604115c5d25e970351/ymusic/787d/df1e/b165/" +
    // "94c0c1975ee8c41afbc855ee79cb9571.mp3".
    QString musicurl;

    // Path to local media resource file.
    QString path;

    // Playing bitrate is like "192".
    QString playbrt;

    // Playing id is like "25642096_1_3712001-bitrate-192".
    QString play_id;

    QString song_id;

    QString song_name;

    // Song type is like "normal".
    QString song_type;

    QString start_time;

    // Type of music source:
    //   * "0" means local music resource.
    //   * "4" means web music resource.
    int type;

    // Album cover url (large images).
    // e.g."http://p4.music.126.net/666304046473315/666304046473315.jpg".
    QString url;

    // Volume delta is like "-2.59".
    QString volume_delta;
  };

  // Get track info currently used.
  TrackInfo* track_info() const;

 signals:
  // Emitted when failed to load media resource.
  void onLoadError(int error_code);

  // Emitted when track info is reset.
  void trackInfoReset();

  void trackInfoChanged();

  void volumeChangedWithAnimation(int volume);

 public slots:
  // Load resource info.
  void load(const QVariantMap& track_info);

  // Update media resource info, like artist name and song name.
  void setInfo(const QVariantMap& track_info);

  void setVolume(int volume);

  // Switch play state between Play and Pause.
  void togglePlayPause();

  void setCrossFadeFactor(double factor);

 private:
  TrackInfo* track_info_;

  // For fade in and fade out
  int original_volume_;
  bool cross_fade_enabled_;
  bool is_fade_out_mode_;
  double cross_fade_factor_;
  QPropertyAnimation* cross_fade_animation_ = nullptr;

  QTimer* loading_timeout_timer_ = nullptr;
  bool loading_timeout_enabled_ = false;

 private slots:
  void fadeOut();
  void fadeIn();

  void handleDurationChanged(qint64 duration);
  void onCrossFadeAnimationFinished();
  void onLoadingMediaTimeout();
  void onPositionChanged();
  void onVolumeChanged(int volume);
};

}  // namespace netease

#endif  // SERVICE_MEDIA_PLAYER_H_
