// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/media_player_cache.h"

#include <QDateTime>
#include <QDebug>

#include "application.h"
#include "base/file_util.h"
#include "service/storage_manager.h"
#include "service/thread_manager.h"

namespace netease {

namespace {

// Default format of cached song.
const char kDefaultFormat[] = "mp3";

// Folder name to keep local cached songs.
const char kCacheDirName[] = "CachedSongs";

// Filename of temporary cache song file.
const char kTempFileName[] = "temp-file";

QDir GetCacheDir() {
  return ConcateDir(GetUserCacheDir(), kCacheDirName);
}

// Get song filename based on its track info.
QString GetSongFileName(const QString& song_id, const QString& bitrate,
                        const QString& md5, const QString& format) {
  QString filename("%1-%2-%3.%4");
  return filename.arg(song_id).arg(bitrate).arg(md5).arg(format);
}

// Remove oldest files.
void ClearOldFiles(QDir root, qint64 total_capacity) {
  const QStringList name_filters = { "*.mp3" };
  QFileInfoList files =
      root.entryInfoList(name_filters, QDir::Files | QDir::NoDotAndDotDot);
  qint64 current_size = 0;
  QMap<uint, QString> file_timestamps;
  for (int i = files.size() - 1; i >= 0; --i) {
    const QFileInfo file = files.at(i);
    file_timestamps.insert(file.lastModified().toTime_t(),
                           file.absoluteFilePath());
    current_size += file.size();
  }

  if (current_size < total_capacity || file_timestamps.size() == 0) {
    // No need to clear local files.
    return;
  }

  // Remove oldest files.
  QList<uint> keys = file_timestamps.keys();
  std::sort(keys.begin(), keys.end(), std::less<uint>());
  for (int i = 0; i < keys.size(); ++i) {
    QFileInfo file_info(file_timestamps.value(keys.at(i)));
    current_size -= file_info.size();
    QFile::remove(file_info.absoluteFilePath());
    if (current_size < total_capacity) {
      break;
    }
  }
}

}  // namespace

MediaPlayerCache::MediaPlayerCache(QObject* parent)
    : SimpleDownloadManager(parent),
      base_dir_(GetCacheDir()),
      temp_filepath_(GetCacheDir().absoluteFilePath(kTempFileName)) {
  this->setObjectName("media_player_cache");
  this->initWorkers(2);
  this->initConnections();
}

MediaPlayerCache::~MediaPlayerCache() {
}

void MediaPlayerCache::clearCachedSongs() {
  qDebug() << "[MediaPlayerCache]::clearCachedSongs()";
  const QStringList name_filters = { "*.mp3" };
  QFileInfoList files =
      base_dir_.entryInfoList(name_filters, QDir::Files | QDir::NoDotAndDotDot);
  for (int i = files.size() - 1; i >= 0; --i) {
    const QFileInfo file = files.at(i);
    QFile::remove(file.absoluteFilePath());
  }
}

QString MediaPlayerCache::getCachedPath(const QString& song_id,
                                        const QString& bitrate,
                                        const QString& md5) const {
  const QString name = GetSongFileName(song_id, bitrate, md5, kDefaultFormat);
  return base_dir_.absoluteFilePath(name);
}

void MediaPlayerCache::handleCacheSongDownloaded(const QString& url,
                                                 const QString& filepath) {
  Q_UNUSED(url);
  Q_UNUSED(filepath);
  // Clear expired files if no more disk capacity is available.
  ClearOldFiles(base_dir_, GetCacheCapacity());
}

void MediaPlayerCache::initConnections() {
  connect(this, &MediaPlayerCache::load,
          this, &MediaPlayerCache::handleLoad);
  connect(this, &MediaPlayerCache::onDownloaded,
          this, &MediaPlayerCache::handleCacheSongDownloaded);
}

bool MediaPlayerCache::isCached(const QString& song_id, const QString& bitrate,
                                const QString& md5) const {
  const QString filename(
      GetSongFileName(song_id, bitrate, md5, kDefaultFormat));
  return QFile::exists(base_dir_.absoluteFilePath(filename));
}

void MediaPlayerCache::handleLoad(const QString& url, const QString& song_id,
                                  const QString& bitrate, const QString& md5,
                                  const QString& format) {
  qDebug() << "[MediaPlayerCache]::load():" << url;
  const QString filename(
      GetSongFileName(song_id, bitrate, md5, format));
  const QString filepath(base_dir_.absoluteFilePath(filename));

  emit this->download(url, filepath);
}


}  // namespace netease
