// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/popup_menu_manager.h"

#include <QCursor>
#include <QDebug>
#include <QIcon>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonValue>
#include <QPixmapCache>
#include <QKeySequence>
#include <QMenu>
#include <QRegExp>

#include "application.h"
#include "base/device_util.h"
#include "channel/orca_reply.h"
#include "service/signal_manager.h"
#include "service/system_tray_icon.h"
#include "ui/views/widgets/popup_menu_style.h"

namespace netease {

namespace {

//const int kMenuTypeMainWindow = 0;
//const int kMenuTypeMiniPlayer = 1;
const int kMenuTypeSystemTray = 2;
//const int kMenuTypeDesktopLyrics = 3;

// Only use black icons in Ubuntu Unity.
QString ReplaceWithBlackIcon(const QString& path) {
  if (IsUbuntuUnity()) {
    if (!path.contains("menu/black/")) {
      return QString(path).replace(QString("menu/"), QString("menu/black/"));
    }
  }
  return path;
}

// Remove keyboard shortcut sequence in text.
QString RemoveShortcut(const QString& text) {
  int index = text.indexOf(QChar('('));
  if (index > 0) {
    return text.left(index).remove(QChar(' '));
  }
  index = text.indexOf("（");
  if (index > 0) {
    return text.left(index).remove(QChar(' '));
  }

  return QString(text).remove(QChar(' '));
}

// Construct menu or submenu based on its configuration.
void ConstructMenu(QMenu* menu, const QJsonArray& content,
                   const QJsonObject& hotkey) {
  for (const QJsonValue& item : content) {
    const QJsonObject obj(item.toObject());
    const bool enabled = obj.value("enable").toBool();
    const bool is_menu = obj.value("menu").toBool();
    const QJsonValue children(obj.value("children"));
    if (! enabled || ! is_menu) {
      continue;
    }
    const bool separator = obj.value("separator").toBool();
    if (separator) {
      menu->addSeparator();

    } else if (children.isNull()) {
      // Append menu item.
      const QString text = RemoveShortcut(obj.value("text").toString());
      const QString image_url =
          ReplaceWithBlackIcon(obj.value("image_path").toString());
      const QString check_image_url =
          ReplaceWithBlackIcon(obj.value("check_image_path").toString());

      QAction* action;
      if (image_url.length() > 0) {
        const QIcon icon(OrcaReply::OrcaToQrc(QUrl(image_url)));
        action = menu->addAction(icon, text);
      } else {
        action = menu->addAction(text);
      }
      action->setCheckable(true);
      action->setChecked(check_image_url.length() > 0);

      if (check_image_url.length() > 0) {
        QPixmap pixmap(OrcaReply::OrcaToQrc(QUrl(check_image_url)));
        QPixmapCache::insert(kCheckedIconCacheKey, pixmap);
      }

      action->setData(obj.value("menu_id").toString());
    } else {
      const QString text(RemoveShortcut(obj.value("text").toString()));
      const QString image_url =
          ReplaceWithBlackIcon(obj.value("image_path").toString());
      QMenu* sub_menu = new QMenu();
      sub_menu->setStyle(new PopupMenuStyle());
      QAction* menu_action = menu->addMenu(sub_menu);
      if (image_url.length() > 0) {
        const QIcon icon(OrcaReply::OrcaToQrc(QUrl(image_url)));
        menu_action->setIcon(icon);
        menu_action->setText(text);
      } else {
        menu_action->setText(text);
      }

      // Construct submenu.
      ConstructMenu(sub_menu, children.toArray(), hotkey);

      if (IsUbuntuUnity()) {
        // Pass through triggered signal to root menu.
        QObject::connect(sub_menu, &QMenu::triggered,
                         menu, &QMenu::triggered);
      }
    }
  }
}

}  // namespace

PopupMenuManager::PopupMenuManager(QObject* parent) : QObject(parent) {
  this->setObjectName("popup_menu_manager");

  menu_ = new QMenu();
  menu_->setStyle(new PopupMenuStyle());

  initConnections();
}

PopupMenuManager::~PopupMenuManager() {
  delete menu_;
  menu_ = nullptr;
}

void PopupMenuManager::handleMenuActionTriggered(QAction* action) {
  const QString menu_id = action->data().toString();
  emit onMenuItemClicked(menu_id);
  if (IsUbuntuUnity() && menu_id == "100") {
    // Show main window in Ubuntu Unity if Music Info item is activated.
    emit app->signal_manager->onRequestChangeToMain();
  }
}

void PopupMenuManager::initConnections() {
  connect(menu_, &QMenu::triggered,
          this, &PopupMenuManager::handleMenuActionTriggered);
  connect(menu_, &QMenu::aboutToHide, this, &PopupMenuManager::onMenuHide);

  if (IsUbuntuUnity()) {
    connect(app->system_tray_icon->contextMenu(), &QMenu::triggered,
            this, &PopupMenuManager::handleMenuActionTriggered);
  }

  connect(this, &PopupMenuManager::hidePopupMenu,
          this, &PopupMenuManager::handleHidePopupMenu);
  connect(this, &PopupMenuManager::popupMenu,
          this, &PopupMenuManager::handlePopupMenu);
}

void PopupMenuManager::handleHidePopupMenu() {
  menu_->hide();
}

void PopupMenuManager::handlePopupMenu(const QVariantMap& conf, int menu_type) {
  const QByteArray content_bytes(conf.value("content").toByteArray());
  const QJsonArray content(QJsonDocument::fromJson(content_bytes).array());

  const QByteArray hotkey_bytes(conf.value("hotkey").toByteArray());
  const QJsonObject hotkey(QJsonDocument::fromJson(hotkey_bytes).object());

  if (IsUbuntuUnity()) {
    if (menu_type == kMenuTypeSystemTray) {
      QMenu* menu = app->system_tray_icon->contextMenu();
      menu->clear();
      ConstructMenu(menu, content, hotkey);
    } else {
      menu_->clear();
      ConstructMenu(menu_, content, hotkey);
      menu_->popup(QCursor::pos());
    }
  } else {
    // If desktop session is not Ubuntu Unity, use default popup menu for
    // system tray.
    menu_->clear();
    ConstructMenu(menu_, content, hotkey);
    menu_->popup(QCursor::pos());
  }
}

}  // namespace netease
