// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_CONTROLLERS_SIGNAL_MANAGER_H_
#define SERVICE_CONTROLLERS_SIGNAL_MANAGER_H_

#include <QObject>

namespace netease {

// SignalManager is used to transfer signals between QObjects.
class SignalManager : public QObject {
  Q_OBJECT

 public:
  explicit SignalManager(QObject* parent = nullptr);
  ~SignalManager() override;

 signals:
  // Request to close desktop lyric window.
  void onLrcRequestClose();

  // Request to move lrc content backward/forward slightly.
  void onLrcRequestBackward();
  void onLrcRequestForward();

  // Request to change font size of lrc.
  void onLrcRequestFontSize(int font_size);

  // Request to lock/unlock desktop lyric window.
  void onLrcRequestLock();
  void onLrcRequestUnlock();

  void onLrcRequestMouseScroll(int button, int offset,
                               int offset_x, int offset_y);

  // Request to popup setting menu.
  void onLrcRequestSettingMenu();

  // Mini player requests to popup context menu on top panel.
  void onMiniPlayerContextMenu();

  // Mini player requests to popup context menu on playlist record.
  void onMiniPlayerMenu(const QString& play_id);

  // Emitted when mini player changes |mode| of playing process.
  // |mode| might be "playorder", "playcycle" or "playrandom".
  void onPlayMode(const QString& mode);

  // Emitted when mini player requests loading next song.
  void onPlayNext();

  // Emitted when mini player requests loading previous song.
  void onPlayPrev();

  // Change to standard mode from mini-player mode.
  void onRequestChangeToMain();

  // Mini player request to change volume.
  void onRequestChangeVolume(int volume);

  // Toggle current program favorite state.
  void onRequestFavorSong(int favour);

  // Add current played song to red-heart playlist or remove from it.
  void onRequestLikeSong();

  // Delete current song from fm list and switch to next one.
  void onRequestDeleteFm();

  // Request to close main window.
  void onRequestMainWindowClose();

  // Pause current playback.
  void onRequestPause();

  // Mini player requests to play song with id |play_id|.
  void onRequestPlay(const QString& play_id);

  // Resume current playback.
  void onRequestResume();

  // Request to switch to detail page in main window.
  void onRequestShowDetail();

  // Set favor mark (+1).
  // Emitted by Player::setFavour().
  void onSetFavorMark(bool favor);

  // Set red-heart mark.
  // Emitted by Player::setLikeMark().
  void onSetLikeMark(int state);

  // Emit this signal to play local media at |media_path|.
  void onPlayLocalMusic(const QString& media_path);

  // Emitted when album cover is updated.
  // |filepath| is refered to default cover image when local cover image
  // is not ready.
  void onAlbumCoverUpdated(const QString& filepath);

  // For LrcCompoundFrame.
  void onRequestRenderLrcImage(const QString& text, const QString& filename);
  void onRenderLrcImageResult(bool ok, int image_width, int image_height);
  void onSetLrcLineMode(bool is_single_line);
  void onSetLrcColor(const QString& base_color1,
                     const QString& base_color2,
                     const QString& played_color1,
                     const QString& played_color2);
  void onSetLrcFont(const QString& font_size, const QString& unused,
                    const QString& font_family);
  void onSetLrcTextAlignment(const QString& primary_alignment,
                         const QString& secondary_alignment);

  // Request to update context menu in system tray.
  // This signal is used in Ubuntu Unity.
  void onUpdateSystemTrayContextMenu();
};

}  // namespace netease

#endif  // SERVICE_CONTROLLERS_SIGNAL_MANAGER_H_
