// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/skin_manager.h"

#include <QDebug>
#include <QDir>
#include <QDomDocument>
#include <QDomElement>
#include <QDomNodeList>
#include <QFileInfo>
#include <QVariant>

#include "application.h"
#include "base/file_util.h"
#include "service/settings_manager.h"

namespace netease {

namespace {

// Placeholder text name used in css style file. This word will be
// replaced by actual absolute folder path in formatStyleFile().
const char kSkinRootNameInCss[] = "SKIN_ROOT";

// Path to root of skin folder in qrc format.
const char kSkinRootPath[] = ":/webapp/style/skin/";

// Group name used in settings file.
const char kSettingsSkinGroupName[] = "ServiceSkinManager";

// Key used to record currently used skin in settings file.
const char kSettingsCurrentSkinName[] = "current-skin";

// File name of info file in each skin packet.
const char kSkinInfoFileName[] = "web/info";

// File name of menu style.
const char kMenuElementFileName[] = ":/webapp/style/skin/%1/menu/element.xml";

// File name of mini playe style.
const char kMiniPlayerSkinFileName[] = ":/webapp/style/skin/%1/mini/skin.xml";

// File name of lrc style.
const char kLrcSkinFileName[] = ":/webapp/style/skin/%1/lrc/skin.xml";

// Type name of local skin.
const char kSkinLocal[] = "default";

// "red" skin packet can be used to customize button colors.
const char kCustomizableSkin[] = "red";

// Text colors to be applied with current theme packet.
const QString kMenuItemTextColor("MENU_ITEM_TEXT_COLOR");
const QString kNoLrcArtistColor("NO_LRC_ARTIST_COLOR");
const QString kNoLrcTitleColor("NO_LRC_TITLE_COLOR");
const QString kInfoPanelArtistColor("INFO_PANEL_ARTIST_COLOR");
const QString kInfoPanelTitleColor("INFO_PANEL_TITLE_COLOR");
const QString kPlaylistItemTextColor("PLAYLIST_ITEM_TEXT_COLOR");
const QString kPlaylistItemSelectedColor("PLAYLIST_ITEM_SELECTED_COLOR");

// Convert QColor to RGBA, QColor::name() can be used to convert that color to
// ARGB, which is not we needed.
QString QColorToRGBA(const QColor& color) {
  QString color_text("rgba(%1, %2, %3, %4)");
  return color_text.arg(color.red())
                   .arg(color.green())
                   .arg(color.blue())
                   .arg(color.alpha() / 255.0);
}

// Update |color| with |color_text|. |color_text| is RGB or RGBA.
//void QColorFromRGBA(QColor& color, const QString& color_text) {
//  if (color_text.length() != 7 && color_text.length() != 9) {
//    qWarning() << "QColorFromRGBA() invalid color:" << color_text;
//    return;
//  }

//  if (color_text.at(0) == '#') {
//    color.setRed(color_text.mid(1, 2).toInt(nullptr, 16));
//    color.setGreen(color_text.mid(3, 2).toInt(nullptr, 16));
//    color.setBlue(color_text.mid(5, 2).toInt(nullptr, 16));
//    if (color_text.length() == 9) {
//      color.setAlpha(color_text.mid(7, 2).toInt(nullptr, 16));
//    }
//  }
//}

}  // namespace

SkinManager::SkinManager(QObject* parent)
    : QObject(parent),
      current_skin_(),
      button_color_() {
  this->setObjectName("skin_manager");
}

SkinManager::~SkinManager() {

}

QString SkinManager::absoluteSkinPath(const QString& path) {
  return QString(path).replace(QString(kSkinRootNameInCss),
                               QString(kSkinRootPath) + currentSkin());
}

bool SkinManager::changeSkin(const QString& skin_type,
                             const QString& skin_name,
                             const QColor& button_color) {
  button_color_ = button_color;

  if (skin_type == kSkinLocal) {
    QString filepath(kSkinRootPath);
    filepath.append(skin_name);
    QDir skin_dir(filepath);
    if (skin_dir.exists()) {
      current_skin_ = skin_name;
      SettingsSetValue(kSettingsSkinGroupName,
                       kSettingsCurrentSkinName,
                       skin_name);

      this->updateTextColors();

      emit this->onUpdated(current_skin_);
      return true;
    } else {
      qWarning() << "[SkinManager]::changeSkin() skin does not exist!"
                 << ", skin_name: " << skin_name
                 << ", filepath: " << filepath;
      return false;
    }
  } else {
    qWarning() << "[SkinManager]::changeSkin() is not local skin:"
               << ", skin_type: " << skin_type
               << ", skin_name: " << skin_name;
    return false;
  }
}

QString SkinManager::currentSkin() {
  if (current_skin_.isEmpty()) {
    current_skin_ = SettingsGetValue(kSettingsSkinGroupName,
                                     kSettingsCurrentSkinName).toString();
  }
  return current_skin_;
}

QString SkinManager::formatStyle(const QString &content) {
  QString style(content);
  style.replace(kMenuItemTextColor,
                QColorToRGBA(text_colors.menu_item));

  style.replace(kNoLrcArtistColor,
                QColorToRGBA(text_colors.no_lrc_artist));
  style.replace(kNoLrcTitleColor,
                QColorToRGBA(text_colors.no_lrc_title));

  style.replace(kInfoPanelArtistColor,
                QColorToRGBA(text_colors.info_panel_artist));
  style.replace(kInfoPanelTitleColor,
                QColorToRGBA(text_colors.info_panel_title));

  style.replace(kPlaylistItemTextColor,
                QColorToRGBA(text_colors.playlist_text));
  style.replace(kPlaylistItemSelectedColor,
                QColorToRGBA(text_colors.playlist_text));
  return style;
}

QString SkinManager::formatStyleFile(const QString& filepath) {
  const QString style = absoluteSkinPath(ReadTextFileContent(filepath));
  return this->formatStyle(style);
}

QList<SkinManager::Skin> SkinManager::getLocalSkins() {
  QList<Skin> skin_list;

  const QDir root_dir(kSkinRootPath);
  if (root_dir.exists()) {
    for (const QFileInfo& file_info : root_dir.entryInfoList(QDir::Dirs)) {
      const QDir skin_dir(file_info.absoluteFilePath());
      const QString info_file(skin_dir.absoluteFilePath(kSkinInfoFileName));
      QString skin_info;
      if (ReadTextFile(info_file, skin_info)) {
        Skin skin = {skin_dir.dirName() , skin_info};
        skin_list.append(skin);
      }
    }
  } else {
    qWarning() << "[SkinManager]::getLocalSkin() is empty!";
  }

  return skin_list;
}

bool SkinManager::isCustomizedSkin() const {
  return current_skin_ == kCustomizableSkin;
}

void SkinManager::updateTextColors() {
  const QString menu_filepath(QString(kMenuElementFileName).arg(currentSkin()));
  QDomDocument menu_doc;
  menu_doc.setContent(ReadTextFileContent(menu_filepath));

  if (! menu_doc.isNull()) {
    QDomNodeList nodes = menu_doc.elementsByTagName("MenuLabel");
    if (nodes.count() == 1) {
      QDomElement element = nodes.at(0).toElement();
      text_colors.menu_item = QColor(element.attribute("textcolor"));
      text_colors.menu_disabled_item
          = QColor(element.attribute("disabledtextcolor"));
    } else {
      qWarning() << "[SkinManager]::updateTextColors() failed to find .MenuLabel";
    }
  }

  const QString mini_player_filepath(
      QString(kMiniPlayerSkinFileName).arg(currentSkin()));
  QDomDocument mini_player_doc;
  mini_player_doc.setContent(ReadTextFileContent(mini_player_filepath));

  if (mini_player_doc.isNull()) {
    qWarning() << "[SkinManager]::updateTextColors() failed to parse skin.xml";
    return;
  }

  QDomNodeList playlist_nodes = mini_player_doc.elementsByTagName("PlayList");
  if (playlist_nodes.count() == 1) {
    QDomElement element = playlist_nodes.at(0).toElement();
    text_colors.playlist_hot = QColor(element.attribute("itemhottextcolor"));
    text_colors.playlist_selected
        = QColor(element.attribute("itemselectedtextcolor"));
    text_colors.playlist_text = QColor(element.attribute("itemtextcolor"));
  } else {
    qWarning() << "[SkinManager]::updateTextColors() failed to find .PlayList";
  }

  QDomNodeList lrc_nodes = mini_player_doc.elementsByTagName("SlideText");
  if (lrc_nodes.count() == 1) {
    QDomElement element = lrc_nodes.at(0).toElement();
    text_colors.lrc_text = QColor(element.attribute("textcolor"));
  } else {
    qWarning() << "[SkinManager]::updateTextColors() failed to find .SlideText";
  }

  QDomNodeList label_nodes = mini_player_doc.elementsByTagName("Label");
  for (int i = 0; i < label_nodes.length(); ++i) {
    QDomElement element = label_nodes.at(i).toElement();
    const QString label_name = element.attribute("name");
    if (label_name == "wordbar_info") {
      text_colors.info_panel_artist = QColor(element.attribute("textcolor1"));
      text_colors.info_panel_title = QColor(element.attribute("textcolor"));
    } else if (label_name == "no_lrc_title") {
      text_colors.no_lrc_title = QColor(element.attribute("textcolor"));
    } else if (label_name == "no_lrc_artist") {
      text_colors.no_lrc_artist = QColor(element.attribute("textcolor"));
    }
  }

  // Read default lrc styles from current skin.
  const QString lrc_filepath(
      QString(kLrcSkinFileName).arg(currentSkin()));
  QDomDocument lrc_doc;
  lrc_doc.setContent(ReadTextFileContent(lrc_filepath));

  if (! lrc_doc.isNull()) {
    // Parse shadowColor in "textOutlineContainer".
    QDomNodeList outline_nodes = lrc_doc.elementsByTagName("TextOutline");
    if (outline_nodes.length() == 2) {
      QDomElement primary_element = outline_nodes.at(0).toElement();
      text_colors.lrc_primary_shadow =
          QColor(primary_element.attribute("shadowColor"));
      QDomElement secondary_element = outline_nodes.at(1).toElement();
      text_colors.lrc_secondary_shadow =
          QColor(secondary_element.attribute("shadowColor"));
    }
  }
}

}  // namespace netease
