// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "service/storage_manager.h"

#include <QDateTime>
#include <QDebug>
#include <QDirIterator>
#include <QPixmap>
#include <QStandardPaths>
#include <QThread>

#include "application.h"
#include "base/consts.h"
#include "base/file_util.h"
#include "service/album_cover_cache.h"
#include "service/media_metadata.h"
#include "service/media_player_cache.h"
#include "service/settings_manager.h"
#include "service/backend/database_worker.h"
#include "service/backend/filesystem_watcher.h"

namespace netease {

namespace {

// Folder name of album cover.
// That is $HOME/.cache/cloud-music/AlbumCover.
const char kAlbumCoverDirName[] = "AlbumCover";

// Folder name to store temporary files.
// That is $HOME/.cache/cloud-music/TempFiles.
const char kTempFilesDirName[] = "TempFiles";

// Default capacity used as cache, 1GB.
const double kDefaultCacheCapacity = 1.0;

// Maximum space used as cache, 10GB.
const double kMaximumCacheCapacity = 10.0;

// Minimum space used as cache, 0.5GB.
const double kMinimumCacheCapacity = 0.5;

// Folder name to $Home/Music.
const char kUserMusicFolderName[] = "Music";

// Group name in cloud-music.ini settings file.
const char kStorageSettingsGroupName[] = "ServiceStorage";

// Key of cache_capacity in settings file.
const char kSettingsCacheCapacityName[] = "cache-capacity";

// Key of cache_path in settings file.
const char kSettingsCachePathName[] = "cache-path";

// Key of download_path in settings file.
const char kSettingsDownloadPathName[] = "download-path";

// Days of file expration, 30days.
const int kExpirationDays = 30;

// Check whether file at |filepath| is too old or not.
// If that file is too old or does not exist, returns true.
// Else returns false.
// TODO(xushaohua): Clear old files at an interval.
bool IsFileExpired(const QString& filepath) {
  const QFileInfo file_info(filepath);
  if (! file_info.exists()) {
    // 1. Check file existence.
    return true;
  } else if (file_info.size() == 0) {
    // 2. Check file is empty or not.
    return true;
  } else {
    const int days_delta = int(file_info.lastModified().daysTo(
        QDateTime::currentDateTime()));
    return days_delta >= kExpirationDays;
  }
}

// |filepath| in saveToFile() and readFromFile() contains reversed-slash(\).
// This function converts reversed-slash to dash(-).
QString FilterFilePath(const QString& filepath) {
  return QString(filepath).replace("\\", "-").replace("/", "-");
}

// Walk through |root|.
void TreeWalk(const QFileInfo& root, QVariantList& result) {
  qDebug() << "[StorageManager]::TreeWalk() root:" << root.absoluteFilePath();
  Q_ASSERT(root.isDir());

  QDir root_dir(root.absoluteDir());
  root_dir.setFilter(QDir::NoDotAndDotDot | QDir::Files |
                     QDir::Dirs | QDir::NoSymLinks);
  QFileInfoList info_list = root_dir.entryInfoList();

  for (int i = 0; i < info_list.size(); ++i) {
    QFileInfo file_info = info_list.at(i);
    QVariantMap item;
    const QString abs_filepath = file_info.absoluteFilePath();
    item.insert("name", abs_filepath);
    if (file_info.isDir()) {
      item.insert("type", "path");
      result.append(item);
      TreeWalk(file_info, result);
    } else if (file_info.isFile()) {
      item.insert("type", "file");
      result.append(item);
    }
  }
}

}  // namespace

qint64 GetCacheCapacity() {
  const QVariant cap = SettingsGetValue(kStorageSettingsGroupName,
                                        kSettingsCacheCapacityName);
  return static_cast<qint64>(cap.toDouble() * (1 << 30));
}

QString GetDefaultMusicDownloadPath() {
  QString music_folder(
      QStandardPaths::writableLocation(QStandardPaths::MusicLocation));
  if (music_folder.length() == 0) {
    music_folder = ConcateDir(QDir::home(),
        QObject::tr(kUserMusicFolderName)).absolutePath();
  }
  return ConcateDir(QDir(music_folder),
      QObject::tr(kDefaultMusicDownloadDirName)).absolutePath();
}

QString ScaleImage(const QString& filepath, int width, int height,
                   int scale_type) {
  if (!QFile::exists(filepath)) {
    qWarning() << "[StorageManager]::ScaleImage() file not found:" << filepath;
    return "";
  }
  QPixmap image(filepath);
  QPixmap temp_image;
  if (scale_type == 2) {
    temp_image = image.scaledToWidth(width);
  } else if (scale_type == 1) {
    temp_image = image.scaledToHeight(height);
  } else {
    temp_image = image.scaled(width, height);
  }
  QString temp_path("/tmp/");
  temp_path.append(GetUrlHash(filepath));
  temp_path.append(".jpg");
  const bool ok = temp_image.save(temp_path, "JPG", 80);
  return ok ? temp_path : "";
}

StorageManager::StorageManager(QObject* parent)
    : QObject(parent),
      album_cover_dir_(),
      storage_cache_dir_(),
      temp_files_dir_(),
      download_dir_(),
      db_worker_(new DatabaseWorker(this)),
      fs_watcher_(new FilesystemWatcher()),
      fs_thread_(new QThread()) {
    this->setObjectName("storage_manager");

  const QDir base_dir(GetUserCacheDir());
  album_cover_dir_ = ConcateDir(base_dir, kAlbumCoverDirName);
  storage_cache_dir_ = ConcateDir(base_dir, kStorageCacheDirName);
  temp_files_dir_ = ConcateDir(base_dir, kTempFilesDirName);

  fs_thread_->start();
  fs_watcher_->moveToThread(fs_thread_);

  this->initDB();
  this->initConnections();
}

StorageManager::~StorageManager() {
  if (fs_watcher_ != nullptr) {
    delete fs_watcher_;
    fs_watcher_ = nullptr;
  }
  if (fs_thread_->isRunning()) {
    fs_thread_->quit();
    fs_thread_->wait(3);
    fs_thread_->deleteLater();
    fs_thread_ = nullptr;
  }
}

void StorageManager::handleCancelMoveStorage() {
  qDebug() << Q_FUNC_INFO;
  // Note(xushaoahua): Does nothing on linux platform.

  // Emit moving operation has finished signal.
  emit onMoveStorage(1.0, 0);
}

void StorageManager::handleCheckFilesExist(const QString& task_id,
                                           const QVariantList& files,
                                           const QString& download_path) {
  qDebug() << Q_FUNC_INFO << task_id << files << download_path;
  QVariantList result;
  for (const QVariant& file : files) {
    const QVariantMap info = file.toMap();
    const QVariant id = info.value("id");
    const QString path = info.value("path").toString();
    if (id.isValid()) {
      QVariantMap item;
      item.insert("id", id);
      const bool exist = QFile::exists(download_path + path);
      item.insert("exist", exist);
      result.append(item);
    }
  }

  const int success = 1;
  emit onCheckFilesExist(task_id, success, result);
}

void StorageManager::handleClearCache() {
  qDebug() << Q_FUNC_INFO;
  app->media_player_cache->clearCachedSongs();
  emit this->onClearCache(true);
}

void StorageManager::handleDeleteFile(const QString& task_id,
                                      const QString& path_type,
                                      const QString& tag,
                                      const QString& filepath,
                                      bool delete_empty_dir) {
  qDebug() << Q_FUNC_INFO << task_id << path_type << tag << filepath << delete_empty_dir;
  Q_UNUSED(task_id);
  Q_UNUSED(tag);

  if (path_type == "abs") {
    QFile::remove(filepath);
  } else if (path_type == "rel") {
    QFile::remove(getStorageCacheFilepath(filepath));
  } else {
    qWarning() << "[StorageManager]::deleteFile() failed:" << filepath;
  }

  if (delete_empty_dir) {
    // TODO(xushaohua): delete empty dir.
  }
}

QString StorageManager::getAlbumCover(const QString& url) {
  qDebug() << Q_FUNC_INFO << url;
  const QString filepath(album_cover_dir_.filePath(GetUrlHash(url)));

  if (!IsFileExpired(filepath)) {
    return filepath;
  } else {
    // Call AlbumCoverCache::download() in background thread.
    emit app->album_cover_cache->download(url, filepath);

    // Returns an empty string to notify that album cover is not cached yet.
    return "";
  }
}

QDir StorageManager::getDownloadDir() const {
  qDebug() << Q_FUNC_INFO;
  return download_dir_;
}

QString StorageManager::getStorageCacheFilepath(const QString& filepath) {
  qDebug() << Q_FUNC_INFO << filepath;
  const QString abs_path(storage_cache_dir_.filePath(filepath));
  CreateParentDirs(abs_path);
  return abs_path;
}

QString StorageManager::getLocalStorageFilePath(const QString& filename) {
  qDebug() << Q_FUNC_INFO << filename;
  return temp_files_dir_.absoluteFilePath(FilterFilePath(filename));
}

QString StorageManager::copyTempFile(const QString& filename) {
  qDebug() << Q_FUNC_INFO << filename;
  const QString old_filepath = this->getLocalStorageFilePath(filename);
  const QString new_filepath =
      this->getLocalStorageFilePath(QString("temp-%1").arg(filename));
  QFile::remove(new_filepath);
  QFile::copy(old_filepath, new_filepath);
  return new_filepath;
}

void StorageManager::handleGetTempFile(const QString& url) {
  qDebug() << Q_FUNC_INFO << url;
  const QString content = ReadTextFileContent(getTempFilepath(url));
  emit onGetTempFile(url, content);
}

QString StorageManager::getTempFilepath(const QString& url) {
  qDebug() << Q_FUNC_INFO << url;
  return temp_files_dir_.absoluteFilePath(GetUrlHash(url));
}

void StorageManager::handleInit(const QString download_path,
                                double cache_capacity) {
  qDebug() << Q_FUNC_INFO << download_path << cache_capacity;
  double new_cache_capacity = cache_capacity;
  if (new_cache_capacity < kMinimumCacheCapacity ||
      new_cache_capacity > kMaximumCacheCapacity) {
    new_cache_capacity = kDefaultCacheCapacity;
  }

  SettingsSetValue(kStorageSettingsGroupName,
                   kSettingsDownloadPathName,
                   download_path);

  updateDownloadDir(download_path);

  SettingsSetValue(kStorageSettingsGroupName,
                   kSettingsCacheCapacityName,
                   new_cache_capacity);
}

void StorageManager::initConnections() {
  connect(app->album_cover_cache, &AlbumCoverCache::onDownloaded,
          this, &StorageManager::onAlbumCoverDownloaded);
  connect(app->album_cover_cache, &AlbumCoverCache::onError,
          this, &StorageManager::onAlbumCoverError);

  connect(this, &StorageManager::cancelMoveStorage,
          this, &StorageManager::handleCancelMoveStorage);
  connect(this, &StorageManager::checkFilesExist,
          this, &StorageManager::handleCheckFilesExist);
  connect(this, &StorageManager::clearCache,
          this, &StorageManager::handleClearCache);
  connect(this, &StorageManager::deleteFile,
          this, &StorageManager::handleDeleteFile);
  connect(this, &StorageManager::downloadScanner,
          this, &StorageManager::handleDownloadScanner);
  connect(this, &StorageManager::execSql,
          this, &StorageManager::handleExecSql);
  connect(this, &StorageManager::execTransaction,
          this, &StorageManager::handleExecTransaction);
  connect(this, &StorageManager::getTempFile,
          this, &StorageManager::handleGetTempFile);
  connect(this, &StorageManager::imagesInfo,
          this, &StorageManager::handleImagesInfo);
  connect(this, &StorageManager::init,
          this, &StorageManager::handleInit);
  connect(this, &StorageManager::listFile,
          this, &StorageManager::handleListFile);
  connect(this, &StorageManager::moveFiles,
          this, &StorageManager::handleMoveFiles);
  connect(this, &StorageManager::moveStorage,
          this, &StorageManager::handleMoveStorage);
  connect(this, &StorageManager::queryCacheTracks,
          this, &StorageManager::handleQueryCacheTracks);
  connect(this, &StorageManager::queryNewCacheTracks,
          this, &StorageManager::handleQueryNewCacheTracks);
  connect(this, &StorageManager::readFromFile,
          this, &StorageManager::handleReadFromFile);
  connect(this, &StorageManager::removeTempFile,
          this, &StorageManager::handleRemoveTempFile);
  connect(this, &StorageManager::saveToFile,
          this, &StorageManager::handleSaveToFile);
  connect(this, &StorageManager::scaleImage,
          this, &StorageManager::handleScaleImage);
  connect(this, &StorageManager::scanDirectory,
          this, &StorageManager::handleScanDirectory);
  connect(this, &StorageManager::updateTempFile,
          this, &StorageManager::handleUpdateTempFile);

  connect(fs_watcher_, &FilesystemWatcher::fileAdded,
          this, &StorageManager::handleFileAdded);
  connect(fs_watcher_, &FilesystemWatcher::fileDeleted,
          this, &StorageManager::handleFileDeleted);
}

void StorageManager::initDB() {
  const QString db_file = GetUserConfigDir().filePath(kOnlineDB);
  if (!db_worker_->open(db_file)) {
    qCritical() << Q_FUNC_INFO << "Failed to open database file";
  }
}

void StorageManager::handleDownloadScanner(const QString& path,
                                           int limit,
                                           const QStringList& exist_files) {
  qDebug() << Q_FUNC_INFO << path << limit << exist_files;
  QStringList abs_existed_files;
  for (const QString& file : exist_files) {
    abs_existed_files.append(path + file);
  }
  this->scanMediaFiles(path, abs_existed_files, limit);
}

void StorageManager::scanMediaFiles(const QString& parent_dir,
                                    const QStringList& existed_files,
                                    int limit) {
  qDebug() << Q_FUNC_INFO << parent_dir << existed_files << limit;
  QVariantList items;

  QDirIterator dir_iter(parent_dir,
                        QDir::Files | QDir::NoDotAndDotDot | QDir::NoSymLinks,
                        QDirIterator::Subdirectories);
  while (dir_iter.hasNext()) {
    dir_iter.next();
    qDebug() << Q_FUNC_INFO << "file info:" << dir_iter.filePath();
    if (!dir_iter.fileInfo().isFile()) {
      qDebug() << Q_FUNC_INFO << "is not file:" << dir_iter.filePath();
      continue;
    }
    const QString filepath = dir_iter.filePath();
    qDebug() << Q_FUNC_INFO << "filepath:" << filepath;

    // Ignores existed files.
    if (existed_files.contains(filepath)) {
      qDebug() << Q_FUNC_INFO << "file in existed file list";
      continue;
    }

    // Check file extension first.
    // mp3|mp4|aac|wma|wav|ogg|m4a|ape|flac
    if (!IsMediaFile(filepath)) {
      qDebug() << Q_FUNC_INFO << "is not media file";
      continue;
    }

    // Check encrypted comment data.
    AudioMetadata metadata = ReadAudioInfo(filepath);
    if (!IsValidMediaComment(metadata.comment)) {
      qDebug() << Q_FUNC_INFO << "is not valid media comment" << metadata.comment;
      continue;
    }

    QVariantMap item;
    // Use relative path instead
    item.insert("path", download_dir_.relativeFilePath(filepath));
    item.insert("size", metadata.size);
    item.insert("comment", DecryptMediaComment(metadata.comment));
    items.append(item);

    // Emit onDownloadScanner() signal if needed.
    if (items.length() >= limit) {
      QVariantList tmp_items(items);
      qDebug() << Q_FUNC_INFO << "Emit onDownloadScanner() signal" << items;
      emit this->onDownloadScanner(tmp_items);
      items.clear();
    }
  }
  // Emit remaining items.
  emit this->onDownloadScanner(items);
}

void StorageManager::handleExecSql(const QString& id,
                                   const QString& raw_statement) {
  qDebug() << Q_FUNC_INFO << id << raw_statement;
  bool ok = true;
  const QVariantList result = db_worker_->rawExec(raw_statement, &ok);
  const int err = ok ? 0 : 1;
  emit onExecSqlDone(id, err, result);
}

void StorageManager::handleExecTransaction(const QString& id,
                                           const QString& statement) {
  qDebug() << Q_FUNC_INFO << id << statement;
  bool ok = true;
  db_worker_->execTransaction(statement, &ok);
  const int err = ok ? 0 : 1;
  const QVariantList result;
  emit this->onExecSqlDone(id, err, result);
}

void StorageManager::handleFileAdded(const QString& root_dir,
                                     const QString& filepath) {
  qDebug() << Q_FUNC_INFO << root_dir << filepath;
  AudioMetadata metadata = ReadAudioInfo(filepath);
  if (IsValidMediaComment(metadata.comment)) {
    QVariantList items;
    QVariantMap item;
    // Emit signal only if comment is encrypted with specific method.
    const QString comment = DecryptMediaComment(metadata.comment);
    qDebug() << Q_FUNC_INFO << "decrypted comment:" << comment;
    item.insert("comment", comment);
    const QString filename = filepath.mid(root_dir.length());
    // |path| is relative to music folder.
    item.insert("path", filename);
    item.insert("size", metadata.size);
    items.append(item);
    emit this->onDownloadScanner(items);
  }
}

void StorageManager::handleFileDeleted(const QString& root_dir,
                                       const QString& filepath) {
  qDebug() << Q_FUNC_INFO << root_dir << filepath;
  // Remove folder name prefix.
  const QString filename = filepath.mid(root_dir.length());
  qDebug() << Q_FUNC_INFO << filename;
  emit this->onFileDeleted(filename);
}

void StorageManager::handleListFile(const QString& task_id,
                                    const QString& path_type,
                                    const QString& tag,
                                    const QString& filepath) {
  qDebug() << Q_FUNC_INFO << path_type << tag << filepath;
  Q_UNUSED(tag);
  QVariantList result;

  if (path_type == "abs") {
    TreeWalk(QFileInfo(filepath), result);
  } else if (path_type == "rel") {
    QFileInfo root_path(getStorageCacheFilepath(filepath));
    TreeWalk(root_path, result);
  }

  const int err = 0;
  emit onListFile(task_id, err, result);
}

void StorageManager::handleMoveFiles(const QString& operation,
                                     const QString& src_path_type,
                                     const QString& arg1,
                                     const QStringList& src_files,
                                     const QString& dest_path_type,
                                     const QString& arg2,
                                     const QStringList& dest_files) {
  qDebug() << Q_FUNC_INFO << operation << src_path_type << arg1
           << src_files << dest_path_type << arg2 << dest_files;
  Q_UNUSED(src_path_type);
  Q_UNUSED(dest_path_type);
  Q_UNUSED(arg1);
  Q_UNUSED(arg2);
  const int kCodeError = 1;
  const int kCodeOk = 0;
  const int kStateDone = 0;
  const int kStateInProcess = 1;

  QVariantMap state;
  if (src_files.length() != dest_files.length()) {
    state.insert("code", kCodeError);
    state.insert("state", kStateDone);
    state.insert("errcode", 1);
    qDebug() << "state:" << state;
    emit onMoveProcess(state);
    return;
  }
  if (operation == "move") {
    const int num_files = src_files.length();
    for (int i = 0; i < num_files; ++i) {
      qDebug() << "Create parent dirs:" << dest_files.at(i);
      CreateParentDirs(dest_files.at(i));
      const bool ok = QFile::rename(src_files.at(i), dest_files.at(i));
      const double process = i * 1.0 / num_files;
      state.insert("process", process);
      state.insert("code", ok ? kCodeOk : kCodeError);
      state.insert("errcode", 0);
      state.insert("state",
                   (i + 1 == num_files) ? kStateDone : kStateInProcess);
      const int remain_time = num_files - i - 1;
      state.insert("remain", remain_time);
      state.insert("src", src_files.at(i));
      state.insert("dst", dest_files.at(i));
      qDebug() << "state:" << state;
      emit this->onMoveProcess(state);
    }
  }
}

void StorageManager::handleMoveStorage(const QString& src, const QString& dest,
                                       int capacity) {
  qDebug() << Q_FUNC_INFO << src << dest << capacity;
  Q_UNUSED(src);
  Q_UNUSED(dest);
  Q_UNUSED(capacity);
  // Emit moving operation has finished signal.
//  emit onMoveStorage(1.0, 0);
}

void StorageManager::handleQueryCacheTracks() {
  qDebug() << Q_FUNC_INFO;
  QVariantList result;
  emit onQueryCacheTracks(result);
}

void StorageManager::handleQueryNewCacheTracks(const QString& task_id,
                                               const QVariantMap& song_info) {
  qDebug() << Q_FUNC_INFO << task_id << song_info;
  Q_UNUSED(song_info);
  QVariantList tracks;
  emit onQueryNewCacheTracks(task_id, tracks);
}

void StorageManager::handleReadFromFile(const QString& task_id,
                                        const QString& filepath) {
  qDebug() << Q_FUNC_INFO << task_id << filepath;
  QString content;
  const bool ok = ReadTextFile(getStorageCacheFilepath(filepath),
                                     content);
  int err = ok ? 0 : 1;
  emit onReadFromFileDone(task_id, err, content);
}

void StorageManager::handleRemoveTempFile(const QString& url) {
  qDebug() << Q_FUNC_INFO << url;
  QFile::remove(getTempFilepath(url));
}

void StorageManager::handleSaveToFile(const QString& task_id,
                                      const QString& content,
                                      const QString& tag,
                                      const QString& filepath) {
  qDebug() << Q_FUNC_INFO << task_id << content << tag << filepath;
  Q_UNUSED(tag);
  const bool ok = WriteTextFile(getStorageCacheFilepath(filepath),
                                      content);
  int err = ok ? 0 : 1;
  emit onSaveToFileDone(task_id, err);
}

void StorageManager::handleScaleImage(const QString &task_id,
                                      const QVariantMap &info) {
  qDebug() << "[StorageManager]::handleScaleImage()" << task_id << info;
  const QString path = info.value("path").toString();
  const int width = info.value("width").toInt();
  const int height = info.value("height").toInt();
  const int scale_type = info.value("scale_type").toInt();

  if (!QFile::exists(path)) {
    // File not found error.
    emit this->onScaleImage(task_id, 1, "");
    return;
  }

  const QString temp_path = ScaleImage(path, width, height, scale_type);
  emit this->onScaleImage(task_id, temp_path.isEmpty() ? 1 : 0, temp_path);
}

void StorageManager::handleScanDirectory(const QString& task_id,
                                         const QString& path_type,
                                         int scan_type,
                                         const QString& path) {
  qDebug() << "TODO [StorageManager]::scanDirectory()" << task_id
           << path_type << scan_type << path;
  emit this->onScanDirectory(task_id);
}

void StorageManager::handleUpdateTempFile(const QString& url,
                                          const QString& content,
                                          const QString& mime) {
  qDebug() << Q_FUNC_INFO << url << content << mime;
  Q_UNUSED(mime);
  const bool state = WriteTextFile(getTempFilepath(url), content);
  emit this->onUpdateTempFile(url, state);
}

void StorageManager::updateDownloadDir(const QString& dirpath) {
  // NOTE(xushaohua): No need to remove file system watch.
//  emit fs_watcher_->rmWatchFolder(download_dir_.absolutePath());
  emit fs_watcher_->watchFolder(dirpath, true);
  download_dir_ = QDir(dirpath);
}

void StorageManager::handleImagesInfo(const QString &id,
                                      const QStringList &files) {
  qDebug() << "[StorageManager]::handleImagesInfo()" << id << files;
  QVariantList result;
  for (const QString& file : files) {
    QPixmap image(file);
    QVariantMap obj;
    obj.insert("code", 0);
    obj.insert("width", image.width());
    obj.insert("height", image.height());
    QFileInfo fileinfo(file);
    obj.insert("size", fileinfo.size());
    obj.insert("path", file);
    result.append(obj);
  }

  qDebug() << "[StorageManager]::handleImageInfo() result:" << result;
  emit this->onImagesInfo(id, result);
}

}  // namespace netease
