// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef SERVICE_STORAGE_MANAGER_H_
#define SERVICE_STORAGE_MANAGER_H_

#include <QDir>
#include <QObject>
#include <QVariant>

class QThread;

namespace netease {

class DatabaseWorker;
class FilesystemWatcher;

// Returns capacity settings of cache-folder.
qint64 GetCacheCapacity();

// Get absolute path to default folder which stores downloaded music resources.
QString GetDefaultMusicDownloadPath();

// Scale image at |filepath| to |width|x|height|.
// Save scaled image to /tmp folder with jpeg format.
QString ScaleImage(const QString& filepath, int width, int height, int scale_type);

// Manages resource storage(cache) settings.
// Note that on linux platform, all resources cache are stored at
// $HOME/.cache/cloud-music and does not support moving storage cache.
// So some methods in this class do nothing.
// This object is running in background thread.
class StorageManager : public QObject {
  Q_OBJECT

 public:
  explicit StorageManager(QObject* parent = nullptr);
  ~StorageManager() override;

 signals:
  // Emitted after getAlbumCover() finished.
  void onAlbumCoverDownloaded(const QString& url, const QString& filepath);

  // Emitted after getAlbumCover() goes wrong.
  void onAlbumCoverError(const QString& url, const QString& filepath);

  void onCheckFilesExist(const QString& task_id, int success,
                         const QVariantList& result);

  void onClearCache(bool ok);

  // Emitted after new media files are added to |download_dir_|.
  void onDownloadScanner(const QVariantList& items);

  void onExecSqlDone(const QString& id, int err, const QVariantList& ret);

  // Emitted when songs are added to or deleted from |download_dir_|.
  void onFileDeleted(const QString& file);

  void onGetTempFile(const QString& url, const QString& content);

  void onMoveProcess(const QVariantMap& state);

  // This signal is emitted after moveStorage() is called.
  // |percent| 0.0~1.0, 1.0 means moving storage finished.
  // |est_time| estimated time in seconds. It should be 0 when |percent|
  // is 1.0.
  // This signal is emitted immediatelly after calling moveStorage() on linux.
  void onMoveStorage(double percent, int est_time);

  void onImagesInfo(const QString& id, const QVariantList& result);

  void onListFile(const QString& task_id, int err, const QVariantList& result);

  void onQueryCacheTracks(const QVariantList& result);

  void onQueryNewCacheTracks(const QString& task_id,
                             const QVariantList& tracks);

  // Emitted after readFromFile() is done.
  void onReadFromFileDone(const QString& task_id, int err,
                          const QString& content);

  // Emitted after saveToFile() is done.
  void onSaveToFileDone(const QString& task_id, int err);

  void onScaleImage(const QString& task_id, int result, const QString& path);

  void onScanDirectory(const QString& task_id);

  void onUpdateTempFile(const QString& url, bool ok);

  // Cancel move storage operation.
  // This function is useful when moving resources storage to a different
  // filesystem.
  // This method does nothing on linux platform.
  // Call/emit onmovedstorage() on end.
  void cancelMoveStorage();

  // Check whether |files| exist in |download_path|.
  // Emit onCheckFilesExist() signal.
  void checkFilesExist(const QString& task_id, const QVariantList& files,
                       const QString& download_path);

  void clearCache();

  // Delete a local file.
  // |path_type| might be "abs" or "rel". If it is relative path, find
  // |filepath| in |storage_cache_dir_|.
  // |tag| is not used.
  void deleteFile(const QString& task_id, const QString& path_type,
                  const QString& tag, const QString& filepath,
                  bool delete_empty_dir);

  void downloadScanner(const QString& path, int limit,
                       const QStringList& exist_files);

  void execSql(const QString& id, const QString& statement);
  void execTransaction(const QString& id, const QString& statement);

  // Read temporary file.
  void getTempFile(const QString& url);

  void imagesInfo(const QString& id, const QStringList& files);

  // Initialize |download_path| and |cache_path|.
  // If |download_path| is empty, returns $HOME/Music.
  // |cache_capacity| is space used as local cache, in GB. When user moves
  // capacity slide in webview page, this value is updated and stored to
  // local settings file. If more space is used when application is running,
  // clear some old files.
  // |cache_path| is always set to $HOME/.cache/clous-music and unchangable.
  void init(const QString download_path, double cache_capacity);

  void listFile(const QString& task_id, const QString& path_type,
                const QString& tag, const QString& filepath);

  void moveFiles(const QString& operation, const QString& src_path_type,
                 const QString& arg1, const QStringList& src_files,
                 const QString&  dest_path_type, const QString& arg2,
                 const QStringList& dest_files);

  // Move resources storage from |src| to |dest|.
  // |src| might be empty for the first time.
  // |capacity| is in unit of GB.
  // When move operation finished, emit/call onmovedstorage().
  // This slot does nothing on linux.
  void moveStorage(const QString& src, const QString& dest, int capacity);

  void queryCacheTracks();

  void queryNewCacheTracks(const QString& task_id,
                           const QVariantMap& song_info);

  // Read cache file stored in |storage_cache_dir_| and
  // emit onReadFromFileDone() signal.
  void readFromFile(const QString& task_id, const QString& filepath);

  // Remove temporary file.
  void removeTempFile(const QString& url);

  // Write cache file to |storage_cache_dir_| and emit
  // onSaveToFileDone() signal.
  // |tag| is not used.
  void saveToFile(const QString& task_id, const QString& content,
                  const QString& tag, const QString& filepath);

  void scaleImage(const QString& id, const QVariantMap& info);

  void scanDirectory(const QString& task_id,const QString& path_type,
                     int scan_type, const QString& path);

  // Write |content| to temporary file.
  // |mime| is ignored.
  void updateTempFile(const QString& url, const QString& content,
                      const QString& mime);

 public slots:
  // Get album cover specifying with |url| and returns absolute filepath to it.
  // If that file is not cached on disk, download it, cache it and emit
  // onAlbumDownloaded() signal, or onAlbumError() signal if failed.
  // Note: Call this method in UI thread.
  QString getAlbumCover(const QString& url);

  // Get absolute path to download folder.
  QDir getDownloadDir() const;

  // Return absolute path to local storage file.
  QString getLocalStorageFilePath(const QString& filename);

  // Copy files in TempFiles and returns its path.
  QString copyTempFile(const QString& filename);

  // Get absolute path to cache file to which |url| refers.
  QString getTempFilepath(const QString& url);

 private:
  // Filter |filepath| when writing to |storage_cache_dir_|.
  QString getStorageCacheFilepath(const QString& filepath);

  // Scan media files in |parent_dir| and append to |items|.
  // If items length is greater than |limit|, emit onDownloadScanner()
  // and clear |items|.
  // Media files listed in |existed_files| are ignored.
  void scanMediaFiles(const QString& parent_dir,
                      const QStringList& existed_files,
                      int limit);

  void initConnections();

  void updateDownloadDir(const QString& dirpath);

  QDir album_cover_dir_;
  QDir storage_cache_dir_;
  QDir temp_files_dir_;
  QDir download_dir_;

  DatabaseWorker* db_worker_ = nullptr;
  FilesystemWatcher* fs_watcher_ = nullptr;
  QThread* fs_thread_ = nullptr;

 private slots:
  void handleCancelMoveStorage();

  void handleCheckFilesExist(const QString& task_id, const QVariantList& files,
                             const QString& download_path);

  void handleClearCache();

  void handleDeleteFile(const QString& task_id, const QString& path_type,
                        const QString& tag, const QString& filepath,
                         bool delete_empty_dir);

  void handleDownloadScanner(const QString& path, int limit,
                             const QStringList& exist_files);

  void handleExecSql(const QString& id, const QString& statement);
  void handleExecTransaction(const QString& id, const QString& statement);

  void handleFileAdded(const QString& root_dir, const QString& filepath);
  void handleFileDeleted(const QString& root_dir, const QString& filepath);

  void handleGetTempFile(const QString& url);

  void handleImagesInfo(const QString& id, const QStringList& files);

  void handleInit(const QString download_path, double cache_capacity);

  void handleListFile(const QString& task_id, const QString& path_type,
                      const QString& tag, const QString& filepath);

  void handleMoveFiles(const QString& operation, const QString& src_path_type,
                       const QString& arg1, const QStringList& src_files,
                       const QString&  dest_path_type, const QString& arg2,
                       const QStringList& dest_files);

  void handleMoveStorage(const QString& src, const QString& dest, int capacity);

  void handleQueryCacheTracks();

  void handleQueryNewCacheTracks(const QString& task_id,
                                 const QVariantMap& song_info);

  void handleReadFromFile(const QString& task_id, const QString& filepath);

  void handleRemoveTempFile(const QString& url);

  void handleSaveToFile(const QString& task_id, const QString& content,
                        const QString& tag, const QString& filepath);

  void handleScaleImage(const QString& task_id, const QVariantMap& info);

  void handleScanDirectory(const QString& task_id,const QString& path_type,
                           int scan_type, const QString& path);

  void handleUpdateTempFile(const QString& url, const QString& content,
                            const QString& mime);

  void initDB();
};

}  // namespace netease

#endif  // SERVICE_STORAGE_MANAGER_H_
