/*
 * Copyright (C) 2017 ~ 2017 Deepin Technology Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "tests/vlc_player_window.h"

#include <QDebug>
#include <QFileDialog>
#include <QFileInfo>
#include <QHBoxLayout>

namespace netease {

VlcPlayerWindow::VlcPlayerWindow(QWidget* parent)
    : QFrame(parent),
      player_(new PlayerBackend(this)) {
  this->initUI();
  this->initConnections();
}

void VlcPlayerWindow::initUI() {
  play_button_ = new QPushButton(tr("Play"), this);
  pause_button_ = new QPushButton(tr("Pause"), this);
  duration_label_ = new QLabel(this);
  position_label_ = new QLabel(this);
  slider_ = new QSlider(Qt::Horizontal, this);

  QHBoxLayout* layout = new QHBoxLayout(this);
  layout->addWidget(play_button_);
  layout->addWidget(pause_button_);
  layout->addWidget(slider_);
  layout->addWidget(position_label_);
  layout->addWidget(duration_label_);
  layout->addWidget(duration_label_);
  this->setLayout(layout);

  this->resize(640, 80);

  dialog_shortcut_ = new QShortcut(QKeySequence("Ctrl+O"), this);
}

void VlcPlayerWindow::initConnections() {
  connect(dialog_shortcut_, &QShortcut::activated,
          this, &VlcPlayerWindow::showFileChooserDialog);
  connect(play_button_, &QPushButton::clicked,
          player_, &PlayerBackend::play);
  connect(pause_button_, &QPushButton::clicked,
          player_, &PlayerBackend::pause);
  connect(slider_, &QSlider::sliderMoved,
          this, &VlcPlayerWindow::onSliderMoved);
  connect(player_, &PlayerBackend::positionChanged,
          this, &VlcPlayerWindow::onPlayerPositionChanged);
  connect(player_, &PlayerBackend::durationChanged,
          this, &VlcPlayerWindow::onPlayerDurationChanged);
  connect(player_, &PlayerBackend::mediaStatusChanged,
          this, &VlcPlayerWindow::onPlayerMediaStatusChanged);
}

void VlcPlayerWindow::showFileChooserDialog() {
  QFileDialog dialog;
  dialog.setAcceptMode(QFileDialog::AcceptOpen);
  dialog.setFileMode(QFileDialog::ExistingFiles);
  dialog.setMimeTypeFilters({"*.mp3"});
  dialog.exec();
  const QStringList selectedFiles = dialog.selectedFiles();
  qDebug() << "Selected files:" << selectedFiles;
  if (selectedFiles.isEmpty()) {
    qWarning() << "No media selected!";
  } else {
    audio_files_ = selectedFiles;
    this->loadMedia(audio_files_.first());
  }
}

void VlcPlayerWindow::onSliderMoved(int value) {
  const qint64 pos = value * player_->duration() / 100;
  qDebug() << Q_FUNC_INFO << "new position:" <<  pos;
  player_->setPosition(pos);
}

void VlcPlayerWindow::onPlayerPositionChanged(qint64 position) {
  const qint64 duration = player_->duration();
  if (duration > 0) {
    const int value = static_cast<int>(position * 100 / player_->duration());
    position_label_->setText(QString::number(position));
    if (!slider_->isSliderDown()) {
      slider_->setValue(value);
    } else {
      qCritical() << "Slider is pressed";
    }
  }
}

void VlcPlayerWindow::onPlayerDurationChanged(qint64 duration) {
  duration_label_->setText(QString::number(duration));
}

void VlcPlayerWindow::loadMedia(const QString& filepath) {
  QFileInfo info(filepath);
  this->setWindowTitle(info.fileName());
  player_->setMedia(QUrl::fromLocalFile(filepath));
  player_->play();
}

void
VlcPlayerWindow::onPlayerMediaStatusChanged(PlayerBackend::MediaStatus status) {
  qDebug() << Q_FUNC_INFO << status;
  switch (status) {
    case PlayerBackend::MediaStatus::EndOfMedia: {
//      const QString current_media = player_->media().canonicalUrl().toLocalFile();
      const QString current_media = player_->media().toLocalFile();
      const int index = audio_files_.indexOf(current_media);
      if (index == -1) {
        qWarning() << "Media not found:" << current_media;
      } else if (index == audio_files_.size() - 1) {
        // Reaches end of audio files, starting from the first track.
        this->loadMedia(audio_files_.first());
      } else {
        this->loadMedia(audio_files_.at(index + 1));
      }
      break;
    }
    default: {
      break;
    }
  }
}

}  // namespace netease