// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/delegates/playlist_delegate.h"

#include <QDebug>
#include <QIcon>
#include <QModelIndex>
#include <QPixmap>
#include <QPainter>
#include <QJsonObject>

#include "application.h"
#include "ui/models/song_info.h"
#include "ui/models/playlist_model.h"
#include "ui/views/mini_player_consts.h"
#include "ui/views/widgets/colored_icon_button.h"
#include "service/skin_manager.h"

namespace netease {

const int kDelegateItemWidth = 280;
const int kDelegateItemHeight = 30;
const int kDelegateLeftMargin = 6;
const int kDelegateRightMargin = 30;
const int kPixmapSize = 20;

const char kRadioNLButton[] = "SKIN_ROOT/mini/list/radio_nl.png";
const char kRadioHLButton[] = "SKIN_ROOT/mini/list/radio_hl.png";
//const char kDJNLButton[] = "SKIN_ROOT/mini/list/dj_nl.png";
//const char kDJHLButton[] = "SKIN_ROOT/mini/list/dj_hl.png";
const char kItemHoverBG[] = "SKIN_ROOT/mini/list/hover.png";
const char kItemNormalBG[] = "SKIN_ROOT/mini/list/itmbk.png";
const char kItemPlayingBG[] = "SKIN_ROOT/mini/list/playing.png";
const char kItemSelectedBG[] = "SKIN_ROOT/mini/list/selected.png";

PlaylistDelegate::PlaylistDelegate(QObject* parent)
  : QItemDelegate(parent) {
  updatePixmap();
  connect(app->skin_manager, &SkinManager::onUpdated,
          this, &PlaylistDelegate::updatePixmap);
}

//QWidget*  PlaylistDelegate::createEditor(QWidget* parent,
//const QStyleOptionViewItem &option,
//const QModelIndex &index) const
//{
//    MiniPlaylistItem*  editor = new MiniPlaylistItem(parent);
//    editor->resize(100, 30);

//    return editor;
//}

void PlaylistDelegate::destroyEditor(QWidget* editor,
                                     const QModelIndex& index) const {
  Q_UNUSED(index);
  editor->deleteLater();
}

void PlaylistDelegate::setEditorData(QWidget* editor,
                                     const QModelIndex& index) const {
  Q_UNUSED(editor)
  Q_UNUSED(index)
//    if (!index.isValid()) {
//        return;
//    }
//    QJsonObject obj = index.model()->data(index, Qt::DisplayRole)
//    .toJsonObject();
//    MiniPlaylistItem* item = static_cast<MiniPlaylistItem* >(editor);

//    if (!obj.isEmpty() && item) {
//        SongInfo info;
//        obj >> info;
//        item->setTitle(info.title);
//        item->setSongPlayState(info.play_state);
//        item->setIsRadioProgram(info.program > 0);
//    }
}

void PlaylistDelegate::updateEditorGeometry(QWidget* editor,
                                            const QStyleOptionViewItem& option,
                                            const QModelIndex& index) const {
  Q_UNUSED(index)
  editor->setGeometry(option.rect);
}

void PlaylistDelegate::paint(QPainter* painter,
                             const QStyleOptionViewItem& option,
                             const QModelIndex& index) const {
  QJsonObject obj = index.model()->data(index, Qt::DisplayRole).toJsonObject();
  if (!obj.isEmpty()) {
    SongInfo info;
    obj >> info;

    QString itemBgS;
    QString radioS = kRadioNLButton;
    QColor textColor;
    QPixmap statePixmap;

    //Check item state
    if (option.state & QStyle::State_MouseOver) {
      itemBgS = kItemHoverBG;
      radioS = kRadioHLButton;
      textColor = app->skin_manager->text_colors.playlist_hot;
    }
    else if (option.state & QStyle::State_Selected) {
      itemBgS = kItemSelectedBG;
      textColor = app->skin_manager->text_colors.playlist_selected;
    }
    else {
      itemBgS = kItemNormalBG;
      textColor = app->skin_manager->text_colors.playlist_text;
    }

    //Check play state
    switch (info.play_state) {
    case Playing:
      statePixmap = playing_Pixmap_;
      itemBgS = kItemPlayingBG;
      break;
    case Paused:
      statePixmap = paused_pixmap_;
      break;
    default:
      statePixmap = stoped_pixmap_;
      break;
    }

    //draw background image
    painter->drawPixmap(option.rect,
                        QPixmap(app->skin_manager->absoluteSkinPath(itemBgS)));

    //draw state pixmap
    QRect sr(kDelegateLeftMargin,
             option.rect.y() + (option.rect.height() - kPixmapSize) / 2,
             kPixmapSize,
             kPixmapSize);
    painter->drawPixmap(sr, statePixmap);

    //draw title
    QRect tr(sr.width() + sr.x(),
             option.rect.y(),
             option.rect.width() - kPixmapSize * 3 - kDelegateRightMargin,
             option.rect.height());
    painter->setPen(QPen(textColor));
    painter->drawText(tr, Qt::AlignLeft | Qt::AlignVCenter, info.title);

    //draw radio pixmap if exist
    if (info.program > 0) {
      QPixmap radioPixmap(app->skin_manager->absoluteSkinPath(radioS));
      QRect rr(option.rect.width() - radioPixmap.width() - kDelegateRightMargin,
               option.rect.y()
               + (option.rect.height() - radioPixmap.height()) / 2,
               radioPixmap.width(),
               radioPixmap.height());
      painter->drawPixmap(rr, radioPixmap);
    }

    //TODO(match-yang) draw dj pixmap if exist
  }
}

QSize PlaylistDelegate::sizeHint(const QStyleOptionViewItem& option,
                                 const QModelIndex& index) const {
  Q_UNUSED(option)
  Q_UNUSED(index)

  return QSize(kDelegateItemWidth, kDelegateItemHeight);
}

void PlaylistDelegate::updatePixmap() {
  if (app->skin_manager->isCustomizedSkin()) {
    stoped_pixmap_ = ColoredIconButton(0, 0, 0, 0, 20, 1).normalPixmap();
    paused_pixmap_ = ColoredIconButton(100, 154, 100, 154, 20, 20).normalPixmap();
    playing_Pixmap_ = ColoredIconButton(65, 96, 65, 96, 20, 20).normalPixmap();
  }
  else {
    const QString buttons_filepath(
      app->skin_manager->absoluteSkinPath(kCollectionButtons));
    const QPixmap buttons(buttons_filepath);
    // Height of null pixmap is 1px.
    stoped_pixmap_ = buttons.copy(0, 0, 20, 1);
    paused_pixmap_ = buttons.copy(100, 154, 20, 20);
    playing_Pixmap_ = buttons.copy(65, 96, 20, 20);
  }
}

}  // namespace netease
