// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/views/desktop_lyrics.h"

#include <QCloseEvent>
#include <QCoreApplication>
#include <QDebug>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QScreen>
#include <QTimer>
#include <QWindow>

#include "application.h"
#include "service/lrc_manager.h"
#include "service/signal_manager.h"
#include "service/skin_manager.h"
#include "service/settings_manager.h"
#include "ui/utils/xutil.h"
#include "ui/views/desktop_lyrics_unlock_window.h"
#include "ui/views/lrc_compound_frame.h"
#include "ui/views/widgets/icon_button.h"

namespace netease {

namespace {

const int kDefaultBorderWidth = 10;

const int kWindowMaxWidth = 9999;  // screen width.
const int kWindowMinWidth = 542;

//const int kWindowDefaultWidth = 662;
//const int kWindowDefaultHeight = 80;
const int kControlButtonsHeight = 35;

const int kSeparatorWidth = 4;
const int kBackgroundLabelBorderWidth = 6;

const char kButtonsFile[] = "SKIN_ROOT/lrc/button.png";
//const char kFrameFile[] = "SKIN_ROOT/lrc/frame.png";
const char kSeparatorFile[] = "SKIN_ROOT/lrc/seperator.png";
//const char kShadowFile[] = "SKIN_ROOT/lrc/shadow.png";

const char kStyleSheetFile[] = ":/styles/desktop-lyrics.css";

// Emit enter event after 500ms.
const int kCursorEnterDelay = 500;
const int kCursorLeaveDelay = 500;

const int kCursorHoverTimeout = 50;

const char kSettingsGroupName[] = "UIDesktopLyrics";
const char kSettingsWindowPositionName[] = "window-position";

}  // namespace

DesktopLyrics::DesktopLyrics(QWidget* parent)
    : QWidget(parent),
      is_locked_(false),
      is_topmost_(true),
      is_lyrics_visible_(true),
      hover_poll_timer_(new QTimer()),
      cursor_enter_time_(0),
      cursor_leave_time_(0) {
  this->setObjectName("desktop_lyrics");

  this->hover_poll_timer_->setInterval(kCursorHoverTimeout);
  this->installEventFilter(this);
  this->initUI();
  this->updateSkin();
  this->initConnections();
  this->handleLrcUpdated();
}

DesktopLyrics::~DesktopLyrics() {
  this->hover_poll_timer_->stop();
  delete this->hover_poll_timer_;
  this->hover_poll_timer_ = nullptr;
  delete unlock_window_;
  unlock_window_ = nullptr;
}

void DesktopLyrics::applyInputTransparent() {
  utils::SetMouseTransparent(this, this->is_locked_);
}

void DesktopLyrics::closeEvent(QCloseEvent* event) {
  event->ignore();
}

bool DesktopLyrics::eventFilter(QObject* obj, QEvent* event) {
  if (event->type() == QEvent::HoverMove) {
    QHoverEvent* hover_event = static_cast<QHoverEvent*>(event);
    const QPoint pos = hover_event->pos();
    if (lrc_compound_frame_->isClickableAt(pos)) {
      utils::SetCursorShape(this, XC_hand1);
    } else if (!isMouseHoverChild()){
      if (!utils::UpdateCursorShape(this, pos.x(), pos.y(),
                                    kDefaultBorderWidth)) {
        utils::SetCursorShape(this, XC_fleur);
      }
    }
  }
  return QWidget::eventFilter(obj, event);
}

void DesktopLyrics::handleCursorEnterEvent() {
  if (is_locked_) {
    // Show locked-window
    setUnlockWindowVisible(true);
  } else {
    // Reset cursor leave time.
    cursor_leave_time_ = 0;
    if (cursor_enter_time_ > 0) {
      return;
    }

    // Check LrcCompoundFrame rect contains current cursor.
    if (lrc_compound_frame_->getContentRegion().contains(QCursor::pos(windowHandle()->screen()))) {
      cursor_enter_time_ = static_cast<int>(time(nullptr));
      // Displays control buttons and disable cursor event pass through.
      QTimer::singleShot(kCursorEnterDelay, this,
                         SLOT(handleCursorEnterEventDelay()));
    }
  }
}

void DesktopLyrics::handleCursorEnterEventDelay() {
  if (cursor_enter_time_ > 0 && ! is_locked_) {
    setControlButtonsVisible(true);
    utils::SetMouseTransparent(this, false);
    // Also reset cursor shape to default.
    utils::ResetCursorShape(this);
  }
}

void DesktopLyrics::handleCursorLeaveEvent() {
  if (is_locked_) {
    setUnlockWindowVisible(false);
  } else {
    // Reset cursor enter time.
    cursor_enter_time_ = 0;
    if (cursor_leave_time_ == 0) {
      cursor_leave_time_ = static_cast<int>(time(nullptr));
      // Hide control buttons after a while and enables cursor event
      // pass through.
      QTimer::singleShot(kCursorLeaveDelay, this,
                         SLOT(handleCursorLeaveEventDelay()));
    }
  }
}

void DesktopLyrics::handleCursorLeaveEventDelay() {
  if (cursor_leave_time_ > 0) {
    setControlButtonsVisible(false);
    utils::SetMouseTransparent(this, true);
  }
}

void DesktopLyrics::handleHoverPollTimerTimeout() {
  const bool hover = this->geometry().contains(QCursor::pos());
  if (hover) {
    this->handleCursorEnterEvent();
  } else {
    this->handleCursorLeaveEvent();
  }
}

void DesktopLyrics::handleLrcFontSizeChanged() {
  this->resize(this->width(),
               lrc_compound_frame_->height() +
               kBackgroundLabelBorderWidth +
               kControlButtonsHeight);
}

void DesktopLyrics::handleLrcUpdated() {
  const bool is_empty = app->lrc_manager->isLrcEmpty();
  if (is_empty) {
    minus_offset_button_->setEnabled(false);
    plus_offset_button_->setEnabled(false);
    minus_offset_button_->updateIconPaths(kButtonsFile, 120, 60,
                                          kButtonsFile, 120, 60);
    plus_offset_button_->updateIconPaths(kButtonsFile, 140, 60,
                                         kButtonsFile, 140, 60);
  } else {
    minus_offset_button_->setEnabled(true);
    plus_offset_button_->setEnabled(true);
    minus_offset_button_->updateIconPaths(kButtonsFile, 120, 0,
                                          kButtonsFile, 120, 20);
    plus_offset_button_->updateIconPaths(kButtonsFile, 140, 0,
                                         kButtonsFile, 140, 20);
  }
}

void DesktopLyrics::handleMediaPlayerStateChanged(MediaPlayer::State state) {
  if (state == MediaPlayer::PlayingState) {
    play_pause_button_->updateIconPaths(kButtonsFile, 180, 0,
                                        kButtonsFile, 180, 20);
    // Media playback is playing if play_pause_button_ is checked.
    play_pause_button_->setChecked(true);
  } else {
    play_pause_button_->updateIconPaths(kButtonsFile, 40, 0,
                                        kButtonsFile, 40, 20);
    play_pause_button_->setChecked(false);
  }
}

void DesktopLyrics::handlePlayPauseButtonClicked() {
  // TODO(xushaohua): check media play has track.
  if (play_pause_button_->isChecked()) {
    emit app->signal_manager->onRequestResume();
  } else {
    emit app->signal_manager->onRequestPause();
  }
}

void DesktopLyrics::hideEvent(QHideEvent* event) {
  this->hover_poll_timer_->stop();
  QWidget::hideEvent(event);
}

void DesktopLyrics::initConnections() {
  connect(app->skin_manager, &SkinManager::onUpdated,
          this, &DesktopLyrics::updateSkin);

  SignalManager* signal_manager = app->signal_manager;

  connect(detail_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onRequestShowDetail);
  connect(prev_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onPlayPrev);
  connect(play_pause_button_, &IconButton::clicked,
          this, &DesktopLyrics::handlePlayPauseButtonClicked);
  connect(next_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onPlayNext);
  connect(settings_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onLrcRequestSettingMenu);
  connect(lock_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onLrcRequestLock);
  connect(unlock_window_, &DesktopLyricsUnlockWindow::unlocked,
          signal_manager, &SignalManager::onLrcRequestUnlock);
  connect(minus_offset_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onLrcRequestBackward);
  connect(plus_offset_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onLrcRequestForward);
  connect(close_button_, &IconButton::clicked,
          signal_manager, &SignalManager::onLrcRequestClose);

  connect(app->media_player, &MediaPlayer::stateChanged,
          this, &DesktopLyrics::handleMediaPlayerStateChanged);

  connect(lrc_compound_frame_, &LrcCompoundFrame::lrcFontSizeChanged,
          this, &DesktopLyrics::handleLrcFontSizeChanged);
  connect(lrc_compound_frame_, &LrcCompoundFrame::onRenderLRCImageResult,
          signal_manager, &SignalManager::onRenderLrcImageResult);
  connect(signal_manager, &SignalManager::onRequestRenderLrcImage,
          lrc_compound_frame_, &LrcCompoundFrame::renderLRCImage);
  connect(signal_manager, &SignalManager::onSetLrcLineMode,
          lrc_compound_frame_, &LrcCompoundFrame::setLrcLineMode);
  connect(signal_manager, &SignalManager::onSetLrcColor,
          lrc_compound_frame_, &LrcCompoundFrame::setLrcColor);
  connect(signal_manager, &SignalManager::onSetLrcFont,
          lrc_compound_frame_, &LrcCompoundFrame::setLrcFont);
  connect(signal_manager, &SignalManager::onSetLrcTextAlignment,
          lrc_compound_frame_, &LrcCompoundFrame::setLrcTextAlignment);

  connect(app->lrc_manager, &LrcManager::onLrcUpdated,
          this, &DesktopLyrics::handleLrcUpdated);

  connect(this->hover_poll_timer_, &QTimer::timeout,
          this, &DesktopLyrics::handleHoverPollTimerTimeout);
}

void DesktopLyrics::initUI() {
  this->background_label_ = new QLabel(this);
  this->background_label_->setObjectName("background-label");
  this->background_label_->setAttribute(Qt::WA_TransparentForMouseEvents, true);

  detail_button_ = new IconButton(kButtonsFile, 0, 0,
                                  kButtonsFile, 0, 20, 20, 20);
  detail_button_->setToolTip(QObject::tr("Show detail page"));
  sep1_ = new QLabel();
  sep1_->setFixedWidth(kSeparatorWidth);
  prev_button_ = new IconButton(kButtonsFile, 20, 0,
                                kButtonsFile, 20, 20, 20, 20);
  prev_button_->setToolTip(QObject::tr("Load previous song"));
  play_pause_button_ = new IconButton(kButtonsFile, 40, 0,
                                      kButtonsFile, 40, 20, 20, 20);
  play_pause_button_->setToolTip(QObject::tr("Play"));
  play_pause_button_->setCheckable(true);
  next_button_ = new IconButton(kButtonsFile, 60, 0,
                                kButtonsFile, 60, 20, 20, 20);
  next_button_->setToolTip(QObject::tr("Load next song"));
  sep2_ = new QLabel();
  sep2_->setFixedWidth(kSeparatorWidth);
  settings_button_ = new IconButton(kButtonsFile, 80, 0,
                                    kButtonsFile, 80, 20, 20, 20);
  settings_button_->setToolTip(QObject::tr("Settings"));
  lock_button_ = new IconButton(kButtonsFile, 100, 0,
                                kButtonsFile, 100, 20, 20, 20);
  lock_button_->setToolTip(QObject::tr("Lock desktop lyric"));
  sep3_ = new QLabel();
  sep3_->setFixedWidth(kSeparatorWidth);
  minus_offset_button_ = new IconButton(kButtonsFile, 120, 0,
                                        kButtonsFile, 120, 20, 20, 20);
  minus_offset_button_->setToolTip(QObject::tr("Backward 0.5s"));
  plus_offset_button_ = new IconButton(kButtonsFile, 140, 0,
                                       kButtonsFile, 140, 20, 20, 20);
  plus_offset_button_->setToolTip(QObject::tr("Forward 0.5s"));
  sep4_ = new QLabel();
  sep4_->setFixedWidth(kSeparatorWidth);
  close_button_ = new IconButton(kButtonsFile, 160, 0,
                                 kButtonsFile, 160, 20, 20, 20);
  close_button_->setToolTip(QObject::tr("Close desktop lyrics"));

  QHBoxLayout* buttons_layout = new QHBoxLayout();
  buttons_layout->setContentsMargins(10, 10, 10, 0);
  buttons_layout->setSpacing(5);
  buttons_layout->addStretch();
  buttons_layout->addWidget(detail_button_);
  buttons_layout->addWidget(sep1_);
  buttons_layout->addWidget(prev_button_);
  buttons_layout->addWidget(play_pause_button_);
  buttons_layout->addWidget(next_button_);
  buttons_layout->addWidget(sep2_);
  buttons_layout->addWidget(settings_button_);
  buttons_layout->addWidget(lock_button_);
  buttons_layout->addWidget(sep3_);
  buttons_layout->addWidget(minus_offset_button_);
  buttons_layout->addWidget(plus_offset_button_);
  buttons_layout->addWidget(sep4_);
  buttons_layout->addWidget(close_button_);
  buttons_layout->addStretch();

  control_buttons_container_ = new QFrame();
  control_buttons_container_->setContentsMargins(0, 0, 0, 0);
  control_buttons_container_->setLayout(buttons_layout);

  QVBoxLayout* wrapper_layout = new QVBoxLayout();
  wrapper_layout->setContentsMargins(0, 0, 0, 0);
  wrapper_layout->setSpacing(0);
  wrapper_layout->addWidget(control_buttons_container_);

  // container_wrapper is used to hold its position when
  // control_buttons_container_ is set to invisible.
  QFrame* control_buttons_container_wrapper = new QFrame();
  control_buttons_container_wrapper->setContentsMargins(0, 0, 0, 0);
  control_buttons_container_wrapper->setFixedHeight(kControlButtonsHeight);
  control_buttons_container_wrapper->setAttribute(Qt::WA_TranslucentBackground);
  control_buttons_container_wrapper->setLayout(wrapper_layout);

  lrc_compound_frame_ = new LrcCompoundFrame();

  QVBoxLayout* main_layout = new QVBoxLayout();

  // Add margins to hold lrc compound frame.
  main_layout->setContentsMargins(kBackgroundLabelBorderWidth,
                                  0,
                                  kBackgroundLabelBorderWidth,
                                  kBackgroundLabelBorderWidth);
  main_layout->setSpacing(0);
  main_layout->addWidget(control_buttons_container_wrapper);
  main_layout->addWidget(lrc_compound_frame_);

  this->setContentsMargins(0, 0, 0, 0);
  this->setLayout(main_layout);

  unlock_window_ = new DesktopLyricsUnlockWindow();

  setWindowFlags(Qt::FramelessWindowHint |
                 Qt::CustomizeWindowHint);
  setAttribute(Qt::WA_TranslucentBackground, true);
  setAttribute(Qt::WA_Hover, true);
}

bool DesktopLyrics::isMouseHoverChild() {
  for (QWidget* widget : findChildren<QWidget*>()) {
    if (qobject_cast<IconButton*>(widget)) {
      const QRect rect(widget->mapToGlobal(QPoint(0, 0)), widget->size());
      if (rect.contains(QCursor::pos())) {
        return true;
      }
    }
  }

  return false;
}

QPoint DesktopLyrics::lastWindowPos() const {
  const QPoint pos = SettingsGetValue(kSettingsGroupName,
                                      kSettingsWindowPositionName).toPoint();
  for (QScreen* screen : app->screens()) {
    if (screen->availableGeometry().contains(pos)) {
      return pos;
    }
  }

  return SettingsGetValue(kSettingsGroupName,
                          kSettingsWindowPositionName).toPoint();
}

bool DesktopLyrics::locked() const {
  return is_locked_;
}

void DesktopLyrics::mousePressEvent(QMouseEvent* event) {
  if (event->button() == Qt::LeftButton) {
    if (lrc_compound_frame_->isClickableAt(event->pos())) {
      emit app->signal_manager->onRequestShowDetail();
    } else {
      const int x = event->x();
      const int y = event->y();
      if (utils::IsCornerEdge(this, x, y, kDefaultBorderWidth)) {
        int spacing = kControlButtonsHeight + kBackgroundLabelBorderWidth;
        this->setMinimumSize(kWindowMinWidth,
                             lrc_compound_frame_->getMinHeightBaseOnFont() +
                             spacing);
        this->setMaximumSize(kWindowMaxWidth,
                             lrc_compound_frame_->getMaxHeightBaseOnFont() +
                             spacing);
        utils::MoveResizeWindow(this, x, y, kDefaultBorderWidth);
      } else {
        // Disable "set maximized window vertically" in kwin
        //this->setFixedSize(this->size());
        utils::MoveWindow(this);
      }
    }
  }
  QWidget::mousePressEvent(event);
}

void DesktopLyrics::mouseReleaseEvent(QMouseEvent* event) {
  // Reset window type to DOCK.
  utils::SetWindowTypeHint(this, utils::X11WindowType::Utility);

  // Reset fixed-size property, so that this window can response to resizeEvent
  // of lrc compound frame.
  //this->setFixedSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX);
  QWidget::mouseReleaseEvent(event);
}

void DesktopLyrics::moveEvent(QMoveEvent* event) {
  SettingsSetValue(kSettingsGroupName,
                   kSettingsWindowPositionName,
                   this->pos());
  this->moveUnlockWindowAlong();
  QWidget::moveEvent(event);
}

void DesktopLyrics::moveUnlockWindowAlong() {
  Q_ASSERT(this->unlock_window_);

  const auto x = this->x() + this->width() / 2;
  const auto y = this->y();
  this->unlock_window_->move(x, y);
}

void DesktopLyrics::resizeEvent(QResizeEvent* event) {
  if (this->background_label_) {
    this->background_label_->resize(this->size());
  }
  QWidget::resizeEvent(event);
  this->applyInputTransparent();
  this->moveUnlockWindowAlong();

  // 多屏每个屏幕缩放比不一样时，窗口从缩放比为1.0的窗口移动到
  // 缩放比为2.0的窗口后，由于窗口实际大小未改变，因为Qt会更新
  // 控件大小为原来的一半，此处需要保证窗口不小于最小大小
  if (event->size().width() < minimumWidth()) {
      resize(minimumSize());
  }
}

void DesktopLyrics::setTopMost(bool topmost) {
  this->is_topmost_ = topmost;
  utils::SetStayOnTop(this, this->is_topmost_);
}

void DesktopLyrics::setControlButtonsVisible(bool visible) {
  background_label_->setVisible(visible);
  control_buttons_container_->setVisible(visible);
}

void DesktopLyrics::setLocked(bool locked) {
  is_locked_ = locked;

  setControlButtonsVisible(!is_locked_);
  setUnlockWindowVisible(is_locked_);

  this->applyInputTransparent();
  utils::SetStayOnTop(this, this->is_topmost_);
}

void DesktopLyrics::setUnlockWindowVisible(bool visible) {
  if (visible && is_lyrics_visible_) {
    this->moveUnlockWindowAlong();
    unlock_window_->show();
    unlock_window_->activateWindow();
    unlock_window_->raise();
  } else {
    unlock_window_->hide();
  }
}

void DesktopLyrics::show() {
  // Restore window position.
  this->move(lastWindowPos());

  QWidget::show();
}

void DesktopLyrics::showEvent(QShowEvent* event) {
  this->hover_poll_timer_->start();

  // Set window type to DOCK.
  utils::SetWindowTypeHint(this, utils::X11WindowType::Utility);
  utils::SkipTaskbarPager(this);
  this->setTopMost(this->is_topmost_);

  // 保证窗口最小尺寸
  int spacing = kControlButtonsHeight + kBackgroundLabelBorderWidth;
  this->setMinimumSize(kWindowMinWidth,
                       lrc_compound_frame_->getMinHeightBaseOnFont() +
                       spacing);

  QWidget::showEvent(event);
}

void DesktopLyrics::setShowWindow(bool show) {
  is_lyrics_visible_ = show;
  if (show) {
    setUnlockWindowVisible(is_locked_);
    utils::SetStayOnTop(this, this->is_topmost_);
    this->show();
  } else {
    this->hide();
  }
}

void DesktopLyrics::updateSkin() {
  const QString sep_filepath(
      app->skin_manager->absoluteSkinPath(kSeparatorFile));
  sep1_->setPixmap(QPixmap(sep_filepath));
  sep2_->setPixmap(QPixmap(sep_filepath));
  sep3_->setPixmap(QPixmap(sep_filepath));
  sep4_->setPixmap(QPixmap(sep_filepath));

  setStyleSheet(app->skin_manager->formatStyleFile(kStyleSheetFile));
}

void DesktopLyrics::updateTooltips(const QString& prev_tooltip,
                                   const QString& play_pause_tooltip,
                                   const QString& next_tooltip) {
  prev_button_->setToolTip(prev_tooltip);
  play_pause_button_->setToolTip(play_pause_tooltip);
  next_button_->setToolTip(next_tooltip);
}

void DesktopLyrics::wheelEvent(QWheelEvent* event) {
  const QPoint point(event->angleDelta());
  int button = 0;
  switch (event->buttons()) {
  case Qt::LeftButton:
    button = 1;
    break;

  case Qt::MiddleButton:
    button = 2;
    break;

  case Qt::RightButton:
    button = 3;
    break;

  default:
    // Ignores other types of buttons
    break;
  }

  const int offset = point.y();
  emit app->signal_manager->onLrcRequestMouseScroll(button, offset,
                                                    point.x(), point.y());
  QWidget::wheelEvent(event);
}

}  // namespace netease
