// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "lrc_compound_frame.h"

#include <QDebug>
#include <QFileInfo>
#include <QLinearGradient>
#include <QPainter>
#include <QPainterPath>
#include <QImageWriter>

#include "application.h"
#include "service/media_player.h"
#include "service/signal_manager.h"
#include "service/skin_manager.h"

namespace netease {

namespace {

const char kAlignLeft[] = "left";
const char kAlignCenter[] = "center";
const char kAlignRight[] = "right";

const int kDefaultFontSize = 32;
const int kMinimumFontSize = 20;
const int kMaximumFontSize = 96;
const char kDefaultFontFamily[] = "Sans Serif";

const char kPrimaryCacheKey[] = "primary_cache_pixmap";
const char kSecondaryCacheKey[] = "secondary_cache_pixmap";

const double kColorPosition1 = 0.2;
const double kColorPosition2 = 0.7;

const QSize kLRCImageSize = QSize(555, 105);

}  // namespace

LrcCompoundFrame::LrcCompoundFrame(QWidget* parent)
  : QWidget(parent),
    primary_element_(),
    secondary_element_(),
    clickable_rect_(),
    lrc_font_(kDefaultFontFamily, kDefaultFontSize),
    base_color1_(Qt::black),
    base_color2_(Qt::black),
    played_color1_(Qt::black),
    played_color2_(Qt::black),
    primary_alignment_(Qt::AlignLeft),
    secondary_alignment_(Qt::AlignLeft),
    primary_element_up_(true),
    element_changed_(true),
    lrc_font_size_changed_(false),
    is_single_line_(true),
    scroll_lrc_(false),
    should_show_slogan_(true),
    offset_x_(0) {

  connect(app->media_player, &MediaPlayer::positionChanged,
          this, &LrcCompoundFrame::handlePositionChanged);
  connect(app->media_player, &MediaPlayer::stateChanged,
          this, &LrcCompoundFrame::handleMediaPlayerStateChanged);
}

QRect LrcCompoundFrame::getContentRegion() {
  QPoint top_left;
  QPoint bottom_right;
  if (app->lrc_manager->isNoTimeLrc()) {
    const QString first_tip =
        QObject::tr("This Lyrics does not support scrolling , ");
    const QString second_tip = QObject::tr("View all Lyrics");
    top_left = QPoint(0, 0);
    bottom_right = QPoint(QFontMetrics(lrc_font_).width(first_tip + second_tip),
                          getElementHeight());
  } else if (should_show_slogan_) {
    const QString slogan = app->lrc_manager->slogan;
    int h_offset = getHorizontalOffset(true, slogan);
    top_left = QPoint(h_offset, 0);
    bottom_right = QPoint(QFontMetrics(lrc_font_).width(slogan) + h_offset,
                          getElementHeight());

  } else {
    // primary element
    int p_h_offset = getHorizontalOffset(true, primary_element_.content);
    int p_width = scroll_lrc_ ? width() :
                  QFontMetrics(lrc_font_).width(primary_element_.content);

    // Secondly element
    int s_h_offset = getHorizontalOffset(false, secondary_element_.content);
    int s_width = QFontMetrics(lrc_font_).width(secondary_element_.content);

    top_left = QPoint(qMin(p_h_offset, s_h_offset), 0);
    bottom_right = QPoint(qMax((p_h_offset + p_width), (s_h_offset + s_width)),
                          height());
  }

  // Combine
  QRect rect(this->mapToGlobal(top_left), this->mapToGlobal(bottom_right));
  return rect;
}

int LrcCompoundFrame::getMaxHeightBaseOnFont() {
  return getHeightBaseOnFont(kMaximumFontSize);
}

int LrcCompoundFrame::getMinHeightBaseOnFont() {
  return getHeightBaseOnFont(kMinimumFontSize);
}

bool LrcCompoundFrame::isClickableAt(const QPoint& pos) {
  // Parent is DesktopLrics and this function should only use in DesktopLrics
  const QPoint pp = this->mapToParent(QPoint(0, 0));
  if (pp.isNull() || ! app->lrc_manager->isNoTimeLrc()) {
    return false;
  }
  QRect r;
  r.setX(pp.x() + clickable_rect_.x());
  r.setY(pp.y() + clickable_rect_.y());
  r.setWidth(clickable_rect_.width());
  r.setHeight(clickable_rect_.height());
  return r.contains(pos);
}

bool LrcCompoundFrame::isSingleLine() const {
  return is_single_line_;
}

void LrcCompoundFrame::renderLRCImage(const QString& text,
                                      const QString& filepath) {
  QPixmap pixmap(this->size());
  pixmap.fill(Qt::transparent);
  QPainter painter(&pixmap);
  painter.setRenderHint(QPainter::Antialiasing);
  int pixmap_height = getElementHeight();
  // Draw primary line
  int primary_y = (is_single_line_ || primary_element_up_) ? 0 : pixmap_height;
  painter.drawPixmap(0, primary_y, width(), pixmap_height,
                     getElementPixmap(true, text, false, false));

  // Draw secondary line
  int secondary_y = primary_element_up_ ? pixmap_height : 0;
  if (! is_single_line_) {
    painter.drawPixmap(0, secondary_y, width(), pixmap_height,
                       getElementPixmap(false, text, false, false));
  }

  // Save generated image file.
  const bool state = pixmap.copy(0, 0, kLRCImageSize.width(),
                                 kLRCImageSize.height()).save(filepath);

  emit this->onRenderLRCImageResult(state, kLRCImageSize.width(),
                                    kLRCImageSize.height());
}

void LrcCompoundFrame::setLrcColor(const QString& base_color1,
                                   const QString& base_color2,
                                   const QString& played_color1,
                                   const QString& played_color2) {
  const QString prefix("#");
  base_color1_ = QColor(prefix + base_color1);
  base_color2_ = QColor(prefix + base_color2);
  played_color1_ = QColor(prefix + played_color1);
  played_color2_ = QColor(prefix + played_color2);

  element_changed_ = true;
  this->update();
}

void LrcCompoundFrame::setLrcFont(const QString& font_size,
                                  const QString& unused,
                                  const QString& font_family) {
  Q_UNUSED(unused);

  const int font_size_int = font_size.toInt();

  lrc_font_.setPointSize(font_size_int);
  lrc_font_.setFamily(font_family);
  lrc_font_size_changed_ = true;

  // Update window size when lrc font changed.
  this->resizeWindowBasedOnFont();
  element_changed_ = true;
  this->update();
}

void LrcCompoundFrame::setLrcLineMode(bool is_single_line) {
  is_single_line_ = is_single_line;

  // |primary_element_| is always rendered at top in single line mode.
  if (is_single_line_) {
    primary_element_up_ = true;
  }

  // Update window size when lrc line mode changed.
  this->resizeWindowBasedOnFont();
  element_changed_ = true;
  this->update();
}

void LrcCompoundFrame::setLrcTextAlignment(const QString& primary_alignment,
                                           const QString& secondary_alignment) {
  if (primary_alignment == kAlignLeft) {
    primary_alignment_ = Qt::AlignLeft;
  } else if (primary_alignment == kAlignCenter) {
    primary_alignment_ = Qt::AlignCenter;
  } else if (primary_alignment == kAlignRight) {
    primary_alignment_ = Qt::AlignRight;
  }

  if (secondary_alignment == kAlignLeft) {
    secondary_alignment_ = Qt::AlignLeft;
  } else if (secondary_alignment == kAlignCenter) {
    secondary_alignment_ = Qt::AlignCenter;
  } else if (secondary_alignment == kAlignRight) {
    secondary_alignment_ = Qt::AlignRight;
  }

  element_changed_ = true;
  this->update();
}

void LrcCompoundFrame::paintEvent(QPaintEvent* event) {
  Q_UNUSED(event)
  QPainter painter(this);

  if (should_show_slogan_) {
    drawSlogan();
    return;
  } else if (app->lrc_manager->isNoTimeLrc()) {
    drawLrcNoTimeTip();
    return;
  }

  int pixmap_height = getElementHeight();
  // Draw primary line
  int primary_y = (is_single_line_ || primary_element_up_) ? 0 : pixmap_height;
  painter.drawPixmap(0,
                     primary_y,
                     width(),
                     pixmap_height,
                     getElementPixmap(true,
                                      primary_element_.content,
                                      app->lrc_manager->isKrc()));

  // Draw secondary line
  int secondary_y = primary_element_up_ ? pixmap_height : 0;
  if (! is_single_line_) {
    painter.drawPixmap(0,
                       secondary_y,
                       width(),
                       pixmap_height,
                       getElementPixmap(false, secondary_element_.content));
  }

  // Reset element changed state.
  element_changed_ = false;
}

// Change font size when window is resized.
void LrcCompoundFrame::resizeEvent(QResizeEvent* event) {
  element_changed_ = true;

  // Ignore this signal, if font size of lrc is changed by calling
  // setLrcFont().
  if (lrc_font_size_changed_) {
    lrc_font_size_changed_ = false;
    return;
  }

  const int window_height = event->size().height();
  const int last_window_height = event->oldSize().height();
  int font_size = 0;

  if (window_height == last_window_height) {
    return;
  } else if (window_height > last_window_height) {
    // Expand lrc label. Consider single line and dual lines mode.
    while (true) {
      font_size = lrc_font_.pointSize();
      ++font_size;
      if (font_size > kMaximumFontSize) {
        break;
      }
      lrc_font_.setPointSize(font_size);
      if (getHeightBaseOnFont(font_size) >= window_height) {
        break;
      }
    }
  } else {
    // Shrink lrc label
    while (true) {
      font_size = lrc_font_.pointSize();
      --font_size;
      if (font_size < kMinimumFontSize) {
        break;
      }
      lrc_font_.setPointSize(font_size);
      if (getHeightBaseOnFont(font_size) <= window_height) {
        break;
      }
    }
  }

  // Notify webview to update lrc font size.
  emit app->signal_manager->onLrcRequestFontSize(font_size);
  QWidget::resizeEvent(event);
}

void LrcCompoundFrame::handlePositionChanged(qint64 position) {
  handleElementUpdate(position);
  // // Draw Title and Artist once if lrc is empty
  // if ((!element_changed_) && (!lrc_font_size_changed_)
  //         && (! app->lrc_manager->isKrc())) {
  //   return;
  // }
  updateElementWidth(position);
  updateOffsetX();

  this->update();
}

void LrcCompoundFrame::handleMediaPlayerStateChanged(MediaPlayer::State state) {
  element_changed_ = true;
  switch (state) {
    case MediaPlayer::PlayingState: {
      should_show_slogan_ = false;
      update();
      break;
    }

    case MediaPlayer::StoppedState: {
      // If track_info is empty, no media resource is played currently.
      if (app->media_player->track_info()->play_id.isEmpty()) {
        should_show_slogan_ = true;
        update();
      }
      break;
    }

    default:
      break;
  }
}

void LrcCompoundFrame::handleElementUpdate(qint64 position) {
  LrcManager::LrcElement pe = app->lrc_manager->isLrcEmpty() ?
                              app->lrc_manager->titleElement() :
                              app->lrc_manager->searchElement(position);
  LrcManager::LrcElement se =
    app->lrc_manager->nextElement();

  const qint64 last_position = primary_element_.start;
  if (last_position != pe.start) {
    primary_element_up_ = (! primary_element_up_);
  }

  if (primary_element_.content != pe.content ||
      secondary_element_.content != se.content) {
    element_changed_ = true;
  }

  primary_element_ = pe;
  secondary_element_ = se;
}

void LrcCompoundFrame::updateElementWidth(qint64 position) {
  // Calculate played width of current element.
  QFontMetrics fm(lrc_font_);
  element_played_width_ =
      fm.width(primary_element_.content) * elementPlayedPercent(position);
}

void LrcCompoundFrame::updateOffsetX() {
  if (app->lrc_manager->isKrc() || app->lrc_manager->isLrc()) {
    const int window_width = width();
    const int half_window_width = window_width * 0.5;
    const int element_width =
        QFontMetrics(lrc_font_).width(primary_element_.content);

    if (element_width <= window_width || window_width < 0) {
      scroll_lrc_ = false;
      offset_x_ = 0;
    } else {
      scroll_lrc_ = true;
      if (element_width - element_played_width_ < half_window_width) {
        // Stop scroll, right alignment.
        offset_x_ = window_width - element_width;
      } else if (element_played_width_ > half_window_width) {
        // Start scroll.
        offset_x_ = half_window_width - element_played_width_;
      } else {
        offset_x_ = 0;
      }
    }
  } else {
    // No scroll.
    scroll_lrc_ = false;
    offset_x_ = 0;
  }
}

double LrcCompoundFrame::elementPlayedPercent(qint64 position) {
  qint64 sustain_time = 0;
  if (app->lrc_manager->isKrc() || app->lrc_manager->isLrc()) {
    if (primary_element_.sustain_time > 0) {
      sustain_time = primary_element_.sustain_time;
    } else if (secondary_element_.start > 0) {
      sustain_time = secondary_element_.start - primary_element_.start;
    } else {
      // Last lrc element.
      sustain_time = app->media_player->duration() - primary_element_.start;
    }
  }

  const qint64 time_delta = position - primary_element_.start;
  if (sustain_time == 0) {
    return 0;
  } else {
    return time_delta * 1.0 / sustain_time;
  }
}

int LrcCompoundFrame::getElementHeight() {
  if (is_single_line_) {
    return height();
  } else {
    return height() / 2;
  }
}

int LrcCompoundFrame::getHeightBaseOnFont(int point_size) {
  const QFontMetrics fm(QFont(lrc_font_.family(), point_size));
  int height = 0;
  if (is_single_line_) {
    height = fm.height();
  } else {
    height = fm.height() * 2;
  }

  return height;
}

int LrcCompoundFrame::getHorizontalOffset(bool is_primary_element,
                                          const QString& content) {
  const QFontMetrics fm(lrc_font_);
  const int window_width = this->width();
  int x = 0;

  if (is_primary_element) {
    const int words_width = fm.width(content);

    if (scroll_lrc_) {
      x = offset_x_;
    } else {
      Qt::Alignment alignment;
      if (primary_element_up_ || is_single_line_) {
        alignment = primary_alignment_;
      } else {
        alignment = secondary_alignment_;
      }
      // TODO(xushaohua): Scroll primary_element_.
      switch (alignment) {
      case Qt::AlignCenter:
        x = (window_width - words_width) / 2;
        break;

      case Qt::AlignRight:
        x = window_width - words_width;
        break;

      default:
        // x = 0;
        break;
      }
    }
  } else {
    const int words_width = fm.width(content);
    Qt::Alignment alignment;
    if (!primary_element_up_ || is_single_line_) {
      alignment = primary_alignment_;
    } else {
      alignment = secondary_alignment_;
    }
    switch (alignment) {
    case Qt::AlignCenter:
      x = (window_width - words_width) / 2;
      break;

    case Qt::AlignRight:
      x = window_width - words_width;
      break;

    default:
      // x = 0;
      break;
    }
  }

  return x;
}

int LrcCompoundFrame::getVerticalOffset(bool is_primary_element) {
  const QFontMetrics fm(lrc_font_);
  const int base_line = fm.ascent();

  if (is_single_line_) {
    return base_line;
  }
  if (is_primary_element == primary_element_up_) {
    return base_line;
  } else {
    // Add vertical padding.
    return base_line * 2.2;
  }
}

void LrcCompoundFrame::resizeWindowBasedOnFont() {
  this->resize(this->width(), getHeightBaseOnFont(lrc_font_.pointSize()));

  // Notify parent window to update its own size.
  emit this->lrcFontSizeChanged();
}

void LrcCompoundFrame::drawLrcNoTimeTip() {
  const QString first_tip =
      QObject::tr("This Lyrics does not support scrolling , ");
  const QString second_tip = QObject::tr("View all Lyrics");
  const int line_top_spacing = 10;
  // Always draw the tip at the top of the frame
  const int y = getElementHeight() - line_top_spacing;

  QPainter painter(this);
  painter.setRenderHint(QPainter::Antialiasing);
  painter.setFont(lrc_font_);

  // Draw first tip
  QPainterPath first_path;
  first_path.addText(0, y, lrc_font_, first_tip);
  painter.fillPath(first_path, getElementBrush(true, true));

  // Draw second tip
  QFont f = lrc_font_;
  f.setUnderline(true);
  painter.setFont(f);
  QFontMetrics fm(f);
  QPainterPath second_path;
  second_path.addText(fm.width(first_tip), y, f, second_tip);
  painter.fillPath(second_path, getElementBrush(true, true));
  painter.setFont(lrc_font_);

  clickable_rect_ = QRect(fm.width(first_tip),
                          0,
                          fm.width(second_tip),
                          fm.height());
}

void LrcCompoundFrame::drawSlogan() {
  const int line_top_spacing = 5;
  int h_offset = getHorizontalOffset(true, app->lrc_manager->slogan);
  // Always draw the slogan at the top of the frame
  const int y = getElementHeight() - line_top_spacing;

  QPainter painter(this);
  painter.setRenderHint(QPainter::Antialiasing);
  painter.setFont(lrc_font_);

  QPainterPath path;
  path.addText(h_offset, y, lrc_font_, app->lrc_manager->slogan);
  painter.fillPath(path, getElementBrush(true, true));
}

QPixmap LrcCompoundFrame::getElementPixmap(bool is_primary,
                                           const QString& content,
                                           bool is_krc,
                                           bool use_cache) {
  int h_offset = getHorizontalOffset(is_primary, content);
  int pixmap_width = 0;
  if (scroll_lrc_) {
    int fw = QFontMetrics(lrc_font_).width(content);
    pixmap_width = qMax(fw, width());
  }
  else {
      pixmap_width = width();
  }
  int pixmap_height = getElementHeight();
  const QString cache_key = is_primary ? kPrimaryCacheKey : kSecondaryCacheKey;

  qreal device_pixel_ratio = devicePixelRatioF();

  QPixmap pixmap(pixmap_width * device_pixel_ratio, pixmap_height * device_pixel_ratio);
  pixmap.fill(Qt::transparent);
  pixmap.setDevicePixelRatio(device_pixel_ratio);

  // Cache after a period of time will be cleared
  if (!use_cache || element_changed_ || !QPixmapCache::find(cache_key, &pixmap)) {
    QPainter painter(&pixmap);
    painter.setRenderHint(QPainter::Antialiasing);

    QPainterPath pp;
    int path_y = QFontMetrics(lrc_font_).ascent();
    pp.addText(h_offset, path_y, lrc_font_, content);

    // Fill background, the color is same between primary and secondary
    // Auto fill played brush if it is not Krc
    const bool is_played = is_krc ? false : true;
    painter.fillPath(pp, getElementBrush(is_primary, is_played));
    // Fill shadow
    painter.fillPath(getTextShadowPath(pp),
                     QBrush(app->skin_manager->text_colors.lrc_primary_shadow));
    // Cache path
    if (is_primary && use_cache) {
      primary_cache_path_ = pp;
    }
    painter.end();

    // Update cache pixmap
    if (use_cache) {
      QPixmapCache::insert(cache_key, pixmap);
    }
  }

  if (is_primary) {
    // Draw krc played gradient
    if (is_krc) {
      QPainter painter(&pixmap);
      painter.setRenderHint(QPainter::Antialiasing);
      int clip_width = element_played_width_;
      clip_width += h_offset > 0 ? h_offset : 0;
      painter.setClipRect(0, 0, clip_width, pixmap_height);
      painter.fillPath(primary_cache_path_, getElementBrush(true, true));
    }

    pixmap = pixmap.copy(- offset_x_ * device_pixel_ratio, 0, width() * device_pixel_ratio, pixmap_height * device_pixel_ratio);
  } else {
    pixmap = pixmap.copy(0, 0, width() * device_pixel_ratio, pixmap_height * device_pixel_ratio);
  }

  pixmap.setDevicePixelRatio(device_pixel_ratio);

  return pixmap;
}

QPainterPath LrcCompoundFrame::getTextShadowPath(const QPainterPath& path) {
  int shadow_width = 1;
  if (lrc_font_.pointSize() > (kMinimumFontSize + kMaximumFontSize) / 2) {
    shadow_width = 2;
  }
  QPainterPathStroker stroker;
  stroker.setWidth(shadow_width);
  stroker.setJoinStyle(Qt::RoundJoin);
  return stroker.createStroke(path);
}

QBrush LrcCompoundFrame::getElementBrush(bool is_primary, bool is_played) {
  QLinearGradient gradient(0, 0, 0, QFontMetrics(lrc_font_).ascent());
  if (is_primary && is_played) {
    gradient.setColorAt(kColorPosition1, played_color1_);
    gradient.setColorAt(kColorPosition2, played_color2_);
  } else {
    gradient.setColorAt(kColorPosition1, base_color1_);
    gradient.setColorAt(kColorPosition2, base_color2_);
  }

  return QBrush(gradient);
}

}  // namespace netease
