// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef UI_VIEWS_LRC_COMPOUND_FRAME_H_
#define UI_VIEWS_LRC_COMPOUND_FRAME_H_

#include <QWidget>
#include <QResizeEvent>
#include <QPixmapCache>

#include "service/lrc_manager.h"
#include "service/media_player.h"

namespace netease {

class LrcCompoundFrame : public QWidget {
  Q_OBJECT
 public:
  explicit LrcCompoundFrame(QWidget* parent = 0);
  // Returns content region relative to parent window.
  QRect getContentRegion();

  int getMaxHeightBaseOnFont();
  int getMinHeightBaseOnFont();

  // Returns true if current lrc has no time tag.
  bool isClickableAt(const QPoint& pos);

  bool isSingleLine() const;

 signals:
  // This signal is emitted when font size changed.
  // Parent window shall handle this signal and resize itself.
  void lrcFontSizeChanged();

  // Emitted after calling renderLRCImage().
  void onRenderLRCImageResult(bool ok, int width, int height);

 public slots:
  // Called in Player::renderLRCImage().
  void renderLRCImage(const QString& text, const QString& filepath);

  void setLrcColor(const QString& base_color1, const QString& base_color2,
                   const QString& played_color1, const QString& played_color2);

  void setLrcFont(const QString& font_size, const QString& unused,
                  const QString& font_family);

  // Displays single line or dual lines.
  void setLrcLineMode(bool is_single_line);

  void setLrcTextAlignment(const QString& primary_alignment,
                           const QString& secondary_alignment);

 protected:
  void paintEvent(QPaintEvent* event);
  void resizeEvent(QResizeEvent* event);

 private:
  void updateElementWidth(qint64 position);
  void updateOffsetX();

  double elementPlayedPercent(qint64 position);

  int getElementHeight();

  int getHeightBaseOnFont(int point_size);

  // Calculate vertical or horizontal element offset.
  // Consider line mode, time offset(horizontal scroll).
  int getHorizontalOffset(bool is_primary_element, const QString& content);

  // Consider line mode only.
  int getVerticalOffset(bool is_primary_element);

  // Resize frame height based on lrc font size and line mode.
  void resizeWindowBasedOnFont();

  // Draw tip if lrc has no time tag.
  void drawLrcNoTimeTip();
  // Draw slogan if nothing is playing
  void drawSlogan();

  QPixmap getElementPixmap(bool is_primary, const QString& content,
                           bool is_krc = false, bool use_cache = true);
  QPainterPath getTextShadowPath(const QPainterPath& path);
  QBrush getElementBrush(bool is_primary, bool is_played);

  LrcManager::LrcElement primary_element_;
  LrcManager::LrcElement secondary_element_;

  // For no time lrc tip click
  QRect clickable_rect_;

  QPainterPath primary_cache_path_;

  QFont lrc_font_;
  QColor base_color1_;
  QColor base_color2_;
  QColor played_color1_;
  QColor played_color2_;
  Qt::Alignment primary_alignment_;
  Qt::Alignment secondary_alignment_;

  // |primary_element_| is at top or bottom of container.
  bool primary_element_up_;
  // Indicate that the painter should redraw whold background area
  bool element_changed_;
  bool lrc_font_size_changed_;
  bool is_single_line_;
  bool scroll_lrc_;
  bool should_show_slogan_;

  // The length of the current element has been played, for krc
  int element_played_width_;
  int offset_x_;

 private slots:
  void handlePositionChanged(qint64 position);
  void handleMediaPlayerStateChanged(MediaPlayer::State state);
  void handleElementUpdate(qint64 position);
};

}  // namespace netease

#endif // UI_VIEWS_LRC_COMPOUND_FRAME_H_
