// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/views/lrc_plain_frame.h"

#include <QDebug>
#include <QFontMetrics>
#include <QPainter>
#include <QPaintEvent>
#include <QResizeEvent>

#include "application.h"
#include "base/file_util.h"
#include "service/media_player.h"
#include "service/skin_manager.h"

namespace netease {

namespace {

const char kDefaultFontFamily[] = "Sans";
const int kDefaultFontSize = 10;

}  // namespace

LrcPlainFrame::LrcPlainFrame(QWidget* parent)
    : QLabel(parent),
      lrc_font_(kDefaultFontFamily, kDefaultFontSize),
      offset_x_(0),
      scroll_lrc_(false),
      window_size_() {
  initConnections();
}

void LrcPlainFrame::resizeEvent(QResizeEvent *event) {
  this->window_size_ = event->size();
  QLabel::resizeEvent(event);
}

void LrcPlainFrame::handlePositionChanged(qint64 position) {
  // * check overlay
  // * repaint if necessary.

  LrcManager* lrc_manager = app->lrc_manager;

  if (lrc_manager->isKrc() || lrc_manager->isLrc()) {
    lrc_element_ = lrc_manager->currentElement();
    LrcManager::LrcElement next_element = lrc_manager->nextElement();
    qint64 sustain_time = 0;
    if (lrc_element_.sustain_time > 0) {
      sustain_time = lrc_element_.sustain_time;
    } else if (next_element.start > 0) {
      sustain_time = next_element.start - lrc_element_.start;
    } else {
      // Last lrc element.
      sustain_time = app->media_player->duration() - lrc_element_.start;
    }

    const qint64 time_delta = position - lrc_element_.start;
    const double percent = time_delta * 1.0 / sustain_time;
    QFontMetrics font_metrics(lrc_font_);
    const int words_width = font_metrics.width(lrc_element_.content);
    const int word_offset = words_width * percent;
    const int window_width = window_size_.width();
    const int half_window_width = window_width * 0.5;

    if (words_width <= window_width || window_width < 0) {
      scroll_lrc_ = false;
      offset_x_ = 0;
    } else {
      scroll_lrc_ = true;
      if (words_width - word_offset < half_window_width) {
        // stop scroll, right alignment.
        offset_x_ = window_width - words_width;
      } else if (word_offset > half_window_width) {
        // Start scroll.
        offset_x_ = half_window_width - word_offset;
      } else {
        offset_x_ = 0;
      }
    }
  } else {
    // No scroll.
    lrc_element_ = lrc_manager->titleElement();
    scroll_lrc_ = false;
    offset_x_ = 0;
  }

  this->update();
}

void LrcPlainFrame::initConnections() {
  connect(app->media_player, &MediaPlayer::positionChanged,
          this, &LrcPlainFrame::handlePositionChanged);
}

void LrcPlainFrame::paintEvent(QPaintEvent* event) {
  Q_UNUSED(event);

  // * get current lrc element.
  // * scroll if necessary.
  QPainter painter(this);
  painter.setFont(lrc_font_);
  painter.setPen(app->skin_manager->text_colors.lrc_text);

  QFontMetrics font_metrics(painter.fontMetrics());

  int words_width = font_metrics.width(lrc_element_.content);
  int words_height = font_metrics.ascent();

  // Align center both vertically and horizontally.
  int x = 0;
  if (scroll_lrc_) {
    // Scroll lrc if necessary.
    x = offset_x_;
  } else {
    x = (this->width() - words_width) / 2;
  }

  int y = (this->height() + words_height) / 2;
  painter.drawText(x, y, lrc_element_.content);
}

}  // namespace netease
