// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/views/mini_player.h"

#include <QDebug>
#include <QGraphicsDropShadowEffect>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QScreen>
#include <QStackedLayout>
#include <QStackedWidget>

#include "application.h"
#include "service/lrc_manager.h"
#include "service/settings_manager.h"
#include "service/signal_manager.h"
#include "service/skin_manager.h"
#include "service/storage_manager.h"
#include "ui/models/song_info.h"
#include "ui/views/mini_player_consts.h"
#include "ui/views/mini_player_info_panel.h"
#include "ui/views/mini_player_lrc.h"
#include "ui/views/mini_player_playlist.h"
#include "ui/views/mini_player_title.h"
#include "ui/views/widgets/colored_icon_button.h"
#include "ui/views/widgets/draggable_area.h"
#include "ui/views/widgets/volume_window.h"
#include "ui/utils/xutil.h"

namespace netease {

namespace {

const int kDefaultBorderWidth = 5;

// SKin file of mini-player.
const char kStyleSheetFile[] = ":/styles/mini-player.css";

// Relative filepath to default album cover.
const char kDefaultAlbumCover[] = "SKIN_ROOT/mini/album/default.png";
const char kAlbumCoverHighlight[] = "SKIN_ROOT/mini/album/highlight.png";
// Width and height of album cover image.
const int kAlbumSize = 42;

const int kControlPanelButtonIndex = 0;
const int kControlPanelTitleIndex = 1;
const int kControlPanelLrcIndex = 2;

// Note that it is typo in filename with "seperator".
//const int kSeparatorWidth = 2;

const char kLikeFile[] = "SKIN_ROOT/btn/like.png";
const char kPraiseFile[] = "SKIN_ROOT/btn/praise.png";
const char kPraiseHoverFile[] = "SKIN_ROOT/btn/praise_hvr.png";
const char kPraisedFile[] = "SKIN_ROOT/btn/praised.png";
const char kVolumeFile[] = "SKIN_ROOT/btn/volume.png";
const char kTrashFile[] = "SKIN_ROOT/btn/delete.png";
const char kTrashHoverFile[] = "SKIN_ROOT/btn/delete_hvr.png";

const int kPlaylistAnimationDuration = 200;
const int kInfoAnimationDuration = 50;

const int kOnlineUnlike = 0;
const int kOnlineLike = 1;
const int kDjRadio = 2;  // Show thumb-up icon if it is DJ Radio.
const int kOfflineUnlike = 4;
const int kOfflineLike = 3;

const char kSettingsGroupName[] = "UIMiniPlayer";
const char kSettingsWindowPositionName[] = "window-position";

}  // namespace

MiniPlayer::MiniPlayer(QWidget* parent)
    : QFrame(parent),
      album_cover_path_(),
      album_cover_url_(),
      is_topmost_(false),
      has_track_(false),
      like_mark_state_(kOnlineUnlike) {
  this->setObjectName("mini_player");

  initUI();
  installEventFilter(this);
  initConnections();
  updateSkin();
}

MiniPlayer::~MiniPlayer() {

}

void MiniPlayer::addSongs(const QList<SongInfo>& song_list) {
  playlist_panel_->addSongs(song_list);
}

void MiniPlayer::clearPlaylist() {
  playlist_panel_->clearPlaylist();
}

void MiniPlayer::closeMainWindow() {
  this->hide();
  emit app->signal_manager->onRequestMainWindowClose();
}

void MiniPlayer::deleteSongs(const QStringList& play_id_list) {
  playlist_panel_->deleteSongs(play_id_list);
}

void MiniPlayer::handleAlbumCoverDownloaded(const QString& url,
                                            const QString& filepath) {
  qDebug() << "[MiniPlayer]::handleAlbumCoverDownloaded()"
           << ", url: " << url
           << ", filepath: " << filepath
           << ", album_cover_url_: " << album_cover_url_;
  if (url == album_cover_url_) {
    album_cover_path_ = filepath;
    this->updateAlbumCover();
  }
}

void MiniPlayer::handleMediaPlayerStateChanged(MediaPlayer::State state) {
  if (state == MediaPlayer::PlayingState) {
    play_pause_button_->updateIconPaths(64, 0, 64, 40);

    play_pause_button_->setToolTip(QObject::tr("Pause"));
    // Media playback is playing if play_pause_butto`n_ is checked.
    play_pause_button_->setChecked(true);

    if ( ! app->lrc_manager->isLrcEmpty() &&
         middle_control_stack_->currentIndex() == kControlPanelTitleIndex) {
      middle_control_stack_->setCurrentIndex(kControlPanelLrcIndex);
    }

    playlist_panel_->setPlayState(true);
  } else {
    play_pause_button_->setToolTip(QObject::tr("Play"));
    play_pause_button_->updateIconPaths(24, 0, 24, 40);
    play_pause_button_->setChecked(false);
    if (has_track_) {
      middle_control_stack_->setCurrentIndex(kControlPanelTitleIndex);
    } else {
      middle_control_stack_->setCurrentIndex(kControlPanelButtonIndex);
    }
    playlist_panel_->setPlayState(false);
  }

}

void MiniPlayer::handlePlaylistButtonChecked(bool toggle) {
  if (toggle) {
    showListPanel();
  } else {
    hideListPanel();
  }
}

void MiniPlayer::handlePlayPauseButtonClicked() {
  if (has_track_) {
    if (play_pause_button_->isChecked()) {
      emit app->signal_manager->onRequestResume();
    } else {
      emit app->signal_manager->onRequestPause();
    }
  }
}

void MiniPlayer::handlePraiseButtonClicked() {
  const int like = praise_button_->isChecked() ? 1 : 0;
  emit app->signal_manager->onRequestFavorSong(like);
}

void MiniPlayer::handleVolumeButtonClicked() {
  if (volume_window_->isVisible()) {
    volume_window_->hide();
  } else {
    QPoint pos(volume_button_->mapToGlobal(QPoint(0, 0)));
    // 12 is half width of volume_button_.
    pos.setX(pos.x() + 12);
    volume_window_->show(pos);
  }
}

void MiniPlayer::hideControlPanel() {
  if (has_track_) {
    if ( ! app->lrc_manager->isLrcEmpty() &&
         app->media_player->state() == MediaPlayer::PlayingState) {
      middle_control_stack_->setCurrentIndex(kControlPanelLrcIndex);
    } else {
      middle_control_stack_->setCurrentIndex(kControlPanelTitleIndex);
    }
    // Only hide info panel if playlist panel is hidden.
    if (! playlist_button_->isChecked()) {
      hideInfoPanel();
    }
  }
}

QPoint MiniPlayer::lastWindowPos() const {
  const QPoint pos = SettingsGetValue(kSettingsGroupName,
                                      kSettingsWindowPositionName).toPoint();
  for (QScreen* screen : app->screens()) {
    if (screen->availableGeometry().contains(pos)) {
      return pos;
    }
  }

  return SettingsGetDefaultValue(kSettingsGroupName,
                                 kSettingsWindowPositionName).toPoint();
}

void MiniPlayer::initConnections() {
  connect(app->skin_manager, &SkinManager::onUpdated,
          this, &MiniPlayer::updateSkin);

  SignalManager* signal_manager = app->signal_manager;

  connect(album_button_, &QPushButton::clicked,
          signal_manager, &SignalManager::onRequestShowDetail);

  connect(prev_button_, &QPushButton::clicked,
          signal_manager, &SignalManager::onPlayPrev);
  connect(next_button_, &QPushButton::clicked,
          signal_manager, &SignalManager::onPlayNext);
  connect(play_pause_button_, &QPushButton::clicked,
          this, &MiniPlayer::handlePlayPauseButtonClicked);

  connect(app->media_player, &MediaPlayer::stateChanged,
          this, &MiniPlayer::handleMediaPlayerStateChanged);

  connect(song_control_buttons_container_, &DraggableArea::mouseLeaved,
          this, &MiniPlayer::hideControlPanel);
  connect(mini_player_lrc_, &MiniPlayerLrc::mouseEntered,
          this, &MiniPlayer::showControlPanel);
  connect(mini_player_title_, &MiniPlayerTitle::mouseEntered,
          this, &MiniPlayer::showControlPanel);

  connect(volume_button_, &QPushButton::clicked,
          this, &MiniPlayer::handleVolumeButtonClicked);

  connect(playlist_button_, &QPushButton::toggled,
          this, &MiniPlayer::handlePlaylistButtonChecked);

  connect(show_main_window_button_, &QPushButton::clicked,
          this, &MiniPlayer::showMainWindow);

  connect(close_button_, &QPushButton::clicked,
          this, &MiniPlayer::closeMainWindow);

  connect(volume_window_, &VolumeWindow::volumeChanged,
          signal_manager, &SignalManager::onRequestChangeVolume);
  connect(app->media_player, &MediaPlayer::volumeChangedWithAnimation,
          volume_window_, &VolumeWindow::setVolume);

  connect(playlist_panel_, &MiniPlayerPlaylist::changeCurrentPlay,
          signal_manager, &SignalManager::onRequestPlay);

  connect(signal_manager, &SignalManager::onSetFavorMark,
          this, &MiniPlayer::setFavour);
  connect(signal_manager, &SignalManager::onSetLikeMark,
          this, &MiniPlayer::setLikeMark);

  connect(praise_button_, &QPushButton::clicked,
          this, &MiniPlayer::handlePraiseButtonClicked);
  connect(like_button_, &QPushButton::clicked,
          signal_manager, &SignalManager::onRequestLikeSong);
  connect(delete_button_, &QPushButton::clicked,
          signal_manager, &SignalManager::onRequestDeleteFm);

  connect(top_panel_, &QFrame::customContextMenuRequested,
          signal_manager, &SignalManager::onMiniPlayerContextMenu);

  connect(app->storage_manager,
          &StorageManager::onAlbumCoverDownloaded,
          this, &MiniPlayer::handleAlbumCoverDownloaded);
}

void MiniPlayer::initUI() {
  initUIAlbum();
  initUISongControl();
  initUISettingControl();
  initUIWindowControl();
  initUITopPanel();
  initUIBottomStack();

  volume_window_ = new VolumeWindow();

  QVBoxLayout* main_layout = new QVBoxLayout(this);
  main_layout->setContentsMargins(0, 0, 0, 0);
  main_layout->setSpacing(0);
  main_layout->addWidget(top_panel_);
//  main_layout->addWidget(bottom_stacked_widget_);
  main_layout->addStretch();

  // The spacing of top_panel_'s bottom shadow
  const int shadow_spacing = 4;
  bottom_stacked_widget_->move(0, kTopPanelHeight - shadow_spacing);

  this->setMinimumWidth(kWindowMinimumWidth);
  this->resize(kWindowMinimumWidth, kTopPanelHeight);
  this->setWindowFlags(Qt::FramelessWindowHint | Qt::CustomizeWindowHint);
  this->setAttribute(Qt::WA_TranslucentBackground, true);
  this->setAttribute(Qt::WA_Hover, true);
}

void MiniPlayer::initUIAlbum() {
  album_button_ = new FlatButton(this);
  album_button_->setFixedSize(kAlbumSize, kAlbumSize);

  QHBoxLayout* layout = new QHBoxLayout(album_button_);
  album_highlight_ = new QLabel(this);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->addWidget(album_highlight_);
}

void MiniPlayer::initUISongControl() {
  prev_button_ = new ColoredIconButton(0, 8, 0, 48, 20, 20);
  prev_button_->setToolTip(QObject::tr("Load previous song"));
  play_pause_button_ = new ColoredIconButton(24, 0, 24, 40, 36, 36);
  play_pause_button_->setToolTip(QObject::tr("Play"));
  play_pause_button_->setCheckable(true);
  next_button_ = new ColoredIconButton(104, 8, 104, 48, 20, 20);
  next_button_->setToolTip(QObject::tr("Load next song"));

  song_control_buttons_container_ = new DraggableArea(this);
  QHBoxLayout* layout = new QHBoxLayout(song_control_buttons_container_);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->addStretch();
  layout->addWidget(prev_button_);
  layout->addWidget(play_pause_button_);
  layout->addWidget(next_button_);
  layout->addStretch();

  mini_player_title_ = new MiniPlayerTitle();
  mini_player_lrc_ = new MiniPlayerLrc();

  middle_control_stack_ = new QStackedLayout();
  middle_control_stack_->addWidget(song_control_buttons_container_);
  middle_control_stack_->addWidget(mini_player_title_);
  middle_control_stack_->addWidget(mini_player_lrc_);
}

void MiniPlayer::initUISettingControl() {
  like_button_ = new IconButton(kLikeFile, 0, 0, kLikeFile, 0, 20, 22, 20);
  like_button_->setToolTip(QObject::tr("Bookmark"));
  like_button_->hide();
  praise_button_ = new IconButton(kPraiseFile, 0, 0,
                                  kPraiseHoverFile, 0, 0, 24, 24);
  praise_button_->setToolTip(QObject::tr("Praise"));
  praise_button_->hide();
  delete_button_ = new IconButton(kTrashFile, 0, 0,
                                  kTrashHoverFile, 0, 0, 21, 22);
  delete_button_->setToolTip(QObject::tr("Delete"));
  delete_button_->hide();
  volume_button_ = new IconButton(kVolumeFile, 0, 0,
                                  kVolumeFile, 0, 24, 22, 24);
  volume_button_->setToolTip(QObject::tr("Volume"));
  playlist_button_ = new IconButton(kCollectionButtons, 10, 95,
                                    kCollectionButtons, 40, 95, 22, 21);
  playlist_button_->setCheckable(true);
  playlist_button_->setToolTip(QObject::tr("Show playlist"));

  setting_buttons_layout_ = new QHBoxLayout();
  setting_buttons_layout_->setContentsMargins(0, 0, 0, 0);
  // Add 5px between buttons.
  setting_buttons_layout_->setSpacing(5);
  setting_buttons_layout_->addWidget(like_button_);
  setting_buttons_layout_->addWidget(praise_button_);
  setting_buttons_layout_->addWidget(delete_button_);
  setting_buttons_layout_->addWidget(volume_button_);
  setting_buttons_layout_->addWidget(playlist_button_);
}

void MiniPlayer::initUIWindowControl() {
  close_button_ = new IconButton(kCollectionButtons, 0, 80,
                                 kCollectionButtons, 20, 80, 10, 10);
  close_button_->setToolTip(QObject::tr("Close"));
  show_main_window_button_ = new IconButton(kCollectionButtons, 40, 80,
                                            kCollectionButtons, 60, 80, 10, 10);
  show_main_window_button_->setToolTip(QObject::tr("Show main window"));

  window_control_layout_ = new QVBoxLayout();
  window_control_layout_->setContentsMargins(0, 5, 5, 0);
  window_control_layout_->setSpacing(2);
  window_control_layout_->addWidget(close_button_);
  window_control_layout_->addWidget(show_main_window_button_);
  window_control_layout_->addStretch(1);
}

void MiniPlayer::initUITopPanel() {
  top_panel_ = new QFrame();
  top_panel_->setObjectName("top-panel-background");
  top_panel_->setFixedHeight(kTopPanelHeight);
  // Enable right click signal.
  top_panel_->setContextMenuPolicy(Qt::CustomContextMenu);
  QFrame* top_panel_gradient = new QFrame();
  top_panel_gradient->setObjectName("top-panel-gradient");
  QHBoxLayout* gradient_layout = new QHBoxLayout(top_panel_);
  gradient_layout->setContentsMargins(0, 0, 0, 0);
  gradient_layout->addWidget(top_panel_gradient);

  QHBoxLayout* layout = new QHBoxLayout(top_panel_gradient);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->setSpacing(0);
  layout->addWidget(album_button_);
  layout->addLayout(middle_control_stack_);
  layout->addLayout(setting_buttons_layout_);
  layout->addLayout(window_control_layout_);
}

void MiniPlayer::initUIBottomStack() {
  playlist_panel_ = new MiniPlayerPlaylist();
  info_panel_ = new MiniPlayerInfoPanel();

  bottom_stacked_widget_ = new QStackedWidget(this);
  bottom_stacked_widget_->setContentsMargins(0, 0, 0, 0);
  bottom_stacked_widget_->addWidget(playlist_panel_);
  bottom_stacked_widget_->addWidget(info_panel_);
  bottom_stacked_widget_->resize(kWindowMinimumWidth, 0);
}

void MiniPlayer::moveEvent(QMoveEvent* event) {
  // TODO(LiuLang): Add timer
  SettingsSetValue(kSettingsGroupName,
                   kSettingsWindowPositionName,
                   this->pos());
  QFrame::moveEvent(event);
}

void MiniPlayer::setAlbumCover(const QString& url) {
  album_cover_url_ = url;
  album_cover_path_ = app->storage_manager->getAlbumCover(url);

  this->updateAlbumCover();
}

void MiniPlayer::setAlwaysOnTop(bool topmost) {
  is_topmost_ = topmost;
  utils::SetStayOnTop(this, this->is_topmost_);
}

void MiniPlayer::setCurrentPlay(const QString& play_id) {
  // Reset album cover.
  album_cover_url_ = "";
  album_cover_path_ = "";

  playlist_panel_->setCurrentPlay(play_id);

  SongInfo song_info(playlist_panel_->currentSongInfo());
  auto track_info = app->media_player->track_info();
  mini_player_title_->updateContent(track_info->song_name,
                                    track_info->artist_name);
  info_panel_->updateContent(track_info->song_name, track_info->artist_name);

  // Reset like button state:
//  like_button_->updateIconPaths(kLikeFile, 0, 0, kLikeFile, 0, 20);
  this->updateLikeButton();

  // Reset volume.
  volume_window_->setVolume(app->media_player->volume());

  // Show lrc panel when that song is loaded.
  if (! song_control_buttons_container_->rect().contains(QCursor::pos())) {
    if (app->lrc_manager->isLrcEmpty()) {
      middle_control_stack_->setCurrentIndex(kControlPanelTitleIndex);
    } else {
      middle_control_stack_->setCurrentIndex(kControlPanelLrcIndex);
    }
  }

  has_track_ = true;

  if (song_info.program > 0) {
    like_button_->hide();
    praise_button_->show();
  } else {
    praise_button_->hide();
    like_button_->show();
  }

  if (track_info->song_type == "normal") {
    delete_button_->hide();
    prev_button_->show();
    playlist_button_->show();
  } else if (track_info->song_type == "fm") {
    prev_button_->hide();
    playlist_button_->hide();
    delete_button_->show();
  } else if (track_info->song_type == "djradio") {
    delete_button_->hide();
    prev_button_->show();
    playlist_button_->show();
  } else {
    qWarning() << "[MiniPlayer]::setCurrentPlay() unknown song type:"
               << track_info->song_type;
  }
}

void MiniPlayer::setFavour(bool favour) {
  praise_button_->setChecked(favour);
  if (favour) {
    praise_button_->updateIconPaths(kPraisedFile, 0, 0, kPraisedFile, 0, 0);
  } else {
    praise_button_->updateIconPaths(kPraiseFile, 0, 0, kPraiseHoverFile, 0, 0);
  }
}

void MiniPlayer::setLikeMark(int state) {
  like_mark_state_ = state;
  this->updateLikeButton();
}

void MiniPlayer::show() {
  this->move(lastWindowPos());

  // Make sure bottom_stacked_widget_ will not show with parent
  if (bottom_stacked_widget_->isHidden()) {
    bottom_stacked_widget_->hide();
  }
  QWidget::show();
}

void MiniPlayer::updateLikeButton() {
  if (like_mark_state_ == kOnlineUnlike || like_mark_state_ == kOfflineUnlike) {
    // Unlike.
    like_button_->updateIconPaths(kLikeFile, 0, 0, kLikeFile, 0, 20);
  } else if (like_mark_state_ == kOnlineLike ||
             like_mark_state_ == kOfflineLike) {
    // Like.
    like_button_->updateIconPaths(kLikeFile, 0, 40, kLikeFile, 0, 60);
  }
}

bool MiniPlayer::eventFilter(QObject* obj, QEvent* event) {
  if (event->type() == QEvent::HoverMove) {
    QHoverEvent* hover_event = static_cast<QHoverEvent*>(event);
    const QPoint pos = hover_event->pos();
    const utils::CornerEdge edge =
        utils::GetCornerEdge(this, pos.x(), pos.y(), kDefaultBorderWidth);
    if (edge == utils::CornerEdge::kRight && pos.y() < kTopPanelHeight) {
      utils::UpdateCursorShape(this, pos.x(), pos.y(), kDefaultBorderWidth);
    } else if (! isMouseHoverChild()) {
      utils::ResetCursorShape(this);
    }
  } else if (event->type() == QEvent::HoverEnter) {
    if (volume_window_->isVisible()) {
      volume_window_->hide();
    }
  }

  return QFrame::eventFilter(obj, event);
}

void MiniPlayer::mousePressEvent(QMouseEvent *event) {
  if (volume_window_->isVisible()) {
    volume_window_->hide();
  }
  const utils::CornerEdge edge =
      utils::GetCornerEdge(this, event->x(), event->y(), kDefaultBorderWidth);
  // Only resize in right side
  if (event->button() == Qt::LeftButton && edge == utils::CornerEdge::kRight) {
    // For window resizable
    utils::MoveResizeWindow(this, event->x(), event->y(), kDefaultBorderWidth);
  }
}

void MiniPlayer::setShowMainWindowButtonToolTip(const QString& tooltip) {
  show_main_window_button_->setToolTip(tooltip);
}

void MiniPlayer::setUseDefaultCover() {
  album_cover_url_ = "";
  album_cover_path_ = "";
  this->updateAlbumCover();
}

void MiniPlayer::showControlPanel() {
  middle_control_stack_->setCurrentIndex(kControlPanelButtonIndex);
  // Only displays info panel if playlist_panel is hidden.
  if (! playlist_button_->isChecked()) {
    showInfoPanel();
  }
}

void MiniPlayer::showEvent(QShowEvent* event) {
  // Show controller buttons.
  this->handleMediaPlayerStateChanged(app->media_player->state());

  // Set window type to Utility.
  utils::SetWindowTypeHint(this, utils::X11WindowType::Utility);
  utils::SkipTaskbarPager(this);
  utils::SetStayOnTop(this, true);

  QFrame::showEvent(event);
}

void MiniPlayer::resizeEvent(QResizeEvent *e) {
  QFrame::resizeEvent(e);
  bottom_stacked_widget_->resize(e->size().width(),
                                 bottom_stacked_widget_->height());
}

void MiniPlayer::showMainWindow() {
  emit app->signal_manager->onRequestChangeToMain();
}

void MiniPlayer::updateAlbumCover() {
  if (album_cover_url_.isEmpty() || album_cover_path_.isEmpty()) {
    album_cover_path_ = app->skin_manager->absoluteSkinPath(kDefaultAlbumCover);
  }

  emit app->signal_manager->onAlbumCoverUpdated(album_cover_path_);

  // Scale album cover image to 42x42.
  QPixmap pixmap(album_cover_path_);
  if (!pixmap.isNull()) {
    qreal device_pixel_ratio = devicePixelRatioF();
    pixmap = pixmap.scaled(kAlbumSize * device_pixel_ratio, kAlbumSize * device_pixel_ratio);
    pixmap.setDevicePixelRatio(device_pixel_ratio);

    album_highlight_->setPixmap(pixmap);
  }
}

void MiniPlayer::showInfoPanel() {
  const QSize end_size(this->width(), kTopPanelHeight + kInfoPanelHeight);
  if (this->size() == end_size) {
    return;
  }
  this->resize(end_size);
  bottom_stacked_widget_->setCurrentIndex(1);
  QPropertyAnimation* animation = new QPropertyAnimation(bottom_stacked_widget_,
                                                         "size");
  animation->setDuration(kInfoAnimationDuration);
  animation->setStartValue(QSize(end_size.width(), 0));
  animation->setEndValue(QSize(end_size.width(), kInfoPanelHeight));
  animation->start();
  connect(animation, &QPropertyAnimation::finished,
          animation, &QPropertyAnimation::deleteLater);
}

void MiniPlayer::hideInfoPanel(){
  const QSize end_size(this->width(), kTopPanelHeight);
  if (this->size() == end_size) {
    return;
  }
  bottom_stacked_widget_->setCurrentIndex(1);
  QPropertyAnimation* animation = new QPropertyAnimation(bottom_stacked_widget_,
                                                         "size");
  animation->setDuration(kInfoAnimationDuration);
  animation->setStartValue(QSize(end_size.width(), kInfoPanelHeight));
  animation->setEndValue(QSize(end_size.width(), 0));
  animation->start();
  connect(animation, &QPropertyAnimation::finished, this, [=] {
    animation->deleteLater();
    this->resize(end_size);
  });
}

void MiniPlayer::showListPanel() {
  const QSize end_size(this->width(), kTopPanelHeight + kPlaylistHeight);
  if (this->size() == end_size) {
      return;
  }
  this->resize(end_size);
  bottom_stacked_widget_->setCurrentIndex(0);
  QPropertyAnimation* animation = new QPropertyAnimation(bottom_stacked_widget_,
                                                         "size");
  animation->setDuration(kPlaylistAnimationDuration);
  animation->setStartValue(QSize(end_size.width(), 0));
  animation->setEndValue(QSize(end_size.width(), kPlaylistHeight));
  animation->start();
  connect(animation, &QPropertyAnimation::finished,
          animation, &QPropertyAnimation::deleteLater);
}

void MiniPlayer::hideListPanel() {
  const QSize end_size(this->width(), kTopPanelHeight);
  if (this->size() == end_size) {
      return;
  }
  bottom_stacked_widget_->setCurrentIndex(0);
  QPropertyAnimation* animation = new QPropertyAnimation(bottom_stacked_widget_,
                                                         "size");
  animation->setDuration(kPlaylistAnimationDuration);
  animation->setStartValue(QSize(end_size.width(), kPlaylistHeight));
  animation->setEndValue(QSize(end_size.width(), 0));
  animation->start();
  connect(animation, &QPropertyAnimation::finished, this, [=] {
    animation->deleteLater();
    this->resize(end_size);
  });
}

bool MiniPlayer::isMouseHoverChild() {
  for (QWidget* widget : findChildren<QWidget*>()) {
    if (qobject_cast<FlatButton*>(widget)) {
      const QRect rect(widget->mapToGlobal(QPoint(0, 0)), widget->size());
      if (rect.contains(QCursor::pos())) {
        return true;
      }
    }
  }

  return false;
}

QString MiniPlayer::absoluteSkinPath(const QString &path) const {
  return app->skin_manager->absoluteSkinPath(path);
}

void MiniPlayer::updateSkin() {
  updateAlbumCover();
  const QString highlight_filepath(
      app->skin_manager->absoluteSkinPath(kAlbumCoverHighlight));
  album_highlight_->setPixmap(QPixmap(highlight_filepath));

  this->setStyleSheet(app->skin_manager->formatStyleFile(kStyleSheetFile));
}

}  // namespace netease
