// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef UI_MINI_PLAYER_H_
#define UI_MINI_PLAYER_H_

#include <QFrame>
#include <QList>
#include <QMouseEvent>

#include "service/media_player.h"

class QEvent;
class QHBoxLayout;
class QLabel;
class QStackedLayout;
class QStackedWidget;
class QVBoxLayout;

namespace netease {

class ColoredIconButton;
class DraggableArea;
class FlatButton;
class IconButton;
class MiniPlayerInfoPanel;
class MiniPlayerLrc;
class MiniPlayerPlaylist;
class MiniPlayerTitle;
class VolumeWindow;
struct SongInfo;

class MiniPlayer : public QFrame {
  Q_OBJECT
 public:
  explicit MiniPlayer(QWidget* parent = nullptr);
  ~MiniPlayer() override;

  // Add songs to playlist.
  void addSongs(const QList<SongInfo>& song_list);

  // Remove all songs in playlist.
  void clearPlaylist();

  // Remove songs from playlist.
  void deleteSongs(const QStringList& play_id_list);

  // Set current album cover to image to which |url| refers.
  void setAlbumCover(const QString& url);

  // Set mini player window always on top.
  void setAlwaysOnTop(bool topmost);

  // Update current played song.
  void setCurrentPlay(const QString& play_id);

  // Update tooltip of show_main_window_button_.
  void setShowMainWindowButtonToolTip(const QString& tooltip);

  // Use default album cover.
  void setUseDefaultCover();

 public slots:
  // Set favorite mark (+1).
  void setFavour(bool favour);

  // Set like mark (red-heart).
  void setLikeMark(int state);

  void show();

 protected:
  void mousePressEvent(QMouseEvent* event) override;
  bool eventFilter(QObject* obj, QEvent* event) override;

  // Override this event to record new position of window.
  void moveEvent(QMoveEvent* event) override;

  // Override this event to show control-panel-buttons when window is visible.
  void showEvent(QShowEvent* event) override;

  void resizeEvent(QResizeEvent* e) override;

 private:
  void initConnections();
  void initUI();
  void initUIAlbum();
  void initUISongControl();
  void initUISettingControl();
  void initUIWindowControl();
  void initUITopPanel();
  void initUIBottomStack();

  // Update album cover based on current album_cover_path_;
  void updateAlbumCover();

  void showInfoPanel();
  void hideInfoPanel();
  void showListPanel();
  void hideListPanel();

  bool isMouseHoverChild();
  QString absoluteSkinPath(const QString &path) const;

  // Get last available position
  QPoint lastWindowPos() const;

 // .-----------.
 // | top panel |
 // .-----------.
 // |   info    |
 // |-----------|
 // |           |
 // | playlist  |
 // |           |
 // .-----------.

  QFrame* top_panel_;

  FlatButton* album_button_;
  QLabel* album_highlight_;
  // Absolute path to album cover image file.
  QString album_cover_path_;
  QString album_cover_url_;

  // TODO(xushaohua): remove DraggableArea. Call utils::xutil instead.
  DraggableArea* song_control_buttons_container_;
  QStackedLayout* middle_control_stack_;
  ColoredIconButton* prev_button_;
  ColoredIconButton* play_pause_button_;
  ColoredIconButton* next_button_;
  MiniPlayerTitle* mini_player_title_;
  MiniPlayerLrc* mini_player_lrc_;

  QHBoxLayout* setting_buttons_layout_;
  // Red-heart button.
  IconButton* like_button_;
  // +1 button.
  IconButton* praise_button_;
  IconButton* delete_button_;
  IconButton* volume_button_;
  IconButton* playlist_button_;

  QVBoxLayout* window_control_layout_;
  IconButton* close_button_;
  IconButton* show_main_window_button_;

  QStackedWidget* bottom_stacked_widget_;
  MiniPlayerPlaylist* playlist_panel_;
  MiniPlayerInfoPanel* info_panel_;

  VolumeWindow* volume_window_;

  bool is_topmost_;

  // MediaPlayer has loaded track info.
  bool has_track_;

  // Update icon in like_button based on media state.
  void updateLikeButton();
  int like_mark_state_;

 private slots:
  // Hide mini player window and request to close main webview window.
  void closeMainWindow();

  void handleAlbumCoverDownloaded(const QString& url, const QString& filepath);

  // Update play_pause_button_ icons based on status of media player.
  void handleMediaPlayerStateChanged(MediaPlayer::State state);

  // Show/hide playlist panel.
  void handlePlaylistButtonChecked(bool toggle);

  // Pause/resume media playback when play_pause_button_ is clicked.
  void handlePlayPauseButtonClicked();

  void handlePraiseButtonClicked();

  // Show/hide volume window.
  void handleVolumeButtonClicked();

  // Hide control panel buttons.
  void hideControlPanel();

  // Show control panel buttons.
  void showControlPanel();

  // Hide mini player window and show main webview window.
  void showMainWindow();

  // Update widget style based on current skin.
  void updateSkin();
};

}  // namespace netease

#endif  // UI_MINI_PLAYER_H_
