// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/views/mini_player_playlist.h"

#include <QDebug>
#include <QHeaderView>
#include <QHBoxLayout>
#include <QLabel>
#include <QListView>
#include <QPalette>
#include <QScrollArea>

#include "application.h"
#include "service/signal_manager.h"
#include "service/skin_manager.h"
#include "ui/delegates/playlist_delegate.h"
#include "ui/models/playlist_model.h"
#include "ui/models/song_info.h"
#include "ui/views/mini_player_consts.h"

namespace netease {

namespace {

const int kScrollAreaLeftMargin = 5;
const int kScrollAreaWidth = kWindowMinimumWidth - 2 * kScrollAreaLeftMargin;
const int kScrollAreaHeight = kPlaylistHeight - 10;

const char kStyleSheetFile[] = ":/styles/mini-player-playlist.css";

}  // namespace

MiniPlayerPlaylist::MiniPlayerPlaylist(QWidget* parent) : QFrame(parent) {

  initUI();
  initConnection();
}

void MiniPlayerPlaylist::addSongs(const QList<SongInfo>& song_list) {
  playlist_model_->addSongs(song_list);
}

void MiniPlayerPlaylist::clearPlaylist() {
  playlist_model_->clearPlaylist();
}

SongInfo MiniPlayerPlaylist::currentSongInfo() const {
  //    return playlist_model_->currentSongInfo();
  SongInfo info;
  QJsonObject obj = playlist_model_->data(playlist_view_->currentIndex(),
                                          Qt::DisplayRole).toJsonObject();
  obj >> info;
  return info;
}

void MiniPlayerPlaylist::deleteSongs(const QStringList& play_id_list) {
  playlist_model_->deleteSongs(play_id_list);
}

void MiniPlayerPlaylist::handleContextMenu(QPoint pos) {
  const QModelIndex index(playlist_view_->indexAt(pos));
  const SongInfo song_info(playlist_model_->getSongInfo(index.row()));
  const QString play_id(song_info.play_id);
  if (play_id.length() > 0) {
    emit app->signal_manager->onMiniPlayerMenu(play_id);
  }
}

void MiniPlayerPlaylist::handleListViewDoubleClicked(const QModelIndex& index) {
  // Ignores current played row.
  if (index.row() == playlist_model_->currentPlayRow()) {
    return;
  }

  const SongInfo song_info(playlist_model_->getSongInfo(index.row()));
  qDebug() << "[MiniPlayerPlaylist]::handleListViewDoubleClicked()"
           << ", song_info.play_id: " << song_info.play_id;

  if (song_info.play_id.length() > 0) {
    emit changeCurrentPlay(song_info.play_id);
  }
}

void MiniPlayerPlaylist::initConnection() {
  connect(app->skin_manager, &SkinManager::onUpdated,
          this, &MiniPlayerPlaylist::updateSkin);

  connect(playlist_view_, &QListView::doubleClicked,
          this, &MiniPlayerPlaylist::handleListViewDoubleClicked);
  connect(playlist_view_, &QListView::customContextMenuRequested,
          this, &MiniPlayerPlaylist::handleContextMenu);
}

void MiniPlayerPlaylist::initUI() {
  playlist_delegate_ = new PlaylistDelegate(this);
  playlist_model_ = new PlaylistModel(this);

  playlist_view_ = new QListView(this);
  playlist_view_->setModel(playlist_model_);
  playlist_view_->setEditTriggers(QAbstractItemView::NoEditTriggers);
  playlist_view_->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
  playlist_view_->setItemDelegate(playlist_delegate_);
  playlist_view_->setAlternatingRowColors(true);
  QPalette palette(playlist_view_->palette());
  palette.setBrush(QPalette::Base, Qt::transparent);
  playlist_view_->setPalette(palette);
  playlist_view_->setSelectionMode(QAbstractItemView::SingleSelection);
  playlist_view_->setLayoutMode(QListView::SinglePass);
  playlist_view_->setContextMenuPolicy(Qt::CustomContextMenu);

  QScrollArea* playlist_scroll_area = new QScrollArea(this);
  playlist_scroll_area->setObjectName("playlist-scroll-area");
  playlist_scroll_area->setWidget(playlist_view_);
  playlist_scroll_area->setWidgetResizable(true);
  playlist_scroll_area->setAttribute(Qt::WA_TranslucentBackground);
  playlist_scroll_area->resize(kScrollAreaWidth, kScrollAreaHeight);
  QHBoxLayout* layout = new QHBoxLayout(this);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->addWidget(playlist_scroll_area);


  this->setObjectName("playlist-background");
  this->resize(kWindowMinimumWidth, kPlaylistHeight);
  this->setContentsMargins(0, 0, 0, 0);

  this->updateSkin();
}

void MiniPlayerPlaylist::setCurrentPlay(const QString& play_id) {
  QModelIndex i = playlist_model_->index(playlist_model_->rowOf(play_id));
  playlist_model_->setCurrentRow(i.row());
  playlist_view_->setCurrentIndex(i);
}

void MiniPlayerPlaylist::setPlayState(bool is_playing) {
  playlist_model_->setPlayState(is_playing);
}

void MiniPlayerPlaylist::updateSkin() {
  this->setStyleSheet(app->skin_manager->formatStyleFile(kStyleSheetFile));
}

}  // namespace netease
