// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/views/web_view.h"

#include <QCloseEvent>
#include <QDebug>
#include <QStackedLayout>
#include <QWebChannel>
#include <QTimer>
#include <QWindow>

#include <qcef_web_page.h>
#include <qcef_web_settings.h>
#include <qcef_web_view.h>

#include "application.h"
#include "base/file_util.h"
#include "channel/channel.h"
#include "ui/utils/xutil.h"
#include "ui/web_view_event_delegate.h"
#include "ui/window_manager.h"

namespace netease {

namespace {

//const char kWebFrameRateFile[] = ":/misc/display-webview-frame-rate.js";
//const char kMvControllerFile[] = ":/misc/fullscreen-video-controller.js";
const char kChannelJs[] = ":/channel/channel.js";
const char kChannelJsQrc[] = "qrc://channel/channel.js";

const int kBorderWidth = 10;

bool IsXfceDesktopEnv() {
  const QByteArray kDeskEnv = qgetenv("XDG_CURRENT_DESKTOP");
  return kDeskEnv.toLower().startsWith("xfce");
}

}  // namespace

WebView::WebView()
    : QFrame(),
      is_main_window(false),
      left_pressed_(false),
      resizable_(false),
      resizable_disabled_(false),
      last_window_state_(Qt::WindowNoState),
      event_delegate_(new WebViewEventDelegate()) {
  // Remove title bar and border on non-xfce desktop environment.
  if (!IsXfceDesktopEnv()) {
    this->setWindowFlags(Qt::FramelessWindowHint | Qt::Window);
  }

  QVBoxLayout* layout = new QVBoxLayout(this);
  layout->setContentsMargins(0, 0, 0, 0);
  layout->setSpacing(0);
  qcef_webview_ = new QCefWebView(this);
  qcef_webview_->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

  layout->addWidget(qcef_webview_);
  this->setLayout(layout);
  this->setContentsMargins(0, 0, 0, 0);

  qcef_webview_->page()->setEventDelegate(event_delegate_);

  auto settings = qcef_webview_->page()->settings();
  settings->setWebSecurity(QCefWebSettings::StateDisabled);

  this->initConnections();
  Channel* channel = new Channel(this);
  qcef_webview_->page()->webChannel()->registerObject("channel", channel);

  this->installEventFilter(this);

  utils::AddCefWindowFlag(this);
}

WebView::~WebView() {
  if (event_delegate_ != nullptr) {
    delete event_delegate_;
    event_delegate_ = nullptr;
  }
}

void WebView::closeEvent(QCloseEvent* event) {
  Q_UNUSED(event);
  // Do not destroy this object, only emit onClose() signal to web side instead.
  emit onClose();
}

void WebView::disableResizable() {
  resizable_disabled_ = true;
}

void WebView::enterFullscreen() {
  if (!this->isVisible()) {
    return;
  }
  last_window_state_ = this->windowState();
  this->showFullScreen(true);
}

bool WebView::eventFilter(QObject* obj, QEvent* event) {
  switch (event->type()) {
    case QEvent::WindowActivate: {
      emit this->onFocusIn();
      break;
    }
    case QEvent::WindowDeactivate: {
      emit this->onFocusOut();
      break;
    }
    default: {
      break;
    }
  }
  return QFrame::eventFilter(obj, event);
}

void WebView::exitFullscreen() {
  if (!this->isVisible()) {
    return;
  }

  utils::ShowFullscreenWindow(this, false);
  switch (last_window_state_) {
    case Qt::WindowMaximized: {
      this->showMaximized();
      break;
    }

    case Qt::WindowNoState: {
      this->showNormal();
      break;
    }

    default: {
      break;
    }
  }
}

void WebView::handleCefFullscreenChanged(bool fullscreen) {
  if (fullscreen) {
    enterFullscreen();
  } else {
    exitFullscreen();
  }
}

void WebView::initConnections() {
  connect(qcef_webview_->page(), &QCefWebPage::fullscreenRequested,
          this, &WebView::handleCefFullscreenChanged);
  connect(qcef_webview_->page(), &QCefWebPage::renderContextCreated, [&]() {
    // Register channel object.
    const QString channelJS = ReadTextFileContent(kChannelJs);
    qcef_webview_->page()->runJavaScript(channelJS, kChannelJsQrc);
  });

  connect(this, &WebView::moveWindow, this, &WebView::handleMoveWindow);
  connect(this, &WebView::resizeWindow, this, &WebView::handleResizeWindow);
}

void WebView::updateWebViewZoomLevel() {
  // 废弃
}

void WebView::load(const QUrl& url) {
  qcef_webview_->load(url);

  if (url.path().endsWith("/login.html")) {
      // 防止登录界面显示空白
      qcef_webview_->page()->runJavaScript("console.log('')");
  }
}

void WebView::raiseWindow() {
  this->show();
  this->activateWindow();
  this->raise();
}

void WebView::resizeEvent(QResizeEvent* event) {
  QFrame::resizeEvent(event);
  // Notify qcef webview to resize browser window.
  emit this->windowStateSizeChanged();
}

void WebView::setResizable(bool resizable) {
  if (resizable_disabled_) {
    qWarning() << "[WebView]::setResizable() current window has fixed size!";
    return;
  }

  resizable_ = resizable;
  if (! resizable) {
    emit onSizeWindowDone();
  }
}

void WebView::setTopMost(bool topmost) {
  Qt::WindowFlags flags = this->windowFlags();
  if (topmost) {
    flags = flags | Qt::WindowStaysOnTopHint;
  } else if (flags & Qt::WindowStaysOnTopHint){
    flags = flags ^ Qt::WindowStaysOnTopHint;
  }
  const bool visible = this->isVisible();
  this->setWindowFlags(flags);
  if (visible) {
    this->show();
  }
}

void WebView::setZoomLevel(double level) {
  qcef_webview_->page()->setZoomFactor(level);
}

void WebView::showFullScreen(bool fullscreen) {
  utils::ShowFullscreenWindow(this, fullscreen);
  this->raiseWindow();
}

void WebView::showMaximized() {
  utils::ShowMaximizedWindow(this);
  this->raiseWindow();
}

void WebView::showMinimized() {
  utils::ShowMinimizedWindow(this, true);
}

void WebView::showNormal() {
  utils::ShowNormalWindow(this);
  this->raiseWindow();
}

void WebView::handleMoveWindow() {
  // Reset focus of window.
  qcef_webview_->hide();
  qcef_webview_->show();
  utils::MoveWindow(this);
//  const QPointF pos(QCursor::pos());
//  QMouseEvent* event = new QMouseEvent(QEvent::MouseButtonPress,
//                                       pos,
//                                       Qt::LeftButton,
//                                       Qt::NoButton,
//                                       Qt::NoModifier);
//  QCoreApplication::sendEvent(this, event);
}

void WebView::handleResizeWindow(const QString& corner) {
  qcef_webview_->hide();
  qcef_webview_->show();
  this->setResizable(true);
  utils::ResizeWindow(this->window(), corner);
}

}  // namespace netease
