// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "popup_menu_style.h"

#include <QDebug>
#include <QPixmap>
#include <QPixmapCache>
#include <QPainter>
#include <QPushButton>
#include <QStyleOption>
#include <QStyleOptionMenuItem>
#include <QVBoxLayout>

#include "application.h"
#include "service/skin_manager.h"

namespace netease {

namespace {

//const char kMenuBackground[] = "SKIN_ROOT/menu/bk.png";
const char kItemNormalBackground[] = "SKIN_ROOT/menu/item.png";
const char kItemHoverBackground[] = "SKIN_ROOT/menu/hover.png";
const char kSubMenuIcon[] = "SKIN_ROOT/menu/sub_icon.png";
const char kSeparatorBackground[] = "SKIN_ROOT/menu/separator.png";

const int kItemHeight = 28;
const int kIconPixelSize = 18;
const int kIconLeftMargin = 13;
const int kIconRightMargin = 11;
const int kIconTextSpacing = 17;
const int kSeparatorHeight = 3;
const int kFontPixelSize = 12;

}  // namespace

PopupMenuStyle::PopupMenuStyle() : QProxyStyle() {

}

int PopupMenuStyle::styleHint(QStyle::StyleHint hint,
                              const QStyleOption* option,
                              const QWidget* widget,
                              QStyleHintReturn* returnData) const {
  switch (hint) {
  case QStyle::SH_Menu_Scrollable:
    return 1;
  case QStyle::SH_Menu_KeyboardSearch:
    return 1;
  default:
    return QProxyStyle::styleHint(hint, option, widget, returnData);
  }
}

int PopupMenuStyle::pixelMetric(QStyle::PixelMetric metric,
                                const QStyleOption* option,
                                const QWidget* widget) const {
  switch (metric) {
  case QStyle::PM_MenuScrollerHeight:
    return 15;
  case QStyle::PM_MenuDesktopFrameWidth:
    return 0;
  default:
    return QProxyStyle::pixelMetric(metric, option, widget);
  }
}

void PopupMenuStyle::drawControl(QStyle::ControlElement control,
                                 const QStyleOption* option,
                                 QPainter* painter,
                                 const QWidget* widget) const {
  switch (control) {
  case QStyle::CE_MenuItem:
    drawMenuItem(option, painter);
    break;
  case QStyle::CE_MenuScroller:
    drawScroller(option, painter);
    break;
  default:
    QProxyStyle::drawControl(control, option, painter, widget);
  }
}

QSize PopupMenuStyle::sizeFromContents(QStyle::ContentsType type,
                                       const QStyleOption* option,
                                       const QSize& contentsSize,
                                       const QWidget* widget) const {
  QSize origin_size = QProxyStyle::sizeFromContents(type, option, contentsSize,
                                                    widget);
  switch (type) {
  case QStyle::CT_MenuItem:
    return getItemSize(option, origin_size);
  default:
    return origin_size;
  }
}

void PopupMenuStyle::drawScroller(const QStyleOption *option,
                                  QPainter *painter) const {
  QRect rect = option->rect;
  // Draw background
  painter->drawPixmap(rect, QPixmap(absoluteSkinPath(kItemNormalBackground)));
  // Draw arrow
  QPixmap arrow_pixmap = getArrowPixmap(option->state);
  QRect arrow_rect(rect.x() + (rect.width() - arrow_pixmap.width()) / 2,
                   rect.y() + (rect.height() - arrow_pixmap.height()) / 2,
                   arrow_pixmap.width(),
                   arrow_pixmap.height());
  painter->drawPixmap(arrow_rect, arrow_pixmap);
}

void PopupMenuStyle::drawMenuItem(const QStyleOption* option,
                                  QPainter* painter) const {
  if (const QStyleOptionMenuItem* menuItem =
        qstyleoption_cast<const QStyleOptionMenuItem* >(option)) {

    if (menuItem->menuItemType == QStyleOptionMenuItem::Separator) {
      painter->drawPixmap(option->rect,
                          QPixmap(absoluteSkinPath(kSeparatorBackground)));
    } else {
      // Draw background
      painter->drawPixmap(menuItem->rect, getItemBGPixmap(menuItem->state));

      drawMenuItemIcon(menuItem, painter);
      drawMenuItemText(menuItem, painter);
      drawMenuItemSubIcon(menuItem, painter);
      drawMenuItemCheckedIcon(menuItem, painter);
    }
  }
}

void PopupMenuStyle::drawMenuItemIcon(const QStyleOptionMenuItem* menu_item,
                                      QPainter* painter) const {
  QPixmap pixmap = getIconPixmap(menu_item->state,
                                 menu_item->icon,
                                 QSize(kIconPixelSize,
                                       kIconPixelSize));
  int height = menu_item->rect.height();
  int x = kIconLeftMargin;
  int y = menu_item->rect.y() + (height - pixmap.height()) / 2;
  painter->drawPixmap(x, y, pixmap.width(), pixmap.height(), pixmap);
}

void PopupMenuStyle::drawMenuItemText(const QStyleOptionMenuItem* menu_item,
                                      QPainter* painter) const {
  // Draw text
  int left_spacing = kIconLeftMargin + kIconPixelSize + kIconTextSpacing;
  int right_spacing = kIconRightMargin + kIconPixelSize;
  int x = left_spacing;
  int y = menu_item->rect.y();
  int width = menu_item->rect.width() - left_spacing - right_spacing;
  int height = menu_item->rect.height();

  QPen p(app->skin_manager->text_colors.menu_item);
  QFont f;
  f.setPixelSize(kFontPixelSize);
  painter->setPen(p);
  painter->setFont(f);
  painter->drawText(x, y, width, height, Qt::AlignLeft | Qt::AlignVCenter,
                    menu_item->text);
}

void PopupMenuStyle::drawMenuItemSubIcon(const QStyleOptionMenuItem* menu_item,
    QPainter* painter) const {
  // Draw sub-menu icon
  if (menu_item->menuItemType == QStyleOptionMenuItem::SubMenu) {
    QPixmap pixmap(absoluteSkinPath(kSubMenuIcon));

    int height = menu_item->rect.height();
    int x = menu_item->rect.width() - kIconRightMargin - kIconPixelSize;
    int y = menu_item->rect.y() + (height - pixmap.height()) / 2;
    painter->drawPixmap(x, y, pixmap.width(), pixmap.height(), pixmap);
  }
}

void PopupMenuStyle::drawMenuItemCheckedIcon(
  const QStyleOptionMenuItem* menu_item,
  QPainter* painter) const {
  if (menu_item->checked) {
    QPixmap pixmap;
    QPixmapCache::find(kCheckedIconCacheKey, &pixmap);

    int height = menu_item->rect.height();
    int x = menu_item->rect.width() - kIconRightMargin - kIconPixelSize;
    int y = menu_item->rect.y() + (height - pixmap.height()) / 2;
    painter->drawPixmap(x, y, pixmap.width(), pixmap.height(), pixmap);
  }
}

QSize PopupMenuStyle::getItemSize(const QStyleOption* option,
                                  const QSize& size) const {
  if (const QStyleOptionMenuItem* menu_item =
        qstyleoption_cast<const QStyleOptionMenuItem* >(option)) {
    switch (menu_item->menuItemType) {
    case QStyleOptionMenuItem::Separator:
      return QSize(size.width(), kSeparatorHeight);
    case QStyleOptionMenuItem::Normal:
      return QSize(size.width(), kItemHeight);
    default:
      return size;
    }
  } else {
    return size;
  }
}

QPixmap PopupMenuStyle::getArrowPixmap(const QStyle::State state) const {
  QTransform tf;
  tf.scale(1.5, 0.8);
  if (state & QStyle::State_DownArrow) {
    tf.rotate(90);
  } else {
    tf.rotate(-90);
  }
  return QPixmap(absoluteSkinPath(kSubMenuIcon)).
          transformed(tf, Qt::SmoothTransformation);
}

QPixmap PopupMenuStyle::getItemBGPixmap(const QStyle::State state) const {
  // QStyle::State_MouseOver not working, don't know why
  if (state & QStyle::State_Selected) {
    return QPixmap(absoluteSkinPath(kItemHoverBackground));
  } else {
    return QPixmap(absoluteSkinPath(kItemNormalBackground));
  }
}

QPixmap PopupMenuStyle::getIconPixmap(const QStyle::State state,
                                      const QIcon& icon,
                                      const QSize& size) const {
  QIcon::Mode icon_mode = QIcon::Normal;
  QIcon::State icon_state = QIcon::Off;
  if (state & QStyle::State_Active) {
    icon_mode = QIcon::Active;
  } else if (state & QStyle::State_Selected) {
    icon_mode = QIcon::Selected;
  } else if (!(state & QStyle::State_Enabled)) {
    icon_mode = QIcon::Disabled;
  }

  if (state & QStyle::State_On) {
    icon_state = QIcon::On;
  }

  return icon.pixmap(size, icon_mode, icon_state);
}

QString PopupMenuStyle::absoluteSkinPath(const QString& path) const {
  return app->skin_manager->absoluteSkinPath(path);
}

}  // namespace netease
