// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/views/widgets/volume_window.h"

#include <QDebug>
#include <QLabel>
#include <QMouseEvent>
#include <QPixmap>

#include "application.h"
#include "service/skin_manager.h"
#include "ui/utils/xutil.h"

namespace netease {

namespace {

const char kStyleSheetFile[] = ":/styles/volume-window.css";
const char kBackgroundFile[] = "SKIN_ROOT/mini/volume/background.png";
const char kBackgroundBottomFile[] =
    "SKIN_ROOT/mini/volume/background_bottom.png";
const char kRailFile[] = "SKIN_ROOT/mini/volume/rail.png";

const int kWindowWidth = 124;
const int kWindowHeight = 48;
const int kRailWidth = 120;
const int kRailHeight = 22;
const int kSlideWidth = 32;
const int kSlideHeight = 22;

const int kMousePosLeft = 4;
const int kMousePosRight = 91;
const int kMouseHorizontalRange = kMousePosRight - kMousePosLeft;

const int kOffsetX = 2;
const int kOffsetY = 6;
const int kReversedOffsetX = 2;
const int kReversedOffsetY = 13;

const int kReversedHeightThreshold = 60;
const int kReversedHeightOffset = 25;

const int kVolumeMinimum = 0;
const int kVolumeMaximum = 100;

}  // namespace

VolumeWindow::VolumeWindow(QWidget* parent)
    : QFrame(parent),
      volume_(0),
      is_reversed_(false),
      volume_changable_(false) {
  setObjectName("volume_window");

  initUI();
  updateSkin();
  setVolume(0);
  initConnections();
}

bool VolumeWindow::event(QEvent* event) {
  switch (event->type()) {
    case QEvent::WindowDeactivate: {
      this->hide();
      break;
    }

    // Hide volume window is cursor is not placed in its content area.
    case QEvent::Leave: {
      if (!this->geometry().contains(QCursor::pos())) {
        this->hide();
      }
      break;
    }

    default: {
      break;
    }
  }

  return QFrame::event(event);
}

void VolumeWindow::initConnections() {
  connect(app->skin_manager, &SkinManager::onUpdated,
          this, &VolumeWindow::updateSkin);
}

void VolumeWindow::initUI() {
  background_label_ = new QLabel(this);
  background_label_->setObjectName("background");

  rail_label_ = new QLabel(this);
  rail_label_->setFixedSize(kRailWidth, kRailHeight);
  rail_label_->move(kOffsetX, kOffsetY);
  rail_label_->setObjectName("rail");

  slide_label_ = new QLabel(this);
  slide_label_->setFixedSize(kSlideWidth, kSlideHeight);
  slide_label_->setObjectName("slide");

  setFixedSize(kWindowWidth, kWindowHeight);

  // Do not show in window manager.
  setWindowFlags(Qt::Dialog | Qt::FramelessWindowHint);
  setAttribute(Qt::WA_TranslucentBackground, true);
}

void VolumeWindow::mouseMoveEvent(QMouseEvent* event) {
  if (volume_changable_) {
    int x = event->x() - kSlideWidth / 2;
    if (x > kMousePosRight) {
      x = kMousePosRight;
    } else if (x < kMousePosLeft) {
      x = kMousePosLeft;
    }
    const int volume = static_cast<int>((x - kMousePosLeft) * 100 /
                                        double(kMouseHorizontalRange));

    // Emit volume changed signal and update ui.
    emit volumeChanged(volume);
    setVolume(volume);
  }
  QFrame::mouseMoveEvent(event);
}

void VolumeWindow::mousePressEvent(QMouseEvent* event) {
  if (event->button() == Qt::LeftButton) {
    volume_changable_ = true;
    // Change volume when mouse is pressed.
    mouseMoveEvent(event);
  }
  QFrame::mousePressEvent(event);
}

void VolumeWindow::mouseReleaseEvent(QMouseEvent* event) {
  volume_changable_ = false;
  QFrame::mouseReleaseEvent(event);
}

void VolumeWindow::setVolume(int volume) {
  int temp_volume;
  if (volume > kVolumeMaximum) {
    temp_volume = kVolumeMaximum;
  } else if (volume < kVolumeMinimum) {
    temp_volume = kVolumeMinimum;
  } else {
    temp_volume = volume;
  }

  // Only update volume if its value is changed.
  if (temp_volume != volume_) {
    volume_ = temp_volume;
    updateSlidePos();
  }
}

void VolumeWindow::show(const QPoint& pos) {
  const int x = pos.x() - kWindowWidth / 2;
  int y;
  if (pos.y() > kReversedHeightThreshold) {
    y = pos.y() - kWindowHeight;
    is_reversed_ = false;
  } else {
    // Move to bottom of cursor point.
    y = pos.y() + kReversedHeightOffset;
    is_reversed_ = true;
  }
  move(x, y);

  if (is_reversed_) {
    rail_label_->move(kReversedOffsetX, kReversedOffsetY);
    slide_label_->move(kReversedOffsetX, kReversedOffsetY);
  } else {
    rail_label_->move(kOffsetX, kOffsetY);
    slide_label_->move(kOffsetX, kOffsetY);
  }

  updateSkin();
  updateSlidePos();

  utils::SetWindowTypeHint(this, utils::X11WindowType::Dock);
  utils::SkipTaskbarPager(this);
  utils::SetStayOnTop(this, true);

  QFrame::show();
}

void VolumeWindow::updateSlidePos() {
  const int width = static_cast<int>(kMouseHorizontalRange * volume_ / 100.0);
  rail_label_->setFixedWidth(width + kSlideWidth / 2);
  if (is_reversed_) {
    slide_label_->move(kReversedOffsetX + width, kReversedOffsetY);
  } else {
    slide_label_->move(kOffsetX + width, kOffsetY);
  }
}

void VolumeWindow::updateSkin() {
  if (is_reversed_) {
    const QString background_filepath(
        app->skin_manager->absoluteSkinPath(kBackgroundBottomFile));
    background_label_->setPixmap(QPixmap(background_filepath));
  } else {
    const QString background_filepath(
        app->skin_manager->absoluteSkinPath(kBackgroundFile));
    background_label_->setPixmap(QPixmap(background_filepath));
  }

  const QString rail_filepath(app->skin_manager->absoluteSkinPath(kRailFile));
  rail_label_->setPixmap(QPixmap(rail_filepath));

  setStyleSheet(app->skin_manager->formatStyleFile(kStyleSheetFile));
}

}  // namespace netease
