// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#include "ui/window_manager.h"

#include <unistd.h>
#include <QColorDialog>
#include <QDebug>
#include <QDir>
#include <QFileDialog>

#include "application.h"
#include "base/consts.h"
#include "service/command_line.h"
#include "ui/views/desktop_lyrics.h"
#include "ui/views/mini_player.h"
#include "ui/views/web_view.h"

namespace netease {

WindowManager::WindowManager(QWidget* parent)
    : QObject(parent),
      web_views_() {
  this->setObjectName("window_manager");

  initConnections();
}

WindowManager::~WindowManager() {
}

void WindowManager::closeAllWebWindows() {
  for (WebView* view : web_views_) {
    view->hide();
//    delete view;
  }
  web_views_.clear();
}

void WindowManager::destroyWindow(WebView* web_view) {
  web_view->hide();
  web_view->deleteLater();
  web_views_.removeOne(web_view);
}

QRect WindowManager::getNativeWindowRect(const QString& window_name) const {
  if (window_name == app->desktop_lyrics->objectName()) {
    return app->desktop_lyrics->rect();
  } else if (window_name == app->mini_player->objectName()) {
    return app->mini_player->rect();
  } else {
    qWarning() << "[WindowManager]::getNativeWindowRect unknown window:"
               << window_name;
    return QRect{};
  }
}

void WindowManager::initConnections() const {
  connect(this, &WindowManager::chooseColor,
          this, &WindowManager::handleChooseColor);
  connect(this, &WindowManager::saveFile,
          this, &WindowManager::handleSaveFile);
  connect(this, &WindowManager::selectFiles,
          this, &WindowManager::handleSelectFiles);
  connect(this, &WindowManager::selectFolder,
          this, &WindowManager::handleSelectFolder);
}

void WindowManager::initMainWindow() {
  QUrl url(kIndexPage);
  url.setQuery(CommandLineToQuery());
  qDebug() << Q_FUNC_INFO << url;
  this->launchWindow(url.toString(), true, true, nullptr);
}

void WindowManager::launchWindow(const QString& url, bool visible,
                                 bool resizable, QWidget* parent) {
  qDebug() << Q_FUNC_INFO << url << visible << resizable << parent;
  Q_UNUSED(parent);
  Q_UNUSED(visible);

  // WebView window have not parent widget.
  WebView* web_view = new WebView();
  web_view->load(QUrl(url));
  // Hide web page first. Show main window when its content is ready.
  // Or else a blank web page will show up.
  web_view->hide();

  if (!url.startsWith(kIndexPage)) {
    web_view->setWindowModality(Qt::ApplicationModal);
  }

  if (!resizable) {
    web_view->disableResizable();
  }

  web_views_.append(web_view);
}

void WindowManager::setNativeWindowShow(const QString& window_name,
                                        bool show) const {
  qDebug() << Q_FUNC_INFO << window_name << show;
  if (window_name == app->desktop_lyrics->objectName()) {
    app->desktop_lyrics->setShowWindow(show);
  } else if (window_name == app->mini_player->objectName()) {
    if (show) {
      app->mini_player->show();
    } else {
      app->mini_player->hide();
    }
  } else {
    qWarning() << "[WinHelper]::setNativeWindowShow() unknown window: "
               << window_name;
    return;
  }
}

void WindowManager::handleChooseColor(const QString& element_id,
                                      const QString& color) {
  qDebug() << Q_FUNC_INFO << element_id << color;
  // Convert color string to its value.
  int color_value = color.toInt(0, 16);
  const int red = color_value >> 16;
  color_value = color_value % (1 << 16);
  const int green = color_value >> 8;
  const int blue = color_value % (1 << 8);
  QColorDialog dialog(web_views_.isEmpty() ? nullptr : web_views_.first());
  dialog.setWindowTitle(QObject::tr("Choose Color"));
//  dialog.setOption(QColorDialog::DontUseNativeDialog);
  dialog.setModal(true);
  dialog.setCurrentColor(QColor(red, green, blue));
  const int result = dialog.exec();

  if (result == QDialog::Accepted) {
    const QColor current_color(dialog.currentColor());
    emit this->onChooseColor(element_id, current_color.red(),
                             current_color.green(), current_color.blue());
  }
}

void WindowManager::handleSaveFile(const QString& keyword, const QString& title,
                                   const QString& filter,
                                   const QString& filename) {
  const QString org_path = QDir::home().absoluteFilePath(filename);
  const QString path = QFileDialog::getSaveFileName(web_views_.isEmpty() ? nullptr : web_views_.first(), title, org_path,
     filter, nullptr/*, QFileDialog::DontUseNativeDialog*/);

  if (!path.isEmpty()) {
    emit this->onSaveFile(keyword, 0, path);
  }
}

void WindowManager::handleSelectFiles(const QString &arg1,
                                      const QString &arg2,
                                      const QString &arg3,
                                      const QString &filter,
                                      int limit) {
  Q_UNUSED(arg1);
  Q_UNUSED(arg2);
  Q_UNUSED(arg3);

  const QString mod_filter(filter);
  const QStringList files = QFileDialog::getOpenFileNames(web_views_.isEmpty() ? nullptr : web_views_.first(),
      QObject::tr("Select Images"), "", mod_filter, nullptr/*,
      QFileDialog::DontUseNativeDialog*/);

  if (!files.isEmpty()) {
    QVariantList result;
    for (int i = 0; i < limit && i < files.size(); ++i) {
      QVariantMap obj;
      obj.insert("path", files.at(i));
      result.append(obj);
    }
    emit this->onSelectFiles("keyword", 1, result);
  } else {
    QVariantList result;
    emit this->onSelectFiles("keyword", 0, result);
  }
}

void WindowManager::handleSelectFolder(const QString& keyword,
                                       const QString& title,
                                       const QString& message,
                                       const QString& path) {
  Q_UNUSED(message);
  // If path is empty, file-manager will jump to executable folder.
  // So we redirect it to HOME folder instead.
  QString orig_path(path);
  if (orig_path.isEmpty()) {
    orig_path = QDir::homePath();
  }
  QString new_path = QFileDialog::getExistingDirectory(web_views_.isEmpty() ? nullptr : web_views_.first(),
                                                       title, orig_path/*, QFileDialog::DontUseNativeDialog*/);
  const bool ok = (new_path.length() > 0);
  if (!new_path.endsWith(QChar('/'))) {
    new_path.append(QChar('/'));
  }
  emit this->onSelectFolder(ok, keyword, ok ? new_path : "");
}

}  // namespace netease
