// Copyright (c) 2015 Deepin Ltd. All rights reserved.
// Use of this source is governed by Commercial License that can be found
// in the LICENSE file.

#ifndef UI_WINDOW_MANAGER_H
#define UI_WINDOW_MANAGER_H

#include <QList>
#include <QWidget>
#include <QDesktopWidget>

namespace netease {

class WebView;

// WindowManager is used to manage webview widgets and mini player.
// There can be several webview window displayed at the same time.
// Popup dialogs are displayed in a separate webview widgets.
// WindowManager also manages dialogs, like ColorChooserDialog and
// FileChooserDialog.
class WindowManager : public QObject {
  Q_OBJECT

 public:
  explicit WindowManager(QWidget* parent = nullptr);
  ~WindowManager() override;

  // Close all cef web windows.
  // Call this method before calling CefQuitMessageLoop().
  void closeAllWebWindows();

  void destroyWindow(WebView* web_view);

  // Get properties of native window. |window_name| might be "mini_player" or
  // "desktop_lyrics".
  QRect getNativeWindowRect(const QString& window_name) const;

  // Open a new web page in a new window. Might use this method to open a
  // new popup dialog.
  // |url| is the url to open.
  // |visible| set window visible or not.
  // |resizable| set window resizable or not.
  // |parent| is parent widget of new window.
  void launchWindow(const QString& url, bool visible, bool resizable,
                    QWidget* parent);

  // Change visibility of native window.
  // |window| might be "mini_player" or "desktop_lyrics".
  void setNativeWindowShow(const QString& window_name, bool show) const;

 signals:
  void onChooseColor(const QString& element_id, int red, int green, int blue);
  void chooseColor(const QString& element_id, const QString& color);

  void onSaveFile(const QString& keyword, int err, const QString& path);
  void saveFile(const QString& keyword, const QString& title,
                const QString& filter, const QString& filename);

  void onSelectFiles(const QString& keyword, int ok, const QVariantList& files);
  void selectFiles(const QString& arg1, const QString& arg2,
                   const QString& arg3, const QString& filter,
                   int limit);

  void onSelectFolder(bool ok, const QString& keyword, const QString& new_path);
  void selectFolder(const QString& keyword, const QString& title,
                    const QString& message, const QString& path);

 public slots:
  // Display main window of player.
  void initMainWindow();

 private:
  // Connect signals to slots.
  void initConnections() const;

  QList<WebView*> web_views_;

 private slots:
  // Dialog handlers.
  void handleChooseColor(const QString& element_id, const QString& color);
  void handleSaveFile(const QString& keyword, const QString& title,
                      const QString& filter, const QString& filename);
  void handleSelectFiles(const QString &arg1, const QString &arg2,
                         const QString &arg3, const QString &filter,
                         int limit);
  void handleSelectFolder(const QString& keyword, const QString& title,
                          const QString& message, const QString& path);
};

}  // namespace netease

#endif  // UI_WINDOW_MANAGER_H_
